#include "e_lib.h"
#include "patch.h"


static void p_CalculateSizes(INTERNAL_MESH *mesh, GEN *data);
static void p_RemoveDuplicateNodes(INTERNAL_MESH *mesh);
static void GenGeodStrMat(INTERNAL_MESH *mesh);
static void CheckGenBoundCond(GEN *data, INTERNAL_MESH *mesh, int node, 
                              int *types, double *value);
static void GenAllBoundCond(GEN *data, INTERNAL_MESH *mesh);


static void p_CollectMaterial(int nelem, int *material, int *mattrans, 
                              MAT *Props, int *matindex, char *defmatname);
static void p_OutputFinalMesh(INTERNAL_MESH *mesh, GEN *data,
                              char *filename);


/* ----------------------------------------------------------- */


void e_Generate_Mesh(GEN *Structure, char *filename)
{
  INTERNAL_MESH   iMesh;
  int             i_bloc, i_con;

  /* calculate the number of points and elements,
     this helps to allocate the appropriate vectors, matrices */
  p_CalculateSizes(&iMesh, Structure);

  /* process patches */
  for(i_bloc = 0; i_bloc < Structure->NSurfaces; i_bloc++)
  {
    /* quad surface */
    if(Structure->Patches[i_bloc].Shape == 4)
    {
      p_GenerateCubicPatch(Structure, i_bloc, &iMesh);
    }
    /* triangle surface */
    else if(Structure->Patches[i_bloc].Shape == 3)
    {
      p_GenerateTriangPatch(Structure, i_bloc, &iMesh);
    }
    else
    {
      e_GeneralError("Unknown surface type !");
    }
  }

  /* process connections */
  for(i_con = 0; i_con < Structure->NConnections; i_con++)
  {
    p_GenerateConnections(Structure, i_con, &iMesh);
  }

  printf("\nGeneration finished.");

  /* remove duplicate nodes */
  p_RemoveDuplicateNodes(&iMesh);

  /* generate material for geodesic strings  */
  if(Structure->DefMat && iMesh.NLinks[4] > 0)
  {
    /* uses final mesh */
    GenGeodStrMat(&iMesh);
  }

  /* generate boundary conditions */
  if(Structure->NBCSides || Structure->NBCNodes)
  {
    GenAllBoundCond(Structure, &iMesh);
  }
  else
  {
    iMesh.NBCNodes = 0;
  }

  /* write out the mesh */
  p_OutputFinalMesh(&iMesh, Structure, filename);

}


static void p_CalculateSizes(INTERNAL_MESH *mesh, GEN *data)
{
  int i_bloc, counter, il, type, i_con, i;
  int np;
  int nq;
  int nl[5];
  int nt[4];

  /* zero it out */
  mesh->NPoints    = 0;
  mesh->NQuads     = 0;
  mesh->NTriangles[0] = 0;
  mesh->NTriangles[1] = 0;
  mesh->NTriangles[2] = 0;
  mesh->NTriangles[3] = 0;
  mesh->NLinks[0]   = 0;
  mesh->NLinks[1]   = 0;
  mesh->NLinks[2]   = 0;
  mesh->NLinks[3]   = 0;
  mesh->NLinks[4]   = 0;

  mesh->Duplicate   = 0;

  np = 0;
  nq = 0;
  nl[0] = nl[1] = nl[2] = nl[3] = nl[4] = 0;
  nt[0] = nt[1] = nt[2] = nt[3] = 0;

  /* go through all patches */
  for(i_bloc = 0; i_bloc < data->NSurfaces; i_bloc++)
  {
    /* triangle patches */
    if(data->Patches[i_bloc].Shape == 3)
    {
      /* count registered links, only approximation */
      for(i = 0; i < 3; i++)
      {
        for(il = 0; il < data->Patches[i_bloc].LinkSides[i]; il++)
        {
          type = data->Patches[i_bloc].LinkType[i][il]-1;
          nl[type] += data->Patches[i_bloc].XDivision;
        }
      }

      /* no of meshpoints */
      for(i = 0; i < data->Patches[i_bloc].XDivision+1; i++)
      {
        np += data->Patches[i_bloc].XDivision+1-i;
      }
      
      /* no of elements */
      counter = data->Patches[i_bloc].XDivision *
                data->Patches[i_bloc].XDivision;

      if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG1)
      {
        nt[0] += counter;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG3)
      {
        nt[1] += counter;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG4)
      {
        nt[2] += counter;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG5)
      {
        nt[3] += counter;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_QUAD1)
      {
        e_GeneralError("Quad elements cannot be used in triangle surface!");
      }
    }
    /* cubic patches */
    else if(data->Patches[i_bloc].Shape == 4)
    {
      /* count registered links */
      for(il = 0; il < data->Patches[i_bloc].LinkSides[1]; il++)
      {
        type = data->Patches[i_bloc].LinkType[1][il]-1;
        nl[type] += data->Patches[i_bloc].YDivision;
      }

      for(il = 0; il < data->Patches[i_bloc].LinkSides[0]; il++)
      {
        type = data->Patches[i_bloc].LinkType[0][il]-1;
        nl[type] += data->Patches[i_bloc].XDivision;
      }
      
      /* no of meshpoints */
      np += (data->Patches[i_bloc].XDivision+1) *
            (data->Patches[i_bloc].YDivision+1);
      
      /* no of elements */
      counter = data->Patches[i_bloc].XDivision *
                data->Patches[i_bloc].YDivision;

      if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG1)
      {
        nt[0] += counter * 2;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG3)
      {
        nt[1] += counter * 2;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG4)
      {
        nt[2] += counter * 2;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_TRIANG5)
      {
        nt[3] += counter * 2;
      }
      else if(data->Patches[i_bloc].ElementType == E_ELEM_TYPE_QUAD1)
      {
        nq += counter;
      }
    }
  }

  /* go through all connections */
  for(i_con = 0; i_con < data->NConnections; i_con++)
  {
    type = data->Connections[i_con].ElementType;
    np += data->Connections[i_con].Division+1;
    nl[type] += data->Connections[i_con].Division;
  }
  
  /* allocate the spaces for the generation */

  mesh->Coords    = e_alloc_cont_d_matr(np, 3);
  mesh->Quads     = e_alloc_cont_i_matr(nq, 4);
  mesh->Triangles[0] = e_alloc_cont_i_matr(nt[0], 3);
  mesh->Triangles[1] = e_alloc_cont_i_matr(nt[1], 3);
  mesh->Triangles[2] = e_alloc_cont_i_matr(nt[2], 3);
  mesh->Triangles[3] = e_alloc_cont_i_matr(nt[3], 3);

  mesh->Links[0]  = e_alloc_cont_i_matr(nl[0], 2);
  mesh->Links[1]  = e_alloc_cont_i_matr(nl[1], 2);
  mesh->Links[2]  = e_alloc_cont_i_matr(nl[2], 2);
  mesh->Links[3]  = e_alloc_cont_i_matr(nl[3], 2);
  mesh->Links[4]  = e_alloc_cont_i_matr(nl[4], 2);

  /* material definition */
  if(data->DefMat)
  {
    mesh->MatQuads = e_alloc_i_vect(nq);
    for(i = 0; i < nq; i++)
      mesh->MatQuads[i] = MAT_QUAD1;
    
    mesh->MatTriangles[0] = e_alloc_i_vect(nt[0]);
    for(i = 0; i < nt[0]; i++)
      mesh->MatTriangles[0][i] = MAT_TRIANG1;

    mesh->MatTriangles[1] = e_alloc_i_vect(nt[1]);
    for(i = 0; i < nt[1]; i++)
      mesh->MatTriangles[1][i] = MAT_TRIANG3;

    mesh->MatTriangles[2] = e_alloc_i_vect(nt[2]);
    for(i = 0; i < nt[2]; i++)
      mesh->MatTriangles[2][i] = MAT_TRIANG4;

    mesh->MatTriangles[3] = e_alloc_i_vect(nt[3]);
    for(i = 0; i < nt[3]; i++)
      mesh->MatTriangles[3][i] = MAT_TRIANG5;


    mesh->MatLinks[0] = e_alloc_i_vect(nl[0]);
    for(i = 0; i < nl[0]; i++)
      mesh->MatLinks[0][i] = MAT_LINK1;

    mesh->MatLinks[1] = e_alloc_i_vect(nl[1]);
    for(i = 0; i < nl[1]; i++)
      mesh->MatLinks[1][i] = MAT_LINK2;

    mesh->MatLinks[2] = e_alloc_i_vect(nl[2]);
    for(i = 0; i < nl[2]; i++)
      mesh->MatLinks[2][i] = MAT_LINK3;

    mesh->MatLinks[3] = e_alloc_i_vect(nl[3]);
    for(i = 0; i < nl[3]; i++)
      mesh->MatLinks[3][i] = MAT_LINK4;

    mesh->MatLinks[4] = e_alloc_i_vect(nl[4]);
    for(i = 0; i < nl[4]; i++)
      mesh->MatLinks[4][i] = MAT_MID_LINK5;
  }
}



static void p_RemoveDuplicateNodes(INTERNAL_MESH *mesh)
{
  int     i, j;
  double  dist;
  int     mfact, user_dist, warn, nrem;

  printf("\nDuplicate node removal started ...");
  printf("\nChecking %d nodes\n", mesh->NPoints);

  nrem = 0;
  warn = 0;
  if(mesh->NPoints > 1000)
  {
    warn = 1;
    user_dist = mesh->NPoints / 1000;
  }
  mesh->Mark = e_alloc_i_vect(mesh->NPoints);
  mesh->Translate = e_alloc_i_vect(mesh->NPoints);

  for(i = 0; i < mesh->NPoints; i++)
  {
    /* whether notify the user about progress */
    if(warn && (i % user_dist == 0))
    {
      fprintf(stderr, ".");
    }
  
    if(mesh->Mark[i] == 0)
      mesh->Translate[i] += i+1; /* ensure the same node number, minus translations */

    mfact = 0;
    for(j = i+1; j < mesh->NPoints; j++)
    {
      if(mesh->Mark[j] == 0)
      {
        dist = sqrt( ( (mesh->Coords[i][0] - mesh->Coords[j][0]) *
                       (mesh->Coords[i][0] - mesh->Coords[j][0])   ) +
                     ( (mesh->Coords[i][1] - mesh->Coords[j][1]) *
                       (mesh->Coords[i][1] - mesh->Coords[j][1])   ) +
                     ( (mesh->Coords[i][2] - mesh->Coords[j][2]) *
                       (mesh->Coords[i][2] - mesh->Coords[j][2])   )   );

        /* they are the same */
        if(dist < 1e-8)
        {
          mesh->Mark[j] = 1;
          mesh->Translate[j] = mesh->Translate[i];
          (mesh->Duplicate)++;
          mfact++;
          nrem++;
        }
        else /* remove duplicate nodes, compress upper nodes */
        {
          mesh->Translate[j] -= mfact;
        }
      }
    }
  }
  printf("\nNumber of removed nodes: %d", nrem);
  
}

/* ############################################################################ */

static void GenGeodStrMat(INTERNAL_MESH *mesh)
{
  int  *nelem, **elems;
  int   i, j, k, node, ntotal;

  /* this is a little bit too much, 
     better approximation is possible */
  nelem = e_alloc_i_vect(mesh->NPoints);
  elems = e_init_frag_i_matr(mesh->NPoints);
  ntotal = 0;
  /* for all five type of link */
  for(k = 0; k < 5; k++)
  {
    /* for all element */
    for(i = 0; i < mesh->NLinks[k]; i++)
    {
      /* for both nodes */
      for(j = 0; j < 2; j++)
      {
        /* nodes are already removed !!! */
        node = mesh->Translate[mesh->Links[k][i][j]-1]-1;
        elems[node] = e_incr_i_vect(nelem[node], elems[node], ALLOC_SIZE);
        elems[node][nelem[node]] = i+ntotal;
        nelem[node]++;
      }
    }
    ntotal += mesh->NLinks[k];
  }
    
  /* count non geodesic strings */
  ntotal = 0;
  for(k = 0; k < 4; k++)
    ntotal += mesh->NLinks[k];
  
  /* check all geodesic strings */
  for(i = 0; i < mesh->NLinks[4]; i++)
  {
    /* start node check */
    node = mesh->Translate[mesh->Links[4][i][0]-1]-1;
    /* check for other link elements */
    for(j = 0; j < nelem[node]; j++)
      if(elems[node][j] < ntotal)
        break;
    /* found some other */
    if(j != nelem[node])
    {
      mesh->MatLinks[4][i] = MAT_BEG_LINK5;
    }
    if(nelem[node] == 1)
      e_GeneralWarning("G-string %d does not connect to any other link at its starting node", 
                       i+1);

    /* end node check */
    node = mesh->Translate[mesh->Links[4][i][1]-1]-1;
    /* check for other link elements */
    for(j = 0; j < nelem[node]; j++)
      if(elems[node][j] < ntotal)
        break;
    /* found some other */
    if(j != nelem[node])
    {
      if(mesh->MatLinks[4][i] == MAT_BEG_LINK5)
      {
        e_GeneralWarning("Both ends should be nullified for g-string %d", i+1);
        mesh->MatLinks[4][i] = MAT_NON_LINK5;
      }
      else
      {
        mesh->MatLinks[4][i] = MAT_END_LINK5;
      }
    }
    if(nelem[node] == 1)
      e_GeneralWarning("G-string %d does not connect to any other link at its end node", 
                       i+1);

  }
    
  e_free(nelem);
  e_free_frag_i_matr(mesh->NPoints, elems);
}

/* ############################################################################ */
/*                         Boundary condition genration                         */
/* ############################################################################ */

static void CheckGenBoundCond(GEN *data, INTERNAL_MESH *mesh, int node, 
                              int *types, double *value)
{
  int j, k;

  /* check that this node is already defined, then overwrite */
  for(k = 0; k < mesh->NBCNodes; k++)
  {
    if(mesh->BCNodes[k] == node+1)
      break;
  }
  /* new node */
  if(k == mesh->NBCNodes)
  {
    mesh->BCNodes = e_incr_i_vect(mesh->NBCNodes, 
                                  mesh->BCNodes, ALLOC_SIZE);
    mesh->BCNodes[k] = node+1;
    
    mesh->BCTypes = e_incr_frag_i_matr(mesh->NBCNodes, 
                                       mesh->BCTypes, ALLOC_SIZE);
    mesh->BCTypes[k] = e_alloc_i_vect(E_NDIMS);
    mesh->BCValue = e_incr_frag_d_matr(mesh->NBCNodes, 
                                       mesh->BCValue, ALLOC_SIZE);
    mesh->BCValue[k] = e_alloc_d_vect(E_NDIMS);
    
    mesh->NBCNodes++;
  }
  else
  {
    for(j = 0; j < E_NDIMS; j++)
    {
      if(   types[j] != mesh->BCTypes[k][j]
         || value[j] != mesh->BCValue[k][j])
        break;
    }
    if(j != E_NDIMS)
    {
      e_GeneralWarning("Boundary condition is defined again for node %d",
                       node+1);
    }
  }

  memcpy((void *)(mesh->BCTypes[k]),
         (void *)(types),
         sizeof(int) * E_NDIMS);
  memcpy((void *)(mesh->BCValue[k]),
         (void *)(value),
         sizeof(double) * E_NDIMS);
}



static void GenAllBoundCond(GEN *data, INTERNAL_MESH *mesh)
{
  int i, j, k, jj, patchID, linkID, node;
  int xs, xe, ys, ye, sum;
  int x, y, found;


  mesh->NBCNodes = 0;
  mesh->BCNodes = NULL;
  mesh->BCTypes = NULL;
  mesh->BCValue = NULL;
  
  /* first, generate the boundary condition along the defined sides */
  for(i = 0; i < data->NBCSides; i++)
  {
    /* reference to surface nodes */
    if(data->BCSides[i][0].Node == SURFACE_NODE)
    {
      patchID = data->BCSides[i][0].ElemID-1;

      /* determine the smaller y position */
      if(data->BCSides[i][0].Position[1] < data->BCSides[i][1].Position[1])
      {
        ys = data->BCSides[i][0].Position[1];
        ye = data->BCSides[i][1].Position[1];
      }
      else
      {
        ys = data->BCSides[i][1].Position[1];
        ye = data->BCSides[i][0].Position[1];
      }
      /* determine the smaller x position */
      if(data->BCSides[i][0].Position[0] < data->BCSides[i][1].Position[0])
      {
        xs = data->BCSides[i][0].Position[0];
        xe = data->BCSides[i][1].Position[0];
      }
      else
      {
        xs = data->BCSides[i][1].Position[0];
        xe = data->BCSides[i][0].Position[0];
      }
      
      /* quad patches */
      if(data->Patches[patchID].Shape == 4)
      {
        for(jj = xs; jj <= xe; jj++)
        {
          for(j = ys; j <= ye; j++)
          {
            /* check that this node is already defined, then overwrite */
            node = mesh->Translate[data->Patches[patchID].
                                         RegNodes[j][jj].
                                         Node-1]-1;
            CheckGenBoundCond(data, mesh, node, data->BCSTypes[i], data->BCSValue[i]);
          }
        }
      }
      else if(data->Patches[patchID].Shape == 3)
      {
        /* second side */
        if(data->BCSides[i][0].Position[0]+data->BCSides[i][0].Position[1] ==
           data->BCSides[i][1].Position[0]+data->BCSides[i][1].Position[1] )
        {
          sum = data->BCSides[i][0].Position[0]+data->BCSides[i][0].Position[1];
          for(jj = xs; jj <= xe; jj++)
          {
            node = mesh->Translate[data->Patches[patchID].
                                         RegNodes[sum-jj][jj].
                                         Node-1]-1;
            CheckGenBoundCond(data, mesh, node, data->BCSTypes[i], data->BCSValue[i]);
          }
        
        }
        /* first or third side */
        else
        {
          for(jj = xs; jj <= xe; jj++)
          {
            for(j = ys; j <= ye; j++)
            {
              node = mesh->Translate[data->Patches[patchID].
                                           RegNodes[j][jj].
                                           Node-1]-1;
              CheckGenBoundCond(data, mesh, node, data->BCSTypes[i], data->BCSValue[i]);
            }
          }
        }
      }
      else
      {
        e_GeneralError("invalid surface type");
      }
    }
    /* reference to link nodes */
    else if(data->BCSides[i][0].Node == LINK_NODE)
    {
      linkID = data->BCSides[i][0].ElemID-1;

      if(data->BCSides[i][0].Position[0] < data->BCSides[i][1].Position[0])
      {
        xs = data->BCSides[i][0].Position[0];
        xe = data->BCSides[i][1].Position[0];
      }
      else
      {
        xs = data->BCSides[i][1].Position[0];
        xe = data->BCSides[i][0].Position[0];
      }

      for(jj = xs; jj <= xe; jj++)
      {
        /* check that this node is already defined, then overwrite */
        node = mesh->Translate[data->Connections[linkID].RegNodes[jj].Node-1]-1;
        CheckGenBoundCond(data, mesh, node, data->BCSTypes[i], data->BCSValue[i]);
      }
    }

  } /* all BC side or area definition */

  /* 
   * Boundary nodes comes second.
   * These BCs can redefine the side definitions.
   */
  for(i = 0; i < data->NBCNodes; i++)
  {
    found = FALSE;
    /* reference for a point entity */
    node = data->BCNodes[i];
    /* search all surfaces for this point entity */
    for(j = 0; j < data->NSurfaces && found == FALSE; j++)
    {
      /* quad surface */
      if(data->Patches[j].Shape == 4)
      {
        /* linear definition type */
        if(   data->Patches[j].Definition == DEF_LINEAR
           || data->Patches[j].Definition == DEF_PARAMETRIC )
        {
          for(k = 0; k < 4; k++)
          {
            if(   data->Patches[j].Nodes[k].Node > 0
               && data->Patches[j].Nodes[k].Node == node)
            {
              GetQuadCornerPos(&(data->Patches[j]), k, &x, &y);
              node = mesh->Translate[data->Patches[j].RegNodes[y][x].Node-1]-1;
              CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
              found = TRUE;
              break;
            }
          }
        }
        else
        {
          e_GeneralError("unsupported definition type. This is an INTERNAL ERROR !");
        }
      }
      /* triangular surface */
      else if(data->Patches[j].Shape == 3)
      {
        if(   data->Patches[j].Nodes[0].Node > 0
           && data->Patches[j].Nodes[0].Node == node)
        {
          node = mesh->Translate[data->Patches[j].RegNodes[0][0].Node-1]-1;
          CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
          found = TRUE;
        }
        else if(   data->Patches[j].Nodes[1].Node > 0
                && data->Patches[j].Nodes[1].Node == node)
        {
          node = mesh->Translate[data->Patches[j].RegNodes[0][data->Patches[j].XDivision].Node-1]-1;
          CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
          found = TRUE;
        }
        else if(   data->Patches[j].Nodes[2].Node > 0
                && data->Patches[j].Nodes[2].Node == node)
        {
          node = mesh->Translate[data->Patches[j].RegNodes[data->Patches[j].XDivision][0].Node-1]-1;
          CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
          found = TRUE;
        }

      }

    }
    
    /* check link nodes as well */
    for(j = 0; j < data->NConnections && found == FALSE; j++)
    {
      if(   data->Connections[j].Nodes[0].Node > 0
         && data->Connections[j].Nodes[0].Node == node)
      {
        node = mesh->Translate[data->Connections[j].RegNodes[0].Node-1]-1;
        CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
        found = TRUE;
        break;
      }

      if(   data->Connections[j].Nodes[1].Node > 0
         && data->Connections[j].Nodes[1].Node == node)
      {
        node = mesh->Translate[data->Connections[j].RegNodes[data->Connections[j].Division].Node-1]-1;
        CheckGenBoundCond(data, mesh, node, data->BCNTypes[i], data->BCNValue[i]);
        found = TRUE;
        break;
      }
    }
    
    if(found == FALSE)
    {
      e_GeneralWarning("point boundary condition definition %d is not used", i+1);
    }
  }
}

/* ##################################################################### */
/* generate E-LIB format mesh */
/* ##################################################################### */

static void p_CollectMaterial(int nelem, int *material, int *mattrans, 
                              MAT *Props, int *matindex, char *defmatname)
{
  int i;
  char matname[E_LINE_MAX+1];

  for(i = 0; i < nelem; i++)
  {
    if(mattrans[material[i]] == -1)
    {
      mattrans[material[i]] = *matindex;
      if(material[i] < DEFAULT_N_MATERIAL)
      {
        /* geodesic material beginning */
        if(material[i] == MAT_BEG_LINK5)
        {
          Props[*matindex].Name = e_alloc_c_vect(strlen(MAT_BEG_LINK5_NAME)+1);
          strcpy(Props[*matindex].Name, MAT_BEG_LINK5_NAME);
        }
        /* geodesic material ending */
        else if(material[i] == MAT_END_LINK5)
        {
          Props[*matindex].Name = e_alloc_c_vect(strlen(MAT_END_LINK5_NAME)+1);
          strcpy(Props[*matindex].Name, MAT_END_LINK5_NAME);
        }
        /* geodesic material both ends killed */
        else if(material[i] == MAT_NON_LINK5)
        {
          Props[*matindex].Name = e_alloc_c_vect(strlen(MAT_NON_LINK5_NAME)+1);
          strcpy(Props[*matindex].Name, MAT_NON_LINK5_NAME);
        }
        /* middle material */
        else
        {
          Props[*matindex].Name = e_alloc_c_vect(strlen(defmatname)+1);
          strcpy(Props[*matindex].Name, defmatname);
        }
      }
      /* user defined material */
      else
      {
        sprintf(matname, "%s%03d", USER_DEF_MAT_NAME, 
                material[i] + 1 - DEFAULT_N_MATERIAL);

        Props[*matindex].Name = 
          e_alloc_c_vect(strlen(matname)+1);
        strcpy(Props[*matindex].Name, matname);
      }
      
      (*matindex)++;
    }
  }
}



static void p_OutputFinalMesh(INTERNAL_MESH *mesh, GEN *data,
                              char *filename)
{
  int    i, j, node, np, matindex, telem;
  MESH  *outmesh;
  int   *mattrans;

  printf("\n\nGenerating E-Lib mesh...\n\n");

  outmesh = (MESH *)e_calloc(1, sizeof(MESH));
  
  for(i = 0; i < E_NKEYWORDS; i++)
    outmesh->KeywordSet[i] = FALSE;
  for(i = 0; i < E_NSUBSTRUCTS; i++)
    outmesh->SubStructSet[i] = FALSE;

  outmesh->Title = e_alloc_c_vect(strlen(data->Title)+1);
  strcpy(outmesh->Title, data->Title);
  outmesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;

  outmesh->NMeshPoints = mesh->NPoints - mesh->Duplicate;
  outmesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  outmesh->NNodes = outmesh->NMeshPoints;
  outmesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;

  outmesh->TotalNElems = mesh->NLinks[0] +
                         mesh->NLinks[1] +
                         mesh->NLinks[2] +
                         mesh->NLinks[3] +
                         mesh->NLinks[4] +
                         mesh->NTriangles[0] +
                         mesh->NTriangles[1] +
                         mesh->NTriangles[2] +
                         mesh->NTriangles[3] +
                         mesh->NQuads;
  
  if(data->DefMat)
  {
    outmesh->Mats = e_alloc_vp_vect(outmesh->TotalNElems);
    outmesh->Props = (MAT *)e_calloc(DEFAULT_N_MATERIAL + data->NMaterials, 
                                     sizeof(MAT) );
    mattrans = e_alloc_i_vect(DEFAULT_N_MATERIAL + data->NMaterials);
    for(i = 0; i < DEFAULT_N_MATERIAL + data->NMaterials; i++)
    {
      mattrans[i] = -1;
    }

    matindex = 0;
    if(mesh->NLinks[0] > 0)
    {
      p_CollectMaterial(mesh->NLinks[0], mesh->MatLinks[0], mattrans, 
                        outmesh->Props, &matindex, MAT_LINK1_NAME);
    }

    if(mesh->NLinks[1] > 0)
    {
      p_CollectMaterial(mesh->NLinks[1], mesh->MatLinks[1], mattrans, 
                        outmesh->Props, &matindex, MAT_LINK2_NAME);
    }

    if(mesh->NLinks[2] > 0)
    {
      p_CollectMaterial(mesh->NLinks[2], mesh->MatLinks[2], mattrans, 
                        outmesh->Props, &matindex, MAT_LINK3_NAME);
    }

    if(mesh->NLinks[3] > 0)
    {
      p_CollectMaterial(mesh->NLinks[3], mesh->MatLinks[3], mattrans, 
                        outmesh->Props, &matindex, MAT_LINK4_NAME);
    }

    if(mesh->NLinks[4] > 0)
    {
      p_CollectMaterial(mesh->NLinks[4], mesh->MatLinks[4], mattrans, 
                        outmesh->Props, &matindex, MAT_MID_LINK5_NAME);
    }

    if(mesh->NTriangles[0] > 0)
    {
      p_CollectMaterial(mesh->NTriangles[0], mesh->MatTriangles[0], mattrans, 
                        outmesh->Props, &matindex, MAT_TRIANG1_NAME);
    }

    if(mesh->NTriangles[1] > 0)
    {
      p_CollectMaterial(mesh->NTriangles[1], mesh->MatTriangles[1], mattrans, 
                        outmesh->Props, &matindex, MAT_TRIANG3_NAME);
    }

    if(mesh->NTriangles[2] > 0)
    {
      p_CollectMaterial(mesh->NTriangles[2], mesh->MatTriangles[2], mattrans, 
                        outmesh->Props, &matindex, MAT_TRIANG4_NAME);
    }

    if(mesh->NTriangles[3] > 0)
    {
      p_CollectMaterial(mesh->NTriangles[3], mesh->MatTriangles[3], mattrans, 
                        outmesh->Props, &matindex, MAT_TRIANG5_NAME);
    }

    if(mesh->NQuads > 0)
    {
      p_CollectMaterial(mesh->NQuads, mesh->MatQuads, mattrans, 
                        outmesh->Props, &matindex, MAT_QUAD1_NAME);
    }

    outmesh->NMats = matindex;
    outmesh->NMdfMats = matindex;
    outmesh->KeywordSet[E_KW_NUM_NMATS] = TRUE;
  }

  telem = 1;
  if(mesh->NLinks[0] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_LINK1] = telem;
    telem += mesh->NLinks[0];

    outmesh->NElemsLink1 = mesh->NLinks[0];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_LINK1] = TRUE;
    outmesh->NodesLink1 = 
      e_alloc_cont_i_matr(mesh->NLinks[0], E_NNPE_LINK1);

    for(i = 0; i < mesh->NLinks[0]; i++)
    {
      outmesh->NodesLink1[i][0] = mesh->Translate[mesh->Links[0][i][0]-1];
      outmesh->NodesLink1[i][1] = mesh->Translate[mesh->Links[0][i][1]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_LINK1] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_LINK1] = TRUE;
      outmesh->MatsLink1 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1;

      for(i = 0; i < mesh->NLinks[0]; i++)
      {
        outmesh->MatsLink1[i] = outmesh->Props + 
                                mattrans[mesh->MatLinks[0][i]];
      }
    }

  }

  if(mesh->NLinks[1] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_LINK2] = telem;
    telem += mesh->NLinks[1];

    outmesh->NElemsLink2 = mesh->NLinks[1];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_LINK2] = TRUE;
    outmesh->NodesLink2 = 
      e_alloc_cont_i_matr(mesh->NLinks[1], E_NNPE_LINK2);

    for(i = 0; i < mesh->NLinks[1]; i++)
    {
      outmesh->NodesLink2[i][0] = mesh->Translate[mesh->Links[1][i][0]-1];
      outmesh->NodesLink2[i][1] = mesh->Translate[mesh->Links[1][i][1]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_LINK2] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_LINK2] = TRUE;
      outmesh->MatsLink2 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1;

      for(i = 0; i < mesh->NLinks[1]; i++)
      {
        outmesh->MatsLink2[i] = outmesh->Props + 
                                mattrans[mesh->MatLinks[1][i]];
      }
    }

  }


  if(mesh->NLinks[2] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_LINK3] = telem;
    telem += mesh->NLinks[2];

    outmesh->NElemsLink3 = mesh->NLinks[2];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_LINK3] = TRUE;
    outmesh->NodesLink3 = 
      e_alloc_cont_i_matr(mesh->NLinks[2], E_NNPE_LINK3);

    for(i = 0; i < mesh->NLinks[2]; i++)
    {
      outmesh->NodesLink3[i][0] = mesh->Translate[mesh->Links[2][i][0]-1];
      outmesh->NodesLink3[i][1] = mesh->Translate[mesh->Links[2][i][1]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_LINK3] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_LINK3] = TRUE;
      outmesh->MatsLink3 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_LINK3]-1;

      for(i = 0; i < mesh->NLinks[2]; i++)
      {
        outmesh->MatsLink3[i] = outmesh->Props + 
                                mattrans[mesh->MatLinks[2][i]];
      }
    }

  }

  if(mesh->NLinks[3] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_LINK4] = telem;
    telem += mesh->NLinks[3];

    outmesh->NElemsLink4 = mesh->NLinks[3];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_LINK4] = TRUE;
    outmesh->NodesLink4 = 
      e_alloc_cont_i_matr(mesh->NLinks[3], E_NNPE_LINK4);

    for(i = 0; i < mesh->NLinks[3]; i++)
    {
      outmesh->NodesLink4[i][0] = mesh->Translate[mesh->Links[3][i][0]-1];
      outmesh->NodesLink4[i][1] = mesh->Translate[mesh->Links[3][i][1]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_LINK4] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_LINK4] = TRUE;
      outmesh->MatsLink4 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_LINK4]-1;

      for(i = 0; i < mesh->NLinks[3]; i++)
      {
        outmesh->MatsLink4[i] = outmesh->Props + 
                                mattrans[mesh->MatLinks[3][i]];
      }
    }

  }

  if(mesh->NLinks[4] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_LINK5] = telem;
    telem += mesh->NLinks[4];

    outmesh->NElemsLink5 = mesh->NLinks[4];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_LINK5] = TRUE;
    outmesh->NodesLink5 = 
      e_alloc_cont_i_matr(mesh->NLinks[4], E_NNPE_LINK5);

    for(i = 0; i < mesh->NLinks[4]; i++)
    {
      outmesh->NodesLink5[i][0] = mesh->Translate[mesh->Links[4][i][0]-1];
      outmesh->NodesLink5[i][1] = mesh->Translate[mesh->Links[4][i][1]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_LINK5] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_LINK5] = TRUE;
      outmesh->MatsLink5 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_LINK5]-1;

      for(i = 0; i < mesh->NLinks[4]; i++)
      {
        outmesh->MatsLink5[i] = outmesh->Props + 
                                mattrans[mesh->MatLinks[4][i]];
      }
    }

  }

  if(mesh->NTriangles[0] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1] = telem;
    telem += mesh->NTriangles[0];

    outmesh->NElemsTriang1 = mesh->NTriangles[0];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG1] = TRUE;
    outmesh->NodesTriang1 = 
      e_alloc_cont_i_matr(mesh->NTriangles[0], E_NNPE_TRIANG1);

    for(i = 0; i < mesh->NTriangles[0]; i++)
    {
      outmesh->NodesTriang1[i][0] = mesh->Translate[mesh->Triangles[0][i][0]-1];
      outmesh->NodesTriang1[i][1] = mesh->Translate[mesh->Triangles[0][i][1]-1];
      outmesh->NodesTriang1[i][2] = mesh->Translate[mesh->Triangles[0][i][2]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_TRIANG1] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_TRIANG1] = TRUE;
      outmesh->MatsTriang1 = (MAT **)outmesh->Mats +
                             outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG1]-1;

      for(i = 0; i < mesh->NTriangles[0]; i++)
      {
        outmesh->MatsTriang1[i] = outmesh->Props + 
                                  mattrans[mesh->MatTriangles[0][i]];
      }
    }

  }

  if(mesh->NTriangles[1] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = telem;
    telem += mesh->NTriangles[1];

    outmesh->NElemsTriang3 = mesh->NTriangles[1];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3] = TRUE;
    outmesh->NodesTriang3 = e_alloc_cont_i_matr(mesh->NTriangles[1], E_NNPE_TRIANG3);

    for(i = 0; i < mesh->NTriangles[1]; i++)
    {
      outmesh->NodesTriang3[i][0] = mesh->Translate[mesh->Triangles[1][i][0]-1];
      outmesh->NodesTriang3[i][1] = mesh->Translate[mesh->Triangles[1][i][1]-1];
      outmesh->NodesTriang3[i][2] = mesh->Translate[mesh->Triangles[1][i][2]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_TRIANG3] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_TRIANG3] = TRUE;
      outmesh->MatsTriang3 = (MAT **)outmesh->Mats +
                             outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1;

      for(i = 0; i < mesh->NTriangles[1]; i++)
      {
        outmesh->MatsTriang3[i] = outmesh->Props + 
                                  mattrans[mesh->MatTriangles[1][i]];
      }
    }

  }

  if(mesh->NTriangles[2] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = telem;
    telem += mesh->NTriangles[2];

    outmesh->NElemsTriang4 = mesh->NTriangles[2];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4] = TRUE;
    outmesh->NodesTriang4 = e_alloc_cont_i_matr(mesh->NTriangles[2], E_NNPE_TRIANG4);

    for(i = 0; i < mesh->NTriangles[2]; i++)
    {
      outmesh->NodesTriang4[i][0] = mesh->Translate[mesh->Triangles[2][i][0]-1];
      outmesh->NodesTriang4[i][1] = mesh->Translate[mesh->Triangles[2][i][1]-1];
      outmesh->NodesTriang4[i][2] = mesh->Translate[mesh->Triangles[2][i][2]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_TRIANG4] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_TRIANG4] = TRUE;
      outmesh->MatsTriang4 = (MAT **)outmesh->Mats +
                             outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1;

      for(i = 0; i < mesh->NTriangles[2]; i++)
      {
        outmesh->MatsTriang4[i] = outmesh->Props + 
                                  mattrans[mesh->MatTriangles[2][i]];
      }
    }

  }

  if(mesh->NTriangles[3] > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] = telem;
    telem += mesh->NTriangles[3];

    outmesh->NElemsTriang5 = mesh->NTriangles[3];
    outmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5] = TRUE;
    outmesh->NodesTriang5 = e_alloc_cont_i_matr(mesh->NTriangles[3], E_NNPE_TRIANG5);

    for(i = 0; i < mesh->NTriangles[3]; i++)
    {
      outmesh->NodesTriang5[i][0] = mesh->Translate[mesh->Triangles[3][i][0]-1];
      outmesh->NodesTriang5[i][1] = mesh->Translate[mesh->Triangles[3][i][1]-1];
      outmesh->NodesTriang5[i][2] = mesh->Translate[mesh->Triangles[3][i][2]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_TRIANG5] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_TRIANG5] = TRUE;
      outmesh->MatsTriang5 = (MAT **)outmesh->Mats +
                             outmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]-1;

      for(i = 0; i < mesh->NTriangles[3]; i++)
      {
        outmesh->MatsTriang5[i] = outmesh->Props + 
                                  mattrans[mesh->MatTriangles[3][i]];
      }
    }

  }

  if(mesh->NQuads > 0)
  {
    outmesh->FirstNetIndex[E_ELEM_TYPE_QUAD1] = telem;
    telem += mesh->NQuads;

    outmesh->NElemsQuad1 = mesh->NQuads;
    outmesh->KeywordSet[E_KW_NUM_NELEMS_QUAD1] = TRUE;
    outmesh->NodesQuad1 = e_alloc_cont_i_matr(mesh->NQuads, E_NNPE_QUAD1);
    
    for(i = 0; i < mesh->NQuads; i++)
    {
      outmesh->NodesQuad1[i][0] = mesh->Translate[mesh->Quads[i][0]-1];
      outmesh->NodesQuad1[i][1] = mesh->Translate[mesh->Quads[i][1]-1];
      outmesh->NodesQuad1[i][2] = mesh->Translate[mesh->Quads[i][2]-1];
      outmesh->NodesQuad1[i][3] = mesh->Translate[mesh->Quads[i][3]-1];
    }
    outmesh->KeywordSet[E_KW_NUM_NODES_QUAD1] = TRUE;
    
    if(data->DefMat)
    {
      outmesh->KeywordSet[E_KW_NUM_MATS_QUAD1] = TRUE;
      outmesh->MatsQuad1 = (MAT **)outmesh->Mats +
                           outmesh->FirstNetIndex[E_ELEM_TYPE_QUAD1]-1;

      for(i = 0; i < mesh->NQuads; i++)
      {
        outmesh->MatsQuad1[i] = outmesh->Props + 
                                mattrans[mesh->MatQuads[i]];
      }
    }

  }

  /* point coordinates */
  outmesh->MeshPointCoords = e_alloc_cont_d_matr(outmesh->NMeshPoints, 3);
  np = 0;
  for(i = 0; i < mesh->NPoints; i++)
  {
    if(mesh->Mark[i] == 0)
    {
      outmesh->MeshPointCoords[np][0] = mesh->Coords[i][0];
      outmesh->MeshPointCoords[np][1] = mesh->Coords[i][1];
      outmesh->MeshPointCoords[np][2] = mesh->Coords[i][2];
      np++;
    }
  }
  outmesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;

  /* save the boundary conditions */
  if(mesh->NBCNodes)
  {
    outmesh->NBCNodes = mesh->NBCNodes;
    outmesh->KeywordSet[E_KW_NUM_NBC_NODES] = TRUE;

    outmesh->TotalNodalDOF = outmesh->NMeshPoints * 3;
    outmesh->NodalDOF = e_alloc_i_vect(outmesh->NMeshPoints);
    for(i = 0; i < outmesh->NMeshPoints; i++)
    {
      outmesh->NodalDOF[i] = 3;
    }
  
    outmesh->BCNodes = e_alloc_i_vect(outmesh->NBCNodes);
    outmesh->BCTypes = e_alloc_frag_i_matr(outmesh->NBCNodes, E_DOF_MAX);
    outmesh->BCDispls = e_alloc_frag_d_matr(outmesh->NBCNodes, E_DOF_MAX);
    outmesh->BCSpringConsts = e_alloc_frag_d_matr(outmesh->NBCNodes, E_DOF_MAX);


    memcpy((void *)(outmesh->BCNodes),
           (void *)(mesh->BCNodes),
           sizeof(int) * mesh->NBCNodes);

/*
    for(i = 0; i < mesh->NBCNodes; i++)
    {
      outmesh->BCNodes[i] = mesh->Translate[mesh->BCNodes[i]-1];
    }
*/

    for(i = 0; i < mesh->NBCNodes; i++)
    {
      node = outmesh->BCNodes[i]-1;
      for(j = 0; j < outmesh->NodalDOF[node]; j++)
      {
        outmesh->BCTypes[i][j] = mesh->BCTypes[i][j];
    
        if(outmesh->BCTypes[i][j] == E_BC_NUM_FIXED)
        {
          outmesh->BCDispls[i][j] = mesh->BCValue[i][j];
        }
        if(outmesh->BCTypes[i][j] == E_BC_NUM_SPRING)
        {
          outmesh->BCSpringConsts[i][j] = mesh->BCValue[i][j];
        }
      }
    }

    outmesh->KeywordSet[E_KW_NUM_BCS] = TRUE;
  }

  /* write-out the mesh */
  e_PutMeshData(outmesh, filename);

}




