/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              e_GetGenData()                                */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include <ctype.h>
#include "e_pltfrm.h"
#include "e_lib.h"
#include "patch.h"

/* e_lib extras */
char *OpenInputFile(char *FileName, char *extension, FILE **fp);
char *OpenOutputFile(char *FileName, char *extension, FILE **fp);
int GetKeyword(char *line, char *word, int *LineNumber, 
               char *FullPathName, FILE *fp);

/* PC compatibility */
#ifdef __E_MSWIN__
#define strcasecmp(s1,s2) strcmpi(s1,s2)
#endif

/* local definitions */

#define PrintError(error)     \
    fprintf(stderr, "\n\nPARSE ERROR: %s!\n\n", error); \
    return 0

#define PrintLineError(error) \
  fprintf(stderr, "\n\nPARSE ERROR at line %d: %s!\n\n", \
          LineNumber, error); \
  return 0

#define PrintPosError(error, pos) \
  fprintf(stderr, "\n\nPARSE ERROR in line %d at position %d: %s!\n\n", \
          LineNumber, pos, error); \
  return 0

/* should be listed in IsAKeyword function */

#define E_KW_GEN_TITLE          "TITLE"
#define E_KW_GEN_NNODES         "NNODES"
#define E_KW_GEN_NSURFACES      "NSURFACES"
#define E_KW_GEN_NCONNECTIONS   "NLINKS"
#define E_KW_GEN_GEN_MATERIAL   "GEN_MATERIAL"
#define E_KW_GEN_NMATERIALS     "NMATERIALS"
#define E_KW_GEN_NBC_NODES      "NBC_NODES"
#define E_KW_GEN_NBC_SIDES      "NBC_SIDES"
#define E_KW_GEN_NODES          "NODE_COORDINATES"

#define E_KW_GEN_SURFACE        "SURFACE"
#define E_KW_GEN_SURFACE_TYPE   "SHAPE"
#define E_KW_GEN_ELEM_TYPE      "ELEMENT_TYPE"
#define E_KW_GEN_DEFINITION     "DEFINITION"
#define E_KW_GEN_NDIV           "N_DIVISIONS"
#define E_KW_GEN_DIV_WEIGHTS    "DIVISION_WEIGHTS"
#define E_KW_GEN_LINK_DIRECTION "LINK_DIRECTION"
#define E_KW_GEN_LINK_DIVISION  "LINK_DIVISION"
#define E_KW_GEN_LINK_TYPE      "LINK_TYPE"
#define E_KW_GEN_CONNECT        "LINK"
#define E_KW_GEN_LOCAL_MATERIAL "MATERIAL"
#define E_KW_GEN_DIAGONAL       "DIAGONAL"
#define E_KW_GEN_END_SECTION    "END"
#define E_KW_GEN_BCS            "BOUNDARY_CONDITIONS"

#define E_GEN                   "gen"


#define E_KW_DEF_LINEAR         "LINEAR"
#define E_KW_DEF_PARAMETRIC     "PARAMETRIC"
#define E_KW_DEF_POINT          "POINT"

#define E_KW_DIAGONAL_1         "1"
#define E_KW_DIAGONAL_2         "2"
#define E_KW_DIAGONAL_SHORTEST  "SHORTEST"
#define E_KW_DIAGONAL_LONGEST   "LONGEST"

static int         LineNumber;                /* current line number */
static char        line[E_LINE_MAX+1];        /* file line */
static char       *FullPathName;
static FILE       *fp;                        /* io-file pointer */
static int         CurPatch;
static int         CurConnection;

/* local functions */
static int GetData(GEN *data);
static int WordMatchesKeyword(char *Keyword, char *word);
static int IsAKeyword(char *word);
static int GetString(char **String);
static int GetNumber(int *Nodes);
static int Get2Number(int *N1, int *N2);
static int Get3Number(int *N1, int *N2, int *N3);
static int GetPointCoords(GEN *data);
static int GetNodeRef(char *line, char *pline,
                      int CurPatch, int CurConnection,
                      GEN *data, NODE *Node);
#if 0
static int GetIntegerSeries(char *pline, int n, int *list);
#endif
static int GetDoubleSeries(char *pline, int n, double *list);

static int GetPatches(GEN *data);
static int GetConnections(GEN *data);
static int GetBoundaryConditions(GEN *data);
static int ReadBndCond(char *pline, int index, 
                       int **BCType, double **BCValue);

static void DeallocateAllMemory(GEN *data);

extern char *OpenInputFile();
extern int   GetKeyword();

/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetGenData: To read in mesh generation data. */
/* The return value is 1 if no input errors occur, else 0. */

int e_GetGenData(GEN **data, char *InputFile)
{
  if((*data = (GEN *)e_calloc(1, sizeof(GEN))) == NULL)
    return 0;

  if((FullPathName = OpenInputFile(InputFile, E_GEN, &(fp))) == NULL)
  {
    e_free(FullPathName);
    e_free(*data);
    return 0;
  }

  if(GetData(*data))
  {
    e_free(FullPathName);
    fclose(fp);
    return 1;
  }
  else
  {
    DeallocateAllMemory(*data);
    *data = NULL;
    e_free(FullPathName);
    fclose(fp);
    return 0;
  }
}

/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully or the end of file
   is encountered, else 0.                                                    */

static int GetData(GEN *data)
{
  char word[E_LINE_MAX + 1];      /* potential keyword */
  int  ReadNextLine = FALSE;

  /* get the title */
  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("cannot read title keyword");
  }
  else if(   !WordMatchesKeyword(E_KW_GEN_TITLE,word)
          || !GetString(&(data->Title)) )
    return 0;

  /* get the number of nodes */
  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("cannot read number of nodes keyword");
  }
  else if(   !WordMatchesKeyword(E_KW_GEN_NNODES,word)
          || !GetNumber(&(data->NNodes)) )
    return 0;

  if(data->NNodes <= 0)
  {
    PrintLineError("invalid number of nodes");
  }

  /* number of patches */
  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("cannot read number of patches keyword");
  }
  else if(   !WordMatchesKeyword(E_KW_GEN_NSURFACES, word)
          || !GetNumber(&(data->NSurfaces)) )
    return 0;

  /* number of connections */
  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("cannot read number of connections keyword");
  }
  else if(   !WordMatchesKeyword(E_KW_GEN_NCONNECTIONS, word)
          || !GetNumber(&(data->NConnections)) )
    return 0;

  if(   data->NSurfaces == 0
     && data->NConnections == 0 )
  {
    PrintError("total number of element definitions is ZERO");
  }

  /* create material information, optional */
  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("premature end of file");
  }
  ReadNextLine = FALSE;

  if(strcmp(E_KW_GEN_GEN_MATERIAL, word) == 0)
  {
    data->DefMat = TRUE;
    ReadNextLine = TRUE;
  }
  else
  {
    data->DefMat = FALSE;
  }

  /* number of material definitions, optional */
  if(   ReadNextLine
     && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("premature end of file");
  }
  ReadNextLine = FALSE;

  if(strcmp(E_KW_GEN_NMATERIALS, word) == 0)
  {
    if(data->DefMat == FALSE)
    {
      PrintError("NMATERIALS is defined, but GEN_MATERIAL is not");
    }
    
    if(!GetNumber(&(data->NMaterials)))
      return 0;
    ReadNextLine = TRUE;
  }

  /* number of boundary condition nodes */
  if(   ReadNextLine
     && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("premature end of file");
  }
  ReadNextLine = FALSE;

  if(strcmp(E_KW_GEN_NBC_NODES, word) == 0)
  {
    if(!GetNumber(&(data->NBCNodes)))
      return 0;
    ReadNextLine = TRUE;
  }

  /* number of boundary condition sides */
  if(   ReadNextLine
     && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("premature end of file");
  }
  ReadNextLine = FALSE;

  if(strcmp(E_KW_GEN_NBC_SIDES, word) == 0)
  {
    if(!GetNumber(&(data->NBCSides)))
      return 0;
    ReadNextLine = TRUE;
  }

  /* nodes with coordinates */
  if(   ReadNextLine
     && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("Premature end of file");
  }
  ReadNextLine = FALSE;

  if(   !WordMatchesKeyword(E_KW_GEN_NODES, word)
     || !GetPointCoords(data) )
    return 0;

  /* get patches */
  if(!GetPatches(data))
    return 0;

  /* get connections */
  if(!GetConnections(data))
    return 0;

  /* material name definitions, NOT IMPLEMENTED */
  if(data->NMaterials != 0)
  {
  }

  /* get boundary conditions */
  if(!GetBoundaryConditions(data))
    return 0;


  return 1;
}

/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */

static int WordMatchesKeyword(char *Keyword, char *word)
{
  if(strcmp(Keyword, word) != 0)
  {
    if(IsAKeyword(word))
    {
      PrintLineError("keyword out of sequence");
    }
    else
    {
      PrintLineError("undefined keyword");
    }
  }
  else
    return 1;
}

/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */

static int IsAKeyword(char *word)
{
  if(   strcmp(E_KW_GEN_TITLE, word)          == 0
     || strcmp(E_KW_GEN_NNODES, word)         == 0
     || strcmp(E_KW_GEN_NSURFACES, word)      == 0
     || strcmp(E_KW_GEN_NCONNECTIONS, word)   == 0
     || strcmp(E_KW_GEN_GEN_MATERIAL, word)   == 0
     || strcmp(E_KW_GEN_NMATERIALS, word)     == 0
     || strcmp(E_KW_GEN_NBC_NODES, word)      == 0
     || strcmp(E_KW_GEN_NBC_SIDES, word)      == 0
     || strcmp(E_KW_GEN_NODES, word)          == 0
     || strcmp(E_KW_GEN_SURFACE, word)        == 0
     || strcmp(E_KW_GEN_SURFACE_TYPE, word)   == 0
     || strcmp(E_KW_GEN_ELEM_TYPE, word)      == 0
     || strcmp(E_KW_GEN_DEFINITION, word)     == 0
     || strcmp(E_KW_GEN_NDIV, word)           == 0
     || strcmp(E_KW_GEN_DIV_WEIGHTS, word)    == 0
     || strcmp(E_KW_GEN_LINK_DIRECTION, word) == 0
     || strcmp(E_KW_GEN_LINK_DIVISION, word)  == 0
     || strcmp(E_KW_GEN_LINK_TYPE, word)      == 0
     || strcmp(E_KW_GEN_CONNECT, word)        == 0
     || strcmp(E_KW_GEN_LOCAL_MATERIAL, word) == 0
     || strcmp(E_KW_GEN_DIAGONAL, word)       == 0
     || strcmp(E_KW_GEN_BCS, word)            == 0
     || strcmp(E_KW_GEN_END_SECTION, word)    == 0 )
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */

static int GetString(char **String)
{
  int   StrLength;             /* string length */

  char *pFirstQuote;           /* pointer to first '"' in 'line' */
  char *pLastQuote;            /* point to last '"' in 'line' */


  if(   (pFirstQuote = strchr(line,'"')) != NULL
     && (pLastQuote = strrchr(line,'"')) != pFirstQuote)
  {
    StrLength = (int)(pLastQuote - pFirstQuote) - 1;

    if((*String = e_alloc_c_vect(StrLength + 1)) == NULL)
    {
      PrintError("cannot allocate memory");
    }

    strncpy(*String, pFirstQuote + 1, StrLength);
    (*String)[StrLength] = '\0';
  }
  else
  {
    PrintLineError("cannot read string");
  }

  return 1;
}

/* ////////////////////////////////// */

static int GetNumber(int *Nodes)
{
  if(sscanf(line,"%*s %d",Nodes) == 1)
  {
    if(*Nodes < 0)
    {
      PrintLineError("ill-defined number");
    }
  }
  else
  {
    PrintLineError("cannot read number");
  }

  return 1;
}

/* ////////////////////////////////// */

static int Get2Number(int *N1, int *N2)
{
  if(sscanf(line,"%*s %d %d",N1, N2) == 2)
  {
    if((*N1 < 0) || (*N2 < 0))
    {
      PrintLineError("ill-defined number(s)");
    }
  }
  else
  {
    PrintLineError("cannot read number(s)");
  }

  return 1;
}

/* ////////////////////////////////// */

static int Get3Number(int *N1, int *N2, int *N3)
{
  if(sscanf(line,"%*s %d %d %d", N1, N2, N3) == 3)
  {
    if((*N1 < 0) || (*N2 < 0) || (*N3 < 0))
    {
      PrintLineError("ill-defined number(s)");
    }
  }
  else
  {
    PrintLineError("cannot read number(s)");
  }

  return 1;
}


/* ////////////////////////////////// */

static int GetPointCoords(GEN *data)
{
  int i;
  int CoordsIndex;

  if((data->Coords = e_alloc_cont_d_matr(data->NNodes, E_NDIMS)) == NULL)
    return 0;

  for(i = 0; i < data->NNodes; i++)
  {
    if(e_GetNextLine(line, &LineNumber, FullPathName, fp) == 0)
    {
      return 0;
    }

    if(sscanf(line,"%d %lf %lf %lf",&CoordsIndex,
                                    &(data->Coords[i][0]),
                                    &(data->Coords[i][1]),
                                    &(data->Coords[i][2])  ) == 4)
    {
      if(CoordsIndex != i + 1)
      {
        PrintLineError("coordinate index out of sequence");
      }
    }
    else
    {
      PrintLineError("cannot read node coordinates");
    }
  }

  return 1;
}

/* ////////////////////////////////// */

static int GetNodeRef(char *line, char *pline,
                      int CurPatch, int CurConnection,
                      GEN *data, NODE *Node)
{
  int   i, patchID, linkID, cl;
  char  check[E_LINE_MAX+1];

  pline = GetStrArg(pline, "%s", check);
  if(pline != NULL)
  {
    /* reference to a surface */
    /* #################### */
    if(   strcasecmp("s", check) == 0
       || strcasecmp("surface", check) == 0 )
    {
      Node->Node = SURFACE_NODE;
      /* read surface ID */
      pline = GetStrArg(pline, "%d", &(Node->ElemID) );
      if(pline == NULL)
      {
        PrintLineError("cannot read referenced surface ID");
      }

      /* validate surface ID */
      patchID = Node->ElemID-1;
      if(patchID < 0)
      {
        PrintPosError("surface ID smaller than one", pline-line);
      }

      if(patchID == CurPatch-1)
      {
        PrintPosError("self-reference", pline-line);
      }

      if(patchID > CurPatch-1)
      {
        PrintPosError("node references higher surface ID than itself", pline-line);
      }
  
      /* read 'x' position */
      pline = GetStrArg(pline, "%s", check);
      if(pline == NULL)
      {
        PrintLineError("could not read 'x' position");
      }

      if(   strcasecmp(check, "m") == 0
         || strcasecmp(check, "max") == 0)
      {
        Node->Position[0] = data->Patches[patchID].XDivision;
      }
      else
      {
        if(sscanf(check, "%d", &(Node->Position[0]) ) != 1)
        {
          PrintPosError("not a number", pline-line);
        }
      }

      /* validate 'x' position */
      if(   (Node->Position[0] < 0) 
         || (Node->Position[0] > data->Patches[patchID].XDivision))
      {
        PrintPosError("invalid 'x' position", pline-line);
      }
    
      /* read 'y' position */
      pline = GetStrArg(pline, "%s", check);
      if(pline == NULL)
      {
        PrintLineError("could not read 'x' position");
      }

      if(   strcasecmp(check, "m") == 0
         || strcasecmp(check, "max") == 0)
      {
        if(data->Patches[patchID].Shape == 4)
        {
          Node->Position[1] = data->Patches[patchID].YDivision;
        }
        else if(data->Patches[patchID].Shape == 3)
        {
          Node->Position[1] = 
            data->Patches[patchID].XDivision - Node->Position[0];
        }
      }
      else
      {
        if(sscanf(check, "%d", &(Node->Position[1]) ) != 1)
        {
          PrintPosError("not a number", pline-line);
        }
      }

      /* validate 'y' position */
      if(data->Patches[patchID].Shape == 4)
      {
        if(   (Node->Position[1] < 0) 
           || (Node->Position[1] > data->Patches[patchID].YDivision))
        {
            PrintPosError("invalid 'y' position", pline-line);
        }
      }
      else if(data->Patches[patchID].Shape == 3)
      {
        if(   (Node->Position[1] < 0) 
           || (Node->Position[1] > data->Patches[patchID].XDivision -
                                   Node->Position[0]) )
        {
            PrintPosError("invalid 'y' position", pline-line);
        }
      }
    }
    /* reference to a link */
    /* ################### */
    else if(   strcasecmp("l", check) == 0
            || strcasecmp("link", check) == 0 )
    {
      Node->Node = LINK_NODE;
      /* read surface ID */
      pline = GetStrArg(pline, "%d", &(Node->ElemID) );
      if(pline == NULL)
      {
        PrintLineError("cannot read referenced link ID");
      }

      /* validate link ID */
      linkID = Node->ElemID-1;
      if(linkID < 0)
      {
        PrintPosError("link ID smaller than one", pline-line);
      }

      if(linkID == CurConnection-1)
      {
        PrintPosError("self-reference", pline-line);
      }

      if(linkID > CurConnection-1)
      {
        PrintPosError("node references higher link ID than itself", 
                      pline-line);
      }
  
      /* read 'x' position */
      pline = GetStrArg(pline, "%d", &(Node->Position[0]) );
      if(pline == NULL)
      {
        PrintLineError("could not read 'x' position");
      }

      /* validate 'x' position */
      if(   (Node->Position[0] < 0) 
         || (Node->Position[0] > data->Connections[linkID].Division))
      {
        PrintPosError("invalid 'x' position", pline-line);
      }
    
    }
    /* only node number */
    /* ################ */
    else
    {
      /* validate, that it is an integer number ONLY */
      cl = strlen(check);
      for(i = 0; i < cl; i++)
      {
        if(!isdigit((int)check[i]))
          break;
      }

      if(i == cl)
      {
        sscanf(check, "%d", &(Node->Node));
        if(   (Node->Node < 1)
           || (Node->Node > data->NNodes) )
        {
          PrintPosError("invalid node number", pline-line);
        }
      }
      /* contains invalid character */
      else
      {
        PrintPosError("invalid data format", pline-line);
      }
    }
  }
  else
  {
    PrintLineError("cannot read node");
  }

  return(pline-line);

}

/* ////////////////////////////////// */

#if 0
static int GetIntegerSeries(char *pline, int n, int *list)
{
  char  c;
  int   pos, i, in, s;
  int   dn;

  pos = 0;

  while(1)
  {
    s = sscanf(pline, "%d (%d)", &in, &dn);
    /* short form for a series */
    if(s == 2)
    {
      for(i = 0; i < in; i++)
      {
        if(pos+i == n)
        {
          PrintPosError("too many number", pline-line);
        }
        list[pos+i] = dn;
      }
      pos += in;
      /* go after the read in data */
      while(*pline != ')')
        pline++;
      pline++;
    }
    /* anything else */
    else
    {
      s = sscanf(pline, "%d%c", &dn, &c);
      if(s == 2)
      {
        if(!isspace((int)c))
        {
          PrintPosError("not a number", pline-line);
        }
        if(pos == n)
        {
          PrintLineError("too many number");
        }
        list[pos] = dn;
        pos++;
        /* skip spaces in front */
        while(isspace((int)(*pline)))
          pline++;
        /* skip the number */
        while(!isspace((int)(*pline)))
          pline++;
      }
      else if(s == 1)
      {
        if(pos == n)
        {
          PrintLineError("too many number");
        }
        list[pos] = dn;
        pos++;
      }
      else
      {
        /* remove white spaces */
        while(*pline != '\0' && isspace((int)(*pline)))
          pline++;

        if(*pline != '\0')
        {
          PrintPosError("invalid data", pline-line);
        }
        if(pos < n)
        {
          PrintLineError("not enough number");
        }
        return 1;
      }
    }
  }
}
#endif


static int GetDoubleSeries(char *pline, int n, double *list)
{
  char    c;
  int     pos, i, in, s;
  double  dn;

  pos = 0;

  while(1)
  {
    s = sscanf(pline, "%d (%lf)", &in, &dn);
    /* short form for a series */
    if(s == 2)
    {
      for(i = 0; i < in; i++)
      {
        if(pos+i == n)
        {
          PrintPosError("too many number", pline-line);
        }
        list[pos+i] = dn;
      }
      pos += in;
      /* go after the read in data */
      while(*pline != ')')
        pline++;
      pline++;
    }
    /* anything else */
    else
    {
      s = sscanf(pline, "%lf%c", &dn, &c);
      if(s == 2)
      {
        if(!isspace((int)c))
        {
          PrintPosError("not a number", pline-line);
        }
        if(pos == n)
        {
          PrintLineError("too many number");
        }
        list[pos] = dn;
        pos++;
        /* skip spaces in front */
        while(isspace((int)(*pline)))
          pline++;
        /* skip the number */
        while(!isspace((int)(*pline)))
          pline++;
      }
      else if(s == 1)
      {
        if(pos == n)
        {
          PrintLineError("too many number");
        }
        list[pos] = dn;
        pos++;
      }
      else
      {
        /* remove white spaces */
        while(*pline != '\0' && isspace((int)(*pline)))
          pline++;

        if(*pline != '\0')
        {
          PrintPosError("invalid data", pline-line);
        }
        if(pos < n)
        {
          PrintLineError("not enough number");
        }
        return 1;
      }
    }
  }
}


void GetQuadCornerPos(SURFACE *data, int corner, int *xe, int *ye)
{
  switch(corner)
  {
    case 0:
      *xe = 0;
      *ye = 0;
      break;
    case 1:
      *xe = data->XDivision;
      *ye = 0;
      break;
    case 2:
      *xe = data->XDivision;
      *ye = data->YDivision;
      break;
    case 3:
      *xe = 0;
      *ye = data->YDivision;
      break;
    default:
      e_GeneralError("invalid corner position. This is an INTERNAL ERROR !!!");
      break;
  }
}


/* ##################################################################### */
/*                                Surfaces                               */
/* ##################################################################### */

static int GetPatches(GEN *data)
{
  int   i_block, i_node, i_div, i_side, i_type;
  int   block_num, n, inc, j, xe, ye;
  int   sum, a_div, local_mat;
  int   ReadNextLine;
  char  word[E_LINE_MAX + 1];      /* potential keyword */
  char *pline;


  if(data->NSurfaces == 0)
    return 1;

  if((data->Patches = (SURFACE *)e_calloc(data->NSurfaces, 
                                          sizeof(SURFACE))) == NULL)
    return 0;

  /* for all specified patches */
  for(i_block = 0; i_block < data->NSurfaces; i_block++)
  {
    /* surface ID */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read surface ID keyword");
    }
    else if(   !WordMatchesKeyword(E_KW_GEN_SURFACE, word)
            || !GetNumber(&block_num) )
      return 0;

    if(block_num != i_block+1)
    {
      PrintLineError("surface ID is out of sequence");
    }

    CurPatch = block_num;

    /* surface shape - triangle, cubic 4 node patch */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read surface shape keyword");
    }
    else if(!WordMatchesKeyword(E_KW_GEN_SURFACE_TYPE, word))
      return 0;
    
    if(sscanf(line, "%*s %s", word) == 1)
    {
      if(strcmp("TRIANG", word) == 0)
      {
        data->Patches[i_block].Shape = 3;
      }
      else if(strcmp("QUAD", word) == 0)
      {
        data->Patches[i_block].Shape = 4;
      }
      else if(   (sscanf(word, "%d", &n) == 1)
              && (   (n == 3)
                  || (n == 4) ) )
      {
        data->Patches[i_block].Shape = n;
      }
      else
      {
        PrintLineError("invalid surface shape");
      }
    }
    else
    {
      PrintLineError("cannot read surface shape");
    }

    /* type of element to generate */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read element type keyword");
    }
    else if(!WordMatchesKeyword(E_KW_GEN_ELEM_TYPE, word))
      return 0;

    if(sscanf(line, "%*s %s", word) == 1)
    {
      if(strcmp("TRIANG1", word) == 0)
      {
        data->Patches[i_block].ElementType = E_ELEM_TYPE_TRIANG1;
      }
      else if(strcmp("TRIANG3", word) == 0)
      {
        data->Patches[i_block].ElementType = E_ELEM_TYPE_TRIANG3;
      }
      else if(strcmp("TRIANG4", word) == 0)
      {
        data->Patches[i_block].ElementType = E_ELEM_TYPE_TRIANG4;
      }
      else if(strcmp("TRIANG5", word) == 0)
      {
        data->Patches[i_block].ElementType = E_ELEM_TYPE_TRIANG5;
      }
      else if(strcmp("QUAD1", word) == 0)
      {
        data->Patches[i_block].ElementType = E_ELEM_TYPE_QUAD1;
      }
      else if(   (sscanf(word, "%d", &n) == 1)
              && (   (n == E_ELEM_TYPE_TRIANG1)
                  || (n == E_ELEM_TYPE_TRIANG3)
                  || (n == E_ELEM_TYPE_TRIANG4)
                  || (n == E_ELEM_TYPE_TRIANG5)
                  || (n == E_ELEM_TYPE_QUAD1) ) )
      {
        data->Patches[i_block].ElementType = n;
      }
      else
      {
        PrintLineError("invalid element type");
      }
    }
    else
    {
      PrintLineError("cannot read element type");
    }

    if(   data->Patches[i_block].Shape == 3
       && data->Patches[i_block].ElementType == E_ELEM_TYPE_QUAD1)
    {
      PrintLineError("cannot generate quad elements in triangle surface");
    }

    if(   data->Patches[i_block].ElementType != E_ELEM_TYPE_QUAD1
       && data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG1
       && data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG3
       && data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG4
       && data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG5)
    {
      PrintLineError("invalid element type");
    }

    /* quad surface */
    if(data->Patches[i_block].Shape == 4)
    {
      /* get the definition type */
      if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
      {
        PrintError("cannot read definition type keyword");
      }
      else if(!WordMatchesKeyword(E_KW_GEN_DEFINITION, word))
        return 0;

      if(sscanf(line, "%*s %s", word) == 1)
      {
        if(strcasecmp(E_KW_DEF_LINEAR, word) == 0)
        {
          data->Patches[i_block].Definition = DEF_LINEAR;
        }
        else if(strcasecmp(E_KW_DEF_PARAMETRIC, word) == 0)
        {
          data->Patches[i_block].Definition = DEF_PARAMETRIC;
        }
        else if(strcasecmp(E_KW_DEF_POINT, word) == 0)
        {
          data->Patches[i_block].Definition = DEF_POINT;
        }
        else if(   (sscanf(word, "%d", &n) == 1)
                && (   (n == DEF_LINEAR)
                    || (n == DEF_PARAMETRIC)
                    || (n == DEF_POINT) ) )
        {
          data->Patches[i_block].Definition = n;
        }
        else
        {
          PrintLineError("invalid definition type");
        }
      }
      else
      {
        PrintLineError("cannot read definition type");
      }

      /* linear */
      if(data->Patches[i_block].Definition == DEF_LINEAR)
      {
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
        {
          pline = line;
          for(i_node = 0; i_node < 4; i_node++)
          {
            inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                             data, &(data->Patches[i_block].Nodes[i_node]) );
            if(inc)
              pline = line + inc;
            else
              return 0;            
          }
        }
        else
        {
          PrintError("cannot read nodes for linear definition");
        }
      }
      /* parametric */
      else if(data->Patches[i_block].Definition == DEF_PARAMETRIC)
      {
        /* four lines of definition */
        for(i_node = 0; i_node < 4; i_node++)
        {
          if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
          {
            pline = line;
            inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                             data, &(data->Patches[i_block].Nodes[i_node]) );
            if(inc)
              pline = line + inc;
            else
              return 0;

            if(GetDoubleSeries(pline, 9, data->Patches[i_block].ParamDef[i_node]) == 0)
              return 0;
          }
          else
          {
            PrintError("cannot read data for algebraic parameter definition");
          }
        }
      }
      /* point definition */
      else if(data->Patches[i_block].Definition == DEF_POINT)
      {
        for(i_node = 0; i_node < 4; i_node++)
        {
          if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
          {
            pline = line;
            for(j = 0; j < 4; j++)
            {
              inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                               data, &(data->Patches[i_block].Nodes[i_node*4+j]) );
              if(inc)
                pline = line + inc;
              else
                return 0;
            }
          }
          else
          {
            PrintError("cannot read data for algebraic point definition");
          }
        }
      }
      else
      {
        PrintError("invalid definitions type");
      }
    }
    /* triangle surface */
    else if(data->Patches[i_block].Shape == 3)
    {
      /* get the definition type */
      if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
      {
        PrintError("cannot read definition type keyword");
      }
      else if(!WordMatchesKeyword(E_KW_GEN_DEFINITION, word))
        return 0;

      if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
      {
        pline = line;
        for(i_node = 0; i_node < 3; i_node++)
        {
          inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                           data, &(data->Patches[i_block].Nodes[i_node]) );
          if(inc)
            pline = line + inc;
          else
            return 0;            
        }
      }
      else
      {
        PrintError("cannot read nodes for triangle surface");
      }
    }
    else
    {
      PrintError("invalid surface shape. This is an INTERNAL ERROR!");
    }

    /* number of divisions */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read number of divisions keyword");
    }
    else if(!WordMatchesKeyword(E_KW_GEN_NDIV, word))
      return 0;

    if(data->Patches[i_block].Shape == 3)
    {
      if(!GetNumber(&(data->Patches[i_block].XDivision)))
        return 0;
      if(data->Patches[i_block].XDivision < 1)
      {
        PrintLineError("invalid division, min. value 1");
      }
      if(   ((data->Patches[i_block].XDivWeight = 
              e_alloc_d_vect(data->Patches[i_block].XDivision) ) == NULL)
         || ((data->Patches[i_block].YDivWeight = 
              e_alloc_d_vect(data->Patches[i_block].XDivision) ) == NULL)
         || ((data->Patches[i_block].TDivWeight = 
              e_alloc_d_vect(data->Patches[i_block].XDivision) ) == NULL) )
        return 0;
    }
    else if(data->Patches[i_block].Shape == 4)
    {
      if(!Get2Number(&(data->Patches[i_block].XDivision),
                     &(data->Patches[i_block].YDivision)  ))
        return 0;
      if(   (data->Patches[i_block].XDivision < 1)
         || (data->Patches[i_block].YDivision < 1) )
      {
        PrintLineError("invalid division, min. value 1");
      }

      if(   ((data->Patches[i_block].XDivWeight = 
              e_alloc_d_vect(data->Patches[i_block].XDivision)) == NULL)
         || ((data->Patches[i_block].YDivWeight = 
              e_alloc_d_vect(data->Patches[i_block].YDivision)) == NULL) )
        return 0;
    }

    /* division weights */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(strcmp(E_KW_GEN_DIV_WEIGHTS, word) == 0)
    {
      if(data->Patches[i_block].Shape == 4)
      {
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }

        if(GetDoubleSeries(line, data->Patches[i_block].XDivision,
                           data->Patches[i_block].XDivWeight) == 0)
          return 0;

        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }

        if(GetDoubleSeries(line, data->Patches[i_block].YDivision,
                           data->Patches[i_block].YDivWeight) == 0)
          return 0;
      }
      else if(data->Patches[i_block].Shape == 3)
      {
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }
        if(GetDoubleSeries(line, data->Patches[i_block].XDivision,
                           data->Patches[i_block].XDivWeight) == 0)
        {
          return 0;
        }
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }
        if(GetDoubleSeries(line, data->Patches[i_block].XDivision,
                           data->Patches[i_block].YDivWeight) == 0)
        {
          return 0;
        }
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }
        if(GetDoubleSeries(line, data->Patches[i_block].XDivision,
                           data->Patches[i_block].TDivWeight) == 0)
        {
          return 0;
        }
      }

      ReadNextLine = TRUE;
    }
    else /* default division weight values */
    {
      if(data->Patches[i_block].Shape == 4)
      {
        for(i_div = 0; i_div < data->Patches[i_block].XDivision; i_div++)
          data->Patches[i_block].XDivWeight[i_div] = 1.0;
        for(i_div = 0; i_div < data->Patches[i_block].YDivision; i_div++)
          data->Patches[i_block].YDivWeight[i_div] = 1.0;
      }
      else if(data->Patches[i_block].Shape == 3)
      {
        for(i_div = 0; i_div < data->Patches[i_block].XDivision; i_div++)
          data->Patches[i_block].XDivWeight[i_div] = 1.0;
      }
    }

    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    /* links on the side and internally */
    if(strcmp(E_KW_GEN_LINK_DIRECTION, word) == 0)
    {
      if(data->Patches[i_block].Shape == 4)
      {
        if(!Get2Number(&(data->Patches[i_block].LinkSides[0]),
                       &(data->Patches[i_block].LinkSides[1]) ))
          return 0;

        data->Patches[i_block].LinkSides[2] = 0;
      }
      else if(data->Patches[i_block].Shape == 3)
      {
        if(!Get3Number(&(data->Patches[i_block].LinkSides[0]),
                       &(data->Patches[i_block].LinkSides[1]),
                       &(data->Patches[i_block].LinkSides[2]) ))
          return 0;
      }
      else
      {
        PrintError("invalid surface shape. This is an INTERNAL ERROR!");
      }
    
      ReadNextLine = TRUE;
    }
    else /* default, no links at all */
    {
      data->Patches[i_block].LinkSides[0] = 0;
      data->Patches[i_block].LinkSides[1] = 0;
      data->Patches[i_block].LinkSides[2] = 0;
    }

    if(   (   (data->Patches[i_block].LinkSides[0] != 0)
           && (data->Patches[i_block].LinkSides[0] != 1) )
       || (   (data->Patches[i_block].LinkSides[1] != 0)
           && (data->Patches[i_block].LinkSides[1] != 1) )
       || (   (data->Patches[i_block].LinkSides[2] != 0)
           && (data->Patches[i_block].LinkSides[2] != 1) ) )
    {
      PrintLineError("invalid data type");
    }

    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;
    
    if(   data->Patches[i_block].LinkSides[0]
       || data->Patches[i_block].LinkSides[1]
       || data->Patches[i_block].LinkSides[2] )
    {
      ReadNextLine = TRUE;
      /* link divisions */
      if(!WordMatchesKeyword(E_KW_GEN_LINK_DIVISION, word))
        return 0;

      for(i_side = 0; i_side < 3; i_side++)
      {
        /* if the side is defined */
        data->Patches[i_block].LinkDiv[i_side] = NULL;
        if(data->Patches[i_block].LinkSides[i_side])
        {
          if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
          {
            pline = line;
            data->Patches[i_block].LinkSides[i_side] = 0;

            sum = 0;
            while((pline = GetStrArg(pline, "%d", &a_div)) != NULL)
            {
              data->Patches[i_block].LinkDiv[i_side] = 
                e_incr_i_vect(data->Patches[i_block].LinkSides[i_side],
                              data->Patches[i_block].LinkDiv[i_side], 10);
              sum += a_div;
              data->Patches[i_block].LinkDiv[i_side][data->Patches[i_block].LinkSides[i_side]] = sum;
              data->Patches[i_block].LinkSides[i_side]++;

              if(a_div < 0)
              {
                PrintLineError("negativ division");
              }
              if((data->Patches[i_block].LinkSides[i_side] != 1) && (a_div == 0))
              {
                PrintLineError("duplicate division");
              }
              
            }
            /* check amount */
            if(data->Patches[i_block].Shape == 4)
            {
              if((i_side == 1) && (sum > data->Patches[i_block].XDivision))
              {
                PrintLineError("invalid position for links parallel with the second side");
              }
              else if((i_side == 0) && (sum > data->Patches[i_block].YDivision))
              {
                PrintLineError("invalid position for links parallel with the first side");
              }
            }
            else if(data->Patches[i_block].Shape == 3)
            {
              if(sum >= data->Patches[i_block].XDivision)
              {
                PrintLineError("invalid division sum");
              }
            }
          }
          else
          {
            PrintError("cannot read link divisions");
          }
        }
      }

      /* link types */
      if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
      {
        PrintError("cannot read link type keyword");
      }
      else if (!WordMatchesKeyword(E_KW_GEN_LINK_TYPE, word))
        return 0;
      else
      {
        for(i_side = 0; i_side < 3; i_side++)
        {
          data->Patches[i_block].LinkType[i_side] = NULL;

          if(data->Patches[i_block].LinkSides[i_side])
          {
            if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
            {
              data->Patches[i_block].LinkType[i_side] = 
                e_alloc_i_vect(data->Patches[i_block].LinkSides[i_side]);

              pline = line;
              for(i_type = 0; i_type < data->Patches[i_block].LinkSides[i_side]; i_type++)
              {
                pline = GetStrArg(pline, "%d", 
                                  &(data->Patches[i_block].LinkType[i_side][i_type]));
                if(pline == NULL)
                {
                  PrintLineError("data is missing");
                }
                if(data->Patches[i_block].LinkType[i_side][i_type] < 0)
                {
                  PrintLineError("ill-defined link type");
                }
              }
            }
            else
            {
              PrintError("cannot read link types");
            }
          }
        }
      }
    }

    /* local material definition */
    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(strcmp(E_KW_GEN_LOCAL_MATERIAL, word) == 0)
    {
      if(sscanf(line, "%*s %d", &local_mat) != 1)
      {
        PrintLineError("cannot read material number");
      }

      if(   local_mat < 1
         || local_mat > data->NMaterials)
      {
        PrintLineError("invalid local material");
      }
      
      data->Patches[i_block].Material = local_mat;

      ReadNextLine = TRUE;
    }

    /* EXPERT options */
    /* These options are for those, who know what they are doing */
    /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    /* triangle generation with diagonal control */
    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(strcmp(E_KW_GEN_DIAGONAL, word) == 0)
    {
      if(data->Patches[i_block].Shape != 4)
      {
        PrintLineError("DIAGONAL can only be defined for quad surfaces");
      }
      if(   (data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG1)
         && (data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG3)
         && (data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG4)
         && (data->Patches[i_block].ElementType != E_ELEM_TYPE_TRIANG5) )
      {
        PrintLineError("DIAGONAL can only be defined for triangle element generation");
      }

      if(sscanf(line, "%*s %s", word) == 1)
      {
        if(strcasecmp(E_KW_DIAGONAL_1, word) == 0)
        {
          data->Patches[i_block].Diagonal = E_DIAGONAL_1;
        }
        else if(strcasecmp(E_KW_DIAGONAL_2, word) == 0)
        {
          data->Patches[i_block].Diagonal = E_DIAGONAL_2;
        }
        else if(strcasecmp(E_KW_DIAGONAL_SHORTEST, word) == 0)
        {
          data->Patches[i_block].Diagonal = E_DIAGONAL_SHORTEST;
        }
        else if(strcasecmp(E_KW_DIAGONAL_LONGEST, word) == 0)
        {
          data->Patches[i_block].Diagonal = E_DIAGONAL_LONGEST;
        }
        else if(   (sscanf(word, "%d", &n) == 1)
                && (   (n == E_DIAGONAL_1)
                    || (n == E_DIAGONAL_2)
                    || (n == E_DIAGONAL_SHORTEST) 
                    || (n == E_DIAGONAL_LONGEST) ) )
        {
          data->Patches[i_block].Diagonal = n;
        }
        else
        {
          PrintLineError("invalid diagonal type");
        }
      }
      else
      {
        PrintLineError("cannot read diagonal type");
      }
      
      ReadNextLine = TRUE;

    }
    else
    {
      data->Patches[i_block].Diagonal = E_DIAGONAL_SHORTEST;
    }

    /* END keyword */
    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(!WordMatchesKeyword(E_KW_GEN_END_SECTION, word))
      return 0;

    /* register referenced corner nodes, if linear def. */
    if(data->Patches[i_block].Shape == 4)
    {
      data->Patches[i_block].RegNodes = 
        (NODE **)e_calloc(data->Patches[i_block].YDivision+1, 
                          sizeof(NODE *) );
    
      for(j = 0; j < data->Patches[i_block].YDivision+1; j++)
      {
        data->Patches[i_block].RegNodes[j] =
          (NODE *)e_calloc(data->Patches[i_block].XDivision+1, 
                           sizeof(NODE) );
      }

      if(data->Patches[i_block].Definition == 1)
      {
        for(i_node = 0; i_node < 4; i_node++)
        {
          GetQuadCornerPos(&(data->Patches[i_block]), i_node, &xe, &ye);

          if(data->Patches[i_block].Nodes[i_node].Node < 1)
          {
            data->Patches[i_block].RegNodes[ye][xe].Node = 
              data->Patches[i_block].Nodes[i_node].Node;
            data->Patches[i_block].RegNodes[ye][xe].ElemID =
              data->Patches[i_block].Nodes[i_node].ElemID;
            data->Patches[i_block].RegNodes[ye][xe].Position[0] =
              data->Patches[i_block].Nodes[i_node].Position[0];
          
            if(data->Patches[i_block].Nodes[i_node].Node == SURFACE_NODE)
            {
              data->Patches[i_block].RegNodes[ye][xe].Position[1] =
                data->Patches[i_block].Nodes[i_node].Position[1];
            }
          }
        }
      }
    }
    else if(data->Patches[i_block].Shape == 3)
    {
      data->Patches[i_block].RegNodes = 
        (NODE **)e_calloc(data->Patches[i_block].XDivision+1, 
                          sizeof(NODE *) );
    
      for(j = 0; j < data->Patches[i_block].XDivision+1; j++)
      {
        data->Patches[i_block].RegNodes[j] =
          (NODE *)e_calloc(data->Patches[i_block].XDivision+1, 
                           sizeof(NODE) );
      }

    }
    
    /* check whether surface is attached to another one, NOT IMPLEMENTED */

  }   /* for(i_block = 0; i_block < NBlocks; i_block++) */

  CurPatch++;

  return 1;

}

/* ////////////////////////////////// */

static int GetConnections(GEN *data)
{
  int    i_con, connect_num, i_node, i_div, inc;
  int    n, ReadNextLine, local_mat;
  char   word[E_LINE_MAX + 1];      /* potential keyword */
  char   check[E_LINE_MAX + 1];
  char  *pline;

  if(data->NConnections == 0)
    return 1;

  if( (data->Connections = 
         (CONNECT *)e_calloc(data->NConnections, sizeof(CONNECT))) == NULL )
    return 0;

  /* for all specified connections */
  for(i_con = 0; i_con < data->NConnections; i_con++)
  {
    /* connection ID */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read connection ID keyword");
    }
    else if(   !WordMatchesKeyword(E_KW_GEN_CONNECT, word)
            || !GetNumber(&connect_num) )
      return 0;

    if(connect_num != i_con+1)
    {
      PrintLineError("connection ID is out of sequence");
    }

    CurConnection = connect_num;

    /* connection nodes */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read connection node keyword");
    }
    else if (!WordMatchesKeyword(E_KW_GEN_DEFINITION, word))
      return 0;

    pline = GetStrArg(line, "%s", check);
    for(i_node = 0; i_node < 2; i_node++)
    {
      inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                       data, &(data->Connections[i_con].Nodes[i_node]) );
      if(inc)
        pline = line + inc;
      else
        return 0;            
    }

    /* connection element type */
    if (GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("cannot read connection element type");
    }
    else if(!WordMatchesKeyword(E_KW_GEN_ELEM_TYPE, word))
      return 0;

    if(sscanf(line, "%*s %s", word) == 1)
    {
      if(strcmp("LINK1", word) == 0)
      {
        data->Connections[i_con].ElementType = E_ELEM_TYPE_LINK1;
      }
      else if(strcmp("LINK2", word) == 0)
      {
        data->Connections[i_con].ElementType = E_ELEM_TYPE_LINK2;
      }
      else if(strcmp("LINK3", word) == 0)
      {
        data->Connections[i_con].ElementType = E_ELEM_TYPE_LINK3;
      }
      else if(strcmp("LINK4", word) == 0)
      {
        data->Connections[i_con].ElementType = E_ELEM_TYPE_LINK4;
      }
      else if(strcmp("LINK5", word) == 0)
      {
        data->Connections[i_con].ElementType = E_ELEM_TYPE_LINK5;
      }
      else if(   (sscanf(word, "%d", &n) == 1)
              && (   (n == E_ELEM_TYPE_LINK1)
                  || (n == E_ELEM_TYPE_LINK2)
                  || (n == E_ELEM_TYPE_LINK3)
                  || (n == E_ELEM_TYPE_LINK4)
                  || (n == E_ELEM_TYPE_LINK5) ) )
      {
        data->Connections[i_con].ElementType = n;
      }
      else
      {
        PrintLineError("invalid element type");
      }
    }
    else
    {
      PrintLineError("cannot read element type");
    }

    if(   data->Connections[i_con].ElementType != E_ELEM_TYPE_LINK1
       && data->Connections[i_con].ElementType != E_ELEM_TYPE_LINK2
       && data->Connections[i_con].ElementType != E_ELEM_TYPE_LINK3
       && data->Connections[i_con].ElementType != E_ELEM_TYPE_LINK4
       && data->Connections[i_con].ElementType != E_ELEM_TYPE_LINK5)
    {
      PrintLineError("invalid element type");
    }

    /* no of divisions */
    if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }

    ReadNextLine = FALSE;
    if(strcmp(E_KW_GEN_NDIV, word) == 0)
    {
      if(!GetNumber(&(data->Connections[i_con].Division)))
        return 0;
      
      if(data->Connections[i_con].Division <= 1)
      {
        PrintLineError("invalid division value");
      }

      data->Connections[i_con].DivWeight = 
        e_alloc_d_vect(data->Connections[i_con].Division);

      if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
      {
        PrintError("cannot read division weight keyword");
      }
      
      if(strcmp(E_KW_GEN_DIV_WEIGHTS, word) == 0)
      {
        if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
        {
          PrintLineError("cannot read double series");
        }
      
        if(GetDoubleSeries(line, data->Connections[i_con].Division,
                           data->Connections[i_con].DivWeight) == 0)
          return 0;
        ReadNextLine = TRUE;
      }
      else
      {
        for(i_div = 0; i_div < data->Connections[i_con].Division; i_div++)
        {
          data->Connections[i_con].DivWeight[i_div] = 1.0;
        }
        ReadNextLine = FALSE;
      }

      if(   ReadNextLine
         && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
      {
        PrintError("premature end of file");
      }
      ReadNextLine = FALSE;
    }
    else
    {
      data->Connections[i_con].Division = 1;
      data->Connections[i_con].DivWeight = NULL;
    }

    /* local material definition */
    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(strcmp(E_KW_GEN_LOCAL_MATERIAL, word) == 0)
    {
      if(sscanf(line, "%*s %d", &local_mat) != 1)
      {
        PrintLineError("cannot read material number");
      }

      if(   local_mat < 1
         || local_mat > data->NMaterials)
      {
        PrintLineError("Invalid local material");
      }
      
      data->Connections[i_con].Material = local_mat;

      ReadNextLine = TRUE;
    }

    /* end of connection */
    if(   ReadNextLine
       && GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
    {
      PrintError("premature end of file");
    }
    ReadNextLine = FALSE;

    if(!WordMatchesKeyword(E_KW_GEN_END_SECTION, word))
      return 0;

    data->Connections[i_con].RegNodes = 
      (NODE *)e_calloc(data->Connections[i_con].Division+1, sizeof(NODE) );

  } /* for all connections */

  CurConnection++;

  return 1;
}

/* ////////////////////////////////// */

static int GetBoundaryConditions(GEN *data)
{

  int   i, i_node, inc;
  char *pline;
  int  *BCSet;
  char  word[E_LINE_MAX+1];
  int   patchID, x1, y1, x2, y2;


  if(   (data->NBCNodes == 0)
     && (data->NBCSides == 0) )
    return 1;

  if(GetKeyword(line, word, &LineNumber, FullPathName, fp) == EOF)
  {
    PrintError("cannot read boundary condition keyword");
  }
  else if(!WordMatchesKeyword(E_KW_GEN_BCS, word))
    return 0;

  /* boundary condition nodes */
  if(data->NBCNodes)
  {
    data->BCNodes = e_alloc_i_vect(data->NBCNodes);
    data->BCNTypes = e_alloc_frag_i_matr(data->NBCNodes, E_DOF_MAX);
    data->BCNValue = e_alloc_frag_d_matr(data->NBCNodes, E_DOF_MAX);

    BCSet = e_alloc_i_vect(data->NNodes);

    for(i = 0; i < data->NBCNodes; i++)
    {
      if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp) != 1)
      {
        e_free(BCSet);
        return 0;
      }

      pline = GetStrArg(line, "%d", &(data->BCNodes[i]) );

      if(pline == NULL)
      {
        e_free(BCSet);
        PrintLineError("cannot read nodal boundary conditions");
      }
      else if(data->BCNodes[i] < 1 || data->BCNodes[i] > data->NNodes)
      {
        e_free(BCSet);
        PrintLineError("ill-defined node index");
      }
      else if(BCSet[data->BCNodes[i]-1])
      {
        e_free(BCSet);
        PrintLineError("boundary condition is defined twice");
      }
    
      if(ReadBndCond(pline, i, data->BCNTypes, data->BCNValue) == 0)
      {
        e_free(BCSet);
        return 0;
      }
    
      BCSet[data->BCNodes[i]-1] = TRUE;
    }
  
    e_free(BCSet);
  }

  /* boundary condition sides */
  if(data->NBCSides)
  {
    data->BCSides  = (NODE **)e_calloc(data->NBCSides, sizeof(NODE *) );
    for(i = 0; i < data->NBCSides; i++)
    {
      data->BCSides[i] = (NODE *)e_calloc(2, sizeof(NODE) );
    }
    
    data->BCSTypes = e_alloc_frag_i_matr(data->NBCSides, E_DOF_MAX);
    data->BCSValue = e_alloc_frag_d_matr(data->NBCSides, E_DOF_MAX);

    for(i = 0; i < data->NBCSides; i++)
    {
      if(e_GetLine(line, E_LINE_MAX, &LineNumber, fp))
      {
        pline = line;
        for(i_node = 0; i_node < 2; i_node++)
        {
          inc = GetNodeRef(line, pline, CurPatch, CurConnection,
                           data, &(data->BCSides[i][i_node]) );
          if(inc)
            pline = line + inc;
          else
            return 0;            
        }
      }
      else
      {
        PrintError("cannot read side boundary condition");
      }

      /* check the side definition */

      /* cannot reference to node */
      if(   data->BCSides[i][0].Node > 0
         || data->BCSides[i][1].Node > 0 )
      {
        PrintLineError("only references are allowed for boundary condition sides");
      }
      
      /* cannot reference to two different type of element */
      if(data->BCSides[i][0].Node != data->BCSides[i][1].Node)
      {
        PrintLineError("boundary condition references two different types of element");
      }

      /* cannot reference to two different elements */
      if(data->BCSides[i][0].ElemID != data->BCSides[i][1].ElemID)
      {
        PrintLineError("boundary condition references two different elements");
      }
      
      /* triangle surface */
      patchID = data->BCSides[i][0].ElemID-1;
      if(data->Patches[patchID].Shape == 3)
      {
        x1 = data->BCSides[i][0].Position[0];
        y1 = data->BCSides[i][0].Position[1];

        x2 = data->BCSides[i][1].Position[0];
        y2 = data->BCSides[i][1].Position[1];
        
        if((x1 != x2) && (y1 != y2) &&
           (x1+y1 != x2+y2) )
        {
          PrintLineError("boundary condition does not define a line");
        }
      }
      
      /* read the bc types per degree of freedom */
      if(ReadBndCond(pline, i, data->BCSTypes, data->BCSValue) == 0)
      {
        return 0;
      }
    }

  }

  return 1;
}


static int ReadBndCond(char *pline, int index, 
                       int **BCType, double **BCValue)
{
  int  j;
  char BCTypeWord[E_LINE_MAX+1];

  for(j = 0; j < E_NDIMS; j++)
  {
    pline = GetStrArg(pline, "%s", BCTypeWord);

    if(pline == NULL)
    {
      PrintLineError("cannot read boundary condition type");
    }
    else if(strcmp(E_BC_FREE, BCTypeWord) == 0)
    {
      BCType[index][j] = E_BC_NUM_FREE;
    }
    else if(strcmp(E_BC_FIXED, BCTypeWord) == 0)
    {
      BCType[index][j] = E_BC_NUM_FIXED;
      pline = GetStrArg(pline, "%lf", &(BCValue[index][j]) );

      if(pline == NULL)
      {
        PrintLineError("cannot read fix boundary condition value");
      }
    }
    else if(strcmp(E_BC_SPRING, BCTypeWord) == 0)
    {
      BCType[index][j] = E_BC_NUM_SPRING;
      pline = GetStrArg(pline, "%lf", &(BCValue[index][j]) );

      if(pline == NULL)
      {
        PrintLineError("cannot read spring boundary condition value");
      }
      else if(BCValue[index][j] < 0.0)
      {
        PrintLineError("ill-defined spring constant");
      }
    }
    else
    {
      PrintLineError("undefined boundary condition type");
    }
  }

  return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */

static void DeallocateAllMemory(GEN *data)
{

}





