#ifndef PATCH_H_
#define PATCH_H_


/* type of definition for cubic patch */
#define DEF_LINEAR      1
#define DEF_PARAMETRIC  2
#define DEF_POINT       3

/* referenced nodes */
#define SURFACE_NODE -1
#define LINK_NODE    -2

/* along which diagonal to cut a quad element, EXPERT option */
#define E_DIAGONAL_1            1
#define E_DIAGONAL_2            2
#define E_DIAGONAL_SHORTEST     3
#define E_DIAGONAL_LONGEST      4

/* order is important */
#define MAT_LINK1_NAME     "LNK1"
#define MAT_LINK2_NAME     "LNK2"
#define MAT_LINK3_NAME     "LNK3"
#define MAT_LINK4_NAME     "LNK4"
#define MAT_BEG_LINK5_NAME "BEG5"
#define MAT_MID_LINK5_NAME "LNK5"
#define MAT_END_LINK5_NAME "END5"
#define MAT_NON_LINK5_NAME "NON5"

#define MAT_TRIANG1_NAME "TRI1"
#define MAT_TRIANG3_NAME "TRI3"
#define MAT_TRIANG4_NAME "TRI4"
#define MAT_TRIANG5_NAME "TRI5"

#define MAT_QUAD1_NAME "QUAD1"

/* indices for material definition */
#define MAT_LINK1           0
#define MAT_LINK2           1
#define MAT_LINK3           2
#define MAT_LINK4           3
#define MAT_BEG_LINK5       4
#define MAT_MID_LINK5       5
#define MAT_END_LINK5       6
#define MAT_NON_LINK5       7

#define MAT_TRIANG1         8
#define MAT_TRIANG3         9
#define MAT_TRIANG4         10
#define MAT_TRIANG5         11

#define MAT_QUAD1           12

/* last material number, required for user defined material */
#define DEFAULT_N_MATERIAL  13

/* user defined material name */
#define USER_DEF_MAT_NAME   "MAT"

/* allocation unit */
#define ALLOC_SIZE 16


/* node definition */
typedef struct
{
  int Node;         /* node number in the final mesh or reference type */
  int ElemID;       /* number of the referenced element */
  int Position[2];  /* position of the referenced node inside the element */
}
NODE;

/* surface definition */
typedef struct
{
  int       Shape;          /* shape of the surface, triangle or quad */
  int       ElementType;    /* what kind of element to generate inside the surface */
  int       Definition;     /* how is the surface defined */
  NODE      Nodes[16];      /* 0-3 are corner nodes except for point format */
  double    ParamDef[4][9]; /* parameters for the parametric definition */
  int       XDivision;      /* number of x divisions */
  int       YDivision;      /* number of y divisions */
  double   *XDivWeight;     /* division weights in the x direction */
  double   *YDivWeight;     /* division weights in the y direction */
  double   *TDivWeight;     /* for triangle surfaces, division weights in the third direction */
  
  int       LinkSides[3];   /* links are parallel with which side */
  int      *LinkDiv[3];     /* link position from the particular side */
  int      *LinkType[3];    /* type of the link element parallel with the particular side */
 
  int       Material;       /* type of material, integer value is converted into text */
  int       Diagonal;       /* for triangle generation how to cut a quad element */
  
  NODE    **RegNodes;       /* all boundary and inside nodes are stored for later reference */
}
SURFACE;

/* link/connection definition */
typedef struct
{
  NODE     Nodes[2];     /* end nodes of the link */
  int      ElementType;  /* what kind of element to generate */
  int      Division;     /* number of divisions in the link */
  double  *DivWeight;    /* division weights, they will be normalised */

  int      Material;     /* type of material, integer value is converted into text */

  NODE    *RegNodes;     /* all boundary and inside nodes are stored for later reference */

}
CONNECT;


typedef struct
{
  char    *Title;
  int      NNodes;
  int      NSurfaces;
  int      NConnections;
  int      NMaterials;
  int      NBCNodes;
  int      NBCSides;

  int      DefMat;
  
  double **Coords;
  
  SURFACE *Patches;
  CONNECT *Connections;

  char   **MaterialName;

  int     *BCNodes;
  int    **BCNTypes;
  double **BCNValue;

  NODE   **BCSides;
  int    **BCSTypes;
  double **BCSValue;

}
GEN;


typedef struct
{
  int       NPoints;       /* pointers during generation */
  int       NQuads;        /* number of quadrilateral elements in the mesh */
  int       NTriangles[4]; /* number of triangle elements in the mesh */
  int       NLinks[5];     /* number of link elements in the mesh */
  
  double  **Coords;        /* coordinate values */
  int     **Quads;         /* nodes of quadrilateral elements */
  int     **Triangles[4];  /* nodes of triangle elements */
  int     **Links[5];      /* nodes of quadrilateral elements */

  int      *MatQuads;        /* materials of quadrilateral elements */
  int      *MatTriangles[4]; /* materials of triangle elements */
  int      *MatLinks[5];     /* materials of link elements */

  int      NBCNodes;         /* number of boundary nodes */
  int     *BCNodes;          /* node number of boundary nodes */
  int    **BCTypes;          /* type of the boundary */
  double **BCValue;          /* value of the boundary */

  /* for removing duplicate nodes */
  int      *Mark;
  int      *Translate;
  int       Duplicate;
}
INTERNAL_MESH;


void p_GenerateTriangPatch(GEN *Structure, int CurPatch, 
                           INTERNAL_MESH *mesh);

void p_GenerateCubicPatch(GEN *data, int CurPatch, INTERNAL_MESH *mesh);

void p_GenerateConnections(GEN *data, int CurCon, 
                           INTERNAL_MESH *mesh);

void DetNodeCoord(GEN *Structure, INTERNAL_MESH *mesh, NODE *nref,
                  double *x, double *y, double *z);


void e_Generate_Mesh(GEN *Structure, char *filename);


int e_GetGenData(GEN **gen_data, char *InputFile);
void GetQuadCornerPos(SURFACE *data, int corner, int *xe, int *ye);

void e_VisualGenFile(GEN *Structure, char *filename);



#endif /* #ifndef PATCH_H_ */


