/* 
 * Functions to carry out MPI send and receive operations for vectors of data
 * Communication can be symmetric (same data to send and receive) and
 * asymmetric (when different elements/nodes are sending and receiving
 * data.
 * The communication is non-blocking and can be called as one function or
 * it can be hidden behind some calcualtion.
 *
 * Setup of the communication in four steps:
 *   - determination of sending elements/nodes (any function)
 *   - sending/receiving this data             (SendCommDataInit / RecvCommDataInit)
 *   - conversion of global data to local      (any function)
 *   - final allocation                        (SetupCommDataFinal)
 */

#include "comm_vector.h"

/* in the case of symmetric communication the sent and received data
   should be added together */
#ifdef __SYMMETRIC__
#define OPERATION +=
#endif

/* in the case of asymmetric communication the sent and the received data
   belongs to different elements */
#ifdef __ASYMMETRIC__
#define OPERATION =
#endif

#ifndef OPERATION
#error "Communication type ( __SYMMETRIC__ | __ASYMMETRIC__ ) is not defined."
#endif


void SendCommDataInit(int NSubdomains, COMM_VECTOR_DATA *data, int i_subdom)
{
  int i;

  MPI_Send(data->Par_NElemSend, NSubdomains, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD);

  for(i = 0; i < NSubdomains; i++)
  {
    if(data->Par_NElemSend[i] > 0)
    {
      MPI_Send(data->Par_ElemSend[i], data->Par_NElemSend[i], MPI_INT, 
               i_subdom, 445+i, MPI_COMM_WORLD);
    }
  }

  MPI_Send(data->Par_NElemRecv, NSubdomains, MPI_INT, 
           i_subdom, 446, MPI_COMM_WORLD);

  for(i = 0; i < NSubdomains; i++)
  {
    if(data->Par_NElemRecv[i] > 0)
    {
      MPI_Send(data->Par_ElemRecv[i], data->Par_NElemRecv[i], MPI_INT, 
               i_subdom, 447+i, MPI_COMM_WORLD);
    }
  }

}


void RecvCommDataInit(int NSubdomains, COMM_VECTOR_DATA **data, int i_subdom)
{
  int i;
  MPI_Status status;

  (*data) = (COMM_VECTOR_DATA *)calloc(1, sizeof(COMM_VECTOR_DATA) );

  (*data)->Par_NElemSend = (int *)calloc(NSubdomains, sizeof(int) );
  MPI_Recv((*data)->Par_NElemSend, NSubdomains, MPI_INT, 
           i_subdom, 444, MPI_COMM_WORLD, &status);

  (*data)->Par_ElemSend = (int **)calloc(NSubdomains, sizeof(int *) );
  for(i = 0; i < NSubdomains; i++)
  {
    if((*data)->Par_NElemSend[i] > 0)
    {
      (*data)->Par_ElemSend[i] = (int *)calloc((*data)->Par_NElemSend[i], sizeof(int) );
      MPI_Recv((*data)->Par_ElemSend[i], (*data)->Par_NElemSend[i], MPI_INT, 
               i_subdom, 445+i, MPI_COMM_WORLD, &status);
    }
  }

  (*data)->Par_NElemRecv = (int *)calloc(NSubdomains, sizeof(int) );
  MPI_Recv((*data)->Par_NElemRecv, NSubdomains, MPI_INT, 
           i_subdom, 446, MPI_COMM_WORLD, &status);

  (*data)->Par_ElemRecv = (int **)calloc(NSubdomains, sizeof(int *) );
  for(i = 0; i < NSubdomains; i++)
  {
    if((*data)->Par_NElemRecv[i] > 0)
    {
      (*data)->Par_ElemRecv[i] = (int *)calloc((*data)->Par_NElemRecv[i], sizeof(int) );
      MPI_Recv((*data)->Par_ElemRecv[i], (*data)->Par_NElemRecv[i], MPI_INT, 
               i_subdom, 447+i, MPI_COMM_WORLD, &status);
    }
  }
}


void SetupCommDataFinal(COMM_VECTOR_DATA *data)
{
  int i, j, si, ri;

  data->Par_SubdNoSend  = (int *)calloc(data->Par_NSubdomains, sizeof(int) );
  data->Par_SubdNoRecv  = (int *)calloc(data->Par_NSubdomains, sizeof(int) );
  data->Par_NSubdNoSend = 0;
  data->Par_NSubdNoRecv = 0;

  /* count subdomains to communicate with */
  for(i = 0; i < data->Par_NSubdomains; i++)
  {
    if(data->Par_NElemSend[i] > 0)
    {
      data->Par_SubdNoSend[data->Par_NSubdNoSend] = i+1;
      data->Par_NSubdNoSend++;
    }

    if(data->Par_NElemRecv[i] > 0)
    {
      data->Par_SubdNoRecv[data->Par_NSubdNoRecv] = i+1;
      data->Par_NSubdNoRecv++;
    }
  }

  /* remove zero length vectors, no communication */
  si = 0;
  for(i = 0; i < data->Par_NSubdomains - si; i++)
  {
    if(data->Par_NElemSend[i] == 0)
    {
      for(j = i; j < data->Par_NSubdomains - 1; j++)
      {
        data->Par_NElemSend[j]   = data->Par_NElemSend[j+1];
        data->Par_NElemSend[j+1] = 0;
        data->Par_ElemSend[j]    = data->Par_ElemSend[j+1];
        data->Par_ElemSend[j+1]  = NULL;
      }
      si++;
      i--;
    }
  }

  ri = 0;
  for(i = 0; i < data->Par_NSubdomains - ri; i++)
  {
    if(data->Par_NElemRecv[i] == 0)
    {
      for(j = i; j < data->Par_NSubdomains - 1; j++)
      {
        data->Par_NElemRecv[j]   = data->Par_NElemRecv[j+1];
        data->Par_NElemRecv[j+1] = 0;
        data->Par_ElemRecv[j]    = data->Par_ElemRecv[j+1];
        data->Par_ElemRecv[j+1]  = NULL;
      }
      ri++;
      i--;
    }
  }

  data->Par_SendCommReq =
    (MPI_Request *)calloc(data->Par_NSubdNoSend, sizeof(MPI_Request) );
  data->Par_SendCommStat =
    (MPI_Status *)calloc(data->Par_NSubdNoSend, sizeof(MPI_Status) );
  data->Par_RecvCommReq =
    (MPI_Request *)calloc(data->Par_NSubdNoRecv, sizeof(MPI_Request) );

  /* allocate buffers to send and receive */
  data->Par_BufSend   = (double **)calloc(data->Par_NSubdNoSend, sizeof(double *) );
  data->Par_BufRecv   = (double **)calloc(data->Par_NSubdNoRecv, sizeof(double *) );

  for(i = 0; i < data->Par_NSubdNoSend; i++)
  {
    data->Par_BufSend[i] = 
      (double *)calloc(data->Par_NElemSend[i] * MAX_VECTOR_COMM, sizeof(double) );
  }

  for(i = 0; i < data->Par_NSubdNoRecv; i++)
  {
    data->Par_BufRecv[i] = 
      (double *)calloc(data->Par_NElemRecv[i] * MAX_VECTOR_COMM, sizeof(double) );
  }
}

/* communication functions */
/* ############################################################################ */
void LaunchDataSend(int n, COMM_VECTOR_DATA *data)
{
  int i, j, k;

  /* assemble buffers and start send */
  for(i = 0; i < data->Par_NSubdNoSend; i++)
  {
    for(k = 0; k < n; k++)
    {
      for(j = 0; j < data->Par_NElemSend[i]; j++)
      {
        data->Par_BufSend[i][j+k*data->Par_NElemSend[i]] = 
          data->Values[k][data->Par_ElemSend[i][j]-1];
      }
    }

    /* initiate the sending procedure, non-blocking */
    MPI_Isend(data->Par_BufSend[i], data->Par_NElemSend[i]*n, MPI_DOUBLE,
              data->Par_SubdNoSend[i]-1, 900+data->Par_ProcID, 
              MPI_COMM_WORLD, data->Par_SendCommReq+i);
  }
}

void WaitDataSend(COMM_VECTOR_DATA *data)
{
  MPI_Waitall(data->Par_NSubdNoSend, data->Par_SendCommReq, 
              data->Par_SendCommStat);
}

void LaunchDataRecv(int n, COMM_VECTOR_DATA *data)
{
  int i;
  
  for(i = 0; i < data->Par_NSubdNoRecv; i++)
  {
    MPI_Irecv(data->Par_BufRecv[i], data->Par_NElemRecv[i]*n, MPI_DOUBLE,
              data->Par_SubdNoRecv[i]-1, 900+data->Par_SubdNoRecv[i],
              MPI_COMM_WORLD, data->Par_RecvCommReq+i);
  }
}

void WaitDataRecv(int n, COMM_VECTOR_DATA *data)
{
  int        i, j, k, index;
  MPI_Status status;
  
  for(i = 0; i < data->Par_NSubdNoRecv; i++)
  {
    /* wait for any incoming data */
    MPI_Waitany(data->Par_NSubdNoRecv, data->Par_RecvCommReq,
                &index, &status);
    /* store the data */
    for(k = 0; k < n; k++)
    {
      for(j = 0; j < data->Par_NElemRecv[index]; j++)
      {
        data->Values[k][data->Par_ElemRecv[index][j]-1] OPERATION
          data->Par_BufRecv[index][ j+k*data->Par_NElemRecv[index] ];
      }
    }
  }
}

/* do all communication */
void CommunicateData(int n, COMM_VECTOR_DATA *data)
{
  if(n > MAX_VECTOR_COMM)
  {
    fprintf(stderr, "Too many vectors to communicate !");
    fprintf(stderr, "Change MAX_VECTOR_COMM and recompile!");
    exit(-1);
  }

  LaunchDataSend(n, data);
  LaunchDataRecv(n, data);

  /* Because the sent and received data absolutely independent, 
   * and waiting for all send operations to finish is blocking,
   * we first receive everything then wait for the completion of send
   */
  WaitDataRecv(n, data);
  WaitDataSend(data);
}



