/* Dynamic relaxation calculation routines */

/*
  Link1 -- Cables - Elastic control
  Link2 -- Trusses - Elastic control
  Link3 -- Constant force control
  Link4 -- Constant force density control
  Link5 -- G-strings (Elastic control)

  Triang3 -- Solid triangle    (Elastic control)
  Triang4 -- Membrane triangle (Elastic control)
  Triang5 -- Formfinding triangle (constant stress, isotrop & anysotrop)

*/

/*
  General Notes: 
  --- 1 ---
  The calculation of pressure can be simplified according
  to Barnes' formulation, but that would require an extra
  communication along the subdomain boundaries.
  The simpler formula would be:
    vector[nodes[i][ii]-1][k] += pressure * norm[k] / 6.00;
  and in that case we do not have to multiply "pressure" with
  the normal vector component in the "dr_pressure" function,
  but "pressure" should be communicated with other boundaries.
  --- 2 ---
  Pressure can be simulated in 2D as well. In this case the pressure acts
  perpendicular to the 1D link elements. 2D elements can be in the mesh
  as well, but the pressure does not affects them in 2D. The same
  mechanism is used for the 2D and the 3D pressure and normal calculation.
*/
#include <assert.h>
#include "dr_calc.h"
#include "timer.h"

#ifdef __TIME_ELEM_CALC__
static unsigned long TimeLink4 = 0, TimeTriang5 = 0;
#endif

#define DR_SIGN(x) ((x) > 0 ? 1 : ((x) < 0 ? -1 : 0))


static void dr_link_mass(int dimension, int no, int **nodes, double **coords,
                         MAT  **materials, double *init_length, int type,
                         double **mass);
static void dr_triang_stiffness_mass(int dimension, int no_elem, 
                                     int **nodes, double **coords,
                                     double **init_length,
                                     MAT  **materials, int type, 
                                     double ***Ke, double **mass);

static void dr_link_res(int dimension, int nelems, int **nodes, double **coords, 
                        MAT **material, double *init_length, int type,
                        double **residual);
static void dr_triang_res(int dimension, int nelems, int **nodes, double **coords, 
                          MAT **material, double **init_length, double ***Ke, 
                          int type, double **residual);


static void dr_init(MESH *LocalMesh, DR_DATA *dr_extra_data);
static void dr_step(MESH *LocalMesh, DR_DATA *dr_extra_data);
static void dr_peak(MESH *LocalMesh, DR_DATA *dr_extra_data);

/* ###################################################################### */


static void dr_triang_hyd_pressure(int nelems, int **nodes, double **coords,
                                   double density, MAT **materials,
                                   double **pressure)
{
  int    ielem, inode, n, k;
  double zc, wh1, wh2, pressure_value;
  int    n1, n2, n3;
  double x21, y21, z21, x31, y31, z31, area2;
  double norm[3];
  
  for(ielem = 0; ielem < nelems; ielem++)
  {
    wh1 = materials[ielem]->dParam8;
    wh2 = materials[ielem]->dParam10;
    assert(   (materials[ielem]->dParam8  > 0.0 && materials[ielem]->dParam9  != 0.0)
           || (materials[ielem]->dParam10 > 0.0 && materials[ielem]->dParam11 != 0.0) );
    
    zc = 0.0;
    for(inode = 0; inode < 3; inode++)
    {
      n = nodes[ielem][inode]-1;
      zc += coords[n][2];
    }
    zc /= 3;

    /* if the element is in water from either side */
    if(wh1 > zc || wh2 > zc)
    {
      pressure_value = 0.0;
      if(wh1 > zc)
          pressure_value += (wh1 - zc) * materials[ielem]->dParam9;
      if(wh2 > zc)
          pressure_value += (wh2 - zc) * materials[ielem]->dParam11;
    }
    else
    {
      continue;
    }

    n1 = nodes[ielem][0]-1;
    n2 = nodes[ielem][1]-1;
    n3 = nodes[ielem][2]-1;

    x21 = coords[n2][0] - coords[n1][0];
    y21 = coords[n2][1] - coords[n1][1];
    z21 = coords[n2][2] - coords[n1][2];

    x31 = coords[n3][0] - coords[n1][0];
    y31 = coords[n3][1] - coords[n1][1];
    z31 = coords[n3][2] - coords[n1][2];

    norm[0] = (y21 * z31) - (y31 * z21);
    norm[1] = (z21 * x31) - (z31 * x21);
    norm[2] = (x21 * y31) - (x31 * y21);

    area2 = sqrt(E_SQUARE(norm[0]) + 
                 E_SQUARE(norm[1]) + 
                 E_SQUARE(norm[2])   );

    for(inode = 0; inode < 3; inode++)
    {
      for(k = 0; k < E_NDIMS; k++)
      {
        n = nodes[ielem][inode]-1;
        pressure[n][k] += pressure_value * area2 / 6.00;
      }
    }

  }
}




/* Job:           calculating normal vectors or pressure in 2D
   Reference:     
   Periodicity:   occasionally
   Notes:         only in 2D, when 'pressure_spec' is not zero the 
                  'pressure_value' is used, otherwise it is ignored 
*/
static void dr_link_normal_or_pressure(int nelems, int **nodes, double **coords, 
                                       int pressure_spec, double pressure_value, 
                                       double **vector)
{
  int i, n1, n2;
  double nx, ny, ll;

  for(i = 0; i < nelems; i++)
  {
    n1 = nodes[i][0]-1;
    n2 = nodes[i][1]-1;
    
    /* the negative direction is accounted here */
    nx = coords[n1][1] - coords[n2][1];
    ny = coords[n2][0] - coords[n1][0];

    /* normalise the vector */
    ll = sqrt(nx * nx + ny * ny);
    nx /= ll;
    ny /= ll;
    
    if(pressure_spec)
    {
      vector[n1][0] += pressure_value * ll / 2.00;
      vector[n1][1] += pressure_value * ll / 2.00;

      vector[n2][0] += pressure_value * ll / 2.00;
      vector[n2][1] += pressure_value * ll / 2.00;
    }
    else
    {
      vector[n1][0] += nx;
      vector[n1][1] += ny;

      vector[n2][0] += nx;
      vector[n2][1] += ny;
    }
  }

} /* end of -- dr_link_normal_or_pressure -- function */

/* ###################################################################### */

/* Job:           calculating pressure for each node
   Reference:     
   Periodicity:   occasionally
   Notes:         only in 3D
*/
static void dr_triang_pressure(int nelems, int **nodes, double **coords, 
                               double pressure_value, MAT **materials,
                               double **vector)
{
  int i, ii, k, n1, n2, n3;
  double x21, y21, z21, x31, y31, z31, norm_2area, full_pressure;
  double norm[3];

  for (i = 0; i < nelems; i++)
  {
    full_pressure = pressure_value;
    
    if(materials[i]->KeywordSet[E_MKW_NUM_DPARAM12] == TRUE)
    {
      full_pressure += materials[i]->dParam12;
    }
    
    n1 = nodes[i][0]-1;
    n2 = nodes[i][1]-1;
    n3 = nodes[i][2]-1;

    x21 = coords[n2][0] - coords[n1][0];
    y21 = coords[n2][1] - coords[n1][1];
    z21 = coords[n2][2] - coords[n1][2];

    x31 = coords[n3][0] - coords[n1][0];
    y31 = coords[n3][1] - coords[n1][1];
    z31 = coords[n3][2] - coords[n1][2];

    norm[0] = (y21 * z31) - (y31 * z21);
    norm[1] = (z21 * x31) - (z31 * x21);
    norm[2] = (x21 * y31) - (x31 * y21);

    norm_2area = sqrt(E_SQUARE(norm[0]) + 
                      E_SQUARE(norm[1]) + 
                      E_SQUARE(norm[2])   );

    /* pressure distribution to the nodes - pneumatic structures */
    for (ii = 0; ii < 3; ii++)           /* for all nodes in the triang */
    {
      for (k = 0; k < E_NDIMS; k++)      /* 3D */
      {
        vector[nodes[i][ii]-1][k] += full_pressure * norm_2area / 6.00;
      }
    }
    
  }
} /* end of -- dr_triang_pressure -- function */



/* Job:           calculating normal vectors for each node
   Reference:     
   Periodicity:   occasionally
   Notes:         only in 3D
*/
static void dr_triang_normal(int nelems, int **nodes, double **coords, double **vector)
{
  int i, ii, k, n1, n2, n3;
  double x21, y21, z21, x31, y31, z31, norm_2area;
  double norm[3];

  for (i = 0; i < nelems; i++)
  {
    n1 = nodes[i][0]-1;
    n2 = nodes[i][1]-1;
    n3 = nodes[i][2]-1;

    x21 = coords[n2][0] - coords[n1][0];
    y21 = coords[n2][1] - coords[n1][1];
    z21 = coords[n2][2] - coords[n1][2];

    x31 = coords[n3][0] - coords[n1][0];
    y31 = coords[n3][1] - coords[n1][1];
    z31 = coords[n3][2] - coords[n1][2];

    norm[0] = (y21 * z31) - (y31 * z21);
    norm[1] = (z21 * x31) - (z31 * x21);
    norm[2] = (x21 * y31) - (x31 * y21);

    norm_2area = sqrt(E_SQUARE(norm[0]) + 
                      E_SQUARE(norm[1]) + 
                      E_SQUARE(norm[2])   );

    /* make it unit length and add it to the node */
    for (ii = 0; ii < 3; ii++)           /* for all nodes in the triang */
        for (k = 0; k < E_NDIMS; k++)      /* 3D */
            vector[nodes[i][ii]-1][k] += norm[k] / norm_2area;

  }
} /* end of -- dr_triang_normal -- function */


/* ###################################################################### */

/* Job:           calculating normal vectors for triangles (3D) or links (2D)
   Reference:     
   Periodicity:   occasionally
   Notes:         the function called only when the normals are really
                  necessary
*/
void dr_normal(MESH *mesh, DR_DATA *data)
{
  int   i, k;
  double sum;

  /* nullify the normal vectors */
  /* ------------------------------------------------------------------------ */
  for(i = 0; i < mesh->NMeshPoints; i++)
    for(k = 0; k < data->Dimension; k++)
      data->Normal[i][k] = 0.00;

  /* calculate the normals and sum them up immediately */
  /* ------------------------------------------------------------------------ */
  if(data->Dimension == 2)
  {
    if(mesh->NElemsLink1)
      dr_link_normal_or_pressure(mesh->NElemsLink1,
                                 mesh->NodesLink1,
                                 mesh->MeshPointCoords,
                                 FALSE, 0.0, data->Normal);

    if(mesh->NElemsLink2)
      dr_link_normal_or_pressure(mesh->NElemsLink2,
                                 mesh->NodesLink2,
                                 mesh->MeshPointCoords,
                                 FALSE, 0.0, data->Normal);

    if(mesh->NElemsLink3)
      dr_link_normal_or_pressure(mesh->NElemsLink3,
                                 mesh->NodesLink3,
                                 mesh->MeshPointCoords,
                                 FALSE, 0.0, data->Normal);

    if(mesh->NElemsLink4)
      dr_link_normal_or_pressure(mesh->NElemsLink4,
                                 mesh->NodesLink4,
                                 mesh->MeshPointCoords,
                                 FALSE, 0.0, data->Normal);

    if(mesh->NElemsLink5)
      dr_link_normal_or_pressure(mesh->NElemsLink5,
                                 mesh->NodesLink5,
                                 mesh->MeshPointCoords,
                                 FALSE, 0.0, data->Normal);

  }
  else if(data->Dimension == 3)
  {
    if(mesh->NElemsTriang3)
      dr_triang_normal(mesh->NElemsTriang3, mesh->NodesTriang3, mesh->MeshPointCoords,
                       data->Normal);

    if(mesh->NElemsTriang4)
      dr_triang_normal(mesh->NElemsTriang4, mesh->NodesTriang4, mesh->MeshPointCoords,
                       data->Normal);

    if(mesh->NElemsTriang5)
      dr_triang_normal(mesh->NElemsTriang5, mesh->NodesTriang5, mesh->MeshPointCoords,
                       data->Normal);
  }

#ifdef __E_MPI__

#ifdef COMM_LOG
  MPE_Log_event(1, 0, "start communicate normals");
#endif

  /* communicate the normals */
  if(data->CommReq)
  {
    data->CommData->Values[0] = data->Normal[0];
    CommunicateData(1, data->CommData);
  }

#ifdef COMM_LOG
  MPE_Log_event(2, 0, "end communicate normals");
#endif

#endif /* __E_MPI__ */


  /* normalise the vectors */
  /* ---------------------------------------------------------------------- */
  for (i = 0; i < mesh->NMeshPoints; i++)
  {
    sum = 0.00;
    for (k = 0; k < data->Dimension; k++)
      sum += E_SQUARE(data->Normal[i][k]);
    sum = sqrt(sum);

    /* in case of a node where only link elements connected in 3D */
    if(sum < DR_ZERO)
      sum = 1.0;

    /* normalise the normal vector */
    for (k = 0; k < data->Dimension; k++)
      data->Normal[i][k] /= sum;

  }

} /* end of -- dr_normal -- function */

/* ###################################################################### */

/* Job:           calculating pressure load on triangles
   Reference:     
   Periodicity:   occasionally
   Notes:         in 2D for only links and in 3D only triangles
*/
void dr_pressure(MESH *mesh, DR_DATA *data)
{
  int i, k;

  /* nullify the pressure vectors */
  /* ------------------------------------------------------------------------ */
  for(i = 0; i < mesh->NMeshPoints; i++)
    for(k = 0; k < data->Dimension; k++)
      data->Pressure[i][k] = 0.00;

  /* calculate the pressure proportional to the area */
  /* ------------------------------------------------------------------------ */
  if(data->Dimension == 2)
  {
    if(mesh->NElemsLink1)
      dr_link_normal_or_pressure(mesh->NElemsLink1, 
                                 mesh->NodesLink1,
                                 mesh->MeshPointCoords,
                                 TRUE, data->PressureValue, 
                                 data->Pressure);

    if(mesh->NElemsLink2)
      dr_link_normal_or_pressure(mesh->NElemsLink2,
                                 mesh->NodesLink2,
                                 mesh->MeshPointCoords,
                                 TRUE, data->PressureValue, 
                                 data->Pressure);

    if(mesh->NElemsLink3)
      dr_link_normal_or_pressure(mesh->NElemsLink3, 
                                 mesh->NodesLink3,
                                 mesh->MeshPointCoords,
                                 TRUE, data->PressureValue, 
                                 data->Pressure);

    if(mesh->NElemsLink4)
      dr_link_normal_or_pressure(mesh->NElemsLink4, 
                                 mesh->NodesLink4,
                                 mesh->MeshPointCoords,
                                 TRUE, data->PressureValue, 
                                 data->Pressure);

    if(mesh->NElemsLink5)
      dr_link_normal_or_pressure(mesh->NElemsLink5, 
                                 mesh->NodesLink5,
                                 mesh->MeshPointCoords,
                                 TRUE, data->PressureValue, 
                                 data->Pressure);
  }
  else if(data->Dimension == 3)
  {
    /* TRIANG 3 */
    if(mesh->NElemsTriang3)
    {
      if(data->PressureOption)
      {
        dr_triang_pressure(mesh->NElemsTriang3, 
                           mesh->NodesTriang3,
                           mesh->MeshPointCoords,
                           data->PressureValue, mesh->MatsTriang3,
                           data->Pressure);
      }
      if(data->HydPressOption)
      {
        dr_triang_hyd_pressure(mesh->NElemsTriang3, mesh->NodesTriang3,
                               mesh->MeshPointCoords,
                               data->HydPressDensity, mesh->MatsTriang3,
                               data->Pressure);
      }
    }

    /* TRIANG 4 */
    if(mesh->NElemsTriang4)
    {
      if(data->PressureOption)
      {
        dr_triang_pressure(mesh->NElemsTriang4, 
                           mesh->NodesTriang4,
                           mesh->MeshPointCoords,
                           data->PressureValue, mesh->MatsTriang4,
                           data->Pressure);
      }
      if(data->HydPressOption)
      {
        dr_triang_hyd_pressure(mesh->NElemsTriang4, mesh->NodesTriang4,
                               mesh->MeshPointCoords, 
                               data->HydPressDensity, mesh->MatsTriang4,
                               data->Pressure);
      }
    }

    /* TRIANG 5 */
    if(mesh->NElemsTriang5)
    {
      if(data->PressureOption)
      {
        dr_triang_pressure(mesh->NElemsTriang5,
                           mesh->NodesTriang5,
                           mesh->MeshPointCoords,
                           data->PressureValue, mesh->MatsTriang5,
                           data->Pressure);
      }
      if(data->HydPressOption)
      {
        dr_triang_hyd_pressure(mesh->NElemsTriang5, mesh->NodesTriang5,
                               mesh->MeshPointCoords, 
                               data->HydPressDensity, mesh->MatsTriang5,
                               data->Pressure);
      }
    }
    
  }

  /* calculate the pressure vectors normal to the surface */
  /* ---------------------------------------------------------------------- */
  for(i = 0; i < mesh->NMeshPoints; i++)
    for(k = 0; k < data->Dimension; k++)
      data->Pressure[i][k] *= data->Normal[i][k];

} /* end of -- dr_pressure -- function */



/* ###################################################################### */

/* Job:           calculating masses for each link
   Reference:     Ref. 2 page 157, Ref. 1 page 9
   Periodicity:   occasionally
   Notes:         In the references, due to the possibility of large changes 
                  in the geometry the suggested stiffness for cables consists
                  an elastic stiffness and a geometric stiffness. Here, it
                  is not used hoping that the mass will be enough to compensate
                  the movement of the mesh !
*/
static void dr_link_mass(int dimension, int no, int **nodes, double **coords,
                         MAT  **materials, double *init_length, int type,
                         double **mass)
{
  int n1, n2, k, i;
  double E, A, xrt, curr_length, length_sq, sx;
  double dl[3];

  for(i = 0; i < no; i++)
  {
    E = materials[i]->YMod;
    A  = materials[i]->dParam1;

    n1 = nodes[i][0]-1;
    n2 = nodes[i][1]-1;

    length_sq = 0.0;
    for(k = 0; k < dimension; k++)
    {
      dl[k] = E_ABS(coords[n2][k] - coords[n1][k]);
      length_sq += E_SQUARE(dl[k]);
    }
    curr_length = sqrt(length_sq);
    
    /* stiffness from elasticity */
    sx = E * A / init_length[i];
    switch(type)
    {
      /* truss, cable */
      case 1:
      case 2:
        sx += E * A * (curr_length - init_length[i]) / 
              (init_length[i] * curr_length);
        break;
      /* constant force */
      case 3:
        sx += materials[i]->dParam2 / curr_length;
        break;
      /* constant force density */
      case 4:
        sx += materials[i]->dParam2 * init_length[i] / curr_length;
        break;
      /* g-strings are weightless */
      case 5:
        sx = 0;
        break;
      default:
        e_GeneralError("dr_link_mass: unknown link type");
    }

    for(k = 0; k < dimension; k++)
    {
      xrt = E_ABS(sx) * dl[k] / curr_length;

      /* only for safety and avoid division by zero error */
      if(E_ABS(xrt) < DR_ZERO)
        xrt = 1.00;

      mass[n1][k] += xrt;
      mass[n2][k] += xrt;
    }
  }
} /* end of -- dr_link_mass -- function */

/* ###################################################################### */

/* Job:           calculating stiffnesses and masses for each tirangle
   Reference:     Ref. 1. page 20-22, Ref. 2. page 158
   Periodicity:   occasionally
   Notes:         Since the x coordinate direction is parallel to the first
                  side of triangle, therefore the G matrix can be greatly
                  simplified.
*/
static void dr_triang_stiffness_mass(int dimension, int no_elem, 
                                     int **nodes, double **coords,
                                     double **init_length,
                                     MAT  **materials, int type, 
                                     double ***Ke, double **mass)
{
  int j, i, l, n, ii, jj, kk, n1, n2;
  double det, young, thick, pois, com, area_real, area_init, sum;
  double  ax_2, ax_3, bx_2, bx_3, cx_2, cx_3;
  double  g[3][3];                      /* G matrix */
  double  dd[3][3];                     /* D matrix */
  double  gt[3][3];                     /* transpose of G matrix */
  double  k1[3][3], k2[3][3];           /* temporary matrix */
  double  angle[3];
  double  side_length[3], k_mass[3];
  double  dl[3][3];                     /* side length components */
  int     side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};

  for (l = 0; l < no_elem; l++)
  {
    /* calculating matrix G */
    /* ------------------------------------------------------------------------ */

    /* current side lengths */
    for(i = 0; i < 3; i++)
    {
      n1 = nodes[l][side_n_node[i][0]]-1;
      n2 = nodes[l][side_n_node[i][1]]-1;
      side_length[i] = 0.00;
      for(j = 0; j < dimension; j++)
      {
        dl[i][j] = E_ABS(coords[n2][j] - coords[n1][j]);
        side_length[i] += E_SQUARE(dl[i][j]);
      }
      side_length[i] = sqrt(side_length[i]);
    }

    sum = E_SQUARE(side_length[0]) + 
          E_SQUARE(side_length[1]) + 
          E_SQUARE(side_length[2]);

    /* we have to use the actual side lengths, see Ref. */
    angle[0] = 0.00;
    angle[1] = M_PI - acos((sum - 2.00 * E_SQUARE(side_length[2])) /
                           (2.00 * side_length[0] * side_length[1]));
    angle[2] = acos((sum - 2.00 * E_SQUARE(side_length[1])) /
                    (2.00 * side_length[0] * side_length[2]));

    ax_2 = E_SQUARE(cos(angle[1]));
    ax_3 = E_SQUARE(cos(angle[2]));

    bx_2 = E_SQUARE(sin(angle[1]));
    bx_3 = E_SQUARE(sin(angle[2]));

    cx_2 = cos(angle[1])*sin(angle[1]);
    cx_3 = cos(angle[2])*sin(angle[2]);

    det = bx_2 * cx_3 - cx_2 * bx_3;

    g[0][0] = 1.00;
    g[0][1] = 0.00;
    g[0][2] = 0.00;

    g[1][0] = (ax_3 * cx_2 - ax_2 * cx_3) / det;
    g[1][1] =  cx_3 / det;
    g[1][2] = -cx_2 / det;

    g[2][0] = (ax_2 * bx_3 - ax_3 * bx_2) / det;
    g[2][1] = -bx_3 / det;
    g[2][2] =  bx_2 / det;

    /* calculating matrix D */
    /* ------------------------------------------------------------------------ */
    young = materials[l]->YMod;
    thick = materials[l]->Thickness;
    pois  = materials[l]->PoissonsRatio;
    com   = young / (1.00 - pois * pois);

    dd[0][0] = com;
    dd[0][1] = dd[1][0] = com * pois;
    dd[0][2] = dd[2][0] = 0;
    dd[1][1] = com;
    dd[1][2] = dd[2][1] = 0;
    dd[2][2] = com * (1.00 - pois) / 2.00;
    

    /* calculating area */
    /* ------------------------------------------------------------------------ */
    sum = (init_length[l][0] + init_length[l][1] + init_length[l][2]) / 2.00;
    area_init = sqrt(sum * (sum - init_length[l][0]) * 
                           (sum - init_length[l][1]) *
                           (sum - init_length[l][2]));

    sum = (side_length[0] + side_length[1] + side_length[2]) / 2.00;
    area_real = sqrt(sum * (sum - side_length[0]) * 
                           (sum - side_length[1]) *
                           (sum - side_length[2]));


    /* check for wrinkling or compression in membrane */
    /* ------------------------------------------------------------------------ */
    /* */


    /* calculating element stiffness matrices */
    /* ------------------------------------------------------------------------ */
    /*          t
          K  = G * D * G * V
          -e   =   =   =
    */

    /* transpose G matrix */
    /* FIXME: not necessary, 'g' can be used !!! */
    
    for(ii = 0; ii < 3; ii++)
      for(jj = 0; jj < 3; jj++)
        gt[jj][ii] = g[ii][jj];
    
    
    /*                     t
       multiplication --- G  * D --- 
    */
    for(ii = 0; ii < 3; ii++)
       for(jj = 0; jj < 3; jj++)
       {
          k1[ii][jj] = 0.00;
          for(kk = 0; kk < 3; kk++)
             k1[ii][jj] += gt[ii][kk]*dd[kk][jj];
       }

    /*                     t
       multiplication --- G  * D * G --- 
    */
    for(ii = 0; ii < 3; ii++)
       for(jj = 0; jj < 3; jj++)
       {
          k2[ii][jj] = 0.00;
          for(kk = 0; kk < 3; kk++)
             k2[ii][jj] += k1[ii][kk]*g[kk][jj];
       }

    /* 
       Store the stiffness for Triang3 and Triang4 elements.
       This does not depend on whether the model is two or three
       dimensional, since this stiffness corresponds to the side
       extensions - natural stiffness.
    */
    if((type == 3) || (type == 4))
      for(i = 0; i < 3; i++)
        for(j = 0; j < 3; j++)
          Ke[l][i][j] = k2[i][j] * area_real * thick;


    /* calculating masses: */
    /* ------------------------------------------------------------------------ */
    /* at the moment the program using the Gershgorin bound to determine
       the nodal masses ( Mx = |Sxx| + |Sxy| + |Sxz| for each direction )
       Moreover the triangle is handled as three connected trusses and
       each truss's stiffness is added to the mass.
    */

    for(i = 0; i < 3; i++)
    {
      k_mass[i] = 0.00;
      for(j = 0; j < 3; j++)
        k_mass[i] += E_ABS(k2[i][j] * area_init * thick);
    }

    /* per node, per coordinate axis direction */
    for(i = 0; i < 3; i++)
    {
      n = nodes[l][i]-1;
      for(j = 0; j < dimension; j++)
      {
        mass[n][j] += k_mass[side_n_node[i][0]] * dl[side_n_node[i][0]][j] / side_length[side_n_node[i][0]];
        mass[n][j] += k_mass[side_n_node[i][1]] * dl[side_n_node[i][1]][j] / side_length[side_n_node[i][1]];
      }
    }

  }

} /* end of function -- dr_triang_stiffness_mass -- */

/* ###################################################################### */

/* Job:           calculating masses for all elements
   Reference:     Ref. 2 page 157
   Periodicity:   occasionally
   Notes:         only fix boundary conditions
*/
void dr_mass(MESH *mesh, DR_DATA *data)
{
  int i, k, n;
  double max;

  /* nullify the masses */
  /* ------------------------------------------------------------------------ */
  for(i = 0; i < mesh->NMeshPoints; i++)
    for(k = 0; k < data->Dimension; k++)
      data->Mass[i][k] = 0.00;

  /* here are the links */
  /* ------------------------------------------------------------------------ */
  /* Trusses, elastic control */
  if(mesh->NElemsLink1)
    dr_link_mass(data->Dimension, mesh->NElemsLink1, mesh->NodesLink1,
                 mesh->MeshPointCoords, mesh->MatsLink1,
                 data->InitLengthData->InitLengthLink1, 1, data->Mass);

  /* Cables, elastic control */
  if(mesh->NElemsLink2)
    dr_link_mass(data->Dimension, mesh->NElemsLink2, mesh->NodesLink2,
                 mesh->MeshPointCoords, mesh->MatsLink2,
                 data->InitLengthData->InitLengthLink2, 2, data->Mass);

  /* Fixed force link */
  if(mesh->NElemsLink3)
    dr_link_mass(data->Dimension, mesh->NElemsLink3, mesh->NodesLink3,
                 mesh->MeshPointCoords, mesh->MatsLink3,
                 data->InitLengthData->InitLengthLink3, 3, data->Mass);

  /* Fixed force density link */
  if(mesh->NElemsLink4)
    dr_link_mass(data->Dimension, mesh->NElemsLink4, mesh->NodesLink4,
                 mesh->MeshPointCoords, mesh->MatsLink4,
                 data->InitLengthData->InitLengthLink4, 4, data->Mass);

  /* here are the triangles */
  /* ------------------------------------------------------------------------ */
  /* Solid triangles */
  if(mesh->NElemsTriang3)
    dr_triang_stiffness_mass(data->Dimension, mesh->NElemsTriang3,
                             mesh->NodesTriang3, mesh->MeshPointCoords,
                             data->InitLengthData->InitLengthTriang3,
                             mesh->MatsTriang3, 3, 
                             data->KeTriang3, data->Mass);
  /* Membrane triangles */
  if(mesh->NElemsTriang4)
    dr_triang_stiffness_mass(data->Dimension, mesh->NElemsTriang4,
                             mesh->NodesTriang4, mesh->MeshPointCoords,
                             data->InitLengthData->InitLengthTriang4,
                             mesh->MatsTriang4, 4, 
                             data->KeTriang4, data->Mass);

  /* Constant stress triangles (isotrop & anysotrop) */
  if(mesh->NElemsTriang5)
    dr_triang_stiffness_mass(data->Dimension, mesh->NElemsTriang5,
                             mesh->NodesTriang5, mesh->MeshPointCoords,
                             data->InitLengthData->InitLengthTriang5,
                             mesh->MatsTriang5, 5, 
                             NULL, data->Mass);

#ifdef __E_MPI__

#ifdef COMM_LOG
  MPE_Log_event(3, 0, "start communicate masses");
#endif

  /* communicate the "pure" masses */
  if(data->CommReq)
  {
    data->CommData->Values[0] = data->Mass[0];
    CommunicateData(1, data->CommData);
  }

#ifdef COMM_LOG
  MPE_Log_event(4, 0, "end communicate masses");
#endif

#endif /* __E_MPI__ */

  /* now we have got all the "stiffnesses" in GV_mass, */
  /* here we can play with them to convert to fictitious masses*/
  /* ------------------------------------------------------------------------ */
  /* The greatest possible stiffness is used for each coordinate direction, see Ref. */

  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    max = (data->Mass[i][0] > data->Mass[i][1] ? 
           data->Mass[i][0] : data->Mass[i][1]   );
    if((data->Dimension == 3) && (data->Mass[i][2] > max))
      max = data->Mass[i][2];

    /* Well, if the mass is still zero, then none of the elements has referenced
     * this node. So here we make some mass to avoid division by zero later. With
     * this solution we are risking, that if there is a load on the node, then
     * it will fly away. Otherwise, because the residual will be zero, there will
     * not be velocity, so the node does not affect anything.
     */
    if(max < DR_ZERO)
    {
      max = 1.0;
    }

    for(k = 0; k < data->Dimension; k++)
      data->Mass[i][k] = max * E_SQUARE(mesh->TimeStep) / 2;
  }

  /* additional mass, if the user specified */
  /* ------------------------------------------------------------------------ */
  if(data->AddMassOption)
  {
    /* mass change for _ALL_ points and for _ALL_ directions */
    if(data->GlobalMassOption)
    {
      for(i = 0; i < mesh->NMeshPoints; i++)
        for(k = 0; k < data->Dimension; k++)
          data->Mass[i][k] *= data->GlobalMassMult;
    }

    /* mass change for _SPECIFIC_ points */
    for(i = 0; i < data->NAddMass; i++)
    {
      n = data->AddMassNode[i]-1;
      for(k = 0; k < data->Dimension; k++)
        data->Mass[n][k] *= data->AddMassMult[i][k];
    }
  }

#ifndef DISPLACEMENT_BC
  /* considering boundary conditions */
  /* ---------------------------------------------------------------------- */
  /* currently, only fixed boundaries */
  for(i = 0; i < mesh->NBCNodes; i++)
  {
    n = mesh->BCNodes[i]-1;
    for(k = 0; k < data->Dimension; k++)
    {
      if(mesh->BCTypes[i][k] == E_BC_NUM_FIXED)
        data->Mass[n][k] = DR_BIG_MASS;
    }
  }
#endif

} /* end of function -- dr_mass -- */

/* ###################################################################### */

/* Job:           calculating residuals for links
   Reference:     for geodesic strings  Ref. 1 page 22-23
   Periodicity:   frequently
   Notes:         
*/
static void dr_link_res(int dimension, int nelems, int **nodes, double **coords, 
                        MAT **material, double *init_length, int type,
                        double **residual)
{
  int i, k;
  int n1, n2;
  double link_length, young, area, tt, dr, length_xy;
  double dl[3];

  tt = 0;
  for(i = 0; i < nelems; i++)
  {
    n1 = nodes[i][0]-1;
    n2 = nodes[i][1]-1;
    link_length = 0.0;
    for(k = 0; k < dimension; k++)
    {
      dl[k] = (coords[n2][k] - coords[n1][k]);
      link_length += E_SQUARE(dl[k]);
    }
    link_length = sqrt(link_length);

    switch(type)
    {
      case 1:                           /* Cables */
        young = material[i]->YMod;
        area  = material[i]->dParam1;
        tt = (link_length - init_length[i]) * young * area / init_length[i];
        if(tt < 0)
        {
          tt = 0;
        }
        break;
      case 2:                           /* Trusses */
        young = material[i]->YMod;
        area  = material[i]->dParam1;
        tt = (link_length - init_length[i]) * young * area / init_length[i];
        break;
      case 3:                                   /* Fixed force link */
        tt = material[i]->dParam2;
        /* stresses are defined in the x-y plane for a 3D analysis;
           projected stresses
         */
        if(material[i]->iParam5 == 1 && dimension == 3)
        {
          /* length of the link element in the xy plane */
          length_xy = sqrt(E_SQUARE(dl[0]) + E_SQUARE(dl[1]));
          tt *= link_length / length_xy;
        }
        break;
      case 4:                                   /* Fixed force density link */
        tt = material[i]->dParam2 * link_length;
        break;
      case 5:                                   /* G-string */
        young = material[i]->YMod;
        tt = (link_length - init_length[i]) * young / init_length[i];
        break;
      default:
        e_GeneralError("Unknown LINK type !");
    }

    /* if compressed boundary */
    if(material[i]->iParam4 == -1)
      tt *= -1;

    for(k = 0; k < dimension; k++)
    {
      dr = dl[k] * tt / link_length;

      if((type == 5) && (material[i]->iParam1))
        residual[n1][k] = 0.00;
      else
        residual[n1][k] += dr;
      if((type == 5) && (material[i]->iParam2))
        residual[n2][k] = 0.00;
      else
        residual[n2][k] -= dr;
    }
  }
} /* end of function -- dr_link_res -- */

/* ###################################################################### */

/* Job:           calculating residuals for triangles
   Reference:     Ref. 1 page 15-21
   Periodicity:   frequently
   Notes:         
*/
static void dr_triang_res(int dimension, int nelems, int **nodes, double **coords, 
                          MAT **material, double **init_length, double ***Ke, 
                          int type, double **residual)
{
  int l, k, i, j, n1, n2;
  double t, dr, sum, area;
  double thick = 0, stress_x = 0, stress_y = 0;
  double extension[3], angle[3], side_length[3], dl[3][3];
  int    side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};
  double length_xy[3], angle_xy[3], sum_xy,  area_xy;

  for(l = 0; l < nelems; l++)
  {
    for(i = 0; i < 3; i++)
    {
      n1 = nodes[l][side_n_node[i][0]]-1;
      n2 = nodes[l][side_n_node[i][1]]-1;
      side_length[i] = 0.00;
      for(j = 0; j < dimension; j++)
      {
        dl[i][j] = (coords[n2][j] - coords[n1][j]);
        side_length[i] += E_SQUARE(dl[i][j]);
      }
      side_length[i] = sqrt(side_length[i]);

      length_xy[i] = sqrt(E_SQUARE(dl[i][0]) + E_SQUARE(dl[i][1]));
    }

    sum = (side_length[0] + side_length[1] + side_length[2]) / 2.00;
    area = sqrt(sum * (sum - side_length[0]) * 
                      (sum - side_length[1]) *
                      (sum - side_length[2])   );

    sum_xy  = (length_xy[0] + length_xy[1] + length_xy[2]) / 2.00;
    area_xy = sqrt(sum_xy * (sum_xy - length_xy[0]) * 
                            (sum_xy - length_xy[1]) *
                            (sum_xy - length_xy[2])   );

    switch(type)
    {
      case 3: case 4:                    /* Solid & membrane triangles */
        for (k = 0; k < 3; k++)
        {
          extension[k] = (side_length[k]-init_length[l][k]) / init_length[l][k]; 
        }
        break;
      case 5:                            /* Constant stress triangles */
        /* projected stresses */
        if(material[l]->iParam5 == 1 && dimension == 3)
        {
          sum_xy = E_SQUARE(length_xy[0]) + 
                   E_SQUARE(length_xy[1]) + 
                   E_SQUARE(length_xy[2]);
          angle_xy[0] = acos((sum_xy - 2.00 * E_SQUARE(length_xy[0])) /
                             (2.00 * length_xy[1] * length_xy[2])       );
          angle_xy[1] = acos((sum_xy - 2.00 * E_SQUARE(length_xy[1])) /
                             (2.00 * length_xy[0] * length_xy[2])       );
          angle_xy[2] = acos((sum_xy - 2.00 * E_SQUARE(length_xy[2])) /
                             (2.00 * length_xy[0] * length_xy[1])       );
        }
        else
        {
          sum = E_SQUARE(side_length[0]) + 
                E_SQUARE(side_length[1]) + 
                E_SQUARE(side_length[2]);
          angle[0] = acos((sum - 2.00 * E_SQUARE(side_length[0])) /
                            (2.00 * side_length[1] * side_length[2]));
          angle[1] = acos((sum - 2.00 * E_SQUARE(side_length[1])) /
                            (2.00 * side_length[0] * side_length[2]));
          angle[2] = acos((sum - 2.00 * E_SQUARE(side_length[2])) /
                            (2.00 * side_length[0] * side_length[1]));
        }
        thick = material[l]->Thickness;
        stress_x = material[l]->dParam1;
        stress_y = material[l]->dParam2;
        break;
      default:
        e_GeneralError("Unknown TRIANG type !");
    }

    for(i = 0; i < 3; i++)
    {
      t = 0.00;

      switch(type)
      {
        case 3: case 4:                  /* Solid & membrane triangles */
          for(j = 0; j < 3; j++)
          {
            t += (Ke[l][i][j] * extension[j]);
          }
          /*  force is related to the height of the triangle */
          t /= (side_length[i]);          
          break;
        case 5:                          /* Constant stress triangles */
          /* stresses are defined in the x-y plane for a 3D analysis;
             projected stresses
           */
          if(material[l]->iParam5 == 1 && dimension == 3)
          {
            /* length of the link element in the xy plane */
            t = stress_y * thick * length_xy[i] * cos(angle_xy[i]) / 
                ( 2.00 * (sin(angle_xy[i]) + DR_ZERO) );
            if(E_ABS(stress_x - stress_y) > DR_ZERO)
            {
              if(i == 0)
                t += (area_xy * thick * (stress_x - stress_y) / length_xy[i]);
            }
            t *= side_length[i] / length_xy[i];
          }
          /* normal case */
          else
          {
            /* cotan is used and protected agains zero division */
            t = stress_y * thick * side_length[i] * cos(angle[i]) / 
                ( 2.00 * (sin(angle[i]) + DR_ZERO) );
            /* take into account anisotropy */
            /* FIXME: what about when no g-string on the side */
            if(E_ABS(stress_x - stress_y) > DR_ZERO)
            {
              if(i == 0)
                t += (area * thick * (stress_x - stress_y) / side_length[i]);
            }
          }
          break;
        default:
          e_GeneralError("Unknown TRIANG type !");
      }

      n1 = nodes[l][side_n_node[i][0]]-1;
      n2 = nodes[l][side_n_node[i][1]]-1;

      for(k = 0; k < dimension; k++)
      {
        dr = dl[i][k] * t / side_length[i];
        residual[n1][k] += dr;
        residual[n2][k] -= dr;
      }
    }
    
  }

} /* end of function -- dr_triang_res -- */

/* ###################################################################### */


/* Job:           calculating residuals for all elements
   Reference:
   Periodicity:   frequently
   Notes:         concentrated forces considered here
*/
void dr_residual(MESH *mesh, DR_DATA *data)
{
  double rn;
  int i, k, n;

  /* nullify the residuals */
  /* ------------------------------------------------------------------------ */
  for(i = 0; i < mesh->NMeshPoints; i++)
    for(k = 0; k < data->Dimension; k++)
      data->Residual[i][k] = 0.00;

  /* geodesic strings */
  /* ------------------------------------------------------------------------ */
  if(mesh->NElemsLink5)
  {
    dr_link_res(data->Dimension, mesh->NElemsLink5,
                mesh->NodesLink5, mesh->MeshPointCoords, mesh->MatsLink5, 
                data->InitLengthData->InitLengthLink5, 5, data->Residual);


    /* when it is specified, the g-string behaves as a cable element, well almost */
    if((data->ProjectTask & TASK_WITHOUT_G_STRING) == 0)
    {
      /* subtract normal components from the global residuals */
      if(data->Dimension == 3)
      {
        for(i = 0; i < mesh->NMeshPoints; i++)
        {
          rn = 0.00;
 
          for(k = 0; k < 3; k++)
            rn += data->Normal[i][k] * data->Residual[i][k];
 
          for(k = 0; k < 3; k++)
            data->Residual[i][k] -= (rn * data->Normal[i][k]);
        }
      }
    }


  }

  /* links */
  /* ------------------------------------------------------------------------ */
  if(mesh->NElemsLink1)
    dr_link_res(data->Dimension, mesh->NElemsLink1,
                mesh->NodesLink1, mesh->MeshPointCoords, mesh->MatsLink1,
                data->InitLengthData->InitLengthLink1, 1, data->Residual);

  if(mesh->NElemsLink2)
    dr_link_res(data->Dimension, mesh->NElemsLink2,
                mesh->NodesLink2, mesh->MeshPointCoords, mesh->MatsLink2, 
                data->InitLengthData->InitLengthLink2, 2, data->Residual);

  if(mesh->NElemsLink3)
    dr_link_res(data->Dimension, mesh->NElemsLink3,
                mesh->NodesLink3, mesh->MeshPointCoords, mesh->MatsLink3, 
                NULL, 3, data->Residual);

#ifdef __TIME_ELEM_CALC__
  StartTimer();
#endif

  if(mesh->NElemsLink4)
    dr_link_res(data->Dimension, mesh->NElemsLink4,
                mesh->NodesLink4, mesh->MeshPointCoords, mesh->MatsLink4, 
                NULL, 4, data->Residual);

#ifdef __TIME_ELEM_CALC__
  TimeLink4 += ElapsedTime();
#endif

  /* triangles */
  /* ------------------------------------------------------------------------ */
  if(mesh->NElemsTriang3)
    dr_triang_res(data->Dimension, mesh->NElemsTriang3,
                  mesh->NodesTriang3, mesh->MeshPointCoords, NULL, 
                  data->InitLengthData->InitLengthTriang3, 
                  data->KeTriang3, 3, data->Residual);

  if(mesh->NElemsTriang4)
    dr_triang_res(data->Dimension, mesh->NElemsTriang4, 
                  mesh->NodesTriang4, mesh->MeshPointCoords, NULL, 
                  data->InitLengthData->InitLengthTriang4, 
                  data->KeTriang4, 4, data->Residual);


  /* pressure loading, per element distributed to the nodes */
  /* this should be here, before the parallel communication */
  if(data->PressureOption || data->HydPressOption)
  {
    for(i = 0; i < mesh->NMeshPoints; i++)
      for(k = 0; k < data->Dimension; k++)
        data->Residual[i][k] += data->Pressure[i][k];
  }

  
#ifdef __TIME_ELEM_CALC__
  StartTimer();
#endif

  if(mesh->NElemsTriang5)
    dr_triang_res(data->Dimension, mesh->NElemsTriang5,
                  mesh->NodesTriang5, mesh->MeshPointCoords, mesh->MatsTriang5, 
                  NULL, NULL, 5, data->Residual);

#ifdef __TIME_ELEM_CALC__
  TimeTriang5 += ElapsedTime();
#endif

  /* communicate the internal forces */
#ifdef __E_MPI__ 

#ifdef COMM_LOG
  MPE_Log_event(5, 0, "start communicate residuals");
#endif
  if(data->CommReq)
  {
    data->CommData->Values[0] = data->Residual[0];
    CommunicateData(1, data->CommData);
  }

#ifdef COMM_LOG
  MPE_Log_event(6, 0, "end communicate residuals");
#endif

#endif

  /* concentrated forces */
  for(i = 0; i < mesh->NLoadedNodes; i++)
  {
    n = mesh->LoadedNodes[i]-1;
    for(k = 0; k < data->Dimension; k++)
      data->Residual[n][k] += mesh->Loads[i][k];
  }

  /* compressed boundaries */
  for(i = 0; i < data->NCompNodes; i++)
  {
    n = data->CompNodes[i]-1;
    for(k = 0; k < data->Dimension; k++)
      data->Residual[n][k] *= -1;
  }


} /* end of function -- dr_residuals -- */


/* *********************************************************************** */
/*                             Iteration                                   */
/* *********************************************************************** */


#ifdef DISPLACEMENT_BC
static void dr_disp_bc(MESH *mesh, DR_DATA *data, int inode, double *disp)
{
  int    idim, ibc, n;
  double normal[3];
  double rn, length;

  /* check boundary conditions */
  for(ibc = 0; ibc < mesh->NBCNodes; ibc++)
  {
    n = mesh->BCNodes[ibc]-1;
    /* point is boundary */
    if(n == inode)
    {
      /* there is any fixity in x, y, z direction */
      if(   (mesh->BCTypes[ibc][0] == E_BC_NUM_FIXED)
         || (mesh->BCTypes[ibc][1] == E_BC_NUM_FIXED)
         || (mesh->BCTypes[ibc][2] == E_BC_NUM_FIXED) )
      {
        for(idim = 0; idim < data->Dimension; idim++)
        {
          /* if it is fixed remove displacement and velocity !!! */
          if(mesh->BCTypes[ibc][idim] == E_BC_NUM_FIXED)
          {
            disp[idim] = 0.0;
            data->Velocity[inode][idim] = 0.0;
          }
        }
      }
      /* there is a plane fixity */
      else if(   (mesh->BCTypes[ibc][0] == E_BC_NUM_PLANE)
              || (mesh->BCTypes[ibc][1] == E_BC_NUM_PLANE)
              || (mesh->BCTypes[ibc][2] == E_BC_NUM_PLANE) )
      {
        length = 0.0;
        for(idim = 0; idim < data->Dimension; idim++)
        {
          assert(mesh->BCTypes[ibc][idim] == E_BC_NUM_PLANE);
          
          normal[idim] = mesh->BCDispls[ibc][idim];
          length += normal[idim] * normal[idim];
        }
        assert(E_ABS(length) > DR_ZERO);
        
        /* normalise the length of the normal vector */
        length = sqrt(length);
        for(idim = 0; idim < data->Dimension; idim++)
        {
          normal[idim] /= length;
        }
        
        /* remove the displacement perpendicular to the plane */
        rn = 0.0;
        for(idim = 0; idim < data->Dimension; idim++)
        {
          rn += normal[idim] * disp[idim];
        }
        for(idim = 0; idim < data->Dimension; idim++)
        {
          disp[idim] -= rn * normal[idim];
        }
        /* remove the velocity perpendicular to the plane */
        rn = 0.0;
        for(idim = 0; idim < data->Dimension; idim++)
        {
          rn += normal[idim] * data->Velocity[inode][idim];
        }
        for(idim = 0; idim < data->Dimension; idim++)
        {
          data->Velocity[inode][idim] -= rn * normal[idim];
        }
      }
      break;
    }
  }
}
#endif





/* initial iteration step */

static void dr_init(MESH *mesh, DR_DATA *data)
{
  int    inode, idim;
  double disp[3];

  for(inode = 0; inode < mesh->NMeshPoints; inode++)
  {
    for(idim = 0; idim < data->Dimension; idim++)
    {
      data->Velocity[inode][idim] = 
        data->B * data->Residual[inode][idim] * mesh->TimeStep /
        (data->Mass[inode][idim] * (1 + data->A));
      
      disp[idim] = mesh->TimeStep * data->Velocity[inode][idim];
    }

#ifdef DISPLACEMENT_BC
    dr_disp_bc(mesh, data, inode, disp);
#endif

    for(idim = 0; idim < data->Dimension; idim++)
    {
      mesh->MeshPointCoords[inode][idim] += disp[idim];
    }
  }

} /* end of function -- dr_init -- */


/* ------------------------------------------------------------------------ */


/* iteration step */

static void dr_step(MESH *mesh, DR_DATA *data)
{
  int    inode, idim;
  double disp[3];
  
  data->CurrentEnergy = 0.00;

  for(inode = 0; inode < mesh->NMeshPoints; inode++)
  {
    /* calculate new velocity and displacements */
    for(idim = 0; idim < data->Dimension; idim++)
    {
      data->Velocity[inode][idim] = 
          data->Velocity[inode][idim] * data->A +
          data->Residual[inode][idim] * mesh->TimeStep * data->B /
          data->Mass[inode][idim];
      
      disp[idim] = mesh->TimeStep * data->Velocity[inode][idim];
    }

#ifdef DISPLACEMENT_BC
    dr_disp_bc(mesh, data, inode, disp);
#endif
        
    for(idim = 0; idim < data->Dimension; idim++)
    {
      mesh->MeshPointCoords[inode][idim] += disp[idim];

#ifdef __E_MPI__
      if(data->CommEnergy[inode])
#endif
        data->CurrentEnergy += 
          E_SQUARE(data->Velocity[inode][idim]) * data->Mass[inode][idim];
    }
  }
} /* end of function -- dr_step -- */


/* ------------------------------------------------------------------------ */


/* energy peak occured */
static void dr_peak(MESH *mesh, DR_DATA *data)
{
  int i, k;
  double disp[3];

  data->CurrentEnergy = 0.00;
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    for(k = 0; k < data->Dimension; k++)
    {
      disp[k] = data->Velocity[i][k] * mesh->TimeStep * 1.5 -
                E_SQUARE(mesh->TimeStep) * data->Residual[i][k] / (2 * data->Mass[i][k]);
      /* this is actually recalculated in dr_init */
      data->Velocity[i][k] = mesh->TimeStep * data->Residual[i][k] / (2 * data->Mass[i][k]);
    }

#ifdef DISPLACEMENT_BC
    dr_disp_bc(mesh, data, i, disp);
#endif

    for(k = 0; k < data->Dimension; k++)
    {
      mesh->MeshPointCoords[i][k] -= disp[k];
    }

  }

  /* if it is 3D we need normals for g-strings
     if it is 2D and pressure is specified we need them again */
  if(   (   (data->Dimension == 3) 
         && (   mesh->NElemsTriang3
             || mesh->NElemsTriang4
             || mesh->NElemsTriang5)   )
     || (   (data->Dimension == 2) 
         && (data->PressureOption) )
    )
  {
    dr_normal(mesh, data);
  }

  /* it is enough to check that we want pressure calculation,
     since the normals either zero or setup properly by now */
  if(data->PressureOption || data->HydPressOption)
  {
    dr_pressure(mesh, data);
  }

  dr_mass(mesh, data);
  dr_residual(mesh, data);
  dr_init(mesh, data);

} /* end of function -- dr_peak -- */


/* *********************************************************************** */
/*                                 Main                                    */
/* *********************************************************************** */

int dr_main(MESH *mesh, DR_DATA *data)
{
  int    it_end, it_error;
  int    i, j, iteration = 0;
  double prev_energy = 0;
  long   beg_t, end_t;

#ifdef __E_MPI__
  double  begin_time, end_time;
  double  energy;
#endif

  /* no timestep, so just return */
  if(   (mesh->NExternalTimeSteps == 0)
     && (mesh->NInternalTimeSteps == 0) )
  {
    return(FALSE);
  }

  /* if it is 3D we need normals for g-strings
     if it is 2D and pressure is specified we need them again */
  if(   (   (data->Dimension == 3) 
         && (   mesh->NElemsTriang3
             || mesh->NElemsTriang4
             || mesh->NElemsTriang5)   )
     || (   (data->Dimension == 2) 
         && (data->PressureOption) )
    )
  {
    dr_normal(mesh, data);
  }

  /* it is enough to check that we want pressure calculation,
     since the normals either zero or setup properly by now */
  if(data->PressureOption || data->HydPressOption)
  {
    dr_pressure(mesh, data);
  }

  dr_mass(mesh, data);
  dr_residual(mesh, data);
  dr_init(mesh, data);

  it_end = FALSE;
  it_error = FALSE;

  for(i = 1; (   (i <= mesh->NExternalTimeSteps) 
              && (it_end == FALSE)
              && (it_error == FALSE)             ); i++)
  {

#ifdef __E_MPI__
    begin_time = MPI_Wtime();
#endif

    beg_t = GetTime();

    for(j = 1; (   (j <= mesh->NInternalTimeSteps)
                && (it_end == FALSE)
                && (it_error  == FALSE)            ); j++)
    {
      iteration = (i-1)*mesh->NInternalTimeSteps+j;
      /* caluclate residuals */
      dr_residual(mesh, data);
      
      /* advance structure by a step */
      dr_step(mesh, data);

#ifdef __E_MPI__
      if(data->CommReq)
      {

#ifdef COMM_LOG
        MPE_Log_event(7, 0, "start communicate energy");
#endif

        MPI_Allreduce(&(data->CurrentEnergy), &energy, 1, 
                      MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

#ifdef COMM_LOG
        MPE_Log_event(8, 0, "end communicate energy");
#endif

        data->CurrentEnergy = energy;
      }
#endif /* __E_MPI__ */

      if(mesh->DampingFactor < 0.0001)
      {
        /* kinetic damping */
        if(data->CurrentEnergy < prev_energy)
        {

#ifdef __E_MPI__
          if(   (data->CommData->Par_ProcID == 1)
             && data->CommReq )
#endif
            printf("\nEnergy peak %1.15e at iteration %d", 
                   data->CurrentEnergy, iteration);

          /* if stop for convergence is defined */
          if(   data->StopAtConv
             && (data->CurrentEnergy < data->ConvCrit) )
          {
            it_end = TRUE;
#ifdef __E_MPI__
            if(   (data->CommData->Par_ProcID == 1)
               && data->CommReq )
#endif
            {
              printf("\n*****");
              printf("\n***** Convergence at %d. iteration *****", iteration);
              printf("\n*****");
            }
          }
          else
          {
            dr_peak(mesh, data);
          }
        }
        prev_energy = data->CurrentEnergy;
      }
      /* viscous damping */
      else
      {
        if(   data->StopAtConv
           && (data->CurrentEnergy < data->ConvCrit) )
        {
          it_end = 1;
#ifdef __E_MPI__
          if(   (data->CommData->Par_ProcID == 1)          
             && data->CommReq )
#endif
          {
            printf("\n*****");
            printf("\n***** Convergence at %d. iteration *****", iteration);
            printf("\n*****");
          }
        }
      }


      /* always stop for divergence */
      if(data->CurrentEnergy > DR_MAX_ENERGY)
      {
        it_end = 1;
        it_error = 1;
        printf("\n\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        printf("\n!!!!! Solution is divergent ! !!!!!");
        printf("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
        fflush(stdout);
      }

    }

#ifdef __E_MPI__
    end_time = MPI_Wtime();
#endif

    end_t = GetTime();

#ifdef __E_MPI__
    if(   (data->CommData->Par_ProcID == 1)
       && data->CommReq )
    {
#endif
      printf("\n----- Energy is %1.15e at iteration %d", 
        data->CurrentEnergy, iteration);

#ifdef __E_MPI__
      printf("\n>>>>> execution time is: %7.2f[sec] by MPI", 
        (end_time - begin_time));
#endif

      printf("\n>>>>> execution time is: %7.2f[sec] by OS", 
        (double)(end_t-beg_t)/1000);

#ifdef __E_MPI__
    }
#endif

  }


#ifdef __TIME_ELEM_CALC__
  {
    double calc;
    if(mesh->NElemsLink4)
    {
      calc = (double)(TimeLink4)/1000;
      calc /= (mesh->NElemsLink4);
 
      printf("\nCalculation time for LINK4 elements is: %7.4f[sec]", calc);
    }
 
    if(mesh->NElemsTriang5)
    {
      calc = (double)(TimeTriang5)/1000;
      calc /= (mesh->NElemsTriang5);
 
      printf("\nCalculation time for TRIANG5 elements is: %7.4f[sec]", calc);
    }
  }
#endif

  return(it_error);
}





