#include "dr_comm.h"
#include "dr_data.h"

#ifdef __E_MPI__
#include "mpi.h"
#endif

void SetupSubdomains(int ProcID, int NSubdomains,
                     DR_DATA *Gdata, MESH *Gmesh, 
                     DR_DATA **data, MESH **mesh)
{
  int         i_subdom, i, j, Prev_SD;
  int         NInterSubDomNodes, *InterSubDomNodes;
  int        *NNodeSubDoms, **NodeSubDoms;
  char        SubdomTitle[E_LINE_MAX+1];

  int send, receive;
  MPI_Status status;

  /* master does the data distribution */
  if(ProcID == 1)
  {
    /* create info about interface nodes */
    e_GetInterSubDomNodes(Gmesh, &NInterSubDomNodes, &InterSubDomNodes,
			                    &NNodeSubDoms, &NodeSubDoms);

    /* check the order of the sub-domains for the interface nodes;
       this is some kind of sanity check, and only certain communication schemes
       require it, if at all; leave it for the moment */
    for(i = 0; i < NInterSubDomNodes; i++)
    {
      Prev_SD = NodeSubDoms[i][0];
      for(j = 1; j < NNodeSubDoms[i]; j++)
      {
        if(Prev_SD >= NodeSubDoms[i][j])
        {
          e_GeneralError("\n\nWrong Subdom order for %d-th IntNode which is node %d !\n", 
                         i, InterSubDomNodes[i]);
        }
        Prev_SD = NodeSubDoms[i][j];
      }
    }


    /* generate and distribute data to all subdomains */
    for(i_subdom = NSubdomains; i_subdom > 0; i_subdom--)
    {
      /* create the subdomain mesh */
      sprintf(SubdomTitle, "Subdomain %d", i_subdom);
      e_CreateSubMesh(Gmesh, mesh, SubdomTitle, i_subdom);

      /* create subdomain DR_DATA structure */
      CreateSubDrDataInit(Gdata, Gmesh, data, *mesh,
                          i_subdom, NInterSubDomNodes, 
                          InterSubDomNodes, NNodeSubDoms, 
                          NodeSubDoms);

      /* setup information for communication */
      /* DD_TYPE_COMM, AL_TYPE_COMM */
      SetupSymCommDataInit((*data)->Dimension, NSubdomains, i_subdom, 
                           NInterSubDomNodes, InterSubDomNodes,
                           NNodeSubDoms, NodeSubDoms,
                           &((*data)->CommData));

      /* if not root, communicate with other subdomains */
      if(i_subdom > 1)
      {
        /* FIXME: MPI glitch */
        MPI_Recv(&receive, 1, MPI_INT, i_subdom-1, 2101, MPI_COMM_WORLD, &status);
        /* send the subdomain mesh */
        e_MPI_Send_MeshData(*mesh, i_subdom - 1);
        
        /* send DR_DATA structure */
        SendDrDataInit(*data, *mesh, i_subdom-1);

        /* send communication information */
        /* DD_TYPE_COMM, AL_TYPE_COMM */
        SendCommDataInit(NSubdomains, (*data)->CommData, i_subdom-1);

        /* free-up subdomain mesh and DR_DATA structures */
        e_FreeMeshData(*mesh);
        FreeSymCommDataInit(NSubdomains, (*data)->CommData);
        FreeDrDataInit(*data);

      }
    }
    
    /* all data is gone, this information is not necessary further more */
    e_FreeInterSubDomNodes(NInterSubDomNodes, InterSubDomNodes,
			                     NNodeSubDoms, NodeSubDoms);
  }
  else /* subdomains receive information from master process */
  {
    /* FIXME: MPI glitch */
    send = ProcID;
    MPI_Send(&send, 1, MPI_INT, 0, 2101, MPI_COMM_WORLD);

    /* receive the subdomain mesh */
    e_MPI_Recv_MeshData(mesh, 0);

    /* receive DR_DATA for the subdomain */
    RecvDrDataInit(data, *mesh, 0);
    
    /* receive communication related information */
    /* DD_TYPE_COMM, AL_TYPE_COMM */
    RecvCommDataInit(NSubdomains, &((*data)->CommData), 0);
  }

  /* ############################# */
  /* every subdomain executes this */
  /* ############################# */
  (*data)->CommData->Par_ProcID      = ProcID;
  (*data)->CommData->Par_NSubdomains = NSubdomains;

  /* DD_TYPE_COMM, AL_TYPE_COMM */
  ConvertSymCommDataInit(*mesh, (*data)->CommData, (*data)->Dimension);
  SetupCommDataFinal((*data)->CommData);

  (*data)->CommReq = TRUE;

  if(ProcID == 1)
  {
    printf("\n\nDynamic relaxation uses the %s communication scheme.", COMM_NAME);
  }

}

/* --------------------------------------------------------------------------- */
/* includes the functions for the different communication schemes */

#ifdef DD_TYPE_COMM
#include "dr_comm_dd.c"
#endif

#ifdef AL_TYPE_COMM
#include "dr_comm_al.c"
#endif

#ifdef BU_TYPE_COMM
#include "dr_comm_bu.c"
#endif



