#ifdef DD_TYPE_COMM

/* 
 * These functions are required for the symmetric DD communication and
 * they use the "comm_vector" implementation. This means, that even matrices
 * are accessed as vectors.
 */


void SetupSymCommDataInit(int dimension, int NSubdomains, int i_subdom, 
                          int NInterSubDomNodes, int *InterSubDomNodes,
                          int *NNodeSubDoms, int **NodeSubDoms,
                          COMM_VECTOR_DATA **data)
{
  int i, j, k, ns;

  (*data) = e_calloc(1, sizeof(COMM_VECTOR_DATA) );

  (*data)->Par_NElemSend = e_alloc_i_vect(NSubdomains);
  (*data)->Par_NElemRecv = e_alloc_i_vect(NSubdomains);
  (*data)->Par_ElemSend  = e_init_frag_i_matr(NSubdomains);
  (*data)->Par_ElemRecv  = e_init_frag_i_matr(NSubdomains);

  for(i = 0; i < NInterSubDomNodes; i++)
  {
    for(j = 0; j < NNodeSubDoms[i]; j++)
    {
      /* this node is in the current subdomain */
      if(i_subdom == NodeSubDoms[i][j])
      {
        for(k = 0; k < NNodeSubDoms[i]; k++)
        {
          ns = NodeSubDoms[i][k]-1;
          /* let's not add ourselves */
          if(ns+1 != i_subdom)
          {
            (*data)->Par_ElemSend[ns] = 
              e_incr_i_vect((*data)->Par_NElemSend[ns], (*data)->Par_ElemSend[ns], ALLOC_SIZE*dimension);
            (*data)->Par_ElemSend[ns][(*data)->Par_NElemSend[ns]] = InterSubDomNodes[i];
            (*data)->Par_NElemSend[ns] += dimension;

            (*data)->Par_ElemRecv[ns] = 
              e_incr_i_vect((*data)->Par_NElemRecv[ns], (*data)->Par_ElemRecv[ns], ALLOC_SIZE*dimension);
            (*data)->Par_ElemRecv[ns][(*data)->Par_NElemRecv[ns]] = InterSubDomNodes[i];
            (*data)->Par_NElemRecv[ns] += dimension;
          }
        }
        break;
      }
    }
  }
}


void FreeSymCommDataInit(int NSubdomains, COMM_VECTOR_DATA *data)
{
  int i;

  for(i = 0; i < NSubdomains; i++)
  {
    if(data->Par_NElemSend[i])
    {
      e_free(data->Par_ElemSend[i]);
    }
    if(data->Par_NElemRecv[i])
    {
      e_free(data->Par_ElemRecv[i]);
    }
  }
  e_free(data->Par_NElemSend);
  e_free(data->Par_NElemRecv);
  data->Par_NElemSend = NULL;
  data->Par_NElemRecv = NULL;
  data->Par_ElemSend  = NULL;
  data->Par_ElemRecv  = NULL;

  e_free(data);
}



void ConvertSymCommDataInit(MESH *mesh, COMM_VECTOR_DATA *data, 
                            int dimension)
{
  int i, j, k, node, index;

  for(i = 0; i < data->Par_NSubdomains; i++)
  {
    /* any element ?, empty communication is deleted later */
    if(data->Par_NElemSend[i] > 0)
    {
      for(j = 0; j < data->Par_NElemSend[i]; j+=dimension)
      {
        node = data->Par_ElemSend[i][j];
        for(index = 0; index < mesh->NMeshPoints; index++)
        {
          if(mesh->GNIs[index] == node)
            break;
        }
        /* check that we have the node locally in the mesh */
        if(index == mesh->NMeshPoints)
        {
          e_GeneralError("%d proc: communication node %d found in mesh!", 
                         data->Par_ProcID, node);
        }
        /* check for communication symmetry */
        if(node != data->Par_ElemRecv[i][j])
        {
          e_GeneralError("%d proc: communication is not symmetric for node %d",
                         data->Par_ProcID, node);
        }
        
        /* setup access to the matrices as a vector */
        for(k = 0; k < dimension; k++)
        {
          data->Par_ElemSend[i][j+k] = index * dimension+k+1;
          data->Par_ElemRecv[i][j+k] = index * dimension+k+1;
        }
        
      }
    }
  }

}

/* include the other communication functions */
#include "comm_vector.c"

#endif




