#define DR_CUT

#include "config.h"
#include <math.h>
#include <assert.h>
#include "dr_cut.h"
#include "dr_data.h"
#include "dr_calc.h"
#include "adjmemb.h"
#include "getvertices.h"
#include "nodelem.h"

#include "dr_stress.h"

#ifdef __E_MPI__
#include "mpi.h"
#endif


/* #define __DEBUG_2__ */
/* #define __DEBUG_3__ */


static void ClearMaterial(MESH *mesh);
static void ReplaceMaterial(MESH *mesh, DR_DATA *data);

static void RemoveLoads(MESH *mesh);
static void RemoveBCs(MESH *mesh);
static int AddAdjTriangToStrip(MESH *mesh, int elem_index, 
                               int **strip_triangs, int *len_tri_list, 
                               int strip_index);
static void AddAllTriangsToStrip(MESH *mesh, DR_DATA *data, ADJDATA *adata,
                                 int NElems, int **Nodes, 
                                 int *Subdoms, int ElemType,
                                 CUTTING_PATTERN *cdata);
static void CreateStrips(MESH *mesh, DR_DATA *data, ADJDATA *adata,
                         CUTTING_PATTERN *cdata);
static void CreateLinkMesh(MESH *mesh, CUTTING_PATTERN *cdata);
static void CreateCutPatMesh(MESH *mesh, DR_DATA *data, 
                             CUTTING_PATTERN **cdata);
static int FindMatchingElements(int index, int n, int *list);
static void StripInitLengthFromStress(DR_DATA *Gdata, MESH *Gmesh,
                                      DR_DATA *data, MESH *mesh);
static void ConvertMeshForCutPat(MESH *mesh);
static void SearchMaxDistNode(MESH *mesh, int *n1, int *n2, double *direction);
static void MakeStaticDetMesh(MESH *mesh);
static void DoCreateCutPat(MESH *mesh, DR_DATA *data,
                           CUTTING_PATTERN *cdata);
#ifdef __E_MPI__
static void StripsToMaster(CUTTING_PATTERN *cdata, int ProcID, int NSubdomains);
static void StripAssemble(G_CUTPAT *cdata, int ProcID, int NSubdomains);
static void StripsFromMaster(int ProcID, int NSubdomains, CUTTING_PATTERN *cdata);
static void StripToSubdoms(int ProcID, int NSubdomains, CUTTING_PATTERN *cdata);
static void MergeStrips(MESH *component_mesh, MESH *final_mesh);
#endif



/* ------------------------------------------------------------------------- */
/*                       If material change is required                      */
/* ------------------------------------------------------------------------- */

/* !!!!! It should be an E-lib function !!!!! */
static void ClearMaterial(MESH *mesh)
{
  int i, j;
  
  for(i = 0; i < mesh->NMats; i++)
  {
    mesh->Props[i].MatType = 0;

    mesh->Props[i].Density       = 0.0;
    mesh->Props[i].YMod          = 0.0;
    mesh->Props[i].YMod_x        = 0.0;
    mesh->Props[i].YMod_y        = 0.0;
    mesh->Props[i].YMod_z        = 0.0;
    mesh->Props[i].PoissonsRatio = 0.0;
    mesh->Props[i].Thickness     = 0.0;

    mesh->Props[i].iParam1 = 0;
    mesh->Props[i].iParam2 = 0;
    mesh->Props[i].iParam3 = 0;
    mesh->Props[i].iParam4 = 0;
    mesh->Props[i].iParam5 = 0;
    mesh->Props[i].iParam6 = 0;

    mesh->Props[i].dParam1  = 0.0;
    mesh->Props[i].dParam2  = 0.0;
    mesh->Props[i].dParam3  = 0.0;
    mesh->Props[i].dParam4  = 0.0;
    mesh->Props[i].dParam5  = 0.0;
    mesh->Props[i].dParam6  = 0.0;
    mesh->Props[i].dParam7  = 0.0;
    mesh->Props[i].dParam8  = 0.0;
    mesh->Props[i].dParam9  = 0.0;
    mesh->Props[i].dParam10 = 0.0;
    mesh->Props[i].dParam11 = 0.0;
    mesh->Props[i].dParam12 = 0.0;
    mesh->Props[i].dParam13 = 0.0;
    mesh->Props[i].dParam14 = 0.0;
    mesh->Props[i].dParam15 = 0.0;
    mesh->Props[i].dParam16 = 0.0;
    mesh->Props[i].dParam17 = 0.0;
    mesh->Props[i].dParam18 = 0.0;
    mesh->Props[i].dParam19 = 0.0;
    mesh->Props[i].dParam20 = 0.0;

    for(j = 0; j < E_NMAT_PROP_KEYWORDS; j++)
    {
      mesh->Props[i].KeywordSet[j] = FALSE;
    }
  }
}

static void ReplaceMaterial(MESH *mesh, DR_DATA *data)
{
#ifdef __E_MPI__
  int  i;
  MPI_Status  status;
#endif

  ClearMaterial(mesh);

#ifdef __E_MPI__
  if(data->CommData->Par_ProcID == 1)
  {
#endif
    /* read new material on root */
    e_GetMaterials(mesh, data->CuttingMaterialFile);

#ifdef __E_MPI__
    for (i = 1; i < data->CommData->Par_NSubdomains; i++)
    {
      MPI_Send(&(mesh->NMats), 1, MPI_INT, i, 
               200, MPI_COMM_WORLD);
      e_MPI_Send_Mats(mesh, i);
    }
  }
  else
  {
    MPI_Recv(&(mesh->NMats), 1, MPI_INT, 0, 
               200, MPI_COMM_WORLD, &status);
    e_MPI_Recv_Mats(mesh, 0);
  }
#endif

}

/* ------------------------------------------------------------------------- */
/*                            Correction of mesh                             */
/* ------------------------------------------------------------------------- */

static void RemoveLoads(MESH *mesh)
{
  if(mesh->NLoadedNodes > 0)
  {
    e_free(mesh->LoadedNodes);
    mesh->LoadedNodes = NULL;
    e_free_frag_d_matr(mesh->NLoadedNodes, mesh->Loads);
    mesh->Loads = NULL;
    
    mesh->NLoadedNodes = 0;
    mesh->KeywordSet[E_KW_NUM_NLOADED_NODES] = FALSE;
    mesh->KeywordSet[E_KW_NUM_LOADS] = FALSE;
  }
}

static void RemoveBCs(MESH *mesh)
{
  if(mesh->NBCNodes > 0)
  {
    e_free(mesh->BCNodes);
    mesh->BCNodes = NULL;
    e_free_frag_i_matr(mesh->NBCNodes, mesh->BCTypes);
    mesh->BCTypes = NULL;
    e_free_frag_d_matr(mesh->NBCNodes, mesh->BCDispls);
    mesh->BCDispls = NULL;
    e_free_frag_d_matr(mesh->NBCNodes, mesh->BCSpringConsts);
    mesh->BCSpringConsts = NULL;

    mesh->NBCNodes = 0;
    mesh->KeywordSet[E_KW_NUM_NBC_NODES] = FALSE;
    mesh->KeywordSet[E_KW_NUM_BCS] = FALSE;
  }
}

/* ------------------------------------------------------------------------- */
/*                       Creation of cutting pattern info                    */
/* ------------------------------------------------------------------------- */

/* add an element to a strip of cloth */
static int AddAdjTriangToStrip(MESH *mesh, int elem_index, 
                               int **strip_triangs, int *len_tri_list, 
                               int strip_index)
{
  int boundary = -1;

  /* determine element type */
  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] <= (elem_index+1))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] <= (elem_index+1))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] <= (elem_index+1))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] <= (elem_index+1))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] <= (elem_index+1))
  {
    boundary = 1;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] <= (elem_index+1))
  {
    boundary = 0;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] <= (elem_index+1))
  {
    boundary = 0;
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] <= (elem_index+1))
  {
    boundary = 0;
  }

  if((boundary == 0) && (mesh->Decomp.SubDoms[elem_index] == 0))
  {
    mesh->Decomp.SubDoms[elem_index] = strip_index;
    (*strip_triangs) = e_incr_i_vect(*len_tri_list, *strip_triangs, ALLOC_SIZE);
    (*strip_triangs)[(*len_tri_list)] = elem_index;
    (*len_tri_list)++;
  }

  return(boundary);
}


#ifdef __E_MPI__
#ifdef DD_TYPE_COMM
static int CheckCommNode(DR_DATA *data, int node)
{
  int ii, jj;

  for(ii = 0; ii < data->CommData->Par_NSubdomains; ii++)
  {
    for(jj = 0; jj < data->CommData->Par_NElemSend[ii]; jj += data->Dimension)
    {
      if((data->CommData->Par_ElemSend[ii][jj] / data->Dimension) == node-1)
        return(TRUE);
    }
  }

  return(FALSE);
}
#else
static int CheckCommNode(DR_DATA *data, int node)
{
  e_GeneralError("CheckCommNode: Only with DD comm !");
}
#endif

#endif


/* 
 * Add elements of the same type to strips. 
 * If the current element is adjacent to a different type of element, it is
 * not a problem.
 */
static void AddAllTriangsToStrip(MESH *mesh, DR_DATA *data, ADJDATA *adata,
                                 int NElems, int **Nodes, 
                                 int *Subdoms, int ElemType,
                                 CUTTING_PATTERN *cdata)
{
  int     i, i_side, i_deg, act_tri;
  int     ptr_tri_list, len_tri_list, ns, nx;
  int    *strip_triangs;
  int     nvertices, *vertices;
#ifdef __E_MPI__
  int     n1, n2;
#endif

  for(i = 0; i < NElems; i++)
  {
    /* elements not yet decided */
    if(Subdoms[i] == 0)
    {
      /* new strip starts */
      mesh->Decomp.NSubDoms++;
      cdata->Strips = (STRIP **)e_incr_vp_vect(cdata->NStrips, (void *)(cdata->Strips), ALLOC_SIZE);
      cdata->Strips[cdata->NStrips] = (STRIP *)e_calloc(1, sizeof(STRIP) );
#ifdef __E_MPI__
      cdata->NCommNodeInStrip = e_incr_i_vect(cdata->NStrips, cdata->NCommNodeInStrip, 1);
      cdata->NElemInStrip = e_incr_i_vect(cdata->NStrips, cdata->NElemInStrip, 1);
#endif
      (cdata->NStrips)++;
      /* first element in the strip is explicitly defined here */
      ptr_tri_list = 0;
      len_tri_list = 1;
      strip_triangs = e_alloc_i_vect(ALLOC_SIZE);
      strip_triangs[ptr_tri_list] = mesh->FirstNetIndex[ElemType] - 1 + i;
      Subdoms[i] = mesh->Decomp.NSubDoms;
      
      /* go, until all elements in the list has been processed */
      while(ptr_tri_list != len_tri_list)
      {
        act_tri = strip_triangs[ptr_tri_list++];

        ns = 0;
        /* get nodes of the element */
        e_GetElementVertices(act_tri, mesh, &nvertices, &vertices);
        /* add new triangle elements to the strip */
        for(i_side = 0; i_side < adata->Sides[act_tri]; i_side++)
        {
          nx = nvertices - 1 - i_side;
          for(i_deg = 0; i_deg < adata->Degree[act_tri][nx]; i_deg++)
          {
            AddAdjTriangToStrip(mesh, adata->Adjacency[act_tri][nx][i_deg],
                                &strip_triangs, &len_tri_list, 
                                mesh->Decomp.NSubDoms);
          }
#ifdef __E_MPI__
          /* if there is no element on the side 
             check whether any of the sides is a communicating side, 
             then store it for further use */
          if(adata->Degree[act_tri][nx] == 0)
          {
            n1 = vertices[ns];
            n2 = vertices[nx];
            if(CheckCommNode(data, n1) && CheckCommNode(data, n2))
            {
              cdata->CommNodesInStrip = e_incr_i_vect(cdata->TotalNCommNodeInStrip, 
                                                      cdata->CommNodesInStrip, 2);
              /* store the global node numbers */
              cdata->CommNodesInStrip[cdata->TotalNCommNodeInStrip  ] = mesh->GNIs[n1-1];
              cdata->CommNodesInStrip[cdata->TotalNCommNodeInStrip+1] = mesh->GNIs[n2-1];
              cdata->TotalNCommNodeInStrip += 2;
              cdata->NCommNodeInStrip[cdata->NStrips-1] += 2;
            }
          }
#endif
          ns = nx;
        }

      }

#ifdef __E_MPI__
      cdata->NElemInStrip[cdata->NStrips-1] = len_tri_list;
#endif
      /* free up memory */
      e_free(strip_triangs);
      strip_triangs = NULL;
    }
  }
}



/* create all strips */
static void CreateStrips(MESH *mesh, DR_DATA *data, ADJDATA *adata,
                         CUTTING_PATTERN *cdata)
{
  int  i;
  char title[E_LINE_MAX+1];


  if(mesh->NElemsTriang3 +
     mesh->NElemsTriang4 +
     mesh->NElemsTriang5 +
     mesh->NElemsLink1 +
     mesh->NElemsLink2 +
     mesh->NElemsLink3 +
     mesh->NElemsLink4 +
     mesh->NElemsLink5   != mesh->TotalNElems)
  {
    e_GeneralError("Create_Strips: Unsupported element type !");
  }

  /* setup valid data structure */
  e_SetupDecompData(mesh);
  /* zero it out, becasue this is an indicator */
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    mesh->Decomp.SubDoms[i] = 0;
  }
  mesh->Decomp.NSubDoms = 0;

  /* process valid triangle element types */
  AddAllTriangsToStrip(mesh, data, adata,
                       mesh->NElemsTriang3, mesh->NodesTriang3,
                       mesh->Decomp.SubDomsTriang3,
                       E_ELEM_TYPE_TRIANG3, cdata);
                       
  AddAllTriangsToStrip(mesh, data, adata,
                       mesh->NElemsTriang4, mesh->NodesTriang4,
                       mesh->Decomp.SubDomsTriang4,
                       E_ELEM_TYPE_TRIANG4, cdata);
                       
  AddAllTriangsToStrip(mesh, data, adata,
                       mesh->NElemsTriang5, mesh->NodesTriang5,
                       mesh->Decomp.SubDomsTriang5,
                       E_ELEM_TYPE_TRIANG5, cdata);

  /* create the submeshes */
  for(i = 0; i < cdata->NStrips; i++)
  {
    sprintf(title, "%d strip", i+1);

    e_CreateSubMesh(mesh, &(cdata->Strips[i]->StripMesh), title, i+1);
  }

}

/* create the mesh containing only link element to unstress */
static void CreateLinkMesh(MESH *mesh, CUTTING_PATTERN *cdata)
{
  int  i;

  /* add one extra mesh */
  mesh->Decomp.NSubDoms++;
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    if(mesh->Decomp.SubDoms[i] == 0)
    {
      mesh->Decomp.SubDoms[i] = mesh->Decomp.NSubDoms;
    }
  }

  e_CreateSubMesh(mesh, &(cdata->CableMesh), "Cable mesh", mesh->Decomp.NSubDoms);

  if(cdata->CableMesh->NElemsLink1 +
     cdata->CableMesh->NElemsLink2 +
     cdata->CableMesh->NElemsLink3 +
     cdata->CableMesh->NElemsLink4 +
     cdata->CableMesh->NElemsLink5 != cdata->CableMesh->TotalNElems)
  {
    e_GeneralError("CreateLinkMesh: Not only links in mesh !");
  }
  
}

/* create cutting pattern information */
static void CreateCutPatMesh(MESH *mesh, DR_DATA *data, 
                             CUTTING_PATTERN **cdata)
{
  int   *NNodeElem;
  int  **NodeElem;
  ADJDATA *adata;

  NNodeElem = e_alloc_i_vect(mesh->TotalNElems);
  NodeElem  = e_init_frag_i_matr(mesh->TotalNElems);

  e_InitMeshPointElemConnection(mesh, NNodeElem, NodeElem);
  Set2DElemAdjacency(mesh, &adata, NNodeElem, NodeElem, ADJ_LINK);
  
  (*cdata) = (CUTTING_PATTERN *)e_calloc(1, sizeof(CUTTING_PATTERN) );
  /* create the strips of clothes */
  CreateStrips(mesh, data, adata, *cdata);
  /* create the mesh with cables to unstress */
  CreateLinkMesh(mesh, *cdata);

  e_free_frag_i_matr(mesh->NMeshPoints, NodeElem);
  e_free(NNodeElem);
  e_Free2DElemAdjacency(mesh, adata);

}

/* ------------------------------------------------------------------------- */
/*                     Determine initial length in 3D of strips              */
/* ------------------------------------------------------------------------- */

static int FindMatchingElements(int index, int n, int *list)
{
  int i;
  
  for(i = 0; i < n; i++)
  {
    if(list[i] == index)
      return(i);
  }

  return(-1);
}




/* calculate initial length from element stresses */
static void StripInitLengthFromStress(DR_DATA *Gdata, MESH *Gmesh,
                                      DR_DATA *data, MESH *mesh)
{
  int i, j, n, NElem, elemindex;
  
  if(mesh->NElemsTriang3 +
     mesh->NElemsTriang4 +
     mesh->NElemsTriang5 != mesh->TotalNElems)
  {
    e_GeneralError("StripInitLengthFromStress: Unsupported element in strip !");
  }
  
  /* 
   * We cannot use 'SetupInitLengthData' !!!
   * Setup should be explicit, because otherwise it requires to have the proper
   * element types in the mesh. Unfortunately we cannot transform the elements 
   * into TRIANG3 yet, since then we cannot determine the stresses. The determination
   * of stresses is different for different element types.
   */
  data->InitLengthData = (INIT_LENGTH_DATA *)e_calloc(1, sizeof(INIT_LENGTH_DATA));
  n = (mesh->NElemsTriang3 + mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;
  data->InitLengthData->InitLength = e_alloc_d_vect(n);

  data->InitLengthData->InitLengthTriang3 = (double **)e_alloc_vp_vect(n);
  data->InitLengthData->InitLengthTriang3[0] = data->InitLengthData->InitLength;
  for(i = 1; i < n; i++)
  {
    data->InitLengthData->InitLengthTriang3[i] = 
      data->InitLengthData->InitLengthTriang3[i-1] + 3;
  }

  NElem = 0;
  /* initial length is given explicitly for analysis elements*/
  for(i = 0; i < mesh->NElemsTriang3; i++)
  {
    elemindex = FindMatchingElements(mesh->GEIsTriang3[i], Gmesh->NElemsTriang3, 
                                     Gmesh->GEIsTriang3);
    if(elemindex == -1)
    {
      e_GeneralError("StripInitLengthFromStress: Internal error 1 !");
    }
    for(j = 0; j < 3; j++)
    {
      data->InitLengthData->InitLengthTriang3[NElem][j] =
        Gdata->InitLengthData->InitLengthTriang3[elemindex][j];
    }
    NElem++;
  }

  for(i = 0; i < mesh->NElemsTriang4; i++)
  {
    elemindex = FindMatchingElements(mesh->GEIsTriang4[i], Gmesh->NElemsTriang4,
                                     Gmesh->GEIsTriang4);
    if(elemindex == -1)
    {
      e_GeneralError("StripInitLengthFromStress: Internal error 1 !");
    }
    for(j = 0; j < 3; j++)
    {
      data->InitLengthData->InitLengthTriang4[NElem][j] =
        Gdata->InitLengthData->InitLengthTriang4[elemindex][j];
    }
    NElem++;
  }

  /* form-finding elements should be unstressed */
  for(i = 0; i < mesh->NElemsTriang5; i++)
  {
    Unstress_GenTriangInitLength(mesh->NodesTriang5[i], mesh->MeshPointCoords, 
                                 mesh->MatsTriang5[i]->YMod, mesh->MatsTriang5[i]->PoissonsRatio,
                                 mesh->MatsTriang5[i]->dParam1, /* Sigma X */
                                 mesh->MatsTriang5[i]->dParam2, /* Sigma Y */
                                 mesh->MatsTriang5[i]->dParam3, /* Tau XY */
                                 data->InitLengthData->InitLengthTriang3[NElem]);
    NElem++;
  }

}

/* ------------------------------------------------------------------------- */
/*                     Convert mesh into TRIANG3 mesh                        */
/* ------------------------------------------------------------------------- */

static void ConvertMeshForCutPat(MESH *mesh)
{
  int **nodes;

  /* only these kinds of elements */
  if(mesh->NElemsTriang3 +
     mesh->NElemsTriang4 +
     mesh->NElemsTriang5 != mesh->TotalNElems)
  {
    e_GeneralError("ConvertMeshForCutPat: Unsupported elements !");
  }


  nodes = e_alloc_cont_i_matr(mesh->NElemsTriang3 + 
                              mesh->NElemsTriang4 + 
                              mesh->NElemsTriang5, 3);

  /* all nodes to TRIANG3 */
  if(mesh->NElemsTriang3)
  {
    memcpy((void *)(nodes),
           (void *)(mesh->NodesTriang3[0]),
           sizeof(int) * mesh->NElemsTriang3 * 3);
  }

  if(mesh->NElemsTriang4)
  {
    memcpy((void *)(nodes[mesh->NElemsTriang3]),
           (void *)(mesh->NodesTriang4[0]),
           sizeof(int) * mesh->NElemsTriang4 * 3);
  }

  if(mesh->NElemsTriang5)
  {
    memcpy((void *)(nodes[mesh->NElemsTriang3 + mesh->NElemsTriang4]),
           (void *)(mesh->NodesTriang5[0]),
           sizeof(int) * mesh->NElemsTriang5 * 3);
  }

  if(mesh->NodesTriang3)
  {
    e_free_cont_i_matr(mesh->NElemsTriang3, mesh->NodesTriang3);
  }
  mesh->NodesTriang3 = nodes;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG3] = TRUE;

  if(mesh->NodesTriang4)
  {
    e_free_cont_i_matr(mesh->NElemsTriang4, mesh->NodesTriang4);
  }
  mesh->NodesTriang4 = NULL;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG4] = FALSE;

  if(mesh->NodesTriang5)
  {
    e_free_cont_i_matr(mesh->NElemsTriang5, mesh->NodesTriang5);
  }
  mesh->NodesTriang5 = NULL;
  mesh->KeywordSet[E_KW_NUM_NODES_TRIANG5] = FALSE;

  /* global element numbers  */
  mesh->GEIsTriang3 = mesh->GEIs;
  mesh->GEIsTriang4 = NULL;
  mesh->GEIsTriang5 = NULL;

  /* materials */
  mesh->MatsTriang3 = (MAT **)mesh->Mats;
  mesh->MatsTriang4 = NULL;
  mesh->MatsTriang5 = NULL;

  mesh->KeywordSet[E_KW_NUM_MATS_TRIANG3] = TRUE;
  mesh->KeywordSet[E_KW_NUM_MATS_TRIANG4] = FALSE;
  mesh->KeywordSet[E_KW_NUM_MATS_TRIANG5] = FALSE;

  /* element numbers */
  mesh->NElemsTriang3 += mesh->NElemsTriang4 + mesh->NElemsTriang5;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = 1;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = 0;
  mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] = 0;

  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3] = TRUE;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4] = FALSE;
  mesh->NElemsTriang4 = 0;
  mesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG5] = FALSE;
  mesh->NElemsTriang5 = 0;

  mesh->TotalNElems = mesh->NElemsTriang3;

/*  e_PutMeshData(mesh, "test"); */

}

/* ------------------------------------------------------------------------- */
/*                     Final modifications of strip meshes                   */
/* ------------------------------------------------------------------------- */

/* 
 * Detrmine the longest extension of the strip in X-Y plane, 
 * determine nodes at those positions and pass it back with
 * directional data.
 */
static void SearchMaxDistNode(MESH *mesh, int *n1, int *n2, double *direction)
{
  int     i;
  double  maxx, maxy, minx, miny, x, y, nx, ny, ll;
  int     imaxx, imaxy, iminx, iminy;

  maxx  = -1e8;
  imaxx = -1;
  maxy  = -1e8;
  imaxy = -1;
  minx  =  1e8;
  iminx = -1;
  miny  =  1e8;
  iminy = -1;
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    x = mesh->MeshPointCoords[i][0];
    y = mesh->MeshPointCoords[i][1];
    if(maxx < x)
    {
      maxx = x;
      imaxx = i;
    }
    if(maxy < y)
    {
      maxy = y;
      imaxy = i;
    }
    if(minx > x)
    {
      minx = x;
      iminx = i;
    }
    if(miny > y)
    {
      miny = y;
      iminy = i;
    }
  }

  if(   imaxx == -1
     || imaxy == -1
     || iminx == -1
     || iminy == -1 )
  {
    e_GeneralError("SearchMaxDistNode: Internal error !");
  }
  
  if(E_ABS(maxx-minx) > E_ABS(maxy-miny))
  {
    *n1 = imaxx;
    *n2 = iminx;
  }
  else
  {
    *n1 = imaxy;
    *n2 = iminy;
  }

  nx = mesh->MeshPointCoords[*n1][0] - mesh->MeshPointCoords[*n2][0];
  ny = mesh->MeshPointCoords[*n1][1] - mesh->MeshPointCoords[*n2][1];
  ll = sqrt(nx*nx + ny*ny);
  nx /= ll;
  ny /= ll;

  direction[0] = ny;
  direction[1] = -nx;
}


/* 
 * Make a strip mesh statically determinant, but try to apply
 * the boundary conditions as far as possible.
 *
 * Function assumes there is no boundary condition applied at all.
 */
static void MakeStaticDetMesh(MESH *mesh)
{
  int n1, n2;
  double direction[2];
  int    nbc;

#if 0
  nbc = 1;
  
  mesh->NBCNodes = nbc;
  mesh->KeywordSet[E_KW_NUM_NBC_NODES] = TRUE;
  mesh->KeywordSet[E_KW_NUM_BCS] = TRUE;

  mesh->BCNodes        = e_alloc_i_vect(nbc);
  mesh->BCTypes        = e_alloc_frag_i_matr(nbc, E_DOF_MAX);
  mesh->BCDispls       = e_alloc_frag_d_matr(nbc, E_DOF_MAX);
  mesh->BCSpringConsts = e_alloc_frag_d_matr(nbc, E_DOF_MAX);

  mesh->BCNodes[0] = 1;

  mesh->BCTypes[0][0] = E_BC_NUM_FIXED;
  mesh->BCTypes[0][1] = E_BC_NUM_FIXED;
  mesh->BCTypes[0][2] = E_BC_NUM_FIXED;

  mesh->BCDispls[0][0] = 0;
  mesh->BCDispls[0][1] = 0;
  mesh->BCDispls[0][2] = 0;

  mesh->BCTypes[1][0] = E_BC_NUM_PLANE;
  mesh->BCTypes[1][1] = E_BC_NUM_PLANE;
  mesh->BCTypes[1][2] = E_BC_NUM_PLANE;

  /* determine max extension of strip */
  SearchMaxDistNode(mesh, &n1, &n2, direction);
  
  mesh->BCNodes[0] = n1+1;
  mesh->BCNodes[1] = n2+1;
  assert(direction[0] != 0.0 && direction[1] != 0.0);
  
  /* restricted plane is "parallel" with longest dimension */
  mesh->BCDispls[0][0] = direction[0];
  mesh->BCDispls[0][1] = direction[1];
  mesh->BCDispls[0][2] = 0;

  mesh->BCDispls[1][0] = direction[0];
  mesh->BCDispls[1][1] = direction[1];
  mesh->BCDispls[1][2] = 0;
#endif

}

/* ------------------------------------------------------------------------- */
/*                         Functions for balancing                           */
/* ------------------------------------------------------------------------- */

#ifdef __E_MPI__
/* send strip information to master for balancing if necessary */
void StripsToMaster(CUTTING_PATTERN *cdata, int ProcID, int NSubdomains)
{
  int          i;
  int          strip_info[2];
  MPI_Status   status;

  if(ProcID == 1)
  {
    cdata->MasterData = (G_CUTPAT *)e_calloc(1, sizeof(G_CUTPAT) );
  
    /* allocate for every subdomain */
    cdata->MasterData->StripInfo         = e_alloc_frag_i_matr(NSubdomains, 2);
    cdata->MasterData->NElemPerStrip     = e_init_frag_i_matr(NSubdomains);
    cdata->MasterData->NCommNodePerStrip = e_init_frag_i_matr(NSubdomains);
    cdata->MasterData->CommNodesPerStrip = e_init_frag_i_matr(NSubdomains);

    /* store locally available info */
    cdata->MasterData->StripInfo[0][0] = cdata->NStrips;
    cdata->MasterData->StripInfo[0][1] = cdata->TotalNCommNodeInStrip;
    
    cdata->MasterData->NElemPerStrip[0] = cdata->NElemInStrip;
    if(cdata->MasterData->StripInfo[0][1])
    {
      cdata->MasterData->NCommNodePerStrip[0] = cdata->NCommNodeInStrip;
      cdata->MasterData->CommNodesPerStrip[0] = cdata->CommNodesInStrip;
    }
    else
    {
      cdata->MasterData->CommNodesPerStrip[0] = 
        e_alloc_i_vect(cdata->MasterData->StripInfo[0][0]);
    }
    
    /* for all other subdomain receive the information */
    for(i = 1; i < NSubdomains; i++)
    {
      MPI_Recv(cdata->MasterData->StripInfo[i], 2, MPI_INT, i, 800, 
               MPI_COMM_WORLD, &status);
      cdata->MasterData->NElemPerStrip[i] =
        e_alloc_i_vect(cdata->MasterData->StripInfo[i][0]);
      MPI_Recv(cdata->MasterData->NElemPerStrip[i], cdata->MasterData->StripInfo[i][0], 
               MPI_INT, i, 801, MPI_COMM_WORLD, &status);

      /* it has to be allocated and at least zero for each element */
      cdata->MasterData->NCommNodePerStrip[i] = 
        e_alloc_i_vect(cdata->MasterData->StripInfo[i][0]);
      
      if(cdata->MasterData->StripInfo[i][1])
      {
        cdata->MasterData->CommNodesPerStrip[i] =
          e_alloc_i_vect(cdata->MasterData->StripInfo[i][1]);
      
        MPI_Recv(cdata->MasterData->NCommNodePerStrip[i], 
                 cdata->MasterData->StripInfo[i][0], 
                 MPI_INT, i, 802, MPI_COMM_WORLD, &status);
        MPI_Recv(cdata->MasterData->CommNodesPerStrip[i], 
                 cdata->MasterData->StripInfo[i][1], 
                 MPI_INT, i, 803, MPI_COMM_WORLD, &status);
      }
    }
  }
  else
  {
    /* workers jobs to send all information about strips */
    strip_info[0] = cdata->NStrips;
    strip_info[1] = cdata->TotalNCommNodeInStrip;
    MPI_Send(strip_info, 2, MPI_INT, 0, 800, MPI_COMM_WORLD);
    MPI_Send(cdata->NElemInStrip, cdata->NStrips, MPI_INT, 0, 801, MPI_COMM_WORLD);
    if(cdata->TotalNCommNodeInStrip)
    {
      MPI_Send(cdata->NCommNodeInStrip, cdata->NStrips,
               MPI_INT, 0, 802, MPI_COMM_WORLD);
      MPI_Send(cdata->CommNodesInStrip, cdata->TotalNCommNodeInStrip,
               MPI_INT, 0, 803, MPI_COMM_WORLD);
    }
  }

}

/* 
 * assemble half strips which has been cut into pieces by the initial
 * partitioning 
 */
static void StripAssemble(G_CUTPAT *cdata, int ProcID, int NSubdomains)
{
  int   i, j, k, ii, jj, kk, start_1, start_2, found;
  int   attach1, attach2, index1, index2, ia, ip;
  int   stripno, subdom;
  int **halfstrips;

  halfstrips = e_init_frag_i_matr(NSubdomains);
  for(i = 0; i < NSubdomains; i++)
  {
    halfstrips[i] = e_alloc_i_vect(cdata->StripInfo[i][0]);
  }

  cdata->NAssembledPartStrip = 0;
  for(i = 0; i < NSubdomains - 1; i++)
  {
    start_1 = 0;
    for(j = 0; j < cdata->StripInfo[i][0]; j++)
    {
      for(ii = i+1; ii < NSubdomains; ii++)
      {
        start_2 = 0;
        for(jj = 0; jj < cdata->StripInfo[ii][0]; jj++)
        {
          found = 0;
          /* look for common sides between half strips 
             a side is determined by two common node */
          for(k = 0; k < cdata->NCommNodePerStrip[i][j]; k += 2)
          {
            for(kk = 0; kk < cdata->NCommNodePerStrip[ii][jj]; kk += 2)
            {
              if(   (   (cdata->CommNodesPerStrip[i][start_1 + k]      == 
                         cdata->CommNodesPerStrip[ii][start_2 + kk]        ) 
                     && (cdata->CommNodesPerStrip[i][start_1 + k + 1]  == 
                         cdata->CommNodesPerStrip[ii][start_2 + kk + 1]    ) ) 
                 || (   (cdata->CommNodesPerStrip[i][start_1 + k]      == 
                         cdata->CommNodesPerStrip[ii][start_2 + kk + 1]    ) 
                     && (cdata->CommNodesPerStrip[i][start_1 + k + 1]  == 
                         cdata->CommNodesPerStrip[ii][start_2 + kk]        ) ) )
              {
                found = 1;
                break;
              }
            }
            if(found)
              break;
          }
          if(found)
          {
            halfstrips[i][j]   = TRUE;
            halfstrips[ii][jj] = TRUE;
            attach1 = FALSE;
            attach2 = FALSE;
            index1  = -1;
            index2  = -1;
            for(ia = 0; ia < cdata->NAssembledPartStrip; ia++)
            {
              for(ip = 0; ip < cdata->NPartForStrip[ia]; ip++)
              {
                stripno = cdata->PartNoForStrip[ia][ip];
                subdom  = cdata->PartSubdomForStrip[ia][ip];
                if(subdom == i && stripno == j)
                {
                  index1 = ia;
                  attach1 = TRUE;
                }
                else if(subdom == ii && stripno == jj)
                {
                  index2 = ia;
                  attach2 = TRUE;
                }
              }
            }
            /* we found only one of them */
            if((attach1 == TRUE) && (attach2 == FALSE))
            {
              cdata->PartSubdomForStrip[index1] =
                e_incr_i_vect(cdata->NPartForStrip[index1],
                              cdata->PartSubdomForStrip[index1], 1);
              cdata->PartSubdomForStrip[index1][cdata->NPartForStrip[index1]] = ii;

              cdata->PartNoForStrip[index1] =
                e_incr_i_vect(cdata->NPartForStrip[index1],
                              cdata->PartNoForStrip[index1], 1);
              cdata->PartNoForStrip[index1][cdata->NPartForStrip[index1]] = jj;
          
              cdata->NElemInPartForStrip[index1] =
                e_incr_i_vect(cdata->NPartForStrip[index1],
                              cdata->NElemInPartForStrip[index1], 1);
              cdata->NElemInPartForStrip[index1][cdata->NPartForStrip[index1]] =
                cdata->NElemPerStrip[ii][jj];
          
              cdata->NPartForStrip[index1]++;
            }
            else if((attach1 == FALSE) && (attach2 == TRUE))
            {
              cdata->PartSubdomForStrip[index2] =
                e_incr_i_vect(cdata->NPartForStrip[index2],
                              cdata->PartSubdomForStrip[index2], 1);
              cdata->PartSubdomForStrip[index2][cdata->NPartForStrip[index2]] = i;

              cdata->PartNoForStrip[index2] =
                e_incr_i_vect(cdata->NPartForStrip[index2],
                              cdata->PartNoForStrip[index2], 1);
              cdata->PartNoForStrip[index2][cdata->NPartForStrip[index2]] = j;
          
              cdata->NElemInPartForStrip[index2] =
                e_incr_i_vect(cdata->NPartForStrip[index2],
                              cdata->NElemInPartForStrip[index2], 1);
              cdata->NElemInPartForStrip[index2][cdata->NPartForStrip[index2]] =
                cdata->NElemPerStrip[i][j];
          
              cdata->NPartForStrip[index2]++;
            }
            /* we found both */
            else if(attach1 == TRUE && attach2 == TRUE)
            {
              if(index1 != index2)
              {
                cdata->PartSubdomForStrip[index1] =
                  e_ext_i_vect(cdata->NPartForStrip[index1],
                               cdata->PartSubdomForStrip[index1], 
                               cdata->NPartForStrip[index2] );
                
                memcpy((void *)(cdata->PartSubdomForStrip[index1]+
                                cdata->NPartForStrip[index1]),
                       (void *)(cdata->PartSubdomForStrip[index2]),
                       sizeof(int) * cdata->NPartForStrip[index2] );
 
                cdata->PartNoForStrip[index1] =
                  e_ext_i_vect(cdata->NPartForStrip[index1],
                               cdata->PartNoForStrip[index1],
                               cdata->NPartForStrip[index2] );

                memcpy((void *)(cdata->PartNoForStrip[index1]+
                                cdata->NPartForStrip[index1]),
                       (void *)(cdata->PartNoForStrip[index2]),
                       sizeof(int) * cdata->NPartForStrip[index2] );
          
                cdata->NElemInPartForStrip[index1] =
                  e_ext_i_vect(cdata->NPartForStrip[index1],
                               cdata->NElemInPartForStrip[index1],
                               cdata->NPartForStrip[index2] );

                memcpy((void *)(cdata->NElemInPartForStrip[index1]+
                                cdata->NPartForStrip[index1]),
                       (void *)(cdata->NElemInPartForStrip[index2]),
                       sizeof(int) * cdata->NPartForStrip[index2] );

                cdata->NPartForStrip[index1] += cdata->NPartForStrip[index2];
                
                /* free the removed one */
                e_free(cdata->PartSubdomForStrip[index2]);
                e_free(cdata->PartNoForStrip[index2]);
                e_free(cdata->NElemInPartForStrip[index2]);
                
                cdata->PartSubdomForStrip[index2] = NULL;
                cdata->PartNoForStrip[index2] = NULL;
                cdata->NElemInPartForStrip[index2] = NULL;
                
                cdata->NPartForStrip[index2] = 0;
                
                if(cdata->NAssembledPartStrip-index2-1)
                {
                  memmove((void *)(cdata->PartSubdomForStrip+index2),
                          (void *)(cdata->PartSubdomForStrip+index2+1),
                          sizeof(int *) * (cdata->NAssembledPartStrip-index2-1) );

                  memmove((void *)(cdata->PartNoForStrip+index2),
                          (void *)(cdata->PartNoForStrip+index2+1),
                          sizeof(int *) * (cdata->NAssembledPartStrip-index2-1) );

                  memmove((void *)(cdata->NElemInPartForStrip+index2),
                          (void *)(cdata->NElemInPartForStrip+index2+1),
                          sizeof(int *) * (cdata->NAssembledPartStrip-index2-1) );
                  memmove((void *)(cdata->NPartForStrip+index2),
                          (void *)(cdata->NPartForStrip+index2+1),
                          sizeof(int) * (cdata->NAssembledPartStrip-index2-1) );
                }
                cdata->NAssembledPartStrip--;
                
              }
            }
            /* none of the strips found */
            else if(attach1 == FALSE && attach2 == FALSE)
            {
              cdata->NPartForStrip = e_incr_i_vect(cdata->NAssembledPartStrip,
                                                   cdata->NPartForStrip, 1);
              cdata->NPartForStrip[cdata->NAssembledPartStrip] = 2;
              cdata->PartSubdomForStrip = 
                e_incr_frag_i_matr(cdata->NAssembledPartStrip,
                                   cdata->PartSubdomForStrip, 1);
              cdata->PartSubdomForStrip[cdata->NAssembledPartStrip] = 
                e_alloc_i_vect(2);
              cdata->PartSubdomForStrip[cdata->NAssembledPartStrip][0] = i;
              cdata->PartSubdomForStrip[cdata->NAssembledPartStrip][1] = ii;
              
              cdata->PartNoForStrip =
                e_incr_frag_i_matr(cdata->NAssembledPartStrip,
                                   cdata->PartNoForStrip, 1);
              cdata->PartNoForStrip[cdata->NAssembledPartStrip] = 
                e_alloc_i_vect(2);
              cdata->PartNoForStrip[cdata->NAssembledPartStrip][0] = j;
              cdata->PartNoForStrip[cdata->NAssembledPartStrip][1] = jj;
            
              cdata->NElemInPartForStrip =
                e_incr_frag_i_matr(cdata->NAssembledPartStrip,
                                   cdata->NElemInPartForStrip, 1);
              cdata->NElemInPartForStrip[cdata->NAssembledPartStrip] =
                e_alloc_i_vect(2);
              cdata->NElemInPartForStrip[cdata->NAssembledPartStrip][0] =
                cdata->NElemPerStrip[i][j];
              cdata->NElemInPartForStrip[cdata->NAssembledPartStrip][1] =
                cdata->NElemPerStrip[ii][jj];
              
              cdata->NAssembledPartStrip++;
            }
          }
          start_2 += cdata->NCommNodePerStrip[ii][jj];
        }
      }
      start_1 += cdata->NCommNodePerStrip[i][j];
    }
  }

  /* determine the number of elements coming from the full strips */
  cdata->NElemInSubdom = e_alloc_i_vect(NSubdomains);
  for(i = 0; i < NSubdomains; i++)
  {
    for(j = 0; j < cdata->StripInfo[i][0]; j++)
    {
      if(halfstrips[i][j] == FALSE)
      {
        cdata->NElemInSubdom[i] += cdata->NElemPerStrip[i][j];
      }
    }
    e_free(halfstrips[i]);
  }
  e_free(halfstrips);

}


static void StripsFromMaster(int ProcID, int NSubdomains, CUTTING_PATTERN *cdata)
{
  int           i, index, i_half, strip_no, no_possible_comms, i_store, half_index;
  MPI_Status    status;
  int         **possible_comms = NULL;
  int          *comms_used_in_round;
  int           counter, sender_index, receiver_index;
  int         **m_strip_to_subdom;
  int          *m_no_strip_to_subdom;

  if(ProcID == 1)
  {
    m_strip_to_subdom = e_init_frag_i_matr(NSubdomains);
    m_no_strip_to_subdom = e_alloc_i_vect(NSubdomains);

    no_possible_comms = 0;
    /* go through all assembled strips and 
       assemble information for the subdomains */
    for(i = 0; i < cdata->MasterData->NAssembledPartStrip; i++)
    {
      /* this is the subdomain where all parts should go */
      index = cdata->MasterData->NewSubdomForPartStrip[i];

      possible_comms = e_ext_cont_i_matr(no_possible_comms, 5, 
                                         possible_comms,
                                         cdata->MasterData->NPartForStrip[i]-1);
      
      /* find the part strip which is in the correct subdomain */
      for(i_half = 0; i_half < cdata->MasterData->NPartForStrip[i]; i_half++)
      {
        if(cdata->MasterData->PartSubdomForStrip[i][i_half] == index)
        {
          strip_no = cdata->MasterData->PartNoForStrip[i][i_half];
          half_index = i_half;
          break;
        }
      }

      i_store = 0;
      for(i_half = 0; i_half < cdata->MasterData->NPartForStrip[i]; i_half++)
      {
        if(   (cdata->MasterData->PartSubdomForStrip[i][i_half] != index) 
           || (   (cdata->MasterData->PartSubdomForStrip[i][i_half] == index) 
               && (i_half != half_index) ) )
        {
          /* sender subdom */
          possible_comms[no_possible_comms+i_store][0] = 
            cdata->MasterData->PartSubdomForStrip[i][i_half];
          /* receiver subdom */
          possible_comms[no_possible_comms+i_store][1] = index;
          /* strip to send */
          possible_comms[no_possible_comms+i_store][2] = 
            cdata->MasterData->PartNoForStrip[i][i_half];
          /* strip to attach to the arriving one */
          possible_comms[no_possible_comms+i_store][3] = strip_no;
          i_store++;
        }
      }
      
      /* sanity check */
      if(i_store != cdata->MasterData->NPartForStrip[i]-1)
      {
        e_GeneralError("StripsFromMaster: Wrong strip info for subdomains");
      }
      
      no_possible_comms += cdata->MasterData->NPartForStrip[i]-1;
    }

    comms_used_in_round = e_alloc_i_vect(NSubdomains);

    counter = 0;
    while(counter != no_possible_comms)
    {
      for(i = 0; i < NSubdomains; i++)
        comms_used_in_round[i] = 0;

      for(i = 0; i < no_possible_comms; i++)
      {
        register int s;

        sender_index = possible_comms[i][0];
        receiver_index = possible_comms[i][1];

        if(   (possible_comms[i][4] == 0)               
           && !comms_used_in_round[sender_index]    
           && !comms_used_in_round[receiver_index] )
        {

          /* sender info setup */
          if(sender_index != receiver_index)
          {
            s = m_no_strip_to_subdom[sender_index];
            m_strip_to_subdom[sender_index] =
              e_incr_i_vect(s, m_strip_to_subdom[sender_index], 3);
            m_strip_to_subdom[sender_index][s] = 0;
            s++;
            m_strip_to_subdom[sender_index][s] = receiver_index;
            s++;
            m_strip_to_subdom[sender_index][s] = possible_comms[i][2];
            s++;
            m_no_strip_to_subdom[sender_index] = s;
          }

          /* receiver info setup */
          s = m_no_strip_to_subdom[receiver_index];
          m_strip_to_subdom[receiver_index] =
            e_incr_i_vect(s, m_strip_to_subdom[receiver_index], 3);
          m_strip_to_subdom[receiver_index][s] = possible_comms[i][3]+1;
          s++;
          m_strip_to_subdom[receiver_index][s] = sender_index;
          s++;
          m_strip_to_subdom[receiver_index][s] = possible_comms[i][2];
          s++;
          m_no_strip_to_subdom[receiver_index] = s;

          counter++;
          possible_comms[i][4] = 1;
          comms_used_in_round[sender_index] = 1;
          comms_used_in_round[receiver_index] = 1;
        }
      }
    }

    /* send over the whole info */
/*fprintf(stderr, "\nMASTER: subdom %d", NSubdomains); */
    for (i = 1; i < NSubdomains; i++)
    {
      MPI_Send(&(m_no_strip_to_subdom[i]), 1, MPI_INT, i, 810, MPI_COMM_WORLD);
/*fprintf(stderr, "\nMASTER: sent %d", i+1); */
      if(m_no_strip_to_subdom[i])
      {
        MPI_Send(m_strip_to_subdom[i], m_no_strip_to_subdom[i], 
                 MPI_INT, i, 811, MPI_COMM_WORLD);
      }
    }
    /* local info available */
    cdata->NPartToSend  = m_no_strip_to_subdom[0];
    cdata->PartCommInfo = m_strip_to_subdom[0];
    

    m_strip_to_subdom[0] = NULL;    /* protect this */
    e_free_frag_i_matr(NSubdomains, m_strip_to_subdom);
    e_free(m_no_strip_to_subdom);
    e_free_cont_i_matr(no_possible_comms, possible_comms);
    e_free(comms_used_in_round);

    /* last chance to free up all memory on the master */
    /* the memory which received info from subdoms */
    e_free_frag_i_matr(NSubdomains, cdata->MasterData->StripInfo);
    e_free_frag_i_matr(NSubdomains, cdata->MasterData->NElemPerStrip);
    e_free_frag_i_matr(NSubdomains, cdata->MasterData->NCommNodePerStrip);
    e_free_frag_i_matr(NSubdomains, cdata->MasterData->CommNodesPerStrip);

    e_free(cdata->MasterData->NPartForStrip);
    e_free_frag_i_matr(cdata->MasterData->NAssembledPartStrip, 
                       cdata->MasterData->PartSubdomForStrip);
    e_free_frag_i_matr(cdata->MasterData->NAssembledPartStrip, 
                       cdata->MasterData->PartNoForStrip);
    e_free_frag_i_matr(cdata->MasterData->NAssembledPartStrip, 
                       cdata->MasterData->NElemInPartForStrip);

    e_free(cdata->MasterData->NElemInSubdom);
    e_free(cdata->MasterData->NewSubdomForPartStrip);
    e_free(cdata->MasterData);
    cdata->MasterData = NULL;
  }
  else
  {

    MPI_Recv(&(cdata->NPartToSend), 1, MPI_INT, 0, 810,
             MPI_COMM_WORLD, &status);

    if(cdata->NPartToSend)
    {
      cdata->PartCommInfo = e_alloc_i_vect(cdata->NPartToSend);
      MPI_Recv(cdata->PartCommInfo, cdata->NPartToSend, MPI_INT, 0, 811,
               MPI_COMM_WORLD, &status);
    }

  }
}


static void StripToSubdoms(int ProcID, int NSubdomains, CUTTING_PATTERN *cdata)
{
  int i, check;
  MPI_Status status;
  MESH *arriving_strip;

  /* how many communication */
  for(i = 0; i < cdata->NPartToSend; i += 3)
  {
    /* receive */
    if(cdata->PartCommInfo[i])
    {
      /* is the strip locally available */
      if(cdata->PartCommInfo[i+1] != ProcID-1)
      {
        /* no, we should receive from another subdomain */
        MPI_Recv(&check, 1, MPI_INT, cdata->PartCommInfo[i+1], 
                 888, MPI_COMM_WORLD, &status);

        if(check != cdata->PartCommInfo[i+2]) /* ERROR */
          e_GeneralError("Wrong strip is arriving !");

        e_MPI_Recv_MeshData(&arriving_strip, cdata->PartCommInfo[i+1]);
      }
      else
      {
        /* this strip is local */
        arriving_strip = cdata->Strips[cdata->PartCommInfo[i+2]]->StripMesh;
        cdata->Strips[cdata->PartCommInfo[i+2]]->StripMesh = NULL;
      }
      
      /* Merge the strips together */
      MergeStrips(arriving_strip, cdata->Strips[cdata->PartCommInfo[i]-1]->StripMesh);

{
  char name[256];
  sprintf(name, "out/merge-%d-%d", ProcID,
          cdata->PartCommInfo[i]);
  e_PutMeshData(cdata->Strips[cdata->PartCommInfo[i]-1]->StripMesh, name);
}
      /* free-up the arrived mesh */
      e_FreeMeshData(arriving_strip);
    }
    /* send */
    else
    {
      MPI_Send(&(cdata->PartCommInfo[i+2]), 1, MPI_INT, 
               cdata->PartCommInfo[i+1], 888, MPI_COMM_WORLD);
      e_MPI_Send_MeshData(cdata->Strips[cdata->PartCommInfo[i+2]]->StripMesh,
                          cdata->PartCommInfo[i+1]);
      e_FreeMeshData(cdata->Strips[cdata->PartCommInfo[i+2]]->StripMesh);
      cdata->Strips[cdata->PartCommInfo[i+2]]->StripMesh = NULL;
    }
  }
}

/* merge two strip meshes together */
static void MergeStrips(MESH *component_mesh, MESH *final_mesh)
{
  int   i, j;
  int   node_no, elem_no, index, netindex;
  int  *existing_node, *mat_trans;
  MAT  *final_mat, *newmat;

  /* check for element types */
  if(final_mesh->NElemsTriang3 +
     final_mesh->NElemsTriang4 +
     final_mesh->NElemsTriang5 != final_mesh->TotalNElems)
  {
    e_GeneralError("Invalid local strip !");
  }

  if(component_mesh->NElemsTriang3 +
     component_mesh->NElemsTriang4 +
     component_mesh->NElemsTriang5 != component_mesh->TotalNElems)
  {
    e_GeneralError("Invalid strip arrived !");
  }

  /* check for existing nodes */
  existing_node = e_alloc_i_vect(component_mesh->NNodes);
  for(i = 0; i < component_mesh->NMeshPoints; i++)
  {
    for(j = 0; j < final_mesh->NMeshPoints; j++)
    {
      if(component_mesh->GNIs[i] == final_mesh->GNIs[j])
      {
        existing_node[i] = j+1;
        break;
      }
    }
  }

  /* store new nodes and their new number as well */
  node_no = final_mesh->NMeshPoints;
  for(i = 0; i < component_mesh->NMeshPoints; i++)
  {
    if(existing_node[i] == 0)
    {
      final_mesh->MeshPointCoords = 
        e_incr_cont_d_matr(node_no, 3, final_mesh->MeshPointCoords, 1);
      final_mesh->MeshPointCoords[node_no][0] = 
        component_mesh->MeshPointCoords[i][0];
      final_mesh->MeshPointCoords[node_no][1] = 
        component_mesh->MeshPointCoords[i][1];
      final_mesh->MeshPointCoords[node_no][2] = 
        component_mesh->MeshPointCoords[i][2];

      final_mesh->GNIs = e_incr_i_vect(node_no, final_mesh->GNIs, 1);
      final_mesh->GNIs[node_no] = component_mesh->GNIs[i];

      final_mesh->NodalDOF = e_incr_i_vect(node_no, final_mesh->NodalDOF, 1);
      final_mesh->NodalDOF[node_no] = 3;

      final_mesh->TotalNodalDOF += 3;

      node_no++;
      existing_node[i] = node_no;
    }
  }

  final_mesh->NMeshPoints = node_no;
  final_mesh->NNodes = node_no;

  /* store material pointer */
  final_mat = final_mesh->Props;
  /* material translation vector */
  mat_trans = e_alloc_i_vect(component_mesh->NMats);
  /* check available materials */
  for(i = 0; i < component_mesh->NMats; i++)
  {
    for(j = 0; j < final_mesh->NMats; j++)
    {
      if(strcmp(component_mesh->Props[i].Name,
                final_mesh->Props[j].Name) == 0)
      {
        mat_trans[i] = j;
        break;
      }
    }
    /* this is a new material, store it */
    if(j == final_mesh->NMats)
    {
      newmat = (MAT *)e_calloc(final_mesh->NMats+1, sizeof(MAT));
      memcpy((void *)(newmat),
             (void *)(final_mesh->Props),
             final_mesh->NMats * sizeof(MAT) );
      memcpy((void *)(newmat+final_mesh->NMats),
             (void *)(component_mesh->Props+i),
             sizeof(MAT) );
      final_mesh->Props = newmat;

      mat_trans[i] = final_mesh->NMats;

      final_mesh->NMats++;
    }
  }

  /* unfortunately the material structure has changed, so adjust */
  if(final_mat != final_mesh->Props)
  {
    for(i = 0; i < final_mesh->TotalNElems; i++)
    {
      index = ((MAT *)final_mesh->Mats[i]) - final_mat;
      final_mesh->Mats[i] = (void *)(final_mesh->Props + index);
    }
  }

  /* merge triangle elements */
  elem_no = final_mesh->TotalNElems;
  
  if(component_mesh->NElemsTriang3 > 0)
  {
    final_mesh->NodesTriang3 = 
      e_ext_cont_i_matr(final_mesh->NElemsTriang3, E_NNPE_TRIANG3,
                        final_mesh->NodesTriang3,
                        component_mesh->NElemsTriang3);
    final_mesh->GEIs = e_ext_i_vect(elem_no, final_mesh->GEIs, 
                                    component_mesh->NElemsTriang3);
    final_mesh->Mats = e_ext_vp_vect(elem_no, final_mesh->Mats,
                                     component_mesh->NElemsTriang3);
  }
  for(i = 0; i < component_mesh->NElemsTriang3; i++)
  {
    /* element definition */
    final_mesh->NodesTriang3[elem_no][0] = existing_node[component_mesh->NodesTriang3[i][0]-1];
    final_mesh->NodesTriang3[elem_no][1] = existing_node[component_mesh->NodesTriang3[i][1]-1];
    final_mesh->NodesTriang3[elem_no][2] = existing_node[component_mesh->NodesTriang3[i][2]-1];
    
    /* global element numbers */
    final_mesh->GEIs[elem_no] = component_mesh->GEIs[i];
    
    /* material for the elements */
    final_mesh->Mats[elem_no] = (void *)(final_mesh->Props + 
                                          (mat_trans[((MAT *)component_mesh->Mats[i]) - 
                                                     component_mesh->Props]) );
    elem_no++;
  }

  if(component_mesh->NElemsTriang4 > 0)
  {
    final_mesh->NodesTriang4 = 
      e_ext_cont_i_matr(final_mesh->NElemsTriang4, E_NNPE_TRIANG4,
                        final_mesh->NodesTriang4,
                        component_mesh->NElemsTriang4);
    final_mesh->GEIs = e_ext_i_vect(elem_no, final_mesh->GEIs, 
                                    component_mesh->NElemsTriang4);
    final_mesh->Mats = e_ext_vp_vect(elem_no, final_mesh->Mats,
                                     component_mesh->NElemsTriang4);
  }
  for(i = 0; i < component_mesh->NElemsTriang4; i++)
  {
    final_mesh->NodesTriang4[elem_no][0] = existing_node[component_mesh->NodesTriang4[i][0]-1];
    final_mesh->NodesTriang4[elem_no][1] = existing_node[component_mesh->NodesTriang4[i][1]-1];
    final_mesh->NodesTriang4[elem_no][2] = existing_node[component_mesh->NodesTriang4[i][2]-1];

    final_mesh->GEIs[elem_no] = component_mesh->GEIs[i];

    final_mesh->Mats[elem_no] = (void *)(final_mesh->Props + 
                                          (mat_trans[((MAT *)component_mesh->Mats[i]) - 
                                                     component_mesh->Props]) );
    elem_no++;
  }

  if(component_mesh->NElemsTriang5 > 0)
  {
    final_mesh->NodesTriang5 = 
      e_ext_cont_i_matr(final_mesh->NElemsTriang5, E_NNPE_TRIANG5,
                        final_mesh->NodesTriang5,
                        component_mesh->NElemsTriang5);
    final_mesh->GEIs = e_ext_i_vect(elem_no, final_mesh->GEIs, 
                                     component_mesh->NElemsTriang5);
    final_mesh->Mats = e_ext_vp_vect(elem_no, final_mesh->Mats,
                                      component_mesh->NElemsTriang5);
  }
  for(i = 0; i < component_mesh->NElemsTriang5; i++)
  {
    final_mesh->NodesTriang5[elem_no][0] = existing_node[component_mesh->NodesTriang5[i][0]-1];
    final_mesh->NodesTriang5[elem_no][1] = existing_node[component_mesh->NodesTriang5[i][1]-1];
    final_mesh->NodesTriang5[elem_no][2] = existing_node[component_mesh->NodesTriang5[i][2]-1];

    final_mesh->GEIs[elem_no] = component_mesh->GEIs[i];

    final_mesh->Mats[elem_no] = (void *)(final_mesh->Props + 
                                          (mat_trans[((MAT *)component_mesh->Mats[i]) - 
                                                     component_mesh->Props]) );
    elem_no++;
  }

  /* reset material pointers */
  netindex = 1;
  if(final_mesh->NElemsTriang3 > 0)
  {
    final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = netindex;
    final_mesh->MatsTriang3 = (MAT **)final_mesh->Mats +
                                      final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1;
    netindex += final_mesh->NElemsTriang3;
  }
  if(final_mesh->NElemsTriang4 > 0)
  {
    final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = netindex;
    final_mesh->MatsTriang4 = (MAT **)final_mesh->Mats +
                                      final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1;
    netindex += final_mesh->NElemsTriang4;
  }
  if(final_mesh->NElemsTriang5 > 0)
  {
    final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] = netindex;
    final_mesh->MatsTriang5 = (MAT **)final_mesh->Mats +
                                      final_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]-1;
    netindex += final_mesh->NElemsTriang5;
  }

  /* adjust element numbers */
  final_mesh->NElemsTriang3 += component_mesh->NElemsTriang3;
  final_mesh->NElemsTriang4 += component_mesh->NElemsTriang4;
  final_mesh->NElemsTriang5 += component_mesh->NElemsTriang5;

  final_mesh->TotalNElems = elem_no;

  e_free(existing_node);
  e_free(mat_trans);

}

#endif


/* ------------------------------------------------------------------------- */
/*                              Flattening                                   */
/* ------------------------------------------------------------------------- */

static void DoCreateCutPat(MESH *mesh, DR_DATA *data,
                           CUTTING_PATTERN *cdata)
{
  int i, j;
  DR_DATA *ldata;
  MESH    *lmesh;
  
  for(i = 0; i < cdata->NStrips; i++)
  {
    lmesh = cdata->Strips[i]->StripMesh;
    if(lmesh != NULL)
    {
      /* setup extra dynamic relacation structures */
      ldata = (DR_DATA *)e_calloc(1, sizeof(DR_DATA) );

      ldata->Projectname      = NULL;
      ldata->Outputname       = NULL;
      ldata->ProjectTask      = 0;
      ldata->OutputTask       = 0;
      ldata->Dimension        = 2;     /* !!! */
      ldata->StopAtConv       = TRUE;
      ldata->ConvCrit         = data->ConvCrit;
      ldata->PressureOption   = FALSE;
      ldata->PressureValue    = 0.0;
      /* ????? */
      ldata->AddMassOption    = data->AddMassOption;
      ldata->GlobalMassOption = data->GlobalMassOption;
      ldata->GlobalMassMult   = data->GlobalMassMult;
      ldata->NAddMass         = 0;
      ldata->AddMassNode      = NULL;
      ldata->AddMassMult      = NULL;
      ldata->CuttingMaterialFile = NULL;

      /* setup initial length */
      StripInitLengthFromStress(data, mesh, ldata, lmesh);

      ldata->NCompNodes       = 0;
      ldata->CompNodes        = NULL;

      /* no interaction, whatsoever */
#ifdef __E_MPI__
      ldata->CommReq    = FALSE;
      ldata->CommEnergy = NULL;

      ldata->CommData   = NULL;
#endif

      /* convert all triangle types into TRIANG3 */
      ConvertMeshForCutPat(lmesh);

      /* element types change before */
      SetupDrDataFinal(ldata, lmesh);

      /* make the mesh statically determinant with boundary conditions,
         we do it here, so if strips merged, 
         this one should be done only once */
      MakeStaticDetMesh(lmesh);

#ifdef __DEBUG_3__
      {
        char ff[E_LINE_MAX+1];
  
        sprintf(ff, "%s-before-%d", data->Outputname, i+1);

        PutInitLengthData(ff, ldata->InitLengthData, lmesh);
      }
#endif

      /* push the mesh into the X-Y plane */
      for(j = 0; j < lmesh->NMeshPoints; j++)
      {
        lmesh->MeshPointCoords[j][2] = 0.0;
      }

#ifdef __DEBUG_2__
      {
        char ff[E_LINE_MAX+1];
  
        sprintf(ff, "%s-flat-%d", data->Outputname, i+1);
        e_PutMeshData(lmesh, ff);
      }
#endif

      /* do the cutting pattern generation */
      cdata->Strips[i]->ResultOK = !(dr_main(lmesh, ldata));

#ifdef __DEBUG_4__
      {
        char ff[E_LINE_MAX+1];
  
        sprintf(ff, "%s-after-%d", data->Outputname, i+1);

        GenInitLengthData(ldata->InitLengthData, lmesh, 
                          0, 0, 0, 0, 0, 0, 0, 0);
        PutInitLengthData(ff, ldata->InitLengthData, lmesh);
      }
#endif

      /* free-up memory */
      FreeDrDataInit(ldata);
    }
  }

}


/* -------------------------------------------------------------------------- */

void dr_cutting_pattern(char *ProjectName, DR_DATA *data, MESH *mesh, 
                        CUTTING_PATTERN **cdata)
{

#ifdef __E_MPI__
  int i;
  MPI_Status  status;
  int         contcutpat;
  GA_DATA    *gadata;

#ifndef DD_TYPE_COMM
  e_GeneralWarning("Cutting Pattern generation is NOT implemented for this communication !");
  return;
#endif
#endif

  /* change material */
/*  ReplaceMaterial(mesh, data); */

  /* reset loads, bc conditions */
  RemoveLoads(mesh);
  RemoveBCs(mesh);

  /* create meshes (strips) to flatten */
  CreateCutPatMesh(mesh, data, cdata);

#ifdef __E_MPI__
  /* send strip information to the master */
  StripsToMaster(*cdata, data->CommData->Par_ProcID, 
                 data->CommData->Par_NSubdomains);
  if(data->CommData->Par_ProcID == 1)
  {
    /* assemble half strips, if any */
    StripAssemble((*cdata)->MasterData, data->CommData->Par_ProcID, 
                  data->CommData->Par_NSubdomains);
    
    printf("\n--- Number of assembled strips is: %d\n",
           (*cdata)->MasterData->NAssembledPartStrip);

    /* if there is any half strip which was assembled  */
    if((*cdata)->MasterData->NAssembledPartStrip)
    {
      /* read in genetic algorithm info */
      contcutpat = init_GA_data(ProjectName, &gadata, &((*cdata)->MasterData->Gamma));
    }
    else
    {
      contcutpat = 1;
    }
  }

  /* pass along whether we can proceed */
  if(data->CommData->Par_ProcID == 1)
  {
    for(i = 1; i < data->CommData->Par_NSubdomains; i++)
      MPI_Send(&(contcutpat), 1, MPI_INT, i, 2, MPI_COMM_WORLD);
  }
  else
  {
    MPI_Recv(&(contcutpat), 1, MPI_INT, 0, 2, MPI_COMM_WORLD, &status);
  }

  if(contcutpat == TRUE)
  {
    /* on the master, lets do the strip balancing if necessary */
    if((data->CommData->Par_ProcID == 1) && ((*cdata)->MasterData->NAssembledPartStrip))
    {
      DoBalancing(data->CommData->Par_NSubdomains, *cdata, gadata);
    }
    
    /* send info to subdomains */
    StripsFromMaster(data->CommData->Par_ProcID, 
                     data->CommData->Par_NSubdomains,
                     *cdata);

    /* all subdomains send and receive their half strips */
    StripToSubdoms(data->CommData->Par_ProcID, 
                   data->CommData->Par_NSubdomains,
                   *cdata);

#endif
    
    /* carry out the cutting pattern generation */
    DoCreateCutPat(mesh, data, *cdata);

#ifdef __E_MPI__
  }
#endif

}





