/*
 * Functions to handle the initial part of the DR_DATA structure
 *   - reading from file or keyboard input
 *   - extraction of data for subdomain
 *   - freein-up the structure
 *   - MPI Sending and receiving of the structure
 *
 *   - read additional mass data (at the moment this data is closely coupled
 *     with the DR_DATA structure)
 *
 * Handles all input substructures except the communication related data !!!
 * Calculation related data (like residual, velocity, etc.) are not handled.
 */

#include "dr_data.h"

#include "dr_len.h" /* initial or slack length handling */

#ifdef __E_MPI__
#include "mpi.h"
#endif

/* for error messages in case of keyboard input */
#define KBD_NAME "in keyboard input"

/* this function is in E-lib, but is not shown in header file (why?) */
extern int GetKeyword(char *line, char *word, int *LineNumber,
                      char *FullPathName, FILE *fp);


/* read in the DR_DATA options from file or keyboard input */
void GetDrDataInit(char *FileName, DR_DATA **data)
{
  FILE *InputFP = NULL;
  int   KbdInput = FALSE;
  int   LineCount = 0;
  char  Line[E_LINE_MAX+1];
  char  yes_or_no[E_LINE_MAX+1];

  if(FileName != NULL)
  {
    if((InputFP = fopen(FileName, "r")) != NULL)
    {
      KbdInput = FALSE;
    }
    else
    {
      e_GeneralError("\nERROR: Could not open file ``%s''\n", FileName);
    }
  }
  else
  {
    KbdInput = TRUE;
  }
  
  if(KbdInput)
  {
    printf("\nReading from keyboard input ...\n\n");
    KbdInput = TRUE;
    InputFP  = stdin; /* standard input file handler */
    FileName = e_alloc_c_vect(strlen(KBD_NAME)+1);
    strcpy(FileName, KBD_NAME);
  }

  (*data) = (DR_DATA *)e_calloc(1, sizeof(DR_DATA));

  /* project name */
  if(KbdInput)
  {
    printf("\nProjectname (MDF file without extension) : \n");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);

  (*data)->Projectname = e_alloc_c_vect(strlen(Line)+1);

  if(sscanf(Line, "%s", (*data)->Projectname) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read projectname");
  }

  /* output name */
  if(KbdInput)
  {
    printf("\nName of the result file(s) (no extension): \n");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);

  (*data)->Outputname = e_alloc_c_vect(strlen(Line)+1);

  if(sscanf(Line, "%s", (*data)->Outputname) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read result file name");
  }

  if(strcmp((*data)->Projectname, (*data)->Outputname) == 0)
  {
    e_GeneralError("Projectname and result file name are the same !");
  }

  /* task option */
  if(KbdInput)
  {
    printf("\nFor combined tasks add the numbers together.");
    printf("\n\nTask to do :");
    printf("\n    %d - only form-finding", TASK_FORM_FINDING);
    printf("\n    %d - only cutting pattern generation", TASK_CUTTING_PATTERN);
    printf("\n    %d - only initial length generation\n", TASK_GEN_INIT_LENGTH);
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%d", &((*data)->ProjectTask)) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read task option");
  }

  if(!(   ((*data)->ProjectTask & TASK_FORM_FINDING) 
       || ((*data)->ProjectTask & TASK_CUTTING_PATTERN) 
       || ((*data)->ProjectTask & TASK_GEN_INIT_LENGTH)))
  {
    e_FileLineError(LineCount, FileName, "invalid task option");
  }

  /* dimension of the problem */
  if(KbdInput)
  {
    printf("\n\nDimension of the problem [2/3]: ");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%d", &((*data)->Dimension)) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read dimension of the problem");
  }

  if(!(((*data)->Dimension == 2) || ((*data)->Dimension == 3)))
  {
    e_FileLineError(LineCount, FileName, "invalid dimension");
  }

  /* additional mass */
  if(KbdInput)
  {
    printf("\n\nDid you specify additional masses [0|1|y|n]: ");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%s", yes_or_no) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read additional mass option");
  }

  if(!(   yes_or_no[0] == '0' || yes_or_no[0] == '1'
       || yes_or_no[0] == 'y' || yes_or_no[0] == 'n'
       || yes_or_no[0] == 'Y' || yes_or_no[0] == 'N'))
  {
    e_FileLineError(LineCount, FileName, "invalid additional mass option");
  }

  if(   yes_or_no[0] == '1' 
     || yes_or_no[0] == 'y'
     || yes_or_no[0] == 'Y')
  {
    (*data)->AddMassOption = 1;
  }
  else
  {
    (*data)->AddMassOption = 0;
  }

  /* pressure load */
  if(KbdInput)
  {
    printf("\n\nDo you want to specify pressure load [0|1|y|n] ?");
    printf("\nIf yes, type the pressure load value at the same line.\n");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%s", yes_or_no) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read pressure load option");
  }

  if(!(   yes_or_no[0] == '0' || yes_or_no[0] == '1'
       || yes_or_no[0] == 'y' || yes_or_no[0] == 'n'
       || yes_or_no[0] == 'Y' || yes_or_no[0] == 'N'))
  {
    e_FileLineError(LineCount, FileName, "invalid pressure load option");
  }

  if(   yes_or_no[0] == '1' 
     || yes_or_no[0] == 'y'
     || yes_or_no[0] == 'Y')
  {
    (*data)->PressureOption = 1;
  }
  else
  {
    (*data)->PressureOption = 0;
  }

  if((*data)->PressureOption)
  {
    if(sscanf(Line, "%*s %lf", &((*data)->PressureValue)) != 1)
    {
      e_FileLineError(LineCount, FileName, "cannot read pressure load value");
    }
  }


  /* hydraulic pressure load */
  if(KbdInput)
  {
    printf("\n\nDo you want to specify hydraulic pressure load [0|1|y|n] ?");
    printf("\nIf yes, type the density at the same line.\n");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%s", yes_or_no) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read hydraulic pressure load option");
  }

  if(!(   yes_or_no[0] == '0' || yes_or_no[0] == '1'
       || yes_or_no[0] == 'y' || yes_or_no[0] == 'n'
       || yes_or_no[0] == 'Y' || yes_or_no[0] == 'N'))
  {
    e_FileLineError(LineCount, FileName, "invalid hydraulic pressure load option");
  }

  if(   yes_or_no[0] == '1' 
     || yes_or_no[0] == 'y'
     || yes_or_no[0] == 'Y')
  {
    (*data)->HydPressOption = 1;
  }
  else
  {
    (*data)->HydPressOption = 0;
  }

  if((*data)->HydPressOption)
  {
    if(sscanf(Line, "%*s %lf", &((*data)->HydPressDensity)) != 1)
    {
      e_FileLineError(LineCount, FileName, "cannot read density for hydraulic pressure");
    }
  }

  /* convergence option */
  if(KbdInput)
  {
    printf("\n\nStop at convergence [0|1|y|n] ?");
    printf("\nIf yes, type convergence criteria at the same line.\n");
  }
  e_GetNextLine(Line, &LineCount, FileName, InputFP);
  if(sscanf(Line, "%s", yes_or_no) != 1)
  {
    e_FileLineError(LineCount, FileName, "cannot read convergence option");
  }

  if(!(   yes_or_no[0] == '0' || yes_or_no[0] == '1'
       || yes_or_no[0] == 'y' || yes_or_no[0] == 'n'
       || yes_or_no[0] == 'Y' || yes_or_no[0] == 'N'))
  {
    e_FileLineError(LineCount, FileName, "invalid convergence option");
  }

  if(   yes_or_no[0] == '1' 
     || yes_or_no[0] == 'y'
     || yes_or_no[0] == 'Y')
  {
    (*data)->StopAtConv = 1;
  }
  else
  {
    (*data)->StopAtConv = 0;
  }

  /* convergence criteria */
  if((*data)->StopAtConv)
  {
    if(sscanf(Line, "%*s %lf", &((*data)->ConvCrit)) != 1)
    {
      e_FileLineError(LineCount, FileName, "cannot read convergence criteria");
    }

    if((*data)->ConvCrit < 0.0)
    {
      e_FileLineError(LineCount, FileName, "invalid convergence criteria");
    }
  }

  /* FIXME:
   * NOT USED, because the same material should be used for cutting pattern
   * generation as for form-finding.
   */
#if 0
  /* cutting pattern material file */
  if((*data)->ProjectTask & TASK_CUTTING_PATTERN)
  {
    if(KbdInput)
    {
      printf("\n\nMaterial file for cutting pattern generation (no extension) :");
    }
    e_GetNextLine(Line, &LineCount, FileName, InputFP);
    (*data)->CuttingMaterialFile = e_alloc_c_vect(strlen(Line)+1);
    if(sscanf(Line, "%s", (*data)->CuttingMaterialFile) != 1)
    {
      e_FileLineError(LineCount, FileName, 
                      "cannot read material file name for cutting pattern generation");
    }
  }
#endif

  if(KbdInput)
  {
    e_free(FileName);
  }
  else
  {
    fclose(InputFP);
  }

}

void SetupDrDataFinal(DR_DATA *data, MESH *mesh)
{

  data->A = (1.0 - (mesh->DampingFactor / 2.0)) / 
            (1.0 + (mesh->DampingFactor / 2.0));
  data->B = 1.0 / (1.0 + (mesh->DampingFactor / 2.0));

  data->CurrentEnergy = 0.00;

  data->Pressure = e_alloc_cont_d_matr(mesh->NMeshPoints, data->Dimension);
  data->Normal   = e_alloc_cont_d_matr(mesh->NMeshPoints, data->Dimension);
  data->Residual = e_alloc_cont_d_matr(mesh->NMeshPoints, data->Dimension);
  data->Velocity = e_alloc_cont_d_matr(mesh->NMeshPoints, data->Dimension);
  data->Mass     = e_alloc_cont_d_matr(mesh->NMeshPoints, data->Dimension);

  if(mesh->NElemsTriang3)
    data->KeTriang3 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang3, 3, 3);

  if(mesh->NElemsTriang4)
    data->KeTriang4 = e_alloc_cont_d_3D_matr(mesh->NElemsTriang4, 3, 3);

  if(data->ProjectTask &  TASK_CHECK_STRESSES)
  {
    data->InitCoords = e_alloc_cont_d_matr(mesh->NMeshPoints, 3);
    memcpy((void *)(data->InitCoords[0]), (void *)(mesh->MeshPointCoords[0]),
           sizeof(double) * 3 * mesh->NMeshPoints);
  }
}


/* free-up the DR_DATA structure */
void FreeDrDataInit(DR_DATA *data)
{
  if(data->Projectname)
  {
    e_free(data->Projectname);
  }
  if(data->Outputname)
  {
    e_free(data->Outputname);
  }
  if(data->CuttingMaterialFile)
  {
    e_free(data->CuttingMaterialFile);
  }

  if(data->NAddMass)
  {
    e_free(data->AddMassNode);
    e_free_cont_d_matr(data->Dimension, data->AddMassMult);
  }

  FreeInitLengthData(data->InitLengthData);

  e_free(data);
}


/* for parallel version send and receive the DR_DATA to/from the subdomains */
#ifdef __E_MPI__


/* create DR_DATA structure for the subdomain */
void CreateSubDrDataInit(DR_DATA *Gdata, MESH *Gmesh,
                         DR_DATA **data, MESH *mesh,
                         int i_subdom, int NInterSubDomNodes, 
                         int *InterSubDomNodes, int *NNodeSubDoms, 
                         int **NodeSubDoms)
{
  int i_m, i_n, i, j, node, index;

  (*data) = e_calloc(1, sizeof(DR_DATA));

  (*data)->Projectname      = NULL;
  (*data)->Outputname       = NULL;
  (*data)->ProjectTask      = Gdata->ProjectTask;
  (*data)->OutputTask       = Gdata->OutputTask;
  (*data)->Dimension        = Gdata->Dimension;
  (*data)->StopAtConv       = Gdata->StopAtConv;
  (*data)->ConvCrit         = Gdata->ConvCrit;
  (*data)->PressureOption   = Gdata->PressureOption;
  (*data)->PressureValue    = Gdata->PressureValue;

  (*data)->HydPressOption   = Gdata->HydPressOption;
  (*data)->HydPressDensity  = Gdata->HydPressDensity;

  (*data)->AddMassOption    = Gdata->AddMassOption;
  (*data)->GlobalMassOption = Gdata->GlobalMassOption;
  (*data)->GlobalMassMult   = Gdata->GlobalMassMult;

  /* setup additional masses for the subdomain */
  (*data)->AddMassNode = NULL;
  (*data)->AddMassMult = NULL;
  (*data)->NAddMass    = 0;

  if(Gdata->NAddMass)
  {
    for(i_m = 0; i_m < Gdata->NAddMass; i_m++)
    {
      for(i_n = 0; i_n < mesh->NMeshPoints; i_n++)
      {
        if(Gdata->AddMassNode[i_m] == mesh->GNIs[i_n])
        {
          (*data)->AddMassNode = 
            e_incr_i_vect((*data)->NAddMass, (*data)->AddMassNode, ALLOC_SIZE);
          (*data)->AddMassNode[(*data)->NAddMass] = i_n+1;
          (*data)->AddMassMult = 
            e_incr_cont_d_matr((*data)->NAddMass, Gdata->Dimension, 
                               (*data)->AddMassMult, ALLOC_SIZE);
          memcpy((void *)((*data)->AddMassMult[(*data)->NAddMass]),
                 (void *)(Gdata->AddMassMult[i_m]),
                 sizeof(double) * Gdata->Dimension);
          (*data)->NAddMass++;
          break;
        }
      }
    }
  }

  CreateSubInitLengthData(Gdata->InitLengthData, Gmesh, 
                          &((*data)->InitLengthData), mesh);

  /* determine nodes which are participating in energy calculation */
  (*data)->CommEnergy = e_alloc_i_vect(mesh->NMeshPoints);
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    (*data)->CommEnergy[i] = TRUE;
  }

  for(i = 0; i < NInterSubDomNodes; i++)
  {
    for(j = 0; j < NNodeSubDoms[i]; j++)
    {
      /* this node is in the current subdomain */
      if(i_subdom == NodeSubDoms[i][j])
      {
        node = InterSubDomNodes[i];
        for(index = 0; index < mesh->NMeshPoints; index++)
        {
          if(mesh->GNIs[index] == node)
            break;
        }
        /* check that we have the node locally in the mesh */
        if(index == mesh->NMeshPoints)
        {
          e_GeneralError("Communication node %d found in mesh!", node);
        }

        if(j == 0)
        {
          (*data)->CommEnergy[index] = TRUE;
        }
        else
        {
          (*data)->CommEnergy[index] = FALSE;
        }
      }
    }
  }


}


#define NO_INT_INFO     9
#define NO_DOUBLE_INFO  4

void SendDrDataInit(DR_DATA *data, MESH *mesh, int subdom)
{
  int     idata[NO_INT_INFO];
  double  ddata[NO_DOUBLE_INFO];

  idata[0] = data->ProjectTask;
  idata[1] = data->OutputTask;
  idata[2] = data->Dimension;
  idata[3] = data->StopAtConv;
  idata[4] = data->PressureOption;
  idata[5] = data->AddMassOption;
  idata[6] = data->GlobalMassOption;
  idata[7] = data->NAddMass;
  idata[8] = data->HydPressOption;
  
  ddata[0] = data->ConvCrit;
  ddata[1] = data->PressureValue;
  ddata[2] = data->GlobalMassMult;
  ddata[3] = data->HydPressDensity;
  
  MPI_Send(idata, NO_INT_INFO, MPI_INT,
           subdom, 200, MPI_COMM_WORLD);
  MPI_Send(ddata, NO_DOUBLE_INFO, MPI_DOUBLE, 
           subdom, 201, MPI_COMM_WORLD);
  
  /* if there is additional mass send it */
  if(data->NAddMass)
  {
    MPI_Send(data->AddMassNode, data->NAddMass, MPI_INT, 
             subdom, 203, MPI_COMM_WORLD);
    MPI_Send(data->AddMassMult[0], data->NAddMass * data->Dimension, MPI_DOUBLE, 
             subdom, 204, MPI_COMM_WORLD);
  }

  /* send initial length */
  SendInitLengthData(data->InitLengthData, mesh, subdom);

  /* send info, whether the nodes participates in energy calculation */
  MPI_Send(data->CommEnergy, mesh->NMeshPoints, MPI_INT,
           subdom, 205, MPI_COMM_WORLD);

}


void RecvDrDataInit(DR_DATA **data, MESH *mesh, int subdom)
{
  int        idata[NO_INT_INFO];
  double     ddata[NO_DOUBLE_INFO];
  MPI_Status status;

  MPI_Recv(idata, NO_INT_INFO, MPI_INT,
           subdom, 200, MPI_COMM_WORLD, &status);
  MPI_Recv(ddata, NO_DOUBLE_INFO, MPI_DOUBLE, 
           subdom, 201, MPI_COMM_WORLD, &status);

  (*data) = e_calloc(1, sizeof(DR_DATA) );

  (*data)->ProjectTask      = idata[0];
  (*data)->OutputTask       = idata[1];
  (*data)->Dimension        = idata[2];
  (*data)->StopAtConv       = idata[3];
  (*data)->PressureOption   = idata[4];
  (*data)->AddMassOption    = idata[5];
  (*data)->GlobalMassOption = idata[6];
  (*data)->NAddMass         = idata[7];
  (*data)->HydPressOption   = idata[8];

  (*data)->ConvCrit         = ddata[0];
  (*data)->PressureValue    = ddata[1];
  (*data)->GlobalMassMult   = ddata[2];
  (*data)->HydPressDensity  = ddata[3];

  /* additional mass data */
  if((*data)->NAddMass)
  {
    (*data)->AddMassNode = e_alloc_i_vect((*data)->NAddMass);
    (*data)->AddMassMult = e_alloc_cont_d_matr((*data)->NAddMass, (*data)->Dimension);
    MPI_Recv((*data)->AddMassNode, (*data)->NAddMass, MPI_INT, 
             subdom, 203, MPI_COMM_WORLD, &status);
    MPI_Recv((*data)->AddMassMult[0], (*data)->NAddMass * (*data)->Dimension, MPI_DOUBLE, 
             subdom, 204, MPI_COMM_WORLD, &status);
  }

  /* setup (!) and receive initial length data */
  RecvInitLengthData(&((*data)->InitLengthData), mesh, subdom);

  (*data)->CommEnergy = e_alloc_i_vect(mesh->NMeshPoints);
  MPI_Recv((*data)->CommEnergy, mesh->NMeshPoints, MPI_INT,
           subdom, 205, MPI_COMM_WORLD, &status);

}

#endif

/* ------------------------------------------------------------------------- */
/*                             Additional mass                               */
/* ------------------------------------------------------------------------- */

/* read in the additional mass file */
void GetMassData(DR_DATA *data, int NNodes)
{
  FILE *InputFP;
  int   LineCount = 0;
  char *FileName;
  char  Line[E_LINE_MAX+1];
  char  Word[E_LINE_MAX+1];
  int   ReadNextLine;
  int   i;

  /* no addition mass should be read */
  if(data->AddMassOption == 0)
    return;

  /* in case of empty file */
  data->GlobalMassOption = 0;
  data->NAddMass = 0;

  FileName = e_alloc_c_vect(strlen(data->Projectname) + strlen(E_MAS) + 2);
  sprintf(FileName, "%s.%s", data->Projectname, E_MAS);

  /* open the mass file */
  if((InputFP = fopen(FileName, "r")) == NULL)
  {
    e_GeneralError("Cannot open additional mass file ``%s'' !", FileName);
  }

  /* global multiplication factor */
  if(GetKeyword(Line, Word, &LineCount, FileName, InputFP) == EOF)
  {
    return;
  }
  ReadNextLine = FALSE;
  
  /* if global keyword is specified */
  if(strcmp(E_KW_MAS_GLOBAL, Word) == 0)
  {
    data->GlobalMassOption = 1;
    if(sscanf(Line, "%*s %lf", &(data->GlobalMassMult)) != 1)
    {
      e_FileLineError(LineCount, FileName,
        "in mass file, cannot read global multiplier value");
    }
    ReadNextLine = TRUE;
  }

  /* local multiplication factor */
  if(   ReadNextLine
     && GetKeyword(Line, Word, &LineCount, FileName, InputFP) == EOF)
  {
    return;
  }
  ReadNextLine = FALSE;

  if(strcmp(E_KW_MAS_LOCAL, Word) == 0)
  {
    if(sscanf(Line, "%*s %d", &(data->NAddMass)) != 1)
    {
      e_FileLineError(LineCount, FileName,
        "in mass file, cannot read no of additional masses");
    }
    ReadNextLine = TRUE;
    
    data->AddMassNode  = e_alloc_i_vect(data->NAddMass);
    data->AddMassMult = e_alloc_cont_d_matr(data->NAddMass, data->Dimension);

    for(i = 0; i < data->NAddMass; i++)
    {
      e_GetNextLine(Line, &LineCount, FileName, InputFP);

      if(data->Dimension == 2)
      {
        if(sscanf(Line, "%d %lf %lf", &(data->AddMassNode[i]),
                  &(data->AddMassMult[i][0]), &(data->AddMassMult[i][1])) != 3)
          e_FileLineError(LineCount, FileName,
                          "in mass file, cannot interpret the line");
      }
      else /* dimension == 3 */
      {
        if(sscanf(Line, "%d %lf %lf %lf", &(data->AddMassNode[i]), 
                  &(data->AddMassMult[i][0]), &(data->AddMassMult[i][1]), 
                  &(data->AddMassMult[i][2])                              ) != 4)
          e_FileLineError(LineCount, FileName,
                          "in mass file, cannot interpret the line");
      }

      if(data->AddMassNode[i] < 1 || data->AddMassNode[i] > NNodes)
        e_FileLineError(LineCount, FileName,
                        "invalid node number");
    }
  }

  if(ReadNextLine == FALSE)
  {
    e_FileLineError(LineCount, FileName, "in mass file, invalid line");
  }

  fclose(InputFP);

  e_free(FileName);

  printf("\nThe additional mass information has been loaded in successfuly");
}



