/*
 * All initialisation functions for the dynamic relaxation method
 */

#include "dr_io.h"   /* everything externally visible */
#include "myarg.h"   /* program argument handling */
#include "dr_len.h"  /* initial or slack length handling */
#include "dr_data.h" /* DR_DATA initial part handling */
#include "dr_cut.h"  /* CUTTING_PATTERN data structure */
#include "dr_stress.h"
#include "e_lib.h"

#ifdef __E_MPI__
#include "mpi.h"     /* MPI functions */
#include "dr_comm.h" /* communication functions */
#endif


static void GetUserOptions(int argc, char **argv, DR_DATA **data);

static void SelfWeightTriang(int dimension, int nelems, int **nodes, 
                             double **coords, MAT **material, double **load);
static void GenSelfWeight(int dimension, MESH *mesh);


static void CollectCompNodes(MESH *mesh, DR_DATA *data);
static void CollectLinkCompNodes(int n, int **nodes, MAT **material,
                                 int *NCompNodes, int **CompNodes);


/* ------------------------------------------------------------------------- */
/*                               User options                                */
/* ------------------------------------------------------------------------- */

static void GetUserOptions(int argc, char **argv, DR_DATA **data)
{
  char *FileName;

  /* determine where the input will come from */
  if(GetStringArg(argc, argv, "-i", &FileName) == 0)
  {
    FileName = NULL;
  }

  /* read in user options */
  GetDrDataInit(FileName, data);

  if(FileName != NULL)
  {
    e_free(FileName);
  }

  /* write back options for the user */
  printf("\n-----");
  printf("\n----- Project name is taken as: %s", (*data)->Projectname);
  printf("\n-----");

  if((*data)->ProjectTask & TASK_FORM_FINDING)
    printf("\nThe program WILL do form-finding.");
  if((*data)->ProjectTask & TASK_CUTTING_PATTERN)
    printf("\nThe program WILL generate cutting-pattern.");
  if((*data)->ProjectTask & TASK_GEN_INIT_LENGTH)
    printf("\nThe program WILL generate initial or slack length file.");

  printf("\nDimension of the problem is taken as: %d", (*data)->Dimension);

  if((*data)->AddMassOption)
    printf("\nAdditional mass IS specified.");
  else
    printf("\nAdditional mass is NOT speficied.");

  if((*data)->PressureOption)
    printf("\nPressure load IS specified.");
  else
    printf("\nPressure load is NOT speficied.");

  if((*data)->StopAtConv)
    printf("\nThe program WILL stop after convergence.");
  else
  {
    printf("\nThe program will NOT stop after convergence.");
    printf("\nIt runs for the specified number of iterations.");
  }
  
  printf("\n\n");
}

/* ------------------------------------------------------------------------- */
/*                           Initialise everything                           */
/* ------------------------------------------------------------------------- */

#ifdef __E_MPI__
void InitDR(int argc, char **argv, 
            DR_DATA **Gdata, MESH **Gmesh,
            DR_DATA **data,  MESH **mesh)
#else
void InitDR(int argc, char **argv, 
            DR_DATA **data, MESH **mesh)
#endif
{

#ifdef __E_MPI__
  int NProc, NSubdomains, ProcID;

  /* initialise MPI and determine process ID */
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &(ProcID));
  ProcID += 1;

  /* only master interacts with the user */
  if(ProcID == 1)
  {
#endif
    /* read user options for running the program */
    GetUserOptions(argc, argv, data);

    /* read the mesh */
    e_GetMeshData(mesh, (*data)->Projectname);
    printf("\n\nThe mesh has been loaded in succesfully with the title:\n<%s>", 
           (*mesh)->Title);
    printf("\n    Total number of elements: %d", (*mesh)->TotalNElems);
    printf("\n    Total number of nodes: %d", (*mesh)->NNodes);

    /* if it is kinetic damping, so shall be */
    if((*mesh)->DampingFactor < 0.00001)
    {
      (*mesh)->DampingFactor = 0;
    }
    
    /* read material information */
    if(e_GetMaterials(*mesh, (*data)->Projectname))
      printf("\nThe Material information has been loaded in successfuly");

    /* read in or generate initial length */
    GetInitLengthData((*data)->Projectname, &((*data)->InitLengthData), *mesh);

    /* read in additional mass if required */
    GetMassData(*data, (*mesh)->NMeshPoints);

    /* generate self-weight as a static load, 
       it must be before the parallel initialisation because it is a static weight
       and it will be distributed when the mesh is cut !!! */
    GenSelfWeight((*data)->Dimension, *mesh);

#ifdef __E_MPI__
    /* load subdomain information */
    e_GetDecompData(*mesh, (*data)->Projectname);
    NSubdomains = (*mesh)->Decomp.NSubDoms;

    printf("\nThe total number of subdomains is taken as: %d", NSubdomains);
    
    /* determine number of MPI processes and compare it with number of subdomains 
       to be sure that they are equal */
    MPI_Comm_size(MPI_COMM_WORLD, &NProc);
    if(NProc != NSubdomains)
      e_GeneralError("Number of processors is NOT equal to number of processes !");
    
    /* store the global mesh and dr data */
    (*Gmesh) = *mesh;
    (*mesh)  = NULL;
    (*Gdata) = *data;
    (*data)  = NULL;
  }
  else
  {
    MPI_Comm_size(MPI_COMM_WORLD, &(NSubdomains));
  }

  /* send information to subdomains and setup communication structure as well */
  SetupSubdomains(ProcID, NSubdomains, *Gdata, *Gmesh, data, mesh);
#endif

  /* collect compressed nodes (per subdomain) */
  CollectCompNodes(*mesh, *data);

  /* setup the rest of DR_DATA */
  SetupDrDataFinal(*data, *mesh);

#ifdef __E_MPI__
  printf("\n%d. proc finished initialisation", ProcID);
#else
  printf("\nInitialisation finished.");
#endif


#ifdef __E_MPI__
  /* logging communication and calculation time with MPI */
#ifdef COMM_LOG
  MPE_Init_log();
  if(ProcID == 1)
  {
    MPE_Describe_state(1, 2, "Normals", "red:vlines3");
    MPE_Describe_state(3, 4, "Masses", "blue:gray3");
    MPE_Describe_state(5, 6, "Residuals", "green:light_gray");
    MPE_Describe_state(7, 8, "Energy", "yellow:gray");
  }
#endif

#endif
}



/* calculates the self-weight of one type of triangles */
static void SelfWeightLink(int dimension, int nelems, int **nodes, 
                           double **coords, MAT **material, double **load)
{
  int    l, j;
  int    n1, n2;
  double weight, length;

  /* go through all elements */
  for(l = 0; l < nelems; l++)
  {
    weight = material[l]->dParam5;
    /* if there is self weight per meter */
    if(weight != 0.0 && dimension == 3)
    {
      n1 = nodes[l][0]-1;
      n2 = nodes[l][1]-1;

      length = 0.00;
      for(j = 0; j < dimension; j++)
      {
        length += E_SQUARE(coords[n2][j] - coords[n1][j]);
      }
      length = sqrt(length);
      
      /* distribute it to all nodes of the link */
      load[n1][2] += -weight * length / 2.0;
      load[n2][2] += -weight * length / 2.0;
    }
  }
}


/* calculates the self-weight of one type of triangles */
static void SelfWeightTriang(int dimension, int nelems, int **nodes, 
                             double **coords, MAT **material, double **load)
{
  int    l, i, j;
  int    n1, n2, n3;
  double side_length[3];
  double dl, weight, pressure, sum, area;
  int    side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};
  double x21, y21, z21, x31, y31, z31;
  double norm[3];
  

  /* go through all elements */
  for(l = 0; l < nelems; l++)
  {
    weight = material[l]->dParam5;
    pressure = material[l]->dParam6;
    
    /* if there is self weight per square meter */
    if((weight != 0.0 || pressure != 0) && dimension == 3)
    {
      /* we need it for the calculation of area */
      for(i = 0; i < 3; i++)
      {
        n1 = nodes[l][side_n_node[i][0]]-1;
        n2 = nodes[l][side_n_node[i][1]]-1;
        side_length[i] = 0.00;
        for(j = 0; j < dimension; j++)
        {
          dl = (coords[n2][j] - coords[n1][j]);
          side_length[i] += E_SQUARE(dl);
        }
        side_length[i] = sqrt(side_length[i]);
      }
      
      sum = (side_length[0] + side_length[1] + side_length[2]) / 2.00;
      area = sqrt(sum * (sum - side_length[0]) * 
                        (sum - side_length[1]) *
                        (sum - side_length[2])   );

      n1 = nodes[l][0]-1;
      n2 = nodes[l][1]-1;
      n3 = nodes[l][2]-1;

      x21 = coords[n2][0] - coords[n1][0];
      y21 = coords[n2][1] - coords[n1][1];
      z21 = coords[n2][2] - coords[n1][2];

      x31 = coords[n3][0] - coords[n1][0];
      y31 = coords[n3][1] - coords[n1][1];
      z31 = coords[n3][2] - coords[n1][2];

/*
      x23 = coords[n2][0] - coords[n3][0];
      y23 = coords[n2][1] - coords[n3][1];
      z23 = coords[n2][2] - coords[n3][2];
*/
      
      norm[0] = (y21 * z31) - (y31 * z21);
      norm[1] = (z21 * x31) - (z31 * x21);
      norm[2] = (x21 * y31) - (x31 * y21);

      sum = sqrt(E_SQUARE(norm[0]) + 
                 E_SQUARE(norm[1]) + 
                 E_SQUARE(norm[2])   );
      norm[0] /= sum;
      norm[1] /= sum;
      norm[2] /= sum;

      /* distribute it to all nodes of the triangle */
      if(weight != 0.0)
      {
        for(i = 0; i < 3; i++)
        {
          n1 = nodes[l][i]-1;
          load[n1][2] += -weight * area / 3;
        }
      }
      
      /* pressure is normal to the surface */
      if(pressure != 0.0)
      {
        for(i = 0; i < 3; i++)
        {
          n1 = nodes[l][i]-1;
          for(j = 0; j < dimension; j++)
          {
            load[n1][j] += norm[j] * (-pressure) * area / 3;
          }
        }
      }
      
    }
  }
}


/* calculates the self of membrane elements, 
   self weight is considered as a static load !!! */
static void GenSelfWeight(int dimension, MESH *mesh)
{
  int i, j;
  int *newnodes;
  double **newloads;

  newnodes = e_alloc_i_vect(mesh->NMeshPoints);
  newloads = e_alloc_frag_d_matr(mesh->NMeshPoints, E_DOF_MAX);

  /* store supplied load in new data */
  if(mesh->NLoadedNodes > 0)
  {
    for(i = 0; i < mesh->NLoadedNodes; i++)
    {
      for(j = 0; j < 3; j++)
      {
        newloads[mesh->LoadedNodes[i]-1][j] = mesh->Loads[i][j];
      }
    }
    e_free_frag_d_matr(mesh->NLoadedNodes, mesh->Loads);
    e_free(mesh->LoadedNodes);
  }
  
  /* create a new load data structure, where every node can be loaded */
  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    newnodes[i] = i+1;
  }
  mesh->NLoadedNodes = mesh->NMeshPoints;
  mesh->LoadedNodes  = newnodes;
  mesh->Loads        = newloads;
  mesh->KeywordSet[E_KW_NUM_NLOADED_NODES] = TRUE;
  mesh->KeywordSet[E_KW_NUM_LOADS] = TRUE;

  /* calculate selfweight */
  if(mesh->NElemsLink1)
  {
    SelfWeightLink(dimension, mesh->NElemsLink1, mesh->NodesLink1,
                   mesh->MeshPointCoords, mesh->MatsLink1, mesh->Loads);
  }
  if(mesh->NElemsLink2)
  {
    SelfWeightLink(dimension, mesh->NElemsLink2, mesh->NodesLink2,
                   mesh->MeshPointCoords, mesh->MatsLink2, mesh->Loads);
  }
  if(mesh->NElemsLink3)
  {
    SelfWeightLink(dimension, mesh->NElemsLink3, mesh->NodesLink3,
                   mesh->MeshPointCoords, mesh->MatsLink3, mesh->Loads);
  }
  if(mesh->NElemsLink4)
  {
    SelfWeightLink(dimension, mesh->NElemsLink4, mesh->NodesLink4,
                   mesh->MeshPointCoords, mesh->MatsLink4, mesh->Loads);
  }
  /* LINK5 should not have any self-weight !!! */
  
  if(mesh->NElemsTriang3)
  {
    SelfWeightTriang(dimension, mesh->NElemsTriang3, mesh->NodesTriang3, 
                     mesh->MeshPointCoords, mesh->MatsTriang3, mesh->Loads);
  }
  if(mesh->NElemsTriang4)
  {
    SelfWeightTriang(dimension, mesh->NElemsTriang4, mesh->NodesTriang4, 
                     mesh->MeshPointCoords, mesh->MatsTriang4, mesh->Loads);
  }
  if(mesh->NElemsTriang5)
  {
    SelfWeightTriang(dimension, mesh->NElemsTriang5, mesh->NodesTriang5, 
                     mesh->MeshPointCoords, mesh->MatsTriang5, mesh->Loads);
  }
}



/* collect nodes per link type for compressed boundary */
static void CollectLinkCompNodes(int n, int **nodes, MAT **material,
                                 int *NCompNodes, int **CompNodes)
{
  int i, j, k;

  /* for all element */
  for(i = 0; i < n; i++)
  {
    /* if link is marked as compressed */
    if(material[i]->iParam4 == -1)
    {
      /* there are two nodes in a link */
      for(k = 0; k < 2; k++)
      {
        /* check whether we already have it */
        for(j = 0; j < (*NCompNodes); j++)
        {
          if((*CompNodes)[j] == nodes[i][k])
            break;
        }
        /* if we do not have it add it */
        if(j == (*NCompNodes))
        {
          (*CompNodes) = e_incr_i_vect((*NCompNodes), (*CompNodes), ALLOC_SIZE);
          (*CompNodes)[(*NCompNodes)] = nodes[i][k];
          (*NCompNodes)++;
        }
      }
    }
  }
}

/* collect all nodes where the residual should be 
   reversed due to ``compression'' */
static void CollectCompNodes(MESH *mesh, DR_DATA *data)
{

  data->NCompNodes = 0;
  data->CompNodes = NULL;

  if(mesh->NElemsLink1)
    CollectLinkCompNodes(mesh->NElemsLink1, mesh->NodesLink1,
                         mesh->MatsLink1, &(data->NCompNodes),
                         &(data->CompNodes));

  if(mesh->NElemsLink2)
    CollectLinkCompNodes(mesh->NElemsLink2, mesh->NodesLink2,
                         mesh->MatsLink2, &(data->NCompNodes),
                         &(data->CompNodes));

  if(mesh->NElemsLink3)
    CollectLinkCompNodes(mesh->NElemsLink3, mesh->NodesLink3,
                         mesh->MatsLink3, &(data->NCompNodes),
                         &(data->CompNodes));

  if(mesh->NElemsLink4)
    CollectLinkCompNodes(mesh->NElemsLink4, mesh->NodesLink4,
                         mesh->MatsLink4, &(data->NCompNodes),
                         &(data->CompNodes));
}



/* ------------------------------------------------------------------------- */
/*                           Write out functions                             */
/* ------------------------------------------------------------------------- */

#ifdef __E_MPI__
void WriteResultOut(DR_DATA *Gdata, MESH *Gmesh, DR_DATA *data, MESH *mesh)
#else
void WriteResultOut(DR_DATA *data, MESH *mesh)
#endif
{

#ifdef __E_MPI__
  int        i, j, k;
  int        NNodes, *GNodes, *GElem, gindex;
  double   **RecvBuf;
  double   **Residual;
  double   **Velocity;
  int        NLink, NTriang;
  double    *LinkStress, **TriangStress;
  
  MPI_Status status;
#endif
  
  /* setup the stresses and determine them for all subdomains */
  if(data->ProjectTask & TASK_WRITE_FORCE)
  {
    SetupStressStruct(mesh);
  

    if(mesh->NElemsLink1)
    {
      LinkForce(mesh->NElemsLink1, mesh->NodesLink1, 
                mesh->MeshPointCoords, mesh->MatsLink1,
                data->InitLengthData->InitLengthLink1, FALSE,
                mesh->Stresses.StressesLink1);
    }

    if(mesh->NElemsLink2)
    {
      LinkForce(mesh->NElemsLink2, mesh->NodesLink2,
                mesh->MeshPointCoords, mesh->MatsLink2,
                data->InitLengthData->InitLengthLink2, TRUE,
                mesh->Stresses.StressesLink2);
    }

    if(mesh->NElemsTriang3)
    {
      TriangForce(mesh->NElemsTriang3, mesh->NodesTriang3, 
                  mesh->MeshPointCoords, mesh->MatsTriang3,
                  data->InitLengthData->InitLengthTriang3,
                  mesh->Stresses.StressesTriang3);
      if(data->ProjectTask &  TASK_CHECK_STRESSES)
      {
        CheckTriangForce(mesh->NElemsTriang3, mesh->NodesTriang3,
                         mesh->MeshPointCoords, mesh->MatsTriang3,
                         data->InitCoords,
                         mesh->Stresses.StressesTriang3);
      }
    }
    if(mesh->NElemsTriang4)
    {
      TriangForce(mesh->NElemsTriang4, mesh->NodesTriang4, 
                  mesh->MeshPointCoords, mesh->MatsTriang4,
                  data->InitLengthData->InitLengthTriang4,
                  mesh->Stresses.StressesTriang4);

      if(data->ProjectTask &  TASK_CHECK_STRESSES)
      {
        CheckTriangForce(mesh->NElemsTriang4, mesh->NodesTriang4, 
                         mesh->MeshPointCoords, mesh->MatsTriang4,
                         data->InitCoords,
                         mesh->Stresses.StressesTriang4);
      }
      
    }
  }


  
#ifdef __E_MPI__
  if(data->CommData->Par_ProcID == 1)
  {
    Residual = e_alloc_cont_d_matr(Gmesh->NMeshPoints, Gdata->Dimension);
    Velocity = e_alloc_cont_d_matr(Gmesh->NMeshPoints, Gdata->Dimension);

    if(Gdata->ProjectTask & TASK_WRITE_FORCE)
    {
      SetupStressStruct(Gmesh);

      for(j = 0; j < mesh->NElemsLink1; j++)
      {
        gindex = mesh->GEIsLink1[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK1];
        Gmesh->Stresses.StressesLink1[gindex][0][0] =
            mesh->Stresses.StressesLink1[j][0][0];
      }
      for(j = 0; j < mesh->NElemsLink2; j++)
      {
        gindex = mesh->GEIsLink2[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK2];
        Gmesh->Stresses.StressesLink2[gindex][0][0] =
            mesh->Stresses.StressesLink2[j][0][0];
      }
      for(j = 0; j < mesh->NElemsTriang3; j++)
      {
        for(k = 0; k < 6; k++)
        {
          gindex = mesh->GEIsTriang3[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3];
          Gmesh->Stresses.StressesTriang3[gindex][0][k] =
              mesh->Stresses.StressesTriang3[j][0][k];
        }
      }
      for(j = 0; j < mesh->NElemsTriang4; j++)
      {
        for(k = 0; k < 6; k++)
        {
          gindex = mesh->GEIsTriang4[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4];
          Gmesh->Stresses.StressesTriang4[gindex][0][k] =
              mesh->Stresses.StressesTriang4[j][0][k];
        }
      }
      
    }

    /* mesh on root */
    for(j = 0; j < mesh->NMeshPoints; j++)
    {
      memcpy((void *)(Gmesh->MeshPointCoords[mesh->GNIs[j]-1]),
             (void *)(mesh->MeshPointCoords[j]),
             sizeof(double) * 3);
      memcpy((void *)(Residual[mesh->GNIs[j]-1]),
             (void *)(data->Residual[j]),
             sizeof(double) * data->Dimension);
      memcpy((void *)(Velocity[mesh->GNIs[j]-1]),
             (void *)(data->Velocity[j]),
             sizeof(double) * data->Dimension);
    }
    
    /* other subdomain meshes */
    for(i = 1; i < data->CommData->Par_NSubdomains; i++)
    {
      MPI_Recv(&NNodes, 1, MPI_INT, i, (i+1)+100, MPI_COMM_WORLD, &status);
      RecvBuf = e_alloc_cont_d_matr(NNodes, 3);
      GNodes  = e_alloc_i_vect(NNodes);

      MPI_Recv(GNodes, NNodes, MPI_INT, i, (i+1)+200, MPI_COMM_WORLD, &status);
      /* coordinates */
      MPI_Recv(RecvBuf[0], NNodes * 3, MPI_DOUBLE, i, (i+1)+300, MPI_COMM_WORLD, &status);
      for(j = 0; j < NNodes; j++)
      {
        memcpy((void *)(Gmesh->MeshPointCoords[GNodes[j]-1]),
               (void *)(RecvBuf[j]),
               sizeof(double) * 3);
      }
      e_free_cont_d_matr(NNodes, RecvBuf);

      RecvBuf = e_alloc_cont_d_matr(NNodes, Gdata->Dimension);
      /* residuals */
      MPI_Recv(RecvBuf[0], NNodes * Gdata->Dimension, MPI_DOUBLE, 
               i, (i+1)+400, MPI_COMM_WORLD, &status);
      for(j = 0; j < NNodes; j++)
      {
        memcpy((void *)(Residual[GNodes[j]-1]),
               (void *)(RecvBuf[j]),
               sizeof(double) * Gdata->Dimension);
      }
      /* velocity */
      MPI_Recv(RecvBuf[0], NNodes * Gdata->Dimension, MPI_DOUBLE, 
               i, (i+1)+500, MPI_COMM_WORLD, &status);
      for(j = 0; j < NNodes; j++)
      {
        memcpy((void *)(Velocity[GNodes[j]-1]),
               (void *)(RecvBuf[j]),
               sizeof(double) * Gdata->Dimension);
      }

      e_free(GNodes);
      e_free_cont_d_matr(NNodes, RecvBuf);

      /* stresses*/
      if(Gdata->ProjectTask & TASK_WRITE_FORCE)
      {
        MPI_Recv(&(NLink), 1, MPI_INT, i, (i+1)+601, MPI_COMM_WORLD, &status);
        if(NLink > 0)
        {
          LinkStress = e_alloc_d_vect(NLink);
          GElem =  e_alloc_i_vect(NLink);
          MPI_Recv(GElem, NLink, MPI_INT, i, (i+1)+602, MPI_COMM_WORLD, &status);
          MPI_Recv(LinkStress, NLink, MPI_DOUBLE, i, (i+1)+603, MPI_COMM_WORLD, &status);
          for(j = 0; j < NLink; j++)
          {
            gindex = GElem[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK1];
            Gmesh->Stresses.StressesLink1[gindex][0][0] = LinkStress[j];
          }
          e_free(LinkStress);
        }
        

        MPI_Recv(&(NLink), 1, MPI_INT, i, (i+1)+601, MPI_COMM_WORLD, &status);
        if(NLink > 0)
        {
          LinkStress = e_alloc_d_vect(NLink);
          GElem =  e_alloc_i_vect(NLink);
          MPI_Recv(GElem, NLink, MPI_INT, i, (i+1)+602, MPI_COMM_WORLD, &status);
          MPI_Recv(LinkStress, NLink, MPI_DOUBLE, i, (i+1)+603, MPI_COMM_WORLD, &status);
          for(j = 0; j < NLink; j++)
          {
            gindex = GElem[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK2];
            Gmesh->Stresses.StressesLink2[gindex][0][0] = LinkStress[j];
          }
          e_free(LinkStress);
        }
        

        MPI_Recv(&(NTriang), 1, MPI_INT, i, (i+1)+601, MPI_COMM_WORLD, &status);
        if(NTriang > 0)
        {
          TriangStress = e_alloc_cont_d_matr(NTriang, 6);
          GElem =  e_alloc_i_vect(NTriang);
          MPI_Recv(GElem, NTriang, MPI_INT, i, (i+1)+602, MPI_COMM_WORLD, &status);
          MPI_Recv(TriangStress[0], NTriang*6, MPI_DOUBLE, i, (i+1)+603,
                   MPI_COMM_WORLD, &status);
          for(j = 0; j < NTriang; j++)
          {
            for(k = 0; k < 6; k++)
            {
              gindex = GElem[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3];
              Gmesh->Stresses.StressesTriang3[gindex][0][k] = TriangStress[j][k];
            }
          }
          e_free_cont_d_matr(NTriang, TriangStress);
        }

        MPI_Recv(&(NTriang), 1, MPI_INT, i, (i+1)+601, MPI_COMM_WORLD, &status);
        if(NTriang > 0)
        {
          TriangStress = e_alloc_cont_d_matr(NTriang, 6);
          GElem =  e_alloc_i_vect(NTriang);
          MPI_Recv(GElem, NTriang, MPI_INT, i, (i+1)+602, MPI_COMM_WORLD, &status);
          MPI_Recv(TriangStress[0], NTriang*6, MPI_DOUBLE, i, (i+1)+603,
                   MPI_COMM_WORLD, &status);
          for(j = 0; j < NTriang; j++)
          {
            for(k = 0; k < 6; k++)
            {
              gindex = GElem[j] - Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4];
              Gmesh->Stresses.StressesTriang4[gindex][0][k] = TriangStress[j][k];
            }
          }
          e_free_cont_d_matr(NTriang, TriangStress);
        }
      }
    }

    /* output the mesh in E-lib format */
    e_PutMeshData(Gmesh, Gdata->Outputname);
    if(Gdata->ProjectTask & TASK_WRITE_FORCE)
    {
      e_PutStresses(Gmesh, Gdata->Outputname);
    }
    
    /* output other results */
    PutDataPerNodes(Residual, "RESIDUAL", E_RES, 
                    Gdata->Outputname, Gmesh, Gdata->Dimension);

    PutDataPerNodes(Velocity, "VELOCITY", E_VEL, 
                    Gdata->Outputname, Gmesh, Gdata->Dimension);
  }
  else
  {
    MPI_Send(&(mesh->NMeshPoints), 1, MPI_INT, 
             0, data->CommData->Par_ProcID+100, MPI_COMM_WORLD);
    MPI_Send(mesh->GNIs, mesh->NNodes, MPI_INT, 
             0, data->CommData->Par_ProcID+200, MPI_COMM_WORLD);

    MPI_Send(mesh->MeshPointCoords[0], mesh->NMeshPoints * 3, MPI_DOUBLE,
             0, data->CommData->Par_ProcID+300, MPI_COMM_WORLD);

    MPI_Send(data->Residual[0], mesh->NMeshPoints * data->Dimension, MPI_DOUBLE,
             0, data->CommData->Par_ProcID+400, MPI_COMM_WORLD);
    MPI_Send(data->Velocity[0], mesh->NMeshPoints * data->Dimension, MPI_DOUBLE,
             0, data->CommData->Par_ProcID+500, MPI_COMM_WORLD);

    /* send the stresses and forces */
    if(data->ProjectTask & TASK_WRITE_FORCE)
    {
      MPI_Send(&(mesh->NElemsLink1), 1, MPI_INT, 
               0, data->CommData->Par_ProcID+601, MPI_COMM_WORLD);
      if(mesh->NElemsLink1 > 0)
      {
        LinkStress = e_alloc_d_vect(mesh->NElemsLink1);
        for(j = 0; j < mesh->NElemsLink1; j++)
        {
          LinkStress[j] = mesh->Stresses.StressesLink1[j][0][0];
        }
        MPI_Send(mesh->GEIsLink1, mesh->NElemsLink1, MPI_INT, 
                 0, data->CommData->Par_ProcID+602, MPI_COMM_WORLD);
        MPI_Send(LinkStress, mesh->NElemsLink1, MPI_DOUBLE, 
                 0, data->CommData->Par_ProcID+603, MPI_COMM_WORLD);
        e_free(LinkStress);
      }
      
      MPI_Send(&(mesh->NElemsLink2), 1, MPI_INT, 
               0, data->CommData->Par_ProcID+601, MPI_COMM_WORLD);
      if(mesh->NElemsLink2 > 0)
      {
        LinkStress = e_alloc_d_vect(mesh->NElemsLink2);
        for(j = 0; j < mesh->NElemsLink2; j++)
        {
          LinkStress[j] = mesh->Stresses.StressesLink2[j][0][0];
        }
        MPI_Send(mesh->GEIsLink2, mesh->NElemsLink2, MPI_INT, 
                 0, data->CommData->Par_ProcID+602, MPI_COMM_WORLD);
        MPI_Send(LinkStress, mesh->NElemsLink2, MPI_DOUBLE, 
                 0, data->CommData->Par_ProcID+603, MPI_COMM_WORLD);
        e_free(LinkStress);
      }


      MPI_Send(&(mesh->NElemsTriang3), 1, MPI_INT, 
               0, data->CommData->Par_ProcID+601, MPI_COMM_WORLD);
      if(mesh->NElemsTriang3 > 0)
      {
        TriangStress = e_alloc_cont_d_matr(mesh->NElemsTriang3, 6);
        for(j = 0; j < mesh->NElemsTriang3; j++)
        {
          for(k = 0; k < 6; k++)
          {
            TriangStress[j][k] = mesh->Stresses.StressesTriang3[j][0][k];
          }
        }
        MPI_Send(mesh->GEIsTriang3, mesh->NElemsTriang3, MPI_INT, 
                 0, data->CommData->Par_ProcID+602, MPI_COMM_WORLD);
        MPI_Send(TriangStress[0], mesh->NElemsTriang3 * 6, MPI_DOUBLE, 
                 0, data->CommData->Par_ProcID+603, MPI_COMM_WORLD);
        e_free_cont_d_matr(mesh->NElemsTriang3, TriangStress);
      }

      MPI_Send(&(mesh->NElemsTriang4), 1, MPI_INT, 
               0, data->CommData->Par_ProcID+601, MPI_COMM_WORLD);
      if(mesh->NElemsTriang4 > 0)
      {
        TriangStress = e_alloc_cont_d_matr(mesh->NElemsTriang4, 6);
        for(j = 0; j < mesh->NElemsTriang4; j++)
        {
          for(k = 0; k < 6; k++)
          {
            TriangStress[j][k] = mesh->Stresses.StressesTriang4[j][0][k];
          }
        }
        MPI_Send(mesh->GEIsTriang4, mesh->NElemsTriang4, MPI_INT, 
                 0, data->CommData->Par_ProcID+602, MPI_COMM_WORLD);
        MPI_Send(TriangStress[0], mesh->NElemsTriang4 * 6, MPI_DOUBLE, 
                 0, data->CommData->Par_ProcID+603, MPI_COMM_WORLD);
        e_free_cont_d_matr(mesh->NElemsTriang4, TriangStress);
      }
    }
  }
  

#else
  /* output the mesh in E-lib format */
  e_PutMeshData(mesh, data->Outputname);
  if(data->ProjectTask & TASK_WRITE_FORCE)
  {
    e_PutStresses(mesh, data->Outputname);
  }

  /* output other results */
  PutDataPerNodes(data->Residual, "RESIDUAL", E_RES, 
                  data->Outputname, mesh, data->Dimension);

  PutDataPerNodes(data->Velocity, "VELOCITY", E_VEL, 
                  data->Outputname, mesh, data->Dimension);

  PutDataPerNodes(data->Mass, "MASS", "mas", 
                  data->Outputname, mesh, data->Dimension);

#endif

}

int PutDataPerNodes(double **result, char *keyword, char *extension, 
                    char *name, MESH *mesh, int dimension)
{
  FILE  *InputFP;
  char  *FileName;
  int    i, j;

  FileName = e_alloc_c_vect(strlen(name) + strlen(extension) + 2);
  sprintf(FileName, "%s.%s", name, extension);

  /* open the file */
  if((InputFP = fopen(FileName, "w")) == NULL)
  {
    e_GeneralError("cannot write file: %s", FileName);
    e_free(FileName);
    return 0;
  }

  fprintf(InputFP, "\n%s\n", keyword);

  for(i = 0; i < mesh->NMeshPoints; i++)
  {
    fprintf(InputFP, "%d", i+1);
    for(j = 0; j < dimension; j++)
    {
      fprintf(InputFP, "\t% .8f", result[i][j]);
    }
    fprintf(InputFP, "\n");
  }

  fclose(InputFP);
  e_free(FileName);

  return 1;
}


#ifdef __E_MPI__
void WriteCuttingPattern(DR_DATA *Gdata, MESH *Gmesh,
                         DR_DATA *data,  MESH *mesh,
                         CUTTING_PATTERN *cdata)
#else
void WriteCuttingPattern(DR_DATA *data, CUTTING_PATTERN *cdata)
#endif
{
  int    i;
  MESH  *lmesh;

#ifdef __E_MPI__
  MPI_Status status;
  int        j, k;
  int        NStrip, NElem, *GElem;

  if(data->CommData->Par_ProcID == 1)
  {
    Gmesh->Decomp.NSubDoms++;
    for(i = 0; i < Gmesh->TotalNElems; i++)
    {
      Gmesh->Decomp.SubDoms[i] = Gmesh->Decomp.NSubDoms;
    }
  
    /* local data */
    for(i = 0; i < cdata->NStrips; i++)
    {
      lmesh = cdata->Strips[i]->StripMesh;
      if(lmesh == NULL)
      {
        continue;
      }
      for(j = 0; j < lmesh->TotalNElems; j++)
      {
        Gmesh->Decomp.SubDoms[lmesh->GEIs[j]-1] = 1;
      }
    }

    /* data from other subdomains */
    for(j = 1; j < data->CommData->Par_NSubdomains; j++)
    {
      MPI_Send(&j, 1, MPI_INT, j, 222, MPI_COMM_WORLD);
      MPI_Recv(&NStrip, 1, MPI_INT, j, j+1+332, MPI_COMM_WORLD, &status);

      for(k = 0; k < NStrip; k++)
      {
        MPI_Recv(&NElem, 1, MPI_INT, j, j+1+333, MPI_COMM_WORLD, &status);

        if(NElem == 0)
        {
          continue;
        }
        GElem = e_alloc_i_vect(NElem);
        MPI_Recv(GElem, NElem, MPI_INT, j, j+1+334, MPI_COMM_WORLD, &status);
      
        for(i = 0; i < NElem; i++)
        {
          if(Gmesh->Decomp.SubDoms[GElem[i]-1] != Gmesh->Decomp.NSubDoms)
          {
            e_GeneralWarning("Element %d is in two strips %d %d !", GElem[i],
                             Gmesh->Decomp.SubDoms[GElem[i]-1], j+1);
          }
          Gmesh->Decomp.SubDoms[GElem[i]-1] = j+1;
        }
        e_free(GElem);
      }
    }
    e_PutMeshData(Gmesh, Gdata->Outputname);
    e_PutDecompData(Gmesh, Gdata->Outputname);
  }
  else
  {
    MPI_Recv(&j, 1, MPI_INT, 0, 222, MPI_COMM_WORLD, &status);

    MPI_Send(&(cdata->NStrips), 1, MPI_INT, 
             0, data->CommData->Par_ProcID+332, MPI_COMM_WORLD);

    for(i = 0; i < cdata->NStrips; i++)
    {
      lmesh = cdata->Strips[i]->StripMesh;
      if(lmesh == NULL)
      {
        NElem = 0;
        MPI_Send(&(NElem), 1, MPI_INT, 
                 0, data->CommData->Par_ProcID+333, MPI_COMM_WORLD);
      }
      else
      {
        MPI_Send(&(lmesh->TotalNElems), 1, MPI_INT, 
                 0, data->CommData->Par_ProcID+333, MPI_COMM_WORLD);
        MPI_Send(lmesh->GEIs, lmesh->TotalNElems, MPI_INT, 
                 0, data->CommData->Par_ProcID+334, MPI_COMM_WORLD);
      }
    }


  }

/*
  for(i = 0; i < cdata->NStrips; i++)
  {
    lmesh = cdata->Strips[i]->StripMesh;
    if(lmesh != NULL)
    {
      sprintf(stripfile, "out/strip-%d-%d", data->CommData->Par_ProcID, i+1);
      e_PutMeshData(lmesh, stripfile);
    }
  }
*/

#else
  {
    char stripfile[E_LINE_MAX+1];
    
    for(i = 0; i < cdata->NStrips; i++)
    {
      lmesh = cdata->Strips[i]->StripMesh;
      if(lmesh != NULL)
      {
        sprintf(stripfile, "%s-strip-%d", data->Outputname, i+1);
        e_PutMeshData(lmesh, stripfile);
      }
    }
  }
  
#endif


}



#ifdef __E_MPI__

/* ------------------------------------------------------------------------- */
/*                         Genetic algorithm input                           */
/* ------------------------------------------------------------------------- */

int init_GA_data(char *name, GA_DATA **gadata, double *gamma)
{
  char line[E_LINE_MAX];
  int  answer_not_ok, stat, contcutpat;

  contcutpat = 1;  
  stat = GetGAParam(name, gadata);
  while(!stat)
  {
    printf("\nDo you want the program to retry the initialisation of the");
    printf("\ngenetic algorithm or skip the cutting pattern generation ?");
    printf("\nType 1 for retry and 0 for skipping: ");
    answer_not_ok = TRUE;
    while(answer_not_ok)
    {
      fgets(line, E_LINE_MAX, stdin);
      if(sscanf(line, "%d", &(contcutpat)) != 1)
        printf("\nType only 0 or 1 for clear answer !");
      if((contcutpat != 0) && (contcutpat != 1))
        printf("\nType only 0 or 1 for clear answer !");
      else
        answer_not_ok = FALSE;
    }

    if(contcutpat == 0)
    {
      break;
    }
    else
    {
      printf("\nTrying to read GA initialisation file ...");
      stat = GetGAParam(name, gadata);
    }

  } 

  stat = FALSE;
  while(stat == FALSE)
  {
    printf("\nGamma parameter: ");
    fgets(line, E_LINE_MAX, stdin);
    if(sscanf(line, "%lf", gamma) != 1)
      printf("\nType only a number for clear answer !");
    else if((*gamma < 0) || (*gamma > 1))
      printf("\nType a number between 0 and 1 for clear answer !");
    else
      stat = TRUE;
  }

  return(contcutpat);

}

#endif









/* ######################################################################## */

/* temporarily removed !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
#if 0


void cutting_result()
{
  MESH  *l_mesh;
  int    i, j, i_elem, strip_count, n;
  char   message[E_LINE_MAX];


#ifdef __E_MPI__
  MPI_Status status;

  if(GV_procID == 1)
  {
    n = GV_global_mesh->Decomp.NSubDoms;
    e_FreeDecompData(GV_global_mesh);
    
    e_SetupDecompData(GV_global_mesh);
    
    GV_global_mesh->Decomp.NSubDoms = n;

    for(i = 1; i < GV_total_no_subdomains; i++)
    {
      MPI_Recv(&strip_count, 1, MPI_INT, i, i+1, MPI_COMM_WORLD, &status);

      for(j = 0; j < strip_count; j++)
      {
        e_MPI_Recv_MeshData(&l_mesh, i);

        sprintf(message, "out/%d-%d-strip", i+1, j+1);
        e_PutMeshData(l_mesh, message);

        for(i_elem = 0; i_elem < l_mesh->NElemsTriang4; i_elem++)
        {
          n = l_mesh->GEIs[i_elem] - GV_global_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5];
          GV_global_mesh->Decomp.SubDomsTriang5[n] = i+1;
        }

        e_FreeMeshData(l_mesh);
      }
    }
#endif

/* locally available strips */
    strip_count = 0;
    for(i = 0; i < GV_s_no_strips; i++)
    {
      l_mesh = GV_strip_meshes[i];
      if(l_mesh != NULL)
      {
        sprintf(message, "out/%d-%d-strip", 1, ++strip_count);
        e_PutMeshData(l_mesh, message);

#ifdef __E_MPI__
        for(i_elem = 0; i_elem < l_mesh->NElemsTriang4; i_elem++)
        {
          n = l_mesh->GEIs[i_elem] - GV_global_mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5];
          GV_global_mesh->Decomp.SubDomsTriang5[n] = 1;
        }
#endif

      }
    }

#ifdef __E_MPI__
    e_PutDecompData(GV_global_mesh, GV_projectname);

  }
  else
  {
    strip_count = 0;
    for(i = 0; i < GV_s_no_strips; i++)
    {
      l_mesh = GV_strip_meshes[i];
      if(l_mesh != NULL)
        strip_count++;
    }

    MPI_Send(&strip_count, 1, MPI_INT, 0, GV_procID, MPI_COMM_WORLD);

    for(i = 0; i < GV_s_no_strips; i++)
    {
      l_mesh = GV_strip_meshes[i];
      if(l_mesh != NULL)
      {
        e_MPI_Send_MeshData(l_mesh, 0);
      }
    }
  }
#endif

}


#endif





