/*
 * Functions to handle the initial or slack length data
 *   - initialisation of the data structure
 *   - reading and writing of the file
 *   - generating initial length data
 *   - freeing-up the structure
 *   - MPI sending and receiving of the structure
 */

#include "dr_len.h"

#ifdef __E_MPI__
#include "mpi.h"
#endif

/* this function is in E-lib, but is not shown in header file (why?) */
extern int GetKeyword(char *line, char *word, int *LineNumber,
                      char *FullPathName, FILE *fp);

static void ReadLinkLength(char *Line, int *LineCount, 
                           char *FileName, FILE *InputFP,
                           int NElemsLink, double **coords, int **nodes,
                           MAT **material, int *treated,
                           double *InitLengthLink);
static void ReadTriangLength(char *Line, int *LineCount, char *FileName, 
                             FILE *InputFP, int NElemsTriang, 
                             double **InitLengthTriang, char *elemname);
static void GenInitLengthLink(int no, int **nodes, double **coords, 
                              double *init_length, double div, int *treated);
static void GenInitLengthTriang(int no, int **nodes, double **coords, 
                                double **init_length);


/* ------------------------------------------------------------------------- */
/*            Initial lenght data handling functions                         */
/* ------------------------------------------------------------------------- */


/* setup function for initial length */
/* there is one double vector, and there are pointers for the different 
   elements to reference inside the vector */
void SetupInitLengthData(INIT_LENGTH_DATA **data, MESH *mesh)
{
  int n, inc, i;

  if(mesh->NElemsLink1 + mesh->NElemsLink2 +
     mesh->NElemsLink3 + mesh->NElemsLink4 +
     mesh->NElemsLink5 + mesh->NElemsTriang3 + 
     mesh->NElemsTriang4 + mesh->NElemsTriang5 != mesh->TotalNElems)
  {
    e_GeneralError("SetupInitLengthData: Mesh contain unsupported element !");
  }
  
  (*data) = e_calloc(1, sizeof(INIT_LENGTH_DATA));
  
  n = mesh->NElemsLink1 + mesh->NElemsLink2 +
      mesh->NElemsLink3 + mesh->NElemsLink4 +
      mesh->NElemsLink5 + 
      (mesh->NElemsTriang3 + mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;

  (*data)->InitLength = e_alloc_d_vect(n);

  if(mesh->NElemsLink1)
  {
    (*data)->InitLengthLink1 = (*data)->InitLength + 
                               mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1;
  }
  else
  {
    (*data)->InitLengthLink1 = NULL;
  }
    
  if(mesh->NElemsLink2)
  {
    (*data)->InitLengthLink2 = (*data)->InitLength + 
                               mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1;
  }
  else
  {
    (*data)->InitLengthLink2 = NULL;
  }

  if(mesh->NElemsLink3)
  {
    (*data)->InitLengthLink3 = (*data)->InitLength + 
                               mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]-1;
  }
  else
  {
    (*data)->InitLengthLink3 = NULL;
  }

  if(mesh->NElemsLink4)
  {
    (*data)->InitLengthLink4 = (*data)->InitLength + 
                               mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]-1;
  }
  else
  {
    (*data)->InitLengthLink4 = NULL;
  }

  if(mesh->NElemsLink5)
  {
    (*data)->InitLengthLink5 = (*data)->InitLength + 
                               mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]-1;
  }
  else
  {
    (*data)->InitLengthLink5 = NULL;
  }

  /* in the case of triangles there is an extra vector allocation, but finally
     the references will point to the double vector */
  inc = 0;
  if(mesh->NElemsTriang3)
  {
    (*data)->InitLengthTriang3 = e_calloc(mesh->NElemsTriang3, sizeof(double *) );
    (*data)->InitLengthTriang3[0] = (*data)->InitLength + inc +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1;
    for(i = 1; i < mesh->NElemsTriang3; i++)
    {
      (*data)->InitLengthTriang3[i] = (*data)->InitLengthTriang3[i-1] + 3;
    }
    inc += mesh->NElemsTriang3 * 2;
  }
  else
  {
    (*data)->InitLengthTriang3 = NULL;
  }

  if(mesh->NElemsTriang4)
  {
    (*data)->InitLengthTriang4 = e_calloc(mesh->NElemsTriang4, sizeof(double *) );

    (*data)->InitLengthTriang4[0] = (*data)->InitLength + inc +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1;
    for(i = 1; i < mesh->NElemsTriang4; i++)
    {
      (*data)->InitLengthTriang4[i] = (*data)->InitLengthTriang4[i-1] + 3;
    }
    inc += mesh->NElemsTriang4 * 2;
  }
  else
  {
    (*data)->InitLengthTriang4 = NULL;
  }

  if(mesh->NElemsTriang5)
  {
    (*data)->InitLengthTriang5 = e_calloc(mesh->NElemsTriang5, sizeof(double *) );

    (*data)->InitLengthTriang5[0] = (*data)->InitLength + inc +
                                    mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]-1;
    for(i = 1; i < mesh->NElemsTriang5; i++)
    {
      (*data)->InitLengthTriang5[i] = (*data)->InitLengthTriang5[i-1] + 3;
    }
    inc += mesh->NElemsTriang5 * 2;
  }
  else
  {
    (*data)->InitLengthTriang5 = NULL;
  }

}


/* read initial length data for link elements */
static void ReadLinkLength(char *Line, int *LineCount, 
                           char *FileName, FILE *InputFP,
                           int NElemsLink, double **coords, int **nodes,
                           MAT **material, int *treated,
                           double *InitLengthLink)
{
  int     i_link, n1, n2, k;
  double  lenforce, dl, curlen, E, A;
  char    type;
  char    end[E_LINE_MAX+1];

  while(1)
  {
    /* read in the line */
    e_GetNextLine(Line, LineCount, FileName, InputFP);
    /* try interpret the line */
    if(sscanf(Line, "%d %lf %c ", &i_link, &lenforce, &type) != 3)
    {
      type = 'l';
      if(sscanf(Line, "%d %lf", &i_link, &lenforce) != 2)
      {
        if(sscanf(Line, "%s", end) != 1)
        {
          e_FileLineError(*LineCount, FileName, "could not interpret the line");
        }

        if(strcmp(end, E_KW_LEN_END) == 0)
        {
          break;
        }
        else
        {
          e_FileLineError(*LineCount, FileName, "could not interpret the line");
        }
      }
    }
    /* check the data validity */
    if(i_link <= 0 || i_link > NElemsLink)
    {
      e_FileLineError(*LineCount, FileName, "invalid element number");
    }
    
    i_link--;
    if(type == 'l')
    {
      /* store the length */
      InitLengthLink[i_link] = lenforce;
    }
    else if(type == 'f')
    {
      E = material[i_link]->YMod;
      A  = material[i_link]->dParam1;

      /* calc surrent length */
      n1 = nodes[i_link][0]-1;
      n2 = nodes[i_link][1]-1;
      dl = 0;
      for (k = 0; k < 3; k++)
      {
        dl += E_SQUARE(coords[n1][k] - coords[n2][k]);
      }
      curlen = sqrt(dl);

      /* from force and current length determine initial length */
      lenforce = (1.0 + (lenforce / (E * A)));
      if(lenforce < 1e-8)
      {
        e_FileLineError(*LineCount, FileName, "initial force becomes zero with this force");
      }
      
      InitLengthLink[i_link] = curlen / lenforce;
    }
    else
    {
      e_FileLineError(*LineCount, FileName, "invalid type of prestress");
    }

    treated[i_link] = 1;
  }
}

/* read initial length data for triangular elements */
static void ReadTriangLength(char *Line, int *LineCount, char *FileName, 
                             FILE *InputFP, int NElemsTriang, 
                             double **InitLengthTriang, char *elemname)
{
  int i_triang, dummy;

  for(i_triang = 0; i_triang < NElemsTriang; i_triang++)
  {
    e_GetNextLine(Line, LineCount, FileName, InputFP);
    if(sscanf(Line, "%d %lf %lf %lf", &dummy, &(InitLengthTriang[i_triang][0]),
              &(InitLengthTriang[i_triang][1]), &(InitLengthTriang[i_triang][2])) != 4)
      e_FileLineError(*LineCount, FileName, "could not interpret this line");
    if(dummy != i_triang+1)
      e_FileLineError(*LineCount, FileName, "element number is out of sequence !");
  }
  if(i_triang != NElemsTriang)
    e_GeneralError("not enough initial length data for %s element", elemname);
}


/* reads in the initial length file */
void GetInitLengthData(char *Projectname, INIT_LENGTH_DATA **data, MESH *mesh)
{
  FILE  *InputFP;
  int    LineCount = 0;
  char  *FileName;
  char   Line[E_LINE_MAX+1];
  char   Word[E_LINE_MAX+1];
  int    i_triang3 = 0, i_triang4 = 0, i_triang5 = 0;

  int   *treated; /* for which elements did we read in data */

  FileName = e_alloc_c_vect(strlen(Projectname) + strlen(E_LEN) + 2);
  sprintf(FileName, "%s.%s", Projectname, E_LEN);

  /* setup memory and pointers */
  SetupInitLengthData(data, mesh);

  treated = e_alloc_i_vect(mesh->TotalNElems);

  /* open the file */
  if((InputFP = fopen(FileName, "r")) == NULL)
  {
    e_GeneralWarning("\nThere is NO iniital length file !");

    if((mesh->NElemsLink1)   || (mesh->NElemsLink2) ||
       (mesh->NElemsTriang3) || (mesh->NElemsTriang4))
    {
      e_GeneralWarning("\nAnalysis elements are NOT prestressed !");
    }
  }
  else
  {
    LineCount = 0;
    /* the data for the different elements can be in any order */
    while (1)
    {
      if(GetKeyword(Line, Word, &LineCount, FileName, InputFP) == EOF)
        break;

      /* --- LINK 1 --- */
      if(!(strcmp(Word, E_KW_LEN_LENGTH_LINK1)))
      {
        ReadLinkLength(Line, &LineCount, FileName, InputFP, 
                       mesh->NElemsLink1, mesh->MeshPointCoords,
                       mesh->NodesLink1, mesh->MatsLink1,
                       &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1]),
                       (*data)->InitLengthLink1);
      }
      /* --- LINK 2 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_LINK2)))
      {
        ReadLinkLength(Line, &LineCount, FileName, InputFP, 
                       mesh->NElemsLink2, mesh->MeshPointCoords,
                       mesh->NodesLink2, mesh->MatsLink2,
                       &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1]),
                       (*data)->InitLengthLink2);
      }
      /* --- LINK 3 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_LINK3)))
      {
        ReadLinkLength(Line, &LineCount, FileName, InputFP, 
                       mesh->NElemsLink3, mesh->MeshPointCoords,
                       mesh->NodesLink3, mesh->MatsLink3,
                       &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]-1]),
                       (*data)->InitLengthLink3);
      }
      /* --- LINK 4 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_LINK4)))
      {
        ReadLinkLength(Line, &LineCount, FileName, InputFP, 
                       mesh->NElemsLink4, mesh->MeshPointCoords,
                       mesh->NodesLink4, mesh->MatsLink4,
                       &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]-1]),
                       (*data)->InitLengthLink4);
      }
      /* --- LINK 5 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_LINK5)))
      {
        ReadLinkLength(Line, &LineCount, FileName, InputFP, 
                       mesh->NElemsLink5, mesh->MeshPointCoords,
                       mesh->NodesLink5, mesh->MatsLink5,
                       &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]-1]),
                       (*data)->InitLengthLink5);
      }
      /* --- TRIANG 3 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_TRIANG3)))
      {
        ReadTriangLength(Line, &LineCount, FileName, InputFP, 
                         mesh->NElemsTriang3, (*data)->InitLengthTriang3, "TRIANG3");
        i_triang3 = mesh->NElemsTriang3;
      }
      /* --- TRIANG 4 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_TRIANG4)))
      {
        ReadTriangLength(Line, &LineCount, FileName, InputFP, 
                         mesh->NElemsTriang4, (*data)->InitLengthTriang4, "TRIANG4");
        i_triang4 = mesh->NElemsTriang4;

      }
      /* --- TRIANG 5 --- */
      else if(!(strcmp(Word, E_KW_LEN_LENGTH_TRIANG5)))
      {
        ReadTriangLength(Line, &LineCount, FileName, InputFP, 
                         mesh->NElemsTriang5, (*data)->InitLengthTriang5, "TRIANG5");
        i_triang5 = mesh->NElemsTriang5;
      }
      else
        e_FileLineError(LineCount, FileName, "unrecognised keyword");
    }
    fclose(InputFP);
  }

  e_free(FileName);

  /* If there is no initial length file,
     or that does not contain information for an element type, it will be generated */
  GenInitLengthData(*data, mesh, treated, i_triang3, i_triang4, i_triang5);

  e_free(treated);
}

/* generates initial length for link elements */
static void GenInitLengthLink(int no, int **nodes, double **coords, 
                              double *init_length, double div,
                              int *treated)
{
  int n1, n2, k, i_link;
  double dl;

  for (i_link = 0; i_link < no; i_link++)
  {
    if(treated[i_link] == 0)
    {
      n1 = nodes[i_link][0]-1;
      n2 = nodes[i_link][1]-1;
      dl = 0;
      for (k = 0; k < 3; k++)
      {
        dl += E_SQUARE(coords[n1][k] - coords[n2][k]);
      }
      init_length[i_link] = sqrt(dl) / div;
    }
  }
}

/* generates initial length for triangular elements */
static void GenInitLengthTriang(int no, int **nodes, double **coords, 
                                double **init_length)
{
  int n1, n2, k, ii, i_triang;
  double dl;
  int    side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};

  for (i_triang = 0; i_triang < no; i_triang++)
  {
    for (ii = 0; ii < 3; ii++)
    {
      n1 = nodes[i_triang][side_n_node[ii][0]]-1;
      n2 = nodes[i_triang][side_n_node[ii][1]]-1;
      dl = 0;
      for (k = 0; k < 3; k++)
      {
        dl += E_SQUARE(coords[n1][k] - coords[n2][k]);
      }
      init_length[i_triang][ii] = sqrt(dl);
    }
  }
}

/* generate initial length for all valid element types,
   if link1, link2, ... is zero it is generated if not, it will be skipped */
void GenInitLengthData(INIT_LENGTH_DATA *data, MESH *mesh, 
                       int *treated, int triang3, int triang4, int triang5)
{

  if(mesh->NElemsLink1 + mesh->NElemsLink2 +
     mesh->NElemsLink3 + mesh->NElemsLink4 +
     mesh->NElemsLink5 + mesh->NElemsTriang3 + 
     mesh->NElemsTriang4 + mesh->NElemsTriang5 != mesh->TotalNElems)
  {
    e_GeneralError("GenInitLengthData: Mesh contain unsupported element !");
  }

  /* --- LINK 1 --- */
  if(mesh->NElemsLink1 > 0)
  {
    GenInitLengthLink(mesh->NElemsLink1, mesh->NodesLink1, 
                      mesh->MeshPointCoords, data->InitLengthLink1, 1.0,
                      &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1]));
  }
  /* --- LINK 2 --- */
  if(mesh->NElemsLink2 > 0)
  {
    GenInitLengthLink(mesh->NElemsLink2, mesh->NodesLink2, 
                      mesh->MeshPointCoords, data->InitLengthLink2, 1.0,
                      &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1]));
  }
  /* --- LINK 3 --- */
  if(mesh->NElemsLink3 > 0)
  {
    GenInitLengthLink(mesh->NElemsLink3, mesh->NodesLink3, 
                      mesh->MeshPointCoords, data->InitLengthLink3, 1.0,
                      &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]-1]));
  }
  /* --- LINK 4 --- */
  if(mesh->NElemsLink4 > 0)
  {
    GenInitLengthLink(mesh->NElemsLink4, mesh->NodesLink4, 
                      mesh->MeshPointCoords, data->InitLengthLink4, 1.0,
                      &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]-1]));
  }
  /* --- LINK 5 --- */
  if(mesh->NElemsLink5 > 0)
  {
    /* FIXME */
    GenInitLengthLink(mesh->NElemsLink5, mesh->NodesLink5, 
                      mesh->MeshPointCoords, data->InitLengthLink5, 
                      E_G_STRING_DIV,
                      &(treated[mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]-1]));
  }

  /* --- TRIANG 3 --- */
  if((mesh->NElemsTriang3 > 0) && (triang3 == 0))
  {
    GenInitLengthTriang(mesh->NElemsTriang3, mesh->NodesTriang3,
                        mesh->MeshPointCoords, data->InitLengthTriang3);
  }
  /* --- TRIANG 4 --- */
  if((mesh->NElemsTriang4 > 0) && (triang4 == 0))
  {
    GenInitLengthTriang(mesh->NElemsTriang4, mesh->NodesTriang4,
                        mesh->MeshPointCoords, data->InitLengthTriang4);
  }
  /* --- TRIANG 5 --- */
  if((mesh->NElemsTriang5 > 0) && (triang5 == 0))
  {
    GenInitLengthTriang(mesh->NElemsTriang5, mesh->NodesTriang5,
                        mesh->MeshPointCoords, data->InitLengthTriang5);
  }
}


/* writes out the initial length */
void PutInitLengthData(char *Projectname, INIT_LENGTH_DATA *data, MESH *mesh)
{
  FILE  *InputFP;
  char  *FileName;
  int    i;

  FileName = e_alloc_c_vect(strlen(Projectname) + strlen(E_LEN) + 2);
  sprintf(FileName, "%s.%s", Projectname, E_LEN);

  /* open the file */
  if((InputFP = fopen(FileName, "w")) == NULL)
  {
    e_GeneralError("cannot write file: %s", FileName);
    e_free(FileName);
  }

  if(mesh->NElemsLink1 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_LINK1);
    for(i = 0; i < mesh->NElemsLink1; i++)
    {
      fprintf(InputFP, "%d %.16f\n", i+1, data->InitLengthLink1[i]);
    }
    fprintf(InputFP, "END\n");
  }

  if(mesh->NElemsLink2 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_LINK2);
    for(i = 0; i < mesh->NElemsLink2; i++)
    {
      fprintf(InputFP, "%d %.16f\n", i+1, data->InitLengthLink2[i]);
    }
    fprintf(InputFP, "END\n");
  }

  if(mesh->NElemsLink3 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_LINK3);
    for(i = 0; i < mesh->NElemsLink3; i++)
    {
      fprintf(InputFP, "%d %.16f\n", i+1, data->InitLengthLink3[i]);
    }
    fprintf(InputFP, "END\n");
  }

  if(mesh->NElemsLink4 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_LINK4);
    for(i = 0; i < mesh->NElemsLink4; i++)
    {
      fprintf(InputFP, "%d %.16f\n", i+1, data->InitLengthLink4[i]);
    }
    fprintf(InputFP, "END\n");
  }

  if(mesh->NElemsLink5 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_LINK5);
    for(i = 0; i < mesh->NElemsLink5; i++)
    {
      fprintf(InputFP, "%d %.16f\n", i+1, data->InitLengthLink5[i]);
    }
    fprintf(InputFP, "END\n");
  }

  if(mesh->NElemsTriang3 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_TRIANG3);
    for(i = 0; i < mesh->NElemsTriang3; i++)
    {
      fprintf(InputFP, "%d %.16f %.16f %.16f\n", i+1, 
              data->InitLengthTriang3[i][0],
              data->InitLengthTriang3[i][1],
              data->InitLengthTriang3[i][2]  );
    }
  }

  if(mesh->NElemsTriang4 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_TRIANG4);
    for(i = 0; i < mesh->NElemsTriang4; i++)
    {
      fprintf(InputFP, "%d %.16f %.16f %.16f\n", i+1, 
              data->InitLengthTriang4[i][0],
              data->InitLengthTriang4[i][1],
              data->InitLengthTriang4[i][2]  );
    }
  }


  if(mesh->NElemsTriang5 > 0)
  {
    fprintf(InputFP, "\n%s\n", E_KW_LEN_LENGTH_TRIANG5);
    for(i = 0; i < mesh->NElemsTriang5; i++)
    {
      fprintf(InputFP, "%d %.16f %.16f %.16f\n", i+1, 
              data->InitLengthTriang5[i][0],
              data->InitLengthTriang5[i][1],
              data->InitLengthTriang5[i][2]  );
    }
  }

  fclose(InputFP);
  e_free(FileName);

}

/* free-up the data structure */
void FreeInitLengthData(INIT_LENGTH_DATA *data)
{
  /* free-up extra vectors for triangle elements */
  if(data->InitLengthTriang3 != NULL)
  {
    e_free(data->InitLengthTriang3);
  }
  if(data->InitLengthTriang4 != NULL)
  {
    e_free(data->InitLengthTriang4);
  }
  if(data->InitLengthTriang5 != NULL)
  {
    e_free(data->InitLengthTriang5);
  }
  
  /* free-up the global vector */
  e_free(data->InitLength);
  
  e_free(data);

}


/* for parallel version of the program, 
   to send and receive the data to/from the subdomains */
#ifdef __E_MPI__

void SendInitLengthData(INIT_LENGTH_DATA *data, MESH *mesh, int subdom)
{
  int n;

  n = mesh->NElemsLink1 + mesh->NElemsLink2 +
      mesh->NElemsLink3 + mesh->NElemsLink4 +
      mesh->NElemsLink5 + 
      (mesh->NElemsTriang3 + mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;

  MPI_Send(data->InitLength, n, MPI_DOUBLE,
           subdom, 333, MPI_COMM_WORLD);
}

void RecvInitLengthData(INIT_LENGTH_DATA **data, MESH *mesh, int subdom)
{
  int n;
  MPI_Status status;

  n = mesh->NElemsLink1 + mesh->NElemsLink2 +
      mesh->NElemsLink3 + mesh->NElemsLink4 +
      mesh->NElemsLink5 + 
      (mesh->NElemsTriang3 + mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;

  SetupInitLengthData(data, mesh);

  MPI_Recv((*data)->InitLength, n, MPI_DOUBLE,
           subdom, 333, MPI_COMM_WORLD, &status);

}

/* create subdomain initial length information */
void CreateSubInitLengthData(INIT_LENGTH_DATA *Gdata, MESH *Gmesh,
                             INIT_LENGTH_DATA **data, MESH *mesh)
{
  int i, j;

  /* create the initial length information for the subdomain */
  SetupInitLengthData(data, mesh);
      
  /* copy out the length information for subdomain */
  for(i = 0; i < mesh->NElemsLink1; i++)
  {
    (*data)->InitLengthLink1[i] = 
      Gdata->InitLengthLink1[mesh->GEIsLink1[i] - 
                             Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK1]];
  }
  for(i = 0; i < mesh->NElemsLink2; i++)
  {
    (*data)->InitLengthLink2[i] = 
      Gdata->InitLengthLink2[mesh->GEIsLink2[i] - 
                             Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK2]];
  }
  for(i = 0; i < mesh->NElemsLink3; i++)
  {
    (*data)->InitLengthLink3[i] = 
      Gdata->InitLengthLink3[mesh->GEIsLink3[i] - 
                             Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK3]];
  }
  for(i = 0; i < mesh->NElemsLink4; i++)
  {
    (*data)->InitLengthLink4[i] = 
      Gdata->InitLengthLink4[mesh->GEIsLink4[i] - 
                             Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK4]];
  }
  for(i = 0; i < mesh->NElemsLink5; i++)
  {
    (*data)->InitLengthLink5[i] = 
      Gdata->InitLengthLink5[mesh->GEIsLink5[i] - 
                             Gmesh->FirstNetIndex[E_ELEM_TYPE_LINK5]];
  }
  for(i = 0; i < mesh->NElemsTriang3; i++)
  {
    for(j = 0; j < E_NEPE_TRIANG3; j++)
    {
      (*data)->InitLengthTriang3[i][j] = 
        Gdata->InitLengthTriang3[mesh->GEIsTriang3[i] - 
                                 Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]][j];
    }
  }
  for(i = 0; i < mesh->NElemsTriang4; i++)
  {
    for(j = 0; j < E_NEPE_TRIANG4; j++)
    {
      (*data)->InitLengthTriang4[i][j] = 
        Gdata->InitLengthTriang4[mesh->GEIsTriang4[i] - 
                                 Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]][j];
    }
  }
  for(i = 0; i < mesh->NElemsTriang5; i++)
  {
    for(j = 0; j < E_NEPE_TRIANG5; j++)
    {
      (*data)->InitLengthTriang5[i][j] = 
        Gdata->InitLengthTriang5[mesh->GEIsTriang5[i] - 
                                 Gmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]][j];
    }
  }


}

#endif /* __E_MPI__ */




