#include "config.h"  /* configuration options */

#include "e_lib.h"   /* MESH definition */
#include "dr.h"      /* DR_DATA definition */

#include "dr_io.h"   /* all DR data setup for running */
#include "dr_calc.h" /* functions for DR calculation */
#include "dr_cut.h"  /* functions for cutting pattern generation */
#include "dr_data.h" /* possible tasks */
#include "dr_stress.h" /* */

#ifdef __E_MPI__
#include "mpi.h"     /* MPI functions */

#ifdef COMM_LOG
#include <mpe.h>     /* logging MPI communication */
#endif

#endif


int main(int argc, char *argv[])
{
  DR_DATA *data;
  MESH    *mesh;
  CUTTING_PATTERN *cdata;

#ifdef __E_MPI__
  /* the global mesh, it is NULL in the subprocesses */
  DR_DATA *Gdata = NULL;
  MESH    *Gmesh = NULL;
  
  /* initialise DR process, read all data, setup communication,
     store global data and mesh */
  InitDR(argc, argv, &Gdata, &Gmesh, &data, &mesh);

  /* profiling MPI communication */
#ifdef COMM_LOG
  MPI_Barrier(MPI_COMM_WORLD);
  MPE_Start_log();
#endif

#else /* sequential program */
  /* initialise DR process */
  InitDR(argc, argv, &data, &mesh);
#endif

  /* only generate initial length */
  if(data->ProjectTask & TASK_GEN_INIT_LENGTH)
  {
    PutInitLengthData(data->Outputname, data->InitLengthData, mesh);
  }

  /* ########### */
  /* formfinding */
  /* ########### */
  if(data->ProjectTask & TASK_FORM_FINDING)
  {
    dr_main(mesh, data);

#ifdef __E_MPI__
    WriteResultOut(Gdata, Gmesh, data, mesh);
#else
    WriteResultOut(data, mesh);

    if(data->ProjectTask & TASK_WRITE_INIT_LENGTH)
    {
      PutInitLengthData(data->Outputname, data->InitLengthData, mesh);
    }

    if(data->ProjectTask & TASK_WRITE_ANALYSIS)
    {
      WriteAnalysisFile(data->Outputname, data, mesh);
    }
    
#endif
  }

#ifdef __E_MPI__
#ifdef COMM_LOG
  MPE_Stop_log();

  MPE_Finish_log("dr.log");
#endif
#endif

  /* ########################## */
  /* cutting pattern generation */
  /* ########################## */

  if(data->ProjectTask & TASK_CUTTING_PATTERN)
  {

#ifdef __E_MPI__
    if(data->CommData->Par_ProcID == 1)
    {
      dr_cutting_pattern(Gdata->Projectname, data, mesh, &cdata);
    }
    else
    {
      dr_cutting_pattern(NULL, data, mesh, &cdata);
    }
#else
    dr_cutting_pattern(data->Projectname, data, mesh, &cdata);
#endif

#ifdef __E_MPI__
    WriteCuttingPattern(Gdata, Gmesh, data, mesh, cdata);
#else
    WriteCuttingPattern(data, cdata);
#endif
  }

  /* close down the MPI programs */
#ifdef __E_MPI__
  MPI_Finalize();
  if(data->CommData->Par_ProcID == 1)
#endif
    printf("\n\n");

  return(0);
}



