#include "config.h"
#include "e_lib.h"
#include "dr_data.h"
#include "dr_stress.h"

static void InitLengthForUnstress(INIT_LENGTH_DATA **data, MESH *mesh);

/* 
 * LINK3 and LINK4 is merged with LINK2 
 * TRIANG5 merged with TRIANG4
 *
 * Creates only the initial data structure, but does not fill it.
 * Cannot use the function in dr_len.h, because here the mesh is one with
 * form-finding elements and the mesh is not yet converted into an analysis
 * mesh.
 */
static void InitLengthForUnstress(INIT_LENGTH_DATA **data, MESH *mesh)
{
  int n, i, nelem;

  /* check element type */
  if(mesh->NElemsLink1 + mesh->NElemsLink2 +
     mesh->NElemsLink3 + mesh->NElemsLink4 +
     mesh->NElemsLink5 + mesh->NElemsTriang3 + 
     mesh->NElemsTriang4 + mesh->NElemsTriang5 != mesh->TotalNElems)
  {
    e_GeneralError("InitLengthForUnstress: Mesh contain unsupported element !");
  }
  
  /* allocate the data structure */
  (*data) = e_calloc(1, sizeof(INIT_LENGTH_DATA));
  
  n = mesh->NElemsLink1 + mesh->NElemsLink2 +
      mesh->NElemsLink3 + mesh->NElemsLink4 +
      mesh->NElemsLink5 + 
      (mesh->NElemsTriang3 + mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;

  (*data)->InitLength = e_alloc_d_vect(n);

  nelem = 0;
  if(mesh->NElemsLink1)
  {
    (*data)->InitLengthLink1 = (*data)->InitLength + nelem;
    nelem += mesh->NElemsLink1;
  }
  else
  {
    (*data)->InitLengthLink1 = NULL;
  }

    
  if(mesh->NElemsLink2 + mesh->NElemsLink3 + mesh->NElemsLink4)
  {
    (*data)->InitLengthLink2 = (*data)->InitLength + nelem;
    nelem += mesh->NElemsLink2 + mesh->NElemsLink3 + mesh->NElemsLink4;
  }
  else
  {
    (*data)->InitLengthLink2 = NULL;
  }

  (*data)->InitLengthLink3 = NULL;
  (*data)->InitLengthLink4 = NULL;
  (*data)->InitLengthLink5 = NULL;

  
  if(mesh->NElemsTriang3)
  {
    (*data)->InitLengthTriang3 = e_calloc(mesh->NElemsTriang3, sizeof(double *) );
    (*data)->InitLengthTriang3[0] = (*data)->InitLength + nelem;

    for(i = 1; i < mesh->NElemsTriang3; i++)
    {
      (*data)->InitLengthTriang3[i] = (*data)->InitLengthTriang3[i-1] + 3;
    }
    nelem += (mesh->NElemsTriang3) * 3;
  }
  else
  {
    (*data)->InitLengthTriang3 = NULL;
  }

  if(mesh->NElemsTriang4 + mesh->NElemsTriang5)
  {
    (*data)->InitLengthTriang4 =
        e_calloc(mesh->NElemsTriang4 + mesh->NElemsTriang5, sizeof(double *) );
    (*data)->InitLengthTriang4[0] = (*data)->InitLength + nelem;

    for(i = 1; i < mesh->NElemsTriang4 + mesh->NElemsTriang5; i++)
    {
      (*data)->InitLengthTriang4[i] = (*data)->InitLengthTriang4[i-1] + 3;
    }
    nelem += (mesh->NElemsTriang4 + mesh->NElemsTriang5) * 3;
  }
  else
  {
    (*data)->InitLengthTriang4 = NULL;
  }
}


static void InvertMatr3x3(double matr[3][3], double inv_matr[3][3])
{
  double det;

  /* invert the matrix */
  /* ugly but should work */
  det = matr[0][0] * (matr[1][1] * matr[2][2] - matr[2][1] * matr[1][2])
      - matr[0][1] * (matr[1][0] * matr[2][2] - matr[2][0] * matr[1][2])
      + matr[0][2] * (matr[1][0] * matr[2][1] - matr[2][0] * matr[1][1]);

  inv_matr[0][0] =  (matr[1][1] * matr[2][2] - matr[2][1] * matr[1][2]) / det;
  inv_matr[0][1] = -(matr[0][1] * matr[2][2] - matr[2][1] * matr[0][2]) / det;
  inv_matr[0][2] =  (matr[0][1] * matr[1][2] - matr[1][1] * matr[0][2]) / det;

  inv_matr[1][0] = -(matr[1][0] * matr[2][2] - matr[2][0] * matr[1][2]) / det;
  inv_matr[1][1] =  (matr[0][0] * matr[2][2] - matr[2][0] * matr[0][2]) / det;
  inv_matr[1][2] = -(matr[0][0] * matr[1][2] - matr[1][0] * matr[0][2]) / det;

  inv_matr[2][0] =  (matr[1][0] * matr[2][1] - matr[2][0] * matr[1][1]) / det;
  inv_matr[2][1] = -(matr[0][0] * matr[2][1] - matr[2][0] * matr[0][1]) / det;
  inv_matr[2][2] =  (matr[0][0] * matr[1][1] - matr[1][0] * matr[0][1]) / det;
}



void Unstress_GenTriangInitLength(int *Nodes, double **Coords, 
                                  double YoungMod, double PoissonRatio,
                                  double SigmaX, double SigmaY, double TauXY,
                                  double *TriangLength)
{
  int     j, i, ii, jj, kk, n1, n2;
  int     i_comp, i_side;
  double  det, com, sum;
  double  ax_2, ax_3, bx_2, bx_3, cx_2, cx_3;
  double  g[3][3], dd[3][3], matr[3][3], inv_matr[3][3];
  double  dl[3][3];
  double  angle[3];
  double  side_length[3];
  double  stress[3];
  int     side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};
  double  strain;

  /* current side lengths */
  for(i = 0; i < 3; i++)
  {
    n1 = Nodes[side_n_node[i][0]]-1;
    n2 = Nodes[side_n_node[i][1]]-1;
    side_length[i] = 0.00;
    for(j = 0; j < 3; j++)
    {
      dl[i][j] = E_ABS(Coords[n2][j] - Coords[n1][j]);
      side_length[i] += E_SQUARE(dl[i][j]);
    }
    side_length[i] = sqrt(side_length[i]);
  }

  /* calculating matrix G */
  sum = E_SQUARE(side_length[0]) + 
        E_SQUARE(side_length[1]) + 
        E_SQUARE(side_length[2]);

  angle[0] = 0.00;
  angle[1] = M_PI - acos((sum - 2.00 * E_SQUARE(side_length[2])) /
                         (2.00 * side_length[0] * side_length[1]));
  angle[2] = acos((sum - 2.00 * E_SQUARE(side_length[1])) /
                  (2.00 * side_length[0] * side_length[2]));

  ax_2 = E_SQUARE(cos(angle[1]));
  ax_3 = E_SQUARE(cos(angle[2]));

  bx_2 = E_SQUARE(sin(angle[1]));
  bx_3 = E_SQUARE(sin(angle[2]));

  cx_2 = cos(angle[1]) * sin(angle[1]);
  cx_3 = cos(angle[2]) * sin(angle[2]);

  det = bx_2 * cx_3 - cx_2 * bx_3;

  g[0][0] = 1.00;
  g[0][1] = 0.00;
  g[0][2] = 0.00;

  g[1][0] = (ax_3 * cx_2 - ax_2 * cx_3) / det;
  g[1][1] =  cx_3 / det;
  g[1][2] = -cx_2 / det;

  g[2][0] = (ax_2 * bx_3 - ax_3 * bx_2) / det;
  g[2][1] = -bx_3 / det;
  g[2][2] =  bx_2 / det;

  /* calculating matrix D */
  com   = YoungMod / (1.00 - PoissonRatio * PoissonRatio);

  dd[0][0] = com;
  dd[0][1] = dd[1][0] = com * PoissonRatio;
  dd[0][2] = dd[2][0] = 0;
  dd[1][1] = com;
  dd[1][2] = dd[2][1] = 0;
  dd[2][2] = com * (1.00 - PoissonRatio) / 2.00;

  /*
     multiplication --- D * G --- 
  */
  for (ii = 0; ii < 3; ii++)
  {
     for (jj = 0; jj < 3; jj++)
     {
        matr[ii][jj] = 0.00;
        for (kk = 0; kk < 3; kk++)
        {
           matr[ii][jj] += dd[ii][kk]*g[kk][jj];
        }
     }
  }

  /* invert the matrix */
  InvertMatr3x3(matr, inv_matr);
  
  /* stress components */
  stress[0] = SigmaX;
  stress[1] = SigmaY;
  stress[2] = TauXY;

  /* calculate unstressed lengthes */
  for(i_side = 0; i_side < 3; i_side++)
  {
    strain = 0.0;
    for(i_comp = 0; i_comp < 3; i_comp++)
    {
      strain += inv_matr[i_side][i_comp] * stress[i_comp];
    }

    TriangLength[i_side] = side_length[i_side] / (1.0 + strain);
  }

}


INIT_LENGTH_DATA *Unstress_GenMeshInitLength(DR_DATA *data, MESH *mesh)
{
  INIT_LENGTH_DATA *lendat;
  int    i, j, k;
  int    nlink;
  double E, A, force, dl, curlen;
  int    n1, n2;

  /* initialise the length data */
  InitLengthForUnstress(&lendat, mesh);
  
  /* initial length is explicitly given */
  for(i = 0; i < mesh->NElemsLink1; i++)
  {
    lendat->InitLengthLink1[i] = data->InitLengthData->InitLengthLink1[i];
  }
  
  nlink = 0;
  for(i = 0; i < mesh->NElemsLink2; i++)
  {
    lendat->InitLengthLink2[nlink] = data->InitLengthData->InitLengthLink2[i];
    nlink++;
  }

  /* initial length is calculated from the force */
  for(i = 0; i < mesh->NElemsLink3; i++)
  {
    E     = mesh->MatsLink3[i]->YMod;
    A     = mesh->MatsLink3[i]->dParam1;
    force = mesh->MatsLink3[i]->dParam2;

    n1 = mesh->NodesLink3[i][0]-1;
    n2 = mesh->NodesLink3[i][1]-1;
    dl = 0;
    for (k = 0; k < 3; k++)
    {
      dl += E_SQUARE(mesh->MeshPointCoords[n1][k] - mesh->MeshPointCoords[n2][k]);
    }
    curlen = sqrt(dl);

    lendat->InitLengthLink2[nlink] = curlen / (1.0 + (force / (E * A)));
    nlink++;
  }

  for(i = 0; i < mesh->NElemsLink4; i++)
  {
    E     = mesh->MatsLink4[i]->YMod;
    A     = mesh->MatsLink4[i]->dParam1;
    force = mesh->MatsLink4[i]->dParam2;

    n1 = mesh->NodesLink4[i][0]-1;
    n2 = mesh->NodesLink4[i][1]-1;
    dl = 0;
    for (k = 0; k < 3; k++)
    {
      dl += E_SQUARE(mesh->MeshPointCoords[n1][k] - mesh->MeshPointCoords[n2][k]);
    }
    curlen = sqrt(dl);

    lendat->InitLengthLink2[nlink] = curlen / (1.0 + (force / (E * A)));
    nlink++;
  }

  /* initial length is explicitly given */
  for(i = 0; i < mesh->NElemsTriang3; i++)
  {
    for(j = 0; j < 3; j++)
    {
      lendat->InitLengthTriang3[i][j] = data->InitLengthData->InitLengthTriang3[i][j];
    }
  }

  for(i = 0; i < mesh->NElemsTriang4; i++)
  {
    for(j = 0; j < 3; j++)
    {
      lendat->InitLengthTriang4[i][j] = data->InitLengthData->InitLengthTriang4[i][j];
    }
  }

  /* form-finding elements should be unstressed */
  for(i = 0; i < mesh->NElemsTriang5; i++)
  {
    Unstress_GenTriangInitLength(mesh->NodesTriang5[i], mesh->MeshPointCoords, 
                                 mesh->MatsTriang5[i]->YMod, mesh->MatsTriang5[i]->PoissonsRatio,
                                 mesh->MatsTriang5[i]->dParam1, /* Sigma X */
                                 mesh->MatsTriang5[i]->dParam2, /* Sigma Y */
                                 0, /* Tau XY */ /* FIXME: it should be a real value !!!! */
                                 lendat->InitLengthTriang4[mesh->NElemsTriang4+i]);
  }

  return(lendat);
}


/* 
 * Creates a new mesh from the original by replicating everything, except
 *   - LINK3 and LINK4 is merged with LINK2 
 *   - TRIANG5 merged with TRIANG4
 *   - LINK5 elements are not copied
 *   - loads(!!!) and boundary conditions are preserved 
 */
MESH *Unstress_GenMesh(MESH *mesh)
{
  MESH *newmesh;
  int   nelem, i, j;

  /* allocate new mesh */
  newmesh = e_calloc(1, sizeof(MESH) );

  /* same title */
  newmesh->Title = strdup(mesh->Title);
  newmesh->KeywordSet[E_KW_NUM_TITLE] = TRUE;

  /* points do not change */
  newmesh->NMeshPoints = mesh->NMeshPoints;
  newmesh->KeywordSet[E_KW_NUM_NMESHPOINTS] = TRUE;
  newmesh->NNodes = newmesh->NMeshPoints;
  newmesh->KeywordSet[E_KW_NUM_NNODES] = TRUE;

  newmesh->MeshPointCoords = e_alloc_cont_d_matr(mesh->NMeshPoints, 3);
  memcpy((void *)(newmesh->MeshPointCoords[0]),
         (void *)(mesh->MeshPointCoords[0]),
         sizeof(double) * mesh->NMeshPoints * 3 );
  newmesh->KeywordSet[E_KW_NUM_MESHPOINT_COORDS] = TRUE;
  
  /* copy global node numbers */
  newmesh->GNIs = e_alloc_i_vect(newmesh->NMeshPoints);
  memcpy((void *)(newmesh->GNIs),
         (void *)(mesh->GNIs),
         sizeof(int) * newmesh->NMeshPoints);
  
  /* go through all elements, copy and convert them */
  nelem = 1;
  if(mesh->NElemsLink1)
  {
    newmesh->NElemsLink1 = mesh->NElemsLink1;
    newmesh->KeywordSet[E_KW_NUM_NELEMS_LINK1] = TRUE;
    newmesh->FirstNetIndex[E_ELEM_TYPE_LINK1] = nelem;
    nelem += newmesh->NElemsLink1;

    newmesh->NodesLink1 = e_alloc_cont_i_matr(newmesh->NElemsLink1, 2);
    memcpy((void *)(newmesh->NodesLink1[0]),
           (void *)(mesh->NodesLink1[0]),
           sizeof(int) * newmesh->NElemsLink1 * 2);
    newmesh->KeywordSet[E_KW_NUM_NODES_LINK1] = TRUE;

  }

  if(mesh->NElemsLink2 + mesh->NElemsLink3 + mesh->NElemsLink4)
  {
    newmesh->NElemsLink2 = mesh->NElemsLink2 + mesh->NElemsLink3 + mesh->NElemsLink4;
    newmesh->KeywordSet[E_KW_NUM_NELEMS_LINK2] = TRUE;
    newmesh->FirstNetIndex[E_ELEM_TYPE_LINK2] = nelem;
    nelem += newmesh->NElemsLink2;

    newmesh->NodesLink2 = e_alloc_cont_i_matr(newmesh->NElemsLink2, 2);
    if(mesh->NElemsLink2)
    {
      memcpy((void *)(newmesh->NodesLink2[0]),
             (void *)(mesh->NodesLink2[0]),
             sizeof(int) * mesh->NElemsLink2 * 2);
    }
    if(mesh->NElemsLink3)
    {
      memcpy((void *)(newmesh->NodesLink2[mesh->NElemsLink2]),
             (void *)(mesh->NodesLink3[0]),
             sizeof(int) * mesh->NElemsLink3 * 2);
    }
    if(mesh->NElemsLink4)
    {
      memcpy((void *)(newmesh->NodesLink2[mesh->NElemsLink2+mesh->NElemsLink3]),
             (void *)(mesh->NodesLink4[0]),
             sizeof(int) * mesh->NElemsLink4 * 2);
    }
    newmesh->KeywordSet[E_KW_NUM_NODES_LINK2] = TRUE;

  }

  if(mesh->NElemsTriang3)
  {
    newmesh->NElemsTriang3 = mesh->NElemsTriang3;
    newmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG3] = TRUE;
    newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] = nelem;
    nelem += newmesh->NElemsTriang3;

    newmesh->NodesTriang3 = e_alloc_cont_i_matr(newmesh->NElemsTriang3, 3);
    memcpy((void *)(newmesh->NodesTriang3[0]),
           (void *)(mesh->NodesTriang3[0]),
           sizeof(int) * newmesh->NElemsTriang3 * 3);
    newmesh->KeywordSet[E_KW_NUM_NODES_TRIANG3] = TRUE;
  }

  if(mesh->NElemsTriang4 + mesh->NElemsTriang5)
  {
    newmesh->NElemsTriang4 = mesh->NElemsTriang4 + mesh->NElemsTriang5;
    newmesh->KeywordSet[E_KW_NUM_NELEMS_TRIANG4] = TRUE;
    newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] = nelem;
    nelem += newmesh->NElemsTriang4;

    newmesh->NodesTriang4 = e_alloc_cont_i_matr(newmesh->NElemsTriang4, 3);
    if(mesh->NElemsTriang4)
    {
      memcpy((void *)(newmesh->NodesTriang4[0]),
             (void *)(mesh->NodesTriang4[0]),
             sizeof(int) * mesh->NElemsTriang4 * 3);
    }
    if(mesh->NElemsTriang5)
    {
      memcpy((void *)(newmesh->NodesTriang4[mesh->NElemsTriang4]),
             (void *)(mesh->NodesTriang5[0]),
             sizeof(int) * mesh->NElemsTriang5 * 3);
    }
    newmesh->KeywordSet[E_KW_NUM_NODES_TRIANG4] = TRUE;
  }
  
  newmesh->TotalNElems = newmesh->NElemsLink1 +
                         newmesh->NElemsLink2 +
                         newmesh->NElemsTriang3 +
                         newmesh->NElemsTriang4;
  
  if(mesh->NBCNodes)
  {
    newmesh->NBCNodes = mesh->NBCNodes;
    newmesh->KeywordSet[E_KW_NUM_NBC_NODES] = TRUE;

    newmesh->BCNodes = e_alloc_i_vect(newmesh->NBCNodes);
    memcpy((void *)(newmesh->BCNodes),
           (void *)(mesh->BCNodes),
           sizeof(int) * newmesh->NBCNodes);
    
    newmesh->BCTypes        = e_alloc_frag_i_matr(newmesh->NBCNodes, E_DOF_MAX);
    newmesh->BCDispls       = e_alloc_frag_d_matr(newmesh->NBCNodes, E_DOF_MAX);
    newmesh->BCSpringConsts = e_alloc_frag_d_matr(newmesh->NBCNodes, E_DOF_MAX);
    
    for(i = 0; i < newmesh->NBCNodes; i++)
    {
      for(j = 0; j < E_DOF_MAX; j++)
      {
        newmesh->BCTypes[i][j]        = mesh->BCTypes[i][j];
        newmesh->BCDispls[i][j]       = mesh->BCDispls[i][j];
        newmesh->BCSpringConsts[i][j] = mesh->BCSpringConsts[i][j];
      }
    }

    newmesh->KeywordSet[E_KW_NUM_BCS] = TRUE;
  }
  
  if(mesh->NLoadedNodes)
  {
    newmesh->NLoadedNodes = mesh->NLoadedNodes;
    newmesh->KeywordSet[E_KW_NUM_NLOADED_NODES] = TRUE;

    newmesh->LoadedNodes = e_alloc_i_vect(newmesh->NLoadedNodes);
    memcpy((void *)(newmesh->LoadedNodes),
           (void *)(mesh->LoadedNodes),
           sizeof(int) * newmesh->NLoadedNodes);
    
    newmesh->Loads = e_alloc_frag_d_matr(newmesh->NLoadedNodes, E_DOF_MAX);
    
    for(i = 0; i < newmesh->NLoadedNodes; i++)
    {
      for(j = 0; j < E_DOF_MAX; j++)
      {
        newmesh->Loads[i][j] = mesh->Loads[i][j];
      }
    }

    newmesh->KeywordSet[E_KW_NUM_LOADS] = TRUE;
  }

  /* by now we now the total number of elements, so create GEIs */
  newmesh->GEIs = e_alloc_i_vect(newmesh->TotalNElems);

  /* handle materials, but there will not be g-strings */
  newmesh->NMats = mesh->NMats-3;
  newmesh->NMdfMats = mesh->NMdfMats-3;
  newmesh->KeywordSet[E_KW_NUM_NMATS] = TRUE;

  newmesh->Props = e_calloc(mesh->NMats, sizeof(MAT));
  memcpy((void *)(newmesh->Props),
         (void *)(mesh->Props),
         sizeof(MAT) * mesh->NMats);

  for(i = 0; i < mesh->NMats; i++)
  {
    newmesh->Props[i].Name = strdup(newmesh->Props[i].Name);
  }

  newmesh->Mats = e_alloc_vp_vect(newmesh->TotalNElems);

  if(newmesh->NElemsLink1)
  {
    newmesh->MatsLink1 = (MAT **)(newmesh->Mats + 
                                  newmesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1 );
    newmesh->GEIsLink1 = newmesh->GEIs + 
                         newmesh->FirstNetIndex[E_ELEM_TYPE_LINK1]-1;

    for(i = 0; i < newmesh->NElemsLink1; i++)
    {
      newmesh->MatsLink1[i] = (mesh->MatsLink1[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsLink1[i] = mesh->GEIsLink1[i];
    }
    newmesh->KeywordSet[E_KW_NUM_MATS_LINK1] = TRUE;
  }

  if(newmesh->NElemsLink2)
  {
    newmesh->MatsLink2 = (MAT **)(newmesh->Mats + 
                                  newmesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1 );
    newmesh->GEIsLink2 = newmesh->GEIs + 
                         newmesh->FirstNetIndex[E_ELEM_TYPE_LINK2]-1;

    nelem = 0;

    for(i = 0; i < mesh->NElemsLink2; i++)
    {
      newmesh->MatsLink2[nelem] = (mesh->MatsLink2[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsLink2[nelem] = mesh->GEIsLink2[i];
      nelem++;
    }

    for(i = 0; i < mesh->NElemsLink3; i++)
    {
      newmesh->MatsLink2[nelem] = (mesh->MatsLink3[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsLink2[nelem] = mesh->GEIsLink3[i];
      nelem++;
    }

    for(i = 0; i < mesh->NElemsLink4; i++)
    {
      newmesh->MatsLink2[nelem] = (mesh->MatsLink4[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsLink2[nelem] = mesh->GEIsLink4[i];
      nelem++;
    }
    
    newmesh->KeywordSet[E_KW_NUM_MATS_LINK2] = TRUE;
  }

  if(newmesh->NElemsTriang3)
  {
    newmesh->MatsTriang3 = (MAT **)(newmesh->Mats + 
                                    newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1 );
    newmesh->GEIsTriang3 = newmesh->GEIs + 
                           newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]-1;

    for(i = 0; i < newmesh->NElemsTriang3; i++)
    {
      newmesh->MatsTriang3[i] = (mesh->MatsTriang3[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsTriang3[i] = mesh->GEIsTriang3[i];
    }
    newmesh->KeywordSet[E_KW_NUM_MATS_TRIANG3] = TRUE;
  }


  if(newmesh->NElemsTriang4 + newmesh->NElemsTriang5)
  {
    newmesh->MatsTriang4 = (MAT **)(newmesh->Mats + 
                                    newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1 );
    newmesh->GEIsTriang4 = newmesh->GEIs + 
                           newmesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]-1;
    nelem = 0;
    for(i = 0; i < mesh->NElemsTriang4; i++)
    {
      newmesh->MatsTriang4[nelem] = (mesh->MatsTriang4[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsTriang4[nelem] = mesh->GEIsTriang4[i];
      nelem++;
    }
    for(i = 0; i < mesh->NElemsTriang5; i++)
    {
      newmesh->MatsTriang4[nelem] = (mesh->MatsTriang5[i] - mesh->Props) + newmesh->Props;
      newmesh->GEIsTriang4[nelem] = mesh->GEIsTriang5[i];
      nelem++;
    }
    
    newmesh->KeywordSet[E_KW_NUM_MATS_TRIANG4] = TRUE;
  }

  /* remaining parameters to set */
  newmesh->TimeStep = mesh->TimeStep;
  newmesh->KeywordSet[E_KW_NUM_TIMESTEP] = TRUE;
  
  newmesh->NInternalTimeSteps = mesh->NInternalTimeSteps;
  newmesh->KeywordSet[E_KW_NUM_NINTERNAL_TIMESTEPS] = TRUE;
  
  newmesh->NExternalTimeSteps = mesh->NExternalTimeSteps;
  newmesh->KeywordSet[E_KW_NUM_NEXTERNAL_TIMESTEPS] = TRUE;

  newmesh->DampingFactor = mesh->DampingFactor;
  newmesh->KeywordSet[E_KW_NUM_DAMPING_FACTOR] = TRUE;
  

  newmesh->NodalDOF = e_alloc_i_vect(newmesh->NMeshPoints);
  for(i = 0; i < newmesh->NMeshPoints; i++)
  {
    newmesh->NodalDOF[i] = 3;
  }
  newmesh->TotalNodalDOF = newmesh->NMeshPoints * 3;

  return(newmesh);
}



void WriteAnalysisFile(char *ProjectName, DR_DATA *data, MESH *mesh)
{
  INIT_LENGTH_DATA *lendat;
  MESH  *newmesh;
  char   analysisname[E_LINE_MAX];

  lendat = Unstress_GenMeshInitLength(data, mesh);

  /* convert mesh as well */
  newmesh = Unstress_GenMesh(mesh);
  
  sprintf(analysisname, "%s_anal", ProjectName);
  
  /* write out the mesh as well */
  e_PutMeshData(newmesh, analysisname);

  /* write out the length file*/
  PutInitLengthData(analysisname, lendat, newmesh);
}


/* *************************************************************************** */
/* force and stress determination */
/* *************************************************************************** */

void LinkForce(int NElems, int **Nodes, 
               double **Coords, MAT **Material, double *InitLength,
               int negative, double ***LinkStresses)
{
  double E, A, il, curlen, dl, force;
  int    i, n1, n2, k;

  for(i = 0; i < NElems; i++)
  {
    E  = Material[i]->YMod;
    A  = Material[i]->dParam1;
    il = InitLength[i];
    
    /* calc surrent length */
    n1 = Nodes[i][0]-1;
    n2 = Nodes[i][1]-1;
    dl = 0;
    for (k = 0; k < 3; k++)
    {
      dl += E_SQUARE(Coords[n1][k] - Coords[n2][k]);
    }
    curlen = sqrt(dl);
     
    force = E * A * (curlen-il) / il;
    if(force < 0 && !negative)
      LinkStresses[i][0][0] = 0.0;
    else
        LinkStresses[i][0][0] = force;
  }
}



/* FIXME: very similar to dr_triang_mass and Unstress_GenTriangInitLength */
void TriangForce(int NElems, int **Nodes,
                 double **Coords, MAT **Material, 
                 double **InitLength, double ***TriangStresses)
{
  int j, i, l, ii, jj, kk, n1, n2;
  double det, young, pois, com, sum;
  double aa, cc;
  double  ax_2, ax_3, bx_2, bx_3, cx_2, cx_3;
  double  g[3][3];                      /* G matrix */
  double  dd[3][3];                     /* D matrix */
  double  k1[3][3];                     /* temporary matrix */
  double  angle[3];
  double  side_length[3];
  double  dl[3][3];                     /* side length components */
  int     side_n_node[3][2] = {{1, 2}, {2, 0}, {0, 1}};
  double  stress, strain[3];

  fprintf(stderr, "\n\nCheck how the stress is written !!!\n\n");
  for(l = 0; l < NElems; l++)
  {
    /* current side lengths */
    for(i = 0; i < 3; i++)
    {
      n1 = Nodes[l][side_n_node[i][0]]-1;
      n2 = Nodes[l][side_n_node[i][1]]-1;
      side_length[i] = 0.00;
      for(j = 0; j < 3; j++)
      {
        dl[i][j] = Coords[n2][j] - Coords[n1][j];
        side_length[i] += E_SQUARE(dl[i][j]);
      }
      side_length[i] = sqrt(side_length[i]);
    }

#if 0
    /* !!!!! only in 2D !!!!! */
    {
      double a1, a2, a3, b1, b2, b3, c1, c2, c3;

      a1 = E_SQUARE(dl[0][0] / side_length[0]);
      a2 = E_SQUARE(dl[1][0] / side_length[1]);
      a3 = E_SQUARE(dl[2][0] / side_length[2]);
      
      b1 = E_SQUARE(dl[0][1] / side_length[0]);
      b2 = E_SQUARE(dl[1][1] / side_length[1]);
      b3 = E_SQUARE(dl[2][1] / side_length[2]);
      
      c1 = (dl[0][0] / side_length[0]) * (dl[0][1] / side_length[0]);
      c2 = (dl[1][0] / side_length[1]) * (dl[1][1] / side_length[1]);
      c3 = (dl[2][0] / side_length[2]) * (dl[2][1] / side_length[2]);
      
      det = a1 * (b2 * c3 - b3 * c2) -
          b1 * (a2 * c3 - a3 * c2) +
          c1 * (a2 * b3 - a3 * b2);
      
      g[0][0] = (b2*c3 - b3*c2) / det;
      g[0][1] = (b3*c1 - b1*c3) / det;
      g[0][2] = (b1*c2 - b2*c1) / det;
      
      g[1][0] = (a3*c2 - a2*c3) / det;
      g[1][1] = (a1*c3 - a3*c1) / det;
      g[1][2] = (a2*c1 - a1*c2) / det;
      
      g[2][0] = (a2*b3 - a3*b2) / det;
      g[2][1] = (a3*b1 - a1*b3) / det;
      g[2][2] = (a1*b2 - a2*b1) / det;
    }
#endif
    
    /* stress directions are parallel with first side !!!! */
    sum = E_SQUARE(side_length[0]) + 
          E_SQUARE(side_length[1]) + 
          E_SQUARE(side_length[2]);

    /* we have to use the actual side lengths, see Ref. */
    angle[0] = 0.00;
    angle[1] = M_PI - acos((sum - 2.00 * E_SQUARE(side_length[2])) /
                           (2.00 * side_length[0] * side_length[1]));
    angle[2] = acos((sum - 2.00 * E_SQUARE(side_length[1])) /
                    (2.00 * side_length[0] * side_length[2]));

    ax_2 = E_SQUARE(cos(angle[1]));
    ax_3 = E_SQUARE(cos(angle[2]));

    bx_2 = E_SQUARE(sin(angle[1]));
    bx_3 = E_SQUARE(sin(angle[2]));

    cx_2 = cos(angle[1])*sin(angle[1]);
    cx_3 = cos(angle[2])*sin(angle[2]);

    det = bx_2 * cx_3 - cx_2 * bx_3;
    g[0][0] = 1.00;
    g[0][1] = 0.00;
    g[0][2] = 0.00;

    g[1][0] = (ax_3 * cx_2 - ax_2 * cx_3) / det;
    g[1][1] =  cx_3 / det;
    g[1][2] = -cx_2 / det;

    g[2][0] = (ax_2 * bx_3 - ax_3 * bx_2) / det;
    g[2][1] = -bx_3 / det;
    g[2][2] =  bx_2 / det;

    
    /* calculating matrix D */
    young = Material[l]->YMod;
    pois  = Material[l]->PoissonsRatio;
    com   = young / (1.00 - pois * pois);

    dd[0][0] = com;
    dd[0][1] = dd[1][0] = com * pois;
    dd[0][2] = dd[2][0] = 0;
    dd[1][1] = com;
    dd[1][2] = dd[2][1] = 0;
    dd[2][2] = com * (1.00 - pois) / 2.00;

      /* D * G */
    for(ii = 0; ii < 3; ii++)
    {
       for(jj = 0; jj < 3; jj++)
       {
          k1[ii][jj] = 0.00;
          for(kk = 0; kk < 3; kk++)
             k1[ii][jj] += dd[ii][kk]*g[kk][jj];
       }
    }

    
    /* the strain per side */
    for(ii = 0; ii < 3; ii++)
    {
      strain[ii] = (side_length[ii] - InitLength[l][ii]) / InitLength[l][ii]; 
    }
    
    /* the stress */
    for(ii = 0; ii < 3; ii++)
    {
      stress = 0.0;
      for(jj = 0; jj < 3; jj++)
      {
        stress += k1[ii][jj] * strain[jj];
      }
      TriangStresses[l][0][ii] = stress;
    }

    /* calculate principal stresses */
    cc = (TriangStresses[l][0][0] + TriangStresses[l][0][1]) / 2.0;
    aa = sqrt(((TriangStresses[l][0][0] - TriangStresses[l][0][1]) *
               (TriangStresses[l][0][0] - TriangStresses[l][0][1]) / 4.0) +
              (TriangStresses[l][0][2] * TriangStresses[l][0][2]));
    
    /* store the principal stresses */
    TriangStresses[l][0][4] = cc + aa;
    TriangStresses[l][0][5] = cc - aa;
    
  }

}


void CheckTriangForce(int NElems, int **Nodes, double **Coords, 
                      MAT **Material, double **InitCoords, 
                      double ***TriangStresses)
{
  int i_triang, ii, jj, kk;
  int n1, n2, n3;
  double aj, ak, bj, bk, area2;
  double a[3][6];
  double young, pois, com;
  double dd[3][3];
  double b[3][6];
  double displ[6];
  double stress;

  for(i_triang = 0; i_triang < NElems; i_triang++)
  {
      /* B matrix */
    n1 = Nodes[i_triang][0]-1;
    n2 = Nodes[i_triang][1]-1;
    n3 = Nodes[i_triang][2]-1;

    aj = InitCoords[n2][0] - InitCoords[n1][0];
    ak = InitCoords[n3][0] - InitCoords[n1][0];
    bj = InitCoords[n2][1] - InitCoords[n1][1];
    bk = InitCoords[n3][1] - InitCoords[n1][1];

    area2 = (aj * bk - ak * bj);

    a[0][0] = (bj - bk) / area2;
    a[0][1] = 0.0;
    a[0][2] = bk / area2;
    a[0][3] = 0.0;
    a[0][4] = -bj / area2;
    a[0][5] = 0.0;
    
    a[1][0] = 0.0;
    a[1][1] = (ak - aj) / area2;
    a[1][2] = 0;
    a[1][3] = -ak / area2;
    a[1][4] = 0;
    a[1][5] = aj / area2;

    a[2][0] = (ak - aj) / area2;
    a[2][1] = (bj - bk) / area2;
    a[2][2] = -ak / area2;
    a[2][3] = bk / area2;
    a[2][4] = aj / area2;
    a[2][5] = -bj / area2;

      /* D matrix */
    young = Material[i_triang]->YMod;
    pois  = Material[i_triang]->PoissonsRatio;
    com   = young / (1.00 - pois * pois);

    dd[0][0] = com;
    dd[0][1] = dd[1][0] = com * pois;
    dd[0][2] = dd[2][0] = 0;
    dd[1][1] = com;
    dd[1][2] = dd[2][1] = 0;
    dd[2][2] = com * (1.00 - pois) / 2.00;

      /* B * D multiplication */
    for(ii = 0; ii < 3; ii++)
    {
       for(jj = 0; jj < 6; jj++)
       {
          b[ii][jj] = 0.00;
          for(kk = 0; kk < 3; kk++)
             b[ii][jj] += dd[ii][kk]*a[kk][jj];
       }
    }

      /* displacements of the nodes (global coordinate system) */
    displ[0] = Coords[n1][0] - InitCoords[n1][0];
    displ[1] = Coords[n1][1] - InitCoords[n1][1];
    displ[2] = Coords[n2][0] - InitCoords[n2][0];
    displ[3] = Coords[n2][1] - InitCoords[n2][1];
    displ[4] = Coords[n3][0] - InitCoords[n3][0];
    displ[5] = Coords[n3][1] - InitCoords[n3][1];

      /* calculation of stresses */
    for(ii = 0; ii < 3; ii++)
    {
      int tt;
      
      stress = 0.0;
      for(jj = 0; jj < 6; jj++)
      {
        tt = b[ii][jj] * displ[jj];
        stress += tt;
      }
      TriangStresses[i_triang][0][3+ii] = stress;
    }
  }
}




void SetupStressStruct(MESH *mesh)
{
    if(mesh->NElemsLink1)
    {
      mesh->NStressPointsLink1 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK1] = TRUE;
    }
    if(mesh->NElemsLink2)
    {
      mesh->NStressPointsLink2 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK2] = TRUE;
    }
    if(mesh->NElemsLink3)
    {
      mesh->NStressPointsLink3 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK3] = TRUE;
    }
    if(mesh->NElemsLink4)
    {
      mesh->NStressPointsLink4 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK4] = TRUE;
    }
    if(mesh->NElemsLink5)
    {
      mesh->NStressPointsLink5 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK5] = TRUE;
    }
    if(mesh->NElemsTriang3)
    {
      mesh->NStressPointsTriang3 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG3] = TRUE;
    }
    if(mesh->NElemsTriang4)
    {
      mesh->NStressPointsTriang4 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG4] = TRUE;
    }
    if(mesh->NElemsTriang5)
    {
      mesh->NStressPointsTriang5 = 1;
      mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG5] = TRUE;
    }
    
    e_SetupStresses(mesh);
    mesh->SubStructSet[E_SS_NUM_STRESS] = TRUE;
}


#if 0
void WriteForceFile(char *ProjectName, DR_DATA *data, MESH *mesh)
{
  FILE  *fp;
  char   name[E_LINE_MAX];

  /* setup stresses */
  if(mesh->NElemsLink1)
  {
    mesh->NStressPointsLink1 = 1;
    mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK1] = TRUE;
  }
  if(mesh->NElemsLink2)
  {
    mesh->NStressPointsLink2 = 1;
    mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_LINK2] = TRUE;
  }
  if(mesh->NElemsTriang3)
  {
    mesh->NStressPointsTriang3 = 1;
    mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG3] = TRUE;
  }
  if(mesh->NElemsTriang4)
  {
    mesh->NStressPointsTriang4 = 1;
    mesh->KeywordSet[E_KW_NUM_NSTRESS_POINTS_TRIANG4] = TRUE;
  }

  e_SetupStresses(mesh);
  mesh->SubStructSet[E_SS_NUM_STRESS] = TRUE;

  
  /* force file, little bit different from stress file */
  sprintf(name, "%s.force", ProjectName);
  fp = fopen(name, "w");
  if(fp == NULL)
  {
    e_GeneralError("Cannot open file: %s", name);
  }
  
  if(mesh->NElemsLink1)
  {
    fprintf(fp, "FORCE_LINK1\n");
    WriteLinkForce(fp, mesh->NElemsLink1, mesh->NodesLink1, 
                   mesh->MeshPointCoords, mesh->MatsLink1, 
                   data->InitLengthData->InitLengthLink1, FALSE);
    fprintf(fp, "END\n");
  }
  
  if(mesh->NElemsLink2)
  {
    fprintf(fp, "FORCE_LINK2\n");
    WriteLinkForce(fp, mesh->NElemsLink2, mesh->NodesLink2,
                   mesh->MeshPointCoords, mesh->MatsLink2, 
                   data->InitLengthData->InitLengthLink2, TRUE);
    fprintf(fp, "END\n");
  }
  
  if(mesh->NElemsTriang3)
  {
    fprintf(fp, "FORCE_TRIANG3\n");
    WriteTriangStress(fp, mesh->NElemsTriang3, mesh->NodesTriang3, 
                      mesh->MeshPointCoords, mesh->MatsTriang3,
                      data->InitLengthData->InitLengthTriang3, 
                      mesh->Stresses.StressesTriang3);
    fprintf(fp, "END\n");
  }
  
  if(mesh->NElemsTriang4)
  {
    fprintf(fp, "FORCE_TRIANG4\n");
    WriteTriangStress(fp, mesh->NElemsTriang4, mesh->NodesTriang4, 
                      mesh->MeshPointCoords, mesh->MatsTriang4,
                      data->InitLengthData->InitLengthTriang4,
                      mesh->Stresses.StressesTriang4);
    fprintf(fp, "END\n");
  }

  fclose(fp);
  
  e_PutStresses(mesh, ProjectName);

}

#endif




