/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                  Functions Common to Other Source Modules                  */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */
#include "e_pltfrm.h"

#ifndef __E_T800SAL__

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <e_lib.h>


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           Input-Output Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* OpenInputFile: To open an input file. The extension "extension" is added
   to the filename argument.                                                  */

/* The return value is a pointer to the full file pathname if the file is
   opened successfully, else NULL.                                            */


char *OpenInputFile(char *FileName, char *extension, FILE **fp)
{
   int FullPathNameLength;
   char *FullPathName;


   FullPathNameLength = strlen(FileName) + strlen(extension) + 1;
   FullPathName = e_alloc_c_vect(FullPathNameLength + 1);

   if (FullPathName == NULL)
      return NULL;

   strcpy(FullPathName,FileName);
   strcat(FullPathName,".");
   strcat(FullPathName,extension);

   if ((*fp = fopen(FullPathName,"r")) == NULL)
   {
      e_FileError(FullPathName,"cannot open file");
      e_free(FullPathName);
      return NULL;
   }
   else
      return FullPathName;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* OpenOutputFile: To open an output file. The extension "extension" is added
   to the filename argument.                                                  */

/* The return value is a pointer to the filename argument if the file is
   opened successfully, else NULL.                                            */


char *OpenOutputFile(char *FileName, char *extension, FILE **fp)
{
   int FullPathNameLength;
   char *FullPathName;


   FullPathNameLength = strlen(FileName) + strlen(extension) + 1;
   FullPathName = e_alloc_c_vect(FullPathNameLength + 1);

   if (FullPathName == NULL)
      return NULL;

   strcpy(FullPathName,FileName);
   strcat(FullPathName,".");
   strcat(FullPathName,extension);

   if ((*fp = fopen(FullPathName,"w")) == NULL)
   {
      e_FileError(FullPathName,"cannot open file");
      e_free(FullPathName);
      return NULL;
   }
   else
   {
      e_free(FullPathName);
      return FileName;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetKeyword: To read in a potential keyword. */

/* The return value is 1 if the word is read successfully, EOF if the end of
   file is encountered, else 0. Note that in the event of failure, "word[0]"
   is set to '\0'.                                                            */


int GetKeyword(char *line, char *word, int *LineNumber, char *FullPathName,
                                                                   FILE *fp)
{
   int ReturnValue;             /* return value of 'e_GetLine' */


   if ((ReturnValue = e_GetLine(line,E_LINE_MAX,LineNumber,fp)) == EOF)
   {
      return EOF;
   }
   else if (ReturnValue == 0)
   {
      e_FileError(FullPathName,"system read error");
      *word = '\0';
      return 0;
   }
   else if (sscanf(line,"%s",word) == 1)
   {
      return 1;
   }
   else
   {
      e_FileError(FullPathName,"system read error");
      *word = '\0';
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetStrArg: To read a single formatted argument from a string. */

/* The return value is a pointer to the next argument in the string if the
   current argument is read successfully, else NULL.                          */


char *GetStrArg(char *pLine, char *format, void *argument)
{
   int FormatLength;
   int SscanfReturnValue;

   char c;
   char *NewFormat;


   FormatLength = strlen(format);

   if (   pLine == NULL
       || (NewFormat = e_alloc_c_vect(FormatLength + 3)) == NULL)
      return NULL;

   strcpy(NewFormat,format);
   strcat(&NewFormat[FormatLength],"%c");

   while (isspace((int)(*pLine)))
      pLine++;

   SscanfReturnValue = sscanf(pLine,NewFormat,argument,&c);

   if (SscanfReturnValue == 2)
   {
      if (   !isspace((int)c)
          && c != '\n'
          && ((unsigned int)(c)) != EOF)
      {
         e_free(NewFormat);
         return NULL;
      }
   }
   else if (SscanfReturnValue != 1)
   {
      e_free(NewFormat);
      return NULL;
   }

   while (!isspace((int)(*pLine)) && *pLine != '\0')
      pLine++;

   e_free(NewFormat);
   return pLine;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PrintEng: To output a double precision number in engineering format. */

/* The return value is the number of characters written to the output stream
   if successful, else -1.                                                    */


#define DEFAULT_PRECISION   6
#define PRECISION_MAX      97   /* PRECISION_MAX + 2 has less than 3 digits */ 


int PrintEng(FILE *fp, char *format, double number)
{
   int i;
   int delta;
   int exponent;
   int FieldWidth;
   int NBlanks;
   int NCharactersPrinted;
   int NumberStringLength;
   int precision;
   
   char *NewFormat;
   char *NumberString;  
   char *pExponent;  
   char *pFieldWidth;
   char *pNewFormat;
   char *pNumberString;  
   char *pPrecision;  

      
   if ((pFieldWidth = strchr(format,'%')) != NULL)
   {  
      if (sscanf(++pFieldWidth,"%d",&FieldWidth) != 1)
         FieldWidth = 1;
   }
   else
   {
      e_GeneralError("'PrintEng' format error");
      return -1;
   }
      
   if ((pPrecision = strchr(format,'.')) != NULL)
   {  
      if (   sscanf(++pPrecision,"%d",&precision) != 1
          || precision > PRECISION_MAX                )
      {
         e_GeneralError("'PrintEng' format error");
         return -1;
      }
   }
   else
      precision = DEFAULT_PRECISION;
   
   NumberStringLength
      = (FieldWidth > precision + 8) ? FieldWidth : precision + 8;
   NumberString = NULL;
   NewFormat = NULL;
   
   if (   (NumberString = e_alloc_c_vect(NumberStringLength + 2 + 1)) == NULL
       ||                         (NewFormat = e_alloc_c_vect(6 + 1)) == NULL)
   {
      e_free(NumberString);
      e_free(NewFormat);
      return -1;
   }   
      
   sprintf(NumberString,format,number);
   
   if (   (pExponent = strchr(NumberString,'e')) != NULL
       || (pExponent = strchr(NumberString,'E')) != NULL)
   {  
      if (sscanf(pExponent + 1,"%d",&exponent) != 1)
      {
         e_GeneralError("'PrintEng' format error");
         e_free(NumberString);
         e_free(NewFormat);
         return -1;
      }   
   }
   else
   {
      e_GeneralError("'PrintEng' format error");
      e_free(NumberString);
      e_free(NewFormat);
      return -1;
   }   

   if (exponent < 0)         
   {
      exponent = -exponent;
      delta = exponent % 3;
      
      if (delta == 1)
         delta = 2;
      else if (delta == 2)
         delta = 1;
   }
   else
      delta = exponent % 3;
        
   precision += delta;
  
   for (i = 0; i < delta; i++)
   {
      exponent--;
      number *= 0.1;
   }
   
   pNewFormat = NewFormat;
   *pNewFormat++ = '%';
   *pNewFormat++ = '1';
   *pNewFormat++ = '.';

   sprintf(pNewFormat,"%d",precision);
   
   while (*pNewFormat != '\0')
      pNewFormat++;
   
   *pNewFormat++ = *pExponent;
   *pNewFormat = '\0';
   
   sprintf(NumberString,NewFormat,number);
   
   if (precision > 0)
   {   
      pNumberString = strchr(NumberString,'.');
      
      for (i = 0; i < delta; i++)
      {
         *pNumberString = *(pNumberString + 1);
         pNumberString++;
      }
      
      *pNumberString = '.';
   }

   pNumberString = strchr(NumberString,*pExponent);
   pNumberString++;

   if (exponent == 0 && *pNumberString == '-')
      *pNumberString = '+';
   
   if (precision > 0 && precision == delta)
   {
      pNumberString = strchr(NumberString,'.');
      
      do
      {
         *pNumberString = *(pNumberString + 1);   
         pNumberString++;
      }
      while (*pNumberString != '\0');
   }
   
   NBlanks = FieldWidth - strlen(NumberString);
   
   if (NBlanks < 0)
      NBlanks = 0;
         
   for (i = 0; i < NBlanks; i++)
   {
      if (fprintf(fp," ") < 0)
      {
         e_GeneralError("'PrintEng' output error");
         e_free(NumberString);
         e_free(NewFormat);
         return -1;
      }   
   }

   if ((NCharactersPrinted = fprintf(fp,"%s",NumberString)) >= 0)
   {
      e_free(NumberString);
      e_free(NewFormat);
      return NBlanks + NCharactersPrinted;
   }
   else
   {
      e_GeneralError("'PrintEng' output error");
      e_free(NumberString);
      e_free(NewFormat);
      return -1;
   }   
}


#endif   /* #ifndef __E_T800SAL__ */
