/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                     e_GetGeomModel(), e_PutGeomModel()                     */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>

#include "comsrc.h"
#include "geommod.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetGeomModel: To read in geometric model data. */
/* The return value is 1 if no input errors occur, else 0. */


int e_GetGeomModel(MESH *mesh, char *InputFile)
{
   if (!mesh->KeywordSet[E_KW_NUM_REMESH_DATA])
   {
      e_GeneralError("e_GetGeomModel: remeshing data not defined");
      return 0;
   }
   else if ((FullPathName = OpenInputFile(InputFile,E_GMF,&fp)) == NULL)
      return 0;

   Initialise(mesh);

   if (GetData())
   {
      SetGeomData(mesh);
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      DeallocateAllMemory();
      fclose(fp);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Initialise: To initialise variables. */


static void Initialise(MESH *mesh)
{
   int j,k;


   Geom = &mesh->Geom;
   Geom->NEndPoints = 0;
   Geom->EndPointCoords = NULL;

   for (k = 0; k < Geom->NNurbsCurves; k++)
   {
      Geom->NurbsCurves[k].CntlPoints = NULL;
      Geom->NurbsCurves[k].BCTypes = NULL;
      Geom->NurbsCurves[k].BCDispls = NULL;
      Geom->NurbsCurves[k].BCSpringConsts = NULL;
      Geom->NurbsCurves[k].PointLoad = NULL;
      Geom->NurbsCurves[k].LinLoadGCBeg = NULL;
      Geom->NurbsCurves[k].LinLoadGCEnd = NULL;
      Geom->NurbsCurves[k].LinLoadCCBeg = NULL;
      Geom->NurbsCurves[k].LinLoadCCEnd = NULL;

      for (j = 0; j < E_NC_NLOAD_TYPES; j++)
         Geom->NurbsCurves[k].LoadTypeSet[j] = FALSE;
   }

   LineNumber = 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetData()
{
   int k;
   int NNurbsCurves;
   int NurbsCurveIndex;
   int ReturnValue;

   char *NurbsCurveSet;
   char word[E_LINE_MAX + 1];


   ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

   if (ReturnValue == EOF)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else if (   ReturnValue == 0
            || !WordMatchesKeyword(E_GKW_NENDPOINTS,word)
            || !GetNEndPoints()
            || (Geom->EndPointCoords
                   = e_alloc_cont_d_matr(Geom->NEndPoints,E_NDIMS)) == NULL)
   {
      return 0;
   }

   ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

   if (ReturnValue == EOF)
   {
      e_FileError(FullPathName,"unexpected end of file");
      return 0;
   }
   else if (   ReturnValue == 0
            || !WordMatchesKeyword(E_GKW_ENDPOINT_COORDS,word)
            || !GetEndPointCoords()                           )
   {
      return 0;
   }

   if ((NurbsCurveSet = e_alloc_c_vect(Geom->NNurbsCurves)) == NULL)
      return 0;

   NNurbsCurves = 0;

   while (NNurbsCurves < Geom->NNurbsCurves)
   {
      ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

      if (ReturnValue == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         e_free(NurbsCurveSet);
         return 0;
      }
      else if (   ReturnValue == 0
               || !WordMatchesKeyword(E_GKW_NC,word)
               || (NurbsCurveIndex = GetNurbsCurveIndex()) == 0)
      {
         e_free(NurbsCurveSet);
         return 0;
      }

      k = NurbsCurveIndex - 1;

      if (NurbsCurveSet[k])
      {
         e_FileLineError(LineNumber,FullPathName,
                         "NURBS curve multiply defined");
         e_free(NurbsCurveSet);
         return 0;
      }

      NurbsCurveSet[k] = TRUE;
      NNurbsCurves++;

      ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

      if (ReturnValue == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         e_free(NurbsCurveSet);
         return 0;
      }
      else if (   ReturnValue == 0
               || !WordMatchesKeyword(E_GKW_NC_DEGREE,word)
               || !GetNurbsCurveDegree(&Geom->NurbsCurves[k]))
      {
         e_free(NurbsCurveSet);
         return 0;
      }

      ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

      if (ReturnValue == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         e_free(NurbsCurveSet);
         return 0;
      }
      else if (   ReturnValue == 0
               || !WordMatchesKeyword(E_GKW_NC_CNTL_POINTS,word)
               || !GetNurbsCurveCntlPoints(&Geom->NurbsCurves[k]))
      {
         e_free(NurbsCurveSet);
         return 0;
      }

      ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

      if (ReturnValue == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         e_free(NurbsCurveSet);
         return 0;
      }
      else if (   ReturnValue == 0
               || !WordMatchesKeyword(E_GKW_NC_WEIGHTS,word)
               || !GetNurbsCurveWeights(&Geom->NurbsCurves[k]))
      {
         e_free(NurbsCurveSet);
         return 0;
      }

      ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

      if (ReturnValue == EOF)
      {
         e_FileError(FullPathName,"unexpected end of file");
         e_free(NurbsCurveSet);
         return 0;
      }
      else if (ReturnValue == 0)
      {
         e_free(NurbsCurveSet);
         return 0;
      }

      if (   strcmp(E_GKW_NC_FIXITY_GC,word) == 0
          || strcmp(E_GKW_NC_FIXITY_CC,word) == 0)
      {
         if (!GetFixity(&Geom->NurbsCurves[k]))
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         if (strcmp(E_GKW_NC_FIXITY_GC,word) == 0)
            Geom->NurbsCurves[k].FixityType = E_NC_NUM_FIXITY_GC;
         else
            Geom->NurbsCurves[k].FixityType = E_NC_NUM_FIXITY_CC;

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (ReturnValue == 0)
         {
            e_free(NurbsCurveSet);
            return 0;
         }
      }
      else
      {
         Geom->NurbsCurves[k].FixityType = E_NC_NUM_NO_FIXITY;
      }

      if (strcmp(E_GKW_NC_POINT_LOAD,word) == 0)
      {
         if (!GetLoad(&Geom->NurbsCurves[k].PointLoad))
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (   ReturnValue == 0
                  || !WordMatchesKeyword(E_GKW_NC_POINT_LOAD_PARAM,word)
                  || !GetPointLoadParam(&Geom->NurbsCurves[k])          )
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         Geom->NurbsCurves[k].LoadTypeSet[E_NC_NUM_POINT_LOAD] = TRUE;

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (ReturnValue == 0)
         {
            e_free(NurbsCurveSet);
            return 0;
         }
      }

      if (strcmp(E_GKW_NC_LIN_LOAD_GC_BEG,word) == 0)
      {
         if (!GetLoad(&Geom->NurbsCurves[k].LinLoadGCBeg))
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (   ReturnValue == 0
                  || !WordMatchesKeyword(E_GKW_NC_LIN_LOAD_GC_END,word)
                  || !GetLoad(&Geom->NurbsCurves[k].LinLoadGCEnd)      )
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         Geom->NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_GC] = TRUE;

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (ReturnValue == 0)
         {
            e_free(NurbsCurveSet);
            return 0;
         }
      }

      if (strcmp(E_GKW_NC_LIN_LOAD_CC_BEG,word) == 0)
      {
         if (!GetLoad(&Geom->NurbsCurves[k].LinLoadCCBeg))
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (   ReturnValue == 0
                  || !WordMatchesKeyword(E_GKW_NC_LIN_LOAD_CC_END,word)
                  || !GetLoad(&Geom->NurbsCurves[k].LinLoadCCEnd)      )
         {
            e_free(NurbsCurveSet);
            return 0;
         }

         Geom->NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_CC] = TRUE;

         ReturnValue = GetKeyword(line,word,&LineNumber,FullPathName,fp);

         if (ReturnValue == EOF)
         {
            e_FileError(FullPathName,"unexpected end of file");
            e_free(NurbsCurveSet);
            return 0;
         }
         else if (ReturnValue == 0)
         {
            e_free(NurbsCurveSet);
            return 0;
         }
      }

      if (!GetEnd(word))
      {
         e_free(NurbsCurveSet);
         return 0;
      }
   }

   if (GetKeyword(line,word,&LineNumber,FullPathName,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }

   e_free(NurbsCurveSet);
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetNEndPoints()
{
   if (sscanf(line,"%*s %d",&Geom->NEndPoints) == 1)
   {
      if (Geom->NEndPoints < 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined number of end-points");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read number of end-points");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetEndPointCoords()
{
   int k;
   int EndPointIndex;


   for (k = 0; k < Geom->NEndPoints; k++)
   {
      if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
      {
         return 0;
      }
      else if (sscanf(line,"%d %lf %lf %lf",&EndPointIndex,
                                            &Geom->EndPointCoords[k][0],
                                            &Geom->EndPointCoords[k][1],
                                            &Geom->EndPointCoords[k][2] ) != 4)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read end-point coordinates");
         return 0;
      }
      else if (EndPointIndex != k + 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "end-point coordinates index out of sequence");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetNurbsCurveIndex()
{
   int k;
   char *pFirstQuote;
   char *pLastQuote;
   char *string;
   char  newname[E_LINE_MAX + 1];

   if ((string = e_alloc_c_vect(E_LINE_MAX + 1)) == NULL)
   {
      return 0;
   }
   else if (   sscanf(line,"%*s %s",string) != 1
            || (pFirstQuote = strchr(string,'"')) == NULL
            || (pLastQuote = strrchr(string,'"')) == pFirstQuote
            || ((int)(pLastQuote - pFirstQuote) - 1) == 0)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read NURBS curve name");
      e_free(string);
      return 0;
   }

   /* assemble quote-less string */
   strcpy(newname, string+1);
   newname[strlen(string) - 2] = '\0';

   for (k = 0; k < Geom->NNurbsCurves; k++)
   {
      if(strcmp(Geom->NurbsCurves[k].Name, newname) == 0)
      {
         e_free(string);
         return k + 1;
      }
   }

   e_FileLineError(LineNumber,FullPathName,"undeclared NURBS curve");
   e_free(string);

   return 0;
}


/* ////////////////////////////////// */


static int GetNurbsCurveDegree(NURBSCURVE *NurbsCurve)
{
   if (sscanf(line,"%*s %d",&NurbsCurve->Degree) == 1)
   {
      if (NurbsCurve->Degree != 3)
      {
         e_FileLineError(LineNumber,FullPathName,"ill-defined degree");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"cannot read degree");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetNurbsCurveCntlPoints(NURBSCURVE *NurbsCurve)
{
   int k,l;
   int CntlPointIndex;


   if (   (NurbsCurve->CntlPoints
              = e_alloc_cont_d_matr
                   (NurbsCurve->Degree + 1,E_NDIMS + 1)) == NULL
       || !e_GetNextLine(line,&LineNumber,FullPathName,fp)      )
   {
      return 0;
   }
   else if (sscanf(line,"%d %d",&CntlPointIndex,&NurbsCurve->EndPoint1) != 2)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read control point data");
      return 0;
   }
   else if (CntlPointIndex != 1)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "control point index out of sequence");
      return 0;
   }
   else if (   NurbsCurve->EndPoint1 < 1
            || NurbsCurve->EndPoint1 > Geom->NEndPoints)
   {
      e_FileLineError(LineNumber,FullPathName,"ill-defined end-point");
      return 0;
   }

   l = NurbsCurve->EndPoint1 - 1;
   NurbsCurve->CntlPoints[0][0] = Geom->EndPointCoords[l][0];
   NurbsCurve->CntlPoints[0][1] = Geom->EndPointCoords[l][1];
   NurbsCurve->CntlPoints[0][2] = Geom->EndPointCoords[l][2];

   for (k = 1; k < NurbsCurve->Degree; k++)
   {
      if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
      {
         return 0;
      }
      else if (sscanf(line,"%d %lf %lf %lf",
                      &CntlPointIndex,
                      &NurbsCurve->CntlPoints[k][0],
                      &NurbsCurve->CntlPoints[k][1],
                      &NurbsCurve->CntlPoints[k][2] ) != 4)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read control point data");
         return 0;
      }
      else if (CntlPointIndex != k + 1)
      {
        e_FileLineError(LineNumber,FullPathName,
                        "control point index out of sequence");
         return 0;
      }
   }

   if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
   {
      return 0;
   }
   else if (sscanf(line,"%d %d",&CntlPointIndex,
                                &NurbsCurve->EndPoint2) != 2)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read control point data");
      return 0;
   }
   else if (CntlPointIndex != NurbsCurve->Degree + 1)
   {
      e_FileLineError(LineNumber,FullPathName,
                      "control point index out of sequence");
      return 0;
   }
   else if (   NurbsCurve->EndPoint2 < 1
            || NurbsCurve->EndPoint2 > Geom->NEndPoints)
   {
      e_FileLineError(LineNumber,FullPathName,"ill-defined end-point");
      return 0;
   }

   k = NurbsCurve->Degree;
   l = NurbsCurve->EndPoint2 - 1;
   NurbsCurve->CntlPoints[k][0] = Geom->EndPointCoords[l][0];
   NurbsCurve->CntlPoints[k][1] = Geom->EndPointCoords[l][1];
   NurbsCurve->CntlPoints[k][2] = Geom->EndPointCoords[l][2];

   return 1;
}


/* ////////////////////////////////// */


static int GetNurbsCurveWeights(NURBSCURVE *NurbsCurve)
{
   int k;
   int WeightIndex;


   for (k = 0; k < NurbsCurve->Degree + 1; k++)
   {
      if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
      {
         return 0;
      }
      else if (sscanf(line,"%d %lf",&WeightIndex,
                                    &NurbsCurve->CntlPoints[k][3]) != 2)
      {
         e_FileLineError(LineNumber,FullPathName,"cannot read weight");
         return 0;
      }
      else if (WeightIndex != k + 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "weight index out of sequence");
         return 0;
      }
      else if (NurbsCurve->CntlPoints[k][3] < 0.0)
      {
         e_FileLineError(LineNumber,FullPathName,"ill-defined weight");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetFixity(NURBSCURVE *NurbsCurve)
{
   int j;
   int FixityIndex;
   int NFieldsRead;

   double Arg;

   char string[E_LINE_MAX + 1];


   if (          (NurbsCurve->BCTypes = e_alloc_i_vect(E_DOF_MAX)) == NULL
       ||       (NurbsCurve->BCDispls = e_alloc_d_vect(E_DOF_MAX)) == NULL
       || (NurbsCurve->BCSpringConsts = e_alloc_d_vect(E_DOF_MAX)) == NULL)
   {
      return 0;
   }

   for (j = 0; j < E_DOF_MAX; j++)
   {
      if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
      {
         return 0;
      }
      else if ((NFieldsRead
                   = sscanf(line,"%d %s %lf",&FixityIndex,string,&Arg)) == 3)
      {
         if (FixityIndex != j + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "fixity index out of sequence");
            return 0;
         }
         else if (strcmp(E_BC_FIXED,string) == 0)
         {
            NurbsCurve->BCTypes[j] = E_BC_NUM_FIXED;
            NurbsCurve->BCDispls[j] = Arg;
         }
         else if (strcmp(E_BC_SPRING,string) == 0)
         {
            NurbsCurve->BCTypes[j] = E_BC_NUM_SPRING;
            NurbsCurve->BCSpringConsts[j] = Arg;
         }
         else if (strcmp(E_BC_FREE,string) == 0)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "cannot read boundary condition");
            return 0;
         }
         else
         {
            e_FileLineError(LineNumber,FullPathName,
                            "undefined boundary condition type");
            return 0;
         }
      }
      else if (NFieldsRead == 2)
      {
         if (FixityIndex != j + 1)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "fixity index out of sequence");
            return 0;
         }
         else if (strcmp(E_BC_FREE,string) == 0)
         {
            NurbsCurve->BCTypes[j] = E_BC_NUM_FREE;
         }
         else if (    strcmp(E_BC_FIXED,string) == 0
                  || strcmp(E_BC_SPRING,string) == 0)
         {
            e_FileLineError(LineNumber,FullPathName,
                            "cannot read boundary condition");
            return 0;
         }
         else
         {
            e_FileLineError(LineNumber,FullPathName,
                            "undefined boundary condition type");
            return 0;
         }
      }
      else
      {
         e_FileLineError(LineNumber,FullPathName,
                         "cannot read boundary condition");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetLoad(double **Load)
{
   int j;
   int LoadIndex;


   if ((*Load = e_alloc_d_vect(E_DOF_MAX)) == NULL)
      return 0;

   for (j = 0; j < E_DOF_MAX; j++)
   {
      if (!e_GetNextLine(line,&LineNumber,FullPathName,fp))
      {
         return 0;
      }
      else if (sscanf(line,"%d %lf",&LoadIndex,&(*Load)[j]) != 2)
      {
         e_FileLineError(LineNumber,FullPathName,"cannot read load");
         return 0;
      }
      else if (LoadIndex != j + 1)
      {
         e_FileLineError(LineNumber,FullPathName,"load index out of sequence");
         return 0;
      }
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetPointLoadParam(NURBSCURVE *NurbsCurve)
{
   if (sscanf(line,"%*s %lf",&NurbsCurve->PointLoadParam) == 1)
   {
      if (   NurbsCurve->PointLoadParam < 0.0
          || NurbsCurve->PointLoadParam > 1.0)
      {
         e_FileLineError(LineNumber,FullPathName,"ill-defined curve parameter");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"cannot read curve parameter");
      return 0;
   }

   return 1;
}


/* ////////////////////////////////// */


static int GetEnd(char *word)
{
   if (strcmp(E_GKW_NC_END,word) == 0)
   {
      return 1;
   }
   else if (IsAKeyword(word))
   {
      e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      return 0;
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,"undefined keyword");
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */


static int IsAKeyword(char *word)
{
   if (            strcmp(E_GKW_NENDPOINTS,word) == 0
       ||       strcmp(E_GKW_NNURBS_CURVES,word) == 0
       ||     strcmp(E_GKW_ENDPOINT_COORDS,word) == 0
       ||                  strcmp(E_GKW_NC,word) == 0
       ||           strcmp(E_GKW_NC_DEGREE,word) == 0
       ||      strcmp(E_GKW_NC_CNTL_POINTS,word) == 0
       ||          strcmp(E_GKW_NC_WEIGHTS,word) == 0
       ||        strcmp(E_GKW_NC_FIXITY_GC,word) == 0
       ||        strcmp(E_GKW_NC_FIXITY_CC,word) == 0
       ||       strcmp(E_GKW_NC_POINT_LOAD,word) == 0
       || strcmp(E_GKW_NC_POINT_LOAD_PARAM,word) == 0
       ||  strcmp(E_GKW_NC_LIN_LOAD_GC_BEG,word) == 0
       ||  strcmp(E_GKW_NC_LIN_LOAD_GC_END,word) == 0
       ||  strcmp(E_GKW_NC_LIN_LOAD_CC_BEG,word) == 0
       ||  strcmp(E_GKW_NC_LIN_LOAD_CC_END,word) == 0
       ||              strcmp(E_GKW_NC_END,word) == 0)
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetGeomData: To set the GEOM structure data. */


static void SetGeomData(MESH *mesh)
{
   mesh->SubStructSet[E_SS_NUM_GEOM] = TRUE;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* DeallocateAllMemory: To deallocate all memory. */


static void DeallocateAllMemory()
{
   int k;


   e_free_cont_d_matr(Geom->NEndPoints,Geom->EndPointCoords);

   for (k = 0; k < Geom->NNurbsCurves; k++)
   {
      e_free_cont_d_matr(Geom->NurbsCurves[k].Degree + 1,
                         Geom->NurbsCurves[k].CntlPoints);
      e_free(Geom->NurbsCurves[k].BCTypes);
      e_free(Geom->NurbsCurves[k].BCDispls);
      e_free(Geom->NurbsCurves[k].BCSpringConsts);
      e_free(Geom->NurbsCurves[k].PointLoad);
      e_free(Geom->NurbsCurves[k].LinLoadGCBeg);
      e_free(Geom->NurbsCurves[k].LinLoadGCEnd);
      e_free(Geom->NurbsCurves[k].LinLoadCCBeg);
      e_free(Geom->NurbsCurves[k].LinLoadCCEnd);
   }
}


/* ////////////////////////////////////////////////////////////////////////// */










/* ////////////////////////////////////////////////////////////////////////// */
/* e_PutGeomModel: To output geometric model data. */
/* The return value is 1 if the data are written successfully, else 0. */


int e_PutGeomModel(MESH *mesh, char *OutputFile)
{
   int j,k;

   FILE *fp;            /* output file pointer */


   if (!mesh->SubStructSet[E_SS_NUM_GEOM])
   {
      e_GeneralError("e_PutGeomModel: geometric model data not set");
      return 0;
   }
   else if (OpenOutputFile(OutputFile,E_GMF,&fp) == NULL)
   {
      return 0;
   }

   fprintf(fp,"%s  %d\n",E_GKW_NENDPOINTS,mesh->Geom.NEndPoints);

   fprintf(fp,"\n\n# End-point coordinates\n");
   fprintf(fp,"# point index");
   fprintf(fp,"          x-coord");
   fprintf(fp,"               y-coord");
   fprintf(fp,"               z-coord\n");

   fprintf(fp,"\n%s\n",E_GKW_ENDPOINT_COORDS);

   for (k = 0; k < mesh->Geom.NEndPoints; k++)
   {
      fprintf(fp,"%7d   ",k + 1);
      fprintf(fp,"%20.8f  ",mesh->Geom.EndPointCoords[k][0]);
      fprintf(fp,"%20.8f  ",mesh->Geom.EndPointCoords[k][1]);
      fprintf(fp,"%20.8f\n",mesh->Geom.EndPointCoords[k][2]);
   }

   for (k = 0; k < mesh->Geom.NNurbsCurves; k++)
   {
      fprintf(fp,"\n\n%s  \"%s\"\n",E_GKW_NC,mesh->Geom.NurbsCurves[k].Name);
      fprintf(fp,"%s  %d\n",E_GKW_NC_DEGREE,mesh->Geom.NurbsCurves[k].Degree);
      fprintf(fp,"%s\n",E_GKW_NC_CNTL_POINTS);
      fprintf(fp,"%3d   %d\n",1,mesh->Geom.NurbsCurves[k].EndPoint1);

      for (j = 1; j < mesh->Geom.NurbsCurves[k].Degree; j++)
      {
         fprintf(fp,"%3d   ",j + 1);
         fprintf(fp,"%20.8f  ",mesh->Geom.NurbsCurves[k].CntlPoints[j][0]);
         fprintf(fp,"%20.8f  ",mesh->Geom.NurbsCurves[k].CntlPoints[j][1]);
         fprintf(fp,"%20.8f\n",mesh->Geom.NurbsCurves[k].CntlPoints[j][2]);
      }

      fprintf(fp,"%3d   %d\n",mesh->Geom.NurbsCurves[k].Degree + 1,
                              mesh->Geom.NurbsCurves[k].EndPoint2);
      fprintf(fp,"%s\n",E_GKW_NC_WEIGHTS);

      for (j = 0; j < mesh->Geom.NurbsCurves[k].Degree + 1; j++)
      {
         fprintf(fp,"%3d   ",j + 1);
         fprintf(fp,"%20.8f\n",mesh->Geom.NurbsCurves[k].CntlPoints[j][3]);
      }

      if (mesh->Geom.NurbsCurves[k].FixityType == E_NC_NUM_FIXITY_GC)
         fprintf(fp,"%s\n",E_GKW_NC_FIXITY_GC);
      else if (mesh->Geom.NurbsCurves[k].FixityType == E_NC_NUM_FIXITY_CC)
         fprintf(fp,"%s\n",E_GKW_NC_FIXITY_CC);

      if (   mesh->Geom.NurbsCurves[k].FixityType == E_NC_NUM_FIXITY_GC
          || mesh->Geom.NurbsCurves[k].FixityType == E_NC_NUM_FIXITY_CC)
      {
         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d   ",j + 1);

            if (mesh->Geom.NurbsCurves[k].BCTypes[j] == E_BC_NUM_FREE)
            {
               fprintf(fp,"%s\n",E_BC_FREE);
            }
            else if (mesh->Geom.NurbsCurves[k].BCTypes[j] == E_BC_NUM_FIXED)
            {
               fprintf(fp,"%-8s   %20.8f\n",
                       E_BC_FIXED,mesh->Geom.NurbsCurves[k].BCDispls[j]);
            }
            else
            {
               fprintf(fp,"%-8s   ",E_BC_SPRING);

               if (PrintEng(fp,"%20.6e",
                            mesh->Geom.NurbsCurves[k].BCSpringConsts[j]) < 0)
               {
                  fclose(fp);
                  return 0;
               }

               fprintf(fp,"\n");
           }
         }
      }

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_POINT_LOAD])
      {
         fprintf(fp,"%s\n",E_GKW_NC_POINT_LOAD);

         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d  ",j + 1);

            if (PrintEng(fp,"%16.6e",
                         mesh->Geom.NurbsCurves[k].PointLoad[j]) < 0)
            {
               fclose(fp);
               return 0;
            }

            fprintf(fp,"\n");
         }

         fprintf(fp,"%s  %10.8f\n",E_GKW_NC_POINT_LOAD_PARAM,
                                   mesh->Geom.NurbsCurves[k].PointLoadParam);
      }

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_GC])
      {
         fprintf(fp,"%s\n",E_GKW_NC_LIN_LOAD_GC_BEG);

         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d  ",j + 1);

            if (PrintEng(fp,"%16.6e",
                         mesh->Geom.NurbsCurves[k].LinLoadGCBeg[j]) < 0)
            {
               fclose(fp);
               return 0;
            }

            fprintf(fp,"\n");
         }

         fprintf(fp,"%s\n",E_GKW_NC_LIN_LOAD_GC_END);

         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d  ",j + 1);

            if (PrintEng(fp,"%16.6e",
                         mesh->Geom.NurbsCurves[k].LinLoadGCEnd[j]) < 0)
            {
               fclose(fp);
               return 0;
            }

            fprintf(fp,"\n");
         }
      }

      if (mesh->Geom.NurbsCurves[k].LoadTypeSet[E_NC_NUM_LIN_LOAD_CC])
      {
         fprintf(fp,"%s\n",E_GKW_NC_LIN_LOAD_CC_BEG);

         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d  ",j + 1);

            if (PrintEng(fp,"%16.6e",
                         mesh->Geom.NurbsCurves[k].LinLoadCCBeg[j]) < 0)
            {
               fclose(fp);
               return 0;
            }

            fprintf(fp,"\n");
         }

         fprintf(fp,"%s\n",E_GKW_NC_LIN_LOAD_CC_END);

         for (j = 0; j < E_DOF_MAX; j++)
         {
            fprintf(fp,"%3d  ",j + 1);

            if (PrintEng(fp,"%16.6e",
                         mesh->Geom.NurbsCurves[k].LinLoadCCEnd[j]) < 0)
            {
               fclose(fp);
               return 0;
            }

            fprintf(fp,"\n");
         }
      }

      fprintf(fp,"%s\n",E_GKW_NC_END);
   }

   fclose(fp);
   return 1;
}

