/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           Mathematical Functions                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <float.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <e_lib.h>

#include "maths.h"


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                 e_LUFact()                                 */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To LU factorise the matrix "A". If "pivot" is not equal to NULL then
   partial pivoting is performed and the return value is 1 if "A" is
   non-singular and 0 otherwise. If "pivot" is equal to NULL then partial
   pivoting is not performed and the return value is 1 if the factorisation
   is successful and 0 if the factorised "A" has a zero diagonal element.     */

/* NOTE: "A" is over-written by "LU".
         In the above, zero means smaller in magnitude than the macro
         PIVOT_MIN defined in "maths.h".                                      */


/* ////////////////////////////////////////////////////////////////////////// */


int e_LUFact(int n, double **A, int *pivot)
{
   if (pivot == NULL)
      return Factorise(n,A);
   else
      return PivotFactorise(n,A,pivot);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PivotFactorise: To perform an LU factorisation of matrix "A" using partial
   pivoting.                                                                  */

/* The return value is 1 if "A" is non-singular, else 0. */


static int PivotFactorise(int n, double **A, int *pivot)
{
   int i,k;


   for (i = 0; i < n; i++)
      pivot[i] = i;

   for (k = 0; k < n; k++)
   {
      if (!SetPivot(k,n,A,pivot)) return 0;

      SetPivotMultFactors(k,n,A,pivot);
      SubtractPivotRow(k,n,A,pivot);
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetPivot: To determine the pivots. */

/* The return value is 1 if "A" is non-singular, else 0. */


static int SetPivot(int k, int n, double **A, int *pivot)
{
   int i;
   int MaximumPivot;
   int MaximumPivotPosition;

   double maximum;


   MaximumPivot = pivot[k];
   MaximumPivotPosition = k;
   maximum = fabs(A[pivot[k]][k]);

   for (i = k + 1; i < n; i++)
   {
      if (fabs(A[pivot[i]][k]) > maximum)
      {
         MaximumPivot = pivot[i];
         MaximumPivotPosition = i;
         maximum = fabs(A[pivot[i]][k]);
      }
   }

   if (maximum > PIVOT_MIN)
   {
      if (pivot[k] != MaximumPivot)
      {
         pivot[MaximumPivotPosition] = pivot[k];
         pivot[k] = MaximumPivot;
      }

      return 1;
   }
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetPivotMultFactors: To determine multiplication factors of k-th
   factorisation matrix for partial pivoting.                                 */


static void SetPivotMultFactors(int k, int n, double **A, int *pivot)
{
   int i;

   double DenominatorInverse;


   DenominatorInverse = 1.0/A[pivot[k]][k];

   for (i = k + 1; i < n; i++)
      A[pivot[i]][k] = DenominatorInverse*A[pivot[i]][k];
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SubtractPivotRow: To subtract multiples of k-th row from remaining rows
   for partial pivoting.                                                      */


static void SubtractPivotRow(int k, int n, double **A, int *pivot)
{
   int i,j;

   double A_ik;


   for (i = k + 1; i < n; i++)
   {
      A_ik = A[pivot[i]][k];

      for (j = k + 1; j < n; j++)
         A[pivot[i]][j] -= A_ik*A[pivot[k]][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Factorise: To perform an LU factorisation of matrix "A" with no partial
   pivoting.                                                                  */

/* The return value is 1 if "A" has no zero diagonal elements, else 0. */


static int Factorise(int n, double **A)
{
   int k;


   for (k = 0; k < n; k++)
   {
      if (fabs(A[k][k]) > PIVOT_MIN)
      {
         SetMultFactors(k,n,A);
         SubtractRow(k,n,A);
      }
      else
         return 0;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetMultFactors: To determine multiplication factors of k-th factorisation
   matrix with no partial pivoting.                                           */


static void SetMultFactors(int k, int n, double **A)
{
   int i;

   double DenominatorInverse;


   DenominatorInverse = 1.0/A[k][k];

   for (i = k + 1; i < n; i++)
      A[i][k] = DenominatorInverse*A[i][k];
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SubtractRow: To subtract multiples of k-th row from remaining rows with no
   partial pivoting.                                                          */


static void SubtractRow(int k, int n, double **A)
{
   int i,j;

   double A_ik;


   for (i = k + 1; i < n; i++)
   {
      A_ik = A[i][k];

      for (j = k + 1; j < n; j++)
         A[i][j] -= A_ik*A[k][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                                e_LUSolve()                                 */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_LUSolve: To solve the matrix equation "LUx = b" where "LU" is an
   LU-factorisation matrix. Partial pivoting is performed if "pivot" is not
   equal to NULL, otherwise partial pivoting is performed.                    */

/* The return value is 1 if the equation is solved successfully, else 0.      */

/* NOTE: The vector "b" is over-written by the solution. */


int e_LUSolve(int n, double **LU, double *b, int *pivot)
{
   if (pivot == NULL)
   {
      ForwardSubstitute(n,LU,b);
      BackSubstitute(n,LU,b);
      return 1;
   }
   else
   {
      PivotForwardSubstitute(n,LU,b,pivot);
      PivotBackSubstitute(n,LU,b,pivot);
      return ReOrder(n,b,pivot);
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PivotForwardSubstitute: To perform a forward substitution for partial
   pivoting.                                                                  */


static void PivotForwardSubstitute(int n, double **LU, double *b, int *pivot)
{
   int i,j;

   for (i = 1; i < n; i++)
   {
      for (j = 0; j < i; j++)
         b[pivot[i]] -= LU[pivot[i]][j]*b[pivot[j]];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* PivotBackSubstitute: To perform a backward substitution for partial
   pivoting.                                                                  */


static void PivotBackSubstitute(int n, double **LU, double *b, int *pivot)
{
   int i,j;


   for (i = n - 1; i >= 0; i--)
   {
      for (j = i + 1; j < n; j++)
         b[pivot[i]] -= LU[pivot[i]][j]*b[pivot[j]];

      b[pivot[i]] /= LU[pivot[i]][i];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* ReOrder: To re-order the solution vector "b". */

/* The return value is 1 if the operation is successful, else 0. */


static int ReOrder(int n, double *b, int *pivot)
{
   int k;
   
   double *TempVector;


   if ((TempVector = e_alloc_d_vect(n)) == NULL)
      return 0;

   for (k = 0; k < n; k++)
      TempVector[k] = b[pivot[k]];

   for (k = 0; k < n; k++)
      b[k] = TempVector[k];

   e_free(TempVector);
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* ForwardSubstitute: To perform a forward substitution with no partial
   pivoting.                                                                  */


static void ForwardSubstitute(int n, double **LU, double *b)
{
   int i,j;


   for (i = 1; i < n; i++)
   {
      for (j = 0; j < i; j++)
         b[i] -= LU[i][j]*b[j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* BackSubstitute: To perform a backward substitution with no partial
   pivoting.                                                                  */


static void BackSubstitute(int n, double **LU, double *b)
{
   int i,j;


   for (i = n - 1; i >= 0; i--)
   {
      for (j = i + 1; j < n; j++)
         b[i] -= LU[i][j]*b[j];

      b[i] /= LU[i][i];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                        Matrix and Vector Operations                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


/* ////////////////////////////////////////////////////////////////////////// */
/* e_VectScale: To scale a vector by a constant factor. */


void e_VectScale(unsigned int n, double Factor, double *x, double *y)
{
   register unsigned int i;


   for (i = 0; i < n; i++)
      y[i] = Factor*x[i];
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_VectDotProd: To return the dot product of two vectors. */


double e_VectDotProd(unsigned int n, double *x, double *y)
{
   register unsigned int i;
   register double DotProduct;


   DotProduct = 0.0;

   for (i = 0; i < n; i++)
      DotProduct += x[i]*y[i];

   return DotProduct;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_VectCrossVect: To determine the cross product of two 3-D vectors. */


void e_VectCrossVect(double *a, double *b, double *c)
{
   c[0] = a[1]*b[2] - a[2]*b[1];
   c[1] = a[2]*b[0] - a[0]*b[2];
   c[2] = a[0]*b[1] - a[1]*b[0];
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_MatrVectMult: To pre-multiply a vector by a matrix. */


void e_MatrVectMult(unsigned int n, double **A, double *x, double *y)
{
   register unsigned int i,j;


   for (i = 0; i < n; i++)
   {
      y[i] = 0.0;

      for (j = 0; j < n; j++)
         y[i] += A[i][j]*x[j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_MatrVectMultNS: To pre-multiply a vector by a non-square matrix. */


void e_MatrVectMultNS(unsigned int NRows, unsigned int NCols,
                             double **A, double *x, double *y)
{
   register unsigned int i,j;


   for (i = 0; i < NRows; i++)
   {
      y[i] = 0.0;

      for (j = 0; j < NCols; j++)
         y[i] += A[i][j]*x[j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_VectMatrMult: To post-multiply a vector by a matrix. */


void e_VectMatrMult(unsigned int n, double *x, double **A, double *y)
{
   register unsigned int i,j;


   memset(y,0,n*sizeof(double));

   for (i = 0; i < n; i++)
   {
      for (j = 0; j < n; j++)
         y[j] += x[i]*A[i][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_VectMatrMultNS: To post-multiply a vector by a non-square matrix. */


void e_VectMatrMultNS(unsigned int NRows, unsigned int NCols,
                             double *x, double **A, double *y)
{
   register unsigned int i,j;


   memset(y,0,NCols*sizeof(double));

   for (i = 0; i < NRows; i++)
   {
      for (j = 0; j < NCols; j++)
         y[j] += x[i]*A[i][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_MatrTransp: To transpose a matrix. */


void e_MatrTransp(unsigned int n, double **A, double **B)
{
   register unsigned int i,j;


   for (i = 0; i < n; i++)
   {
      for (j = 0; j < n; j++)
         B[j][i] = A[i][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_MatrTranspNS: To transpose a non-square matrix. */


void e_MatrTranspNS(unsigned int NRows, unsigned int NCols,
                                     double **A, double **B)
{
   register unsigned int i,j;


   for (i = 0; i < NRows; i++)
   {
      for (j = 0; j < NCols; j++)
         B[j][i] = A[i][j];
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_MatrMult: To multiply two matrices. */


void e_MatrMult(unsigned int n, double **A, double **B, double **C)
{
   register unsigned int i,j,k;


   for (i = 0; i < n; i++)
   {
      for (j = 0; j < n; j++)
      {
         C[i][j] = 0.0;

         for (k = 0; k < n; k++)
            C[i][j] += A[i][k]*B[k][j];
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_NonSqrMult: To multiply two non-square matrices. */


void e_MatrMultNS(unsigned int NRowsA, unsigned int NColsA_NRowsB,
                  unsigned int NColsB, double **A, double **B, double **C)
{
   register unsigned int i,j,k;


   for (i = 0; i < NRowsA; i++)
   {
      for (j = 0; j < NColsB; j++)
      {
         C[i][j] = 0.0;

         for (k = 0; k < NColsA_NRowsB; k++)
            C[i][j] += A[i][k]*B[k][j];
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_Factorial: To return the factorial of a number if the argument is no
   greater than 12.                                                           */


unsigned long e_Factorial(unsigned long n)
{
   if (n > 12)
   {
      e_GeneralError("e_Factorial: argument must be no greater than 12");
   }

   if (n == 0)
      return 1UL;
   else
      return n*e_Factorial(n - 1UL);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FactorialDP: To return the double precision factorial of a number. */


double e_FactorialDP(unsigned long n)
{
   if (n == 0)
      return 1.0;
   else
      return n*e_FactorialDP(n - 1UL);
}
