/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                      e_GetMeshData(), e_PutMeshData()                      */
/*                                Header File                                 */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */


static int AllCompulsoryKeywordsSet(void);
static int GetCompPropsType1Index(char *string);
static int GetPropsIndex(char *string);
static int IsAKeyword(char *word);
static int OptionalDataOkay1(void);
static int OptionalDataOkay2(void);
static int SetMeshData(MESH *mesh);
static int WordMatchesKeyword(char *Keyword, char *word, char optional);

static int GetData(void);
static int GetKeywordData(int i);
static int GetTitle(void);
static int GetNMeshPoints(void);
static int GetNNodes(void);
static int GetNElems(int *NElems, int ElemType, int NElemsKeywordNumber,
                                                  int NodesKeywordNumber);
static int GetNBCNodes(void);
static int GetNLoadedNodes(void);
static int GetNMats(void);
static int GetNCompMatsType1(void);
static int GetNNurbsCurves(void);
static int GetTimeStep(void);
static int GetNTimeSteps(void);
static int GetNInternalTimeSteps(void);
static int GetNExternalTimeSteps(void);
static int GetDampingFactor(void);
static int GetBeta(void);
static int GetGamma(void);
static int GetNStress(int *NStressPoint, int KeywordNumber);
static int GetMeshPointCoords(void);
static int GetLinkNodes(int NElems, int ***Nodes,
                        int NNodesPerElem, int VertexNodeDOF,
                        int NElemsKeywordNumber, int NodesKeywordNumber);
static int GetTriangNodes(int NElems, int ***Nodes,
                          int NNodesPerElem, int VertexNodeDOF,
                          int NElemsKeywordNumber, int NodesKeywordNumber);
static int GetNodesTriang2(void);
static int GetNodesQuad1(void);
static int GetNodesQuad3(void);
static int GetNodesTetrah1(void);
static int GetNodesBlock1(void);
static int GetBoundaryConditions(void);
static int GetLoads(void);
static int GetMaterials(int NElements, int ElementType, int KeywordNumber);
static int GetCompMatsType1(int NElements, int ElementType,
                                   int CompMatKeywordNumber);
static int GetRemeshData(void);
static int GetRemeshExterior(int k);
static int GetNurbsCurveName(char *Name);

static void DeallocateAllMemory(MESH *mesh);
static void FreeBoundaryConditions(int NBCNodes);
static void FreeLoads(int NLoadedNodes);
static void FreeRemeshData(void);
static void Initialise(void);


/* ////////////////////////////////////////////////////////////////////////// */
/* Mesh data */

static char *Title;                     /* title */

static int NMeshPoints;                 /* number of mesh-points */
static int NNodes;                      /* number of finite element nodes */
static int NElemsLink1;                 /* number of LINK1 elements */
static int NElemsLink2;                 /* number of LINK2 elements */
static int NElemsLink3;                 /* number of LINK3 elements */
static int NElemsLink4;                 /* number of LINK4 elements */
static int NElemsLink5;                 /* number of LINK5 elements */
static int NElemsTriang1;               /* number of TRIANG1 elements */
static int NElemsTriang2;               /* number of TRIANG2 elements */
static int NElemsTriang3;               /* number of TRIANG3 elements */
static int NElemsTriang4;               /* number of TRIANG4 elements */
static int NElemsTriang5;               /* number of TRIANG5 elements */
static int NElemsQuad1;                 /* number of QUAD1 elements */
static int NElemsQuad3;                 /* number of QUAD3 elements */
static int NElemsTetrah1;               /* number of TETRAH1 elements */
static int NElemsBlock1;                /* number of BLOCK1 elements  */
static int TotalNElems;                 /* total number of elements */
static int NBCNodes;                    /* number of BC nodes */
static int NLoadedNodes;                /* number of loaded nodes */
static int NMats;                       /* number of materials */
static int NCompMatsType1;              /* number of TYPE1 composite mats. */
static int NTimeSteps;                  /* number of time-steps (mut. excl.) */
static int NInternalTimeSteps;          /* number of internal time-steps */
static int NExternalTimeSteps;          /* number of external time-steps */
static int FirstNetIndex[E_NELEM_TYPES];/* first net index of each el. type */
static double TimeStep;                 /* time-step */
static double DampingFactor;            /* damping factor in the case of
                                           viscous damping                */
static double Beta;                     /* Newmark's integration constant */
static double Gamma;                    /* Newmark's integration constant */

static int NStressPointsLink1;
static int NStressPointsLink2;
static int NStressPointsLink3;
static int NStressPointsLink4;
static int NStressPointsLink5;
static int NStressPointsTriang1;
static int NStressPointsTriang2;
static int NStressPointsTriang3;
static int NStressPointsTriang4;
static int NStressPointsTriang5;
static int NStressPointsQuad1;
static int NStressPointsQuad3;
static int NStressPointsTetrah1;
static int NStressPointsBlock1;

static double **MeshPointCoords;        /* mesh-point coordinates */

static int **NodesLink1;                /* LINK1 node indices */
static int **NodesLink2;                /* LINK2 node indices */
static int **NodesLink3;                /* LINK3 node indices */
static int **NodesLink4;                /* LINK4 node indices */
static int **NodesLink5;                /* LINK5 node indices */
static int **NodesTriang1;              /* TRIANG1 node indices */
static int **NodesTriang2;              /* TRIANG2 node indices */
static int **NodesTriang3;              /* TRIANG3 node indices */
static int **NodesTriang4;              /* TRIANG4 node indices */
static int **NodesTriang5;              /* TRIANG5 node indices */
static int **NodesQuad1;                /* QUAD1 node indices  */
static int **NodesQuad3;                /* QUAD3 node indices  */
static int **NodesTetrah1;              /* TETRAH1 node indices */
static int **NodesBlock1;               /* BLOCK1 node indices for type */

static int TotalNodalDOF;               /* total nodal degrees of freedom */
static int *NodalDOF;                   /* nodal degees of freedom */

static int  *BCNodes;                   /* boundary condition nodes */
static int **BCTypes;                   /* boundary condition types */
static double **BCDispls;               /* BC node initial displacements */
static double **BCSpringConsts;         /* BC spring constants */

static int *LoadedNodes;                /* loaded nodes */
static double **Loads;                  /* loads */

static void **Mats;                     /* materials */
static MAT  **MatsLink1;                /* LINK1 materials */
static MAT  **MatsLink2;                /* LINK2 materials */
static MAT  **MatsLink3;                /* LINK3 materials */
static MAT  **MatsLink4;                /* LINK4 materials */
static MAT  **MatsLink5;                /* LINK5 materials */
static MAT  **MatsTriang1;              /* TRIANG1 materials */
static COMPMAT1 **CompMatsTriang2;      /* TRIANG2 composite materials */
static MAT  **MatsTriang3;              /* TRIANG3 materials */
static MAT  **MatsTriang4;              /* TRIANG4 materials */
static MAT  **MatsTriang5;              /* TRIANG5 materials */
static MAT  **MatsQuad1;                /* QUAD1 materials */
static COMPMAT1 **CompMatsQuad3;        /* QUAD3 composite materials */
static MAT  **MatsTetrah1;              /* TETRAH1 materials */
static MAT  **MatsBlock1;               /* BLOCK1 materials */

static int *NodeTypes;
static int *NNurbsCurves;
static double **NurbsCurveParams;
static NURBSCURVE ***NurbsCurves;

static GEOM Geom;                       /* geometric model */
static MAT *Props;                      /* material properties */
static COMPMAT1 *CompPropsType1;        /* TYPE1 composite material props. */

static int KeywordSet[E_NKEYWORDS];     /* keyword set Booleans */


/* ////////////////////////////////////////////////////////////////////////// */

static int CurrentNNurbsCurves;             /* current number of NURBS curves */
static int CurrentNMats;                    /* current number of materials */
static int CurrentNCompMatsType1;           /* current number of TYPE1
                                                composite materials     */
static int DefinedStresses;                 /* number of defined stress point
                                               numbers */
static int LineNumber;                      /* current line number */

static char Keyword[E_NKEYWORDS][E_KEYWORD_MAX + 1];
                                            /* keyword strings */
static char *FullPathName;                  /* io-file full pathname */
static char KeywordIsOptional[E_NKEYWORDS]; /* whether keywords optional */
static char line[E_LINE_MAX + 1];           /* file line */
static char ReadNextLine;                   /* whether to read next line */
static char word[E_LINE_MAX + 1];           /* potential keyword */

static FILE *fp;                            /* io-file pointer */


/* ////////////////////////////////////////////////////////////////////////// */


#define NVECT_MAX0  1000          /* (initial) maximum number of boundary
                                     condition nodes and loaded nodes     */








