#include "e_lib.h"

#define SEGREF_SMALL_MPR 1e-6
#define SEGREF_VERSION   "$Revision: 1.3 $"

static double dist(double *A, double *B);

int e_SegmentRefine(double *start_crd,
                    double start_mpr,
                    double *end_crd,
                    double end_mpr,
                    double ***seg_crd,
                    double **seg_mpr, 
                    int generate_nodes, int mode)
{
  int n;
  int nbar;
  double s;
  double sbar;
  double alfa;
  double alfabar;
  double ab_pow;                /* a power of alfabar */
  double beta;
  double mpr;
  double d;
  int i, j;

  /* checking the input */
  if ((mode != E_SEGREF_NONE) && (mode != E_SEGREF_ODD)
      && (mode != E_SEGREF_EVEN))
    e_GeneralError("e_SegmentRefine: unrecognized mode!");
  if ((start_mpr < SEGREF_SMALL_MPR) || (end_mpr < SEGREF_SMALL_MPR))
    e_GeneralError
      ("e_SegmentRefine: mesh parameters should not be negative or very small!");

  /* first the length of the segment */
  s = 0.0;
  for (i = 0; i < 3; i++)
    s += (end_crd[i] - start_crd[i]) * (end_crd[i] - start_crd[i]);
  s = sqrt(s);
  if ((start_mpr >= s) && (end_mpr >= s))
    return 0;
  if ((start_mpr < s) && (end_mpr < s))
  {
    /* first the number of nodes */
    if (E_ABS(start_mpr - end_mpr) < SEGREF_SMALL_MPR)
      n = (int) ceil(s / start_mpr) - 1;
    else
    {
      /* alfa=(s-start_mpr)/(s-end_mpr); */
      alfa = (s - start_mpr + end_mpr) / s;
      if (E_ABS(alfa - 1.0) < SEGREF_SMALL_MPR)
        n = (int) ceil(s / start_mpr) - 1;
      else
      {
        /* n=(int)ceil(log(end_mpr/start_mpr)/log(alfa)); */
        n = (int) ceil(log(end_mpr / start_mpr) / log(alfa) - 1.0);
      }
    }
    nbar = n;
    if (mode == E_SEGREF_EVEN)
    {
      if (n % 2 != 0)
        nbar = n + 1;
    }
    else if (mode == E_SEGREF_ODD)
    {
      if (n % 2 == 0)
        nbar = n + 1;
    }
    if (!generate_nodes)
      return nbar;

    /* generate the nodes */
    (*seg_mpr) = e_alloc_d_vect(nbar);
    (*seg_crd) = e_alloc_cont_d_matr(nbar, 3);

    if (E_ABS(start_mpr - end_mpr) < SEGREF_SMALL_MPR)
    {
      for (i = 0; i < nbar; i++)
      {
        for (j = 0; j < 3; j++)
          (*seg_crd)[i][j] =
            start_crd[j] + (end_crd[j] - start_crd[j]) / (nbar + 1) * (i + 1);
        (*seg_mpr)[i] = start_mpr;
      }
      return nbar;
    }

  restart:

    alfabar = pow(end_mpr / start_mpr, 1.0 / (double)(nbar+1));
    if (E_ABS(alfabar - 1.0) < SEGREF_SMALL_MPR)
      sbar = (nbar + 1) * start_mpr;
    else
      sbar =
        (start_mpr * (1.0 - pow(alfabar, (double) (nbar + 1)))) / (1.0 -
                                                                   alfabar);
    beta = s / sbar;

    if (start_mpr < end_mpr)
    {
      (*seg_mpr)[0] = start_mpr + (end_mpr - start_mpr) *
        beta * start_mpr / s;
      for (j = 0; j < 3; j++)
        (*seg_crd)[0][j] = start_crd[j] +
          beta * start_mpr / s * (end_crd[j] - start_crd[j]);
      ab_pow = alfabar;
      for (i = 1; i < nbar; i++)
      {
        (*seg_mpr)[i] = (*seg_mpr)[i - 1] + (end_mpr - start_mpr) *
          beta * start_mpr * ab_pow / s;
        for (j = 0; j < 3; j++)
          (*seg_crd)[i][j] = (*seg_crd)[i - 1][j] +
            beta * start_mpr * ab_pow * ((end_crd[j] - start_crd[j]) / s);
        ab_pow *= alfabar;
      }
    }
    else
    {
      (*seg_mpr)[nbar - 1] = end_mpr + (start_mpr - end_mpr) *
        beta * end_mpr / s;
      for (j = 0; j < 3; j++)
        (*seg_crd)[nbar - 1][j] = end_crd[j] +
          beta * end_mpr / s * (start_crd[j] - end_crd[j]);
      ab_pow = 1.0 / alfabar;
      for (i = nbar - 2; i > -1; i--)
      {
        (*seg_mpr)[i] = (*seg_mpr)[i + 1] + (start_mpr - end_mpr) *
          beta * end_mpr / s * ab_pow;
        for (j = 0; j < 3; j++)
          (*seg_crd)[i][j] = (*seg_crd)[i + 1][j] +
            beta * end_mpr * ab_pow * ((start_crd[j] - end_crd[j]) / s);
        ab_pow /= alfabar;
      }
    }

    for (i = 0; i < nbar + 1; i++)
    {
      d =
        dist(i == 0 ? start_crd : (*seg_crd)[i - 1],
             i == nbar ? end_crd : (*seg_crd)[i]);
      if ((d > (i == 0 ? start_mpr : (*seg_mpr)[i - 1]))
          || (d > (i == nbar ? end_mpr : (*seg_mpr)[i])))
        break;
    }
    if (i < nbar + 1)
    {
      if ((mode == E_SEGREF_EVEN) || (mode == E_SEGREF_ODD))
      {
        (*seg_mpr) = e_ext_d_vect(nbar, (*seg_mpr), 2);
        (*seg_crd) = e_ext_cont_d_matr(nbar, 3, (*seg_crd), 2);
        nbar += 2;
        goto restart;
      }
      else
      {
        (*seg_mpr) = e_ext_d_vect(nbar, (*seg_mpr), 1);
        (*seg_crd) = e_ext_cont_d_matr(nbar, 3, (*seg_crd), 1);
        nbar += 1;
        goto restart;
      }
    }
  }
  else                          /* one of the mesh params is >= length */
  {
    sbar = 0.0;                 /* now used as running length */
    n = 0;
    mpr = E_MIN(start_mpr, end_mpr);
    while (sbar < s)
    {
      sbar += mpr;
      n++;
      mpr = E_MIN(start_mpr, end_mpr) + E_ABS(end_mpr - start_mpr) * sbar / s;
    }
    n--;
    nbar = n;
    if (mode == E_SEGREF_EVEN)
    {
      if (n % 2 != 0)
        nbar = n + 1;
    }
    else if (mode == E_SEGREF_ODD)
    {
      if (n % 2 == 0)
        nbar = n + 1;
    }
    if (!generate_nodes)
      return nbar;
    if (nbar > n)
    {
      mpr = E_MIN(start_mpr, end_mpr) + E_ABS(end_mpr - start_mpr) * sbar / s;
      sbar += mpr;
    }
    beta = s / sbar;

    (*seg_mpr) = e_alloc_d_vect(nbar);
    (*seg_crd) = e_alloc_cont_d_matr(nbar, 3);

    if (start_mpr < end_mpr)
    {
      sbar = 0.0;
      mpr = start_mpr;
      sbar += mpr;
      for (j = 0; j < 3; j++)
        (*seg_crd)[0][j] = start_crd[j] +
          beta * (end_crd[j] - start_crd[j]) / s * mpr;
      (*seg_mpr)[0] = start_mpr + (end_mpr - start_mpr) / s * beta * mpr;
      for (i = 1; i < nbar; i++)
      {
        mpr = start_mpr + (end_mpr - start_mpr) * sbar / s;
        sbar += mpr;
        for (j = 0; j < 3; j++)
          (*seg_crd)[i][j] = (*seg_crd)[i - 1][j] +
            beta * (end_crd[j] - start_crd[j]) / s * mpr;
        (*seg_mpr)[i] = start_mpr + (end_mpr - start_mpr) * beta * sbar / s;
      }
    }
    else
    {
      sbar = 0.0;
      mpr = end_mpr;
      sbar += mpr;
      for (j = 0; j < 3; j++)
        (*seg_crd)[nbar - 1][j] = end_crd[j] +
          beta * (start_crd[j] - end_crd[j]) / s * mpr;
      (*seg_mpr)[nbar - 1] = end_mpr + (start_mpr - end_mpr) / s * beta * mpr;
      for (i = nbar - 2; i > -1; i--)
      {
        mpr = end_mpr + (start_mpr - end_mpr) * sbar / s;
        sbar += mpr;
        for (j = 0; j < 3; j++)
          (*seg_crd)[i][j] = (*seg_crd)[i + 1][j] +
            beta * (start_crd[j] - end_crd[j]) / s * mpr;
        (*seg_mpr)[i] = end_mpr + (start_mpr - end_mpr) * beta * sbar / s;
      }
    }
  }

  return nbar;
}

static double dist(double *A, double *B)
{
  /* returns the positive distance between two points */

  double dx2, dy2, dz2;

  dx2 = (A[0] - B[0]);
  dx2 = dx2 * dx2;
  dy2 = (A[1] - B[1]);
  dy2 = dy2 * dy2;
  dz2 = (A[2] - B[2]);
  dz2 = dz2 * dz2;

  return sqrt(dx2 + dy2 + dz2);
}
