/********************************************************************
 circumcenter.c  --
 
 Jelle Muylle <jelle@sect.mce.hw.ac.uk>
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    

 CREDITS: The code in this file was posted by J.R. Schewchuk:
   To:             compgeom-discuss@research.bell-labs.com
   Subject:        Re: circumsphere
   Date:           Wed, 1 Apr 98 0:34:28 EST
   From:           Jonathan R Shewchuk <jrs+@cs.cmu.edu>

 SEE ALSO: see end of the file for the dissussion about circumcentre.

********************************************************************/

/* $Id: circumcenter.c,v 1.3 2002/10/16 19:05:07 peteri Exp $ */

#include <stdlib.h>
#include "eg_lib.h"

#define EXACT 1 /* use Shewchuk's predicates */

/* Shewchuk's predicates: see predicates.c for details */
extern void exactinit();
extern double orient2d(double *pa, double *pb, double *pc);
extern double orient3d(double *pa, double *pb, double *pc, double *pd);
extern double incircle(double *pa, double *pb, double *pc, double *pd);
extern double insphere(double *pa, double *pb, double *pc, double *pd,
                       double *pe);

/****f*
 * NAME
 *   eg_TetCircumcenter
 * DESCRIPTION 
 *                                                                           
 *   Find the circumcenter of a tetrahedron.              
 *                                                                           
 *   The result is returned both in terms of xyz coordinates and xi-eta-zeta  
 *   coordinates, relative to the tetrahedron's point `a' (that is, `a' is    
 *   the origin of both coordinate systems).  Hence, the xyz coordinates      
 *   returned are NOT absolute; one must add the coordinates of `a' to        
 *   find the absolute coordinates of the circumcircle.  However, this means  
 *   that the result is frequently more accurate than would be possible if    
 *   absolute coordinates were returned, due to limited floating-point        
 *   precision.  In general, the circumradius can be computed much more       
 *   accurately.                                                              
 *                                                                           
 *   The xi-eta-zeta coordinate system is defined in terms of the             
 *   tetrahedron.  Point `a' is the origin of the coordinate system.          
 *   The edge `ab' extends one unit along the xi axis.  The edge `ac'         
 *   extends one unit along the eta axis.  The edge `ad' extends one unit     
 *   along the zeta axis.  These coordinate values are useful for linear      
 *   interpolation.                                                           
 *                                                                           
 *   If `xi' is NULL on input, the xi-eta-zeta coordinates will not be        
 *   computed.                                                                
 *                                                                           
 * ARGUMENTS:
 *    a, b, c, d - tetrahedron coordinates
 *    circumecenter - tetrahedron circumcenters (relative)
 *    xi, eta, zeta - circumcenter coordinates in xi-eta-zeta system
 * RETURN VALUE
 *    None 
 ******/
void eg_TetCircumcenter (double a[3], double b[3], double c[3], double d[3], 
                double circumcenter[3], double *xi, double *eta, double *zeta)
{
  double xba, yba, zba, xca, yca, zca, xda, yda, zda;
  double balength, calength, dalength;
  double xcrosscd, ycrosscd, zcrosscd;
  double xcrossdb, ycrossdb, zcrossdb;
  double xcrossbc, ycrossbc, zcrossbc;
  double denominator;
  double xcirca, ycirca, zcirca;

  /* Use coordinates relative to point `a' of the tetrahedron. */
  xba = b[0] - a[0];
  yba = b[1] - a[1];
  zba = b[2] - a[2];
  xca = c[0] - a[0];
  yca = c[1] - a[1];
  zca = c[2] - a[2];
  xda = d[0] - a[0];
  yda = d[1] - a[1];
  zda = d[2] - a[2];
  /* Squares of lengths of the edges incident to `a'. */
  balength = xba * xba + yba * yba + zba * zba;
  calength = xca * xca + yca * yca + zca * zca;
  dalength = xda * xda + yda * yda + zda * zda;
  /* Cross products of these edges. */
  xcrosscd = yca * zda - yda * zca;
  ycrosscd = zca * xda - zda * xca;
  zcrosscd = xca * yda - xda * yca;
  xcrossdb = yda * zba - yba * zda;
  ycrossdb = zda * xba - zba * xda;
  zcrossdb = xda * yba - xba * yda;
  xcrossbc = yba * zca - yca * zba;
  ycrossbc = zba * xca - zca * xba;
  zcrossbc = xba * yca - xca * yba;

  /* Calculate the denominator of the formulae. */
#ifdef EXACT
  /* Use orient3d() from http://www.cs.cmu.edu/~quake/robust.html     */
  /*   to ensure a correctly signed (and reasonably accurate) result, */
  /*   avoiding any possibility of division by zero.                  */
  denominator = 0.5 / orient3d(b, c, d, a);
#else
  /* Take your chances with floating-point roundoff. */
  denominator = 0.5 / (xba * xcrosscd + yba * ycrosscd + zba * zcrosscd);
#endif

  /* Calculate offset (from `a') of circumcenter. */
  xcirca = (balength * xcrosscd + calength * xcrossdb + dalength * xcrossbc) *
           denominator;
  ycirca = (balength * ycrosscd + calength * ycrossdb + dalength * ycrossbc) *
           denominator;
  zcirca = (balength * zcrosscd + calength * zcrossdb + dalength * zcrossbc) *
           denominator;
  circumcenter[0] = xcirca;
  circumcenter[1] = ycirca;
  circumcenter[2] = zcirca;

  if (xi != (double *) NULL) {
    /* To interpolate a linear function at the circumcenter, define a    */
    /*   coordinate system with a xi-axis directed from `a' to `b',      */
    /*   an eta-axis directed from `a' to `c', and a zeta-axis directed  */
    /*   from `a' to `d'.  The values for xi, eta, and zeta are computed */
    /*   by Cramer's Rule for solving systems of linear equations.       */
    *xi = (xcirca * xcrosscd + ycirca * ycrosscd + zcirca * zcrosscd) *
          (2.0 * denominator);
    *eta = (xcirca * xcrossdb + ycirca * ycrossdb + zcirca * zcrossdb) *
           (2.0 * denominator);
    *zeta = (xcirca * xcrossbc + ycirca * ycrossbc + zcirca * zcrossbc) *
            (2.0 * denominator);
  }
}

/****f*
 * NAME
 *  eg_TriCircumcenter2D
 * DESCRIPTION
 *  Find the circumcenter of a triangle.                 
 *                                                                           
 *  The result is returned both in terms of x-y coordinates and xi-eta       
 *  coordinates, relative to the triangle's point `a' (that is, `a' is       
 *  the origin of both coordinate systems).  Hence, the x-y coordinates      
 *  returned are NOT absolute; one must add the coordinates of `a' to        
 *  find the absolute coordinates of the circumcircle.  However, this means  
 *  that the result is frequently more accurate than would be possible if    
 *  absolute coordinates were returned, due to limited floating-point        
 *  precision.  In general, the circumradius can be computed much more       
 *  accurately.                                                              
 *                                                                           
 *  The xi-eta coordinate system is defined in terms of the triangle.        
 *  Point `a' is the origin of the coordinate system.  The edge `ab' extends 
 *  one unit along the xi axis.  The edge `ac' extends one unit along the    
 *  eta axis.  These coordinate values are useful for linear interpolation.  
 *                                                                           
 *  If `xi' is NULL on input, the xi-eta coordinates will not be computed.   
 *                                                                           
 *  ARGUMENTS
 *    a, b, c - two dimiensional points of triangle (IN)
 *    circumcenter - circumcenter of the triangle (OUT)
 *    xi, eta - circumcenter in xi-eta coordiante system (IN/OUT)
 *  RETURN VALUE
 *    NONE
 ******/
void eg_TriCircumcenter2D (double a[2], double b[2], double c[2], 
                           double circumcenter[2], double *xi, double *eta)
{
  double xba, yba, xca, yca;
  double balength, calength;
  double denominator;
  double xcirca, ycirca;

  /* Use coordinates relative to point `a' of the triangle. */
  xba = b[0] - a[0];
  yba = b[1] - a[1];
  xca = c[0] - a[0];
  yca = c[1] - a[1];
  /* Squares of lengths of the edges incident to `a'. */
  balength = xba * xba + yba * yba;
  calength = xca * xca + yca * yca;

  /* Calculate the denominator of the formulae. */
#ifdef EXACT
  /* Use orient2d() from http://www.cs.cmu.edu/~quake/robust.html     */
  /*   to ensure a correctly signed (and reasonably accurate) result, */
  /*   avoiding any possibility of division by zero.                  */
  denominator = 0.5 / orient2d(b, c, a);
#else
  /* Take your chances with floating-point roundoff. */
  denominator = 0.5 / (xba * yca - yba * xca);
#endif

  /* Calculate offset (from `a') of circumcenter. */
  xcirca = (yca * balength - yba * calength) * denominator;  
  ycirca = (xba * calength - xca * balength) * denominator;  
  circumcenter[0] = xcirca;
  circumcenter[1] = ycirca;

  if (xi != (double *) NULL) {
    /* To interpolate a linear function at the circumcenter, define a     */
    /*   coordinate system with a xi-axis directed from `a' to `b' and    */
    /*   an eta-axis directed from `a' to `c'.  The values for xi and eta */
    /*   are computed by Cramer's Rule for solving systems of linear      */
    /*   equations.                                                       */
    *xi = (xcirca * yca - ycirca * xca) * (2.0 * denominator);
    *eta = (ycirca * xba - xcirca * yba) * (2.0 * denominator);
  }
}

/****f*
 * NAME                                                                           
 *  eg_TriCircumcenter3D
 *
 * DESCRIPTION
 *  Find the circumcenter of a triangle in 3D.         
 *                                                                           
 *  The result is returned both in terms of xyz coordinates and xi-eta       
 *  coordinates, relative to the triangle's point `a' (that is, `a' is       
 *  the origin of both coordinate systems).  Hence, the xyz coordinates      
 *  returned are NOT absolute; one must add the coordinates of `a' to        
 *  find the absolute coordinates of the circumcircle.  However, this means  
 *  that the result is frequently more accurate than would be possible if    
 *  absolute coordinates were returned, due to limited floating-point        
 *  precision.  In general, the circumradius can be computed much more       
 *  accurately.                                                              
 *                                                                           
 *  The xi-eta coordinate system is defined in terms of the triangle.        
 *  Point `a' is the origin of the coordinate system.  The edge `ab' extends 
 *  one unit along the xi axis.  The edge `ac' extends one unit along the    
 *  eta axis.  These coordinate values are useful for linear interpolation.  
 *                                                                           
 *  If `xi' is NULL on input, the xi-eta coordinates will not be computed.   
 *                                                                           
 *  ARGUMENTS
 *    a, b, c - three dimiensional points of triangle (IN)
 *    circumcenter - circumcenter of the triangle (OUT)
 *    xi, eta - circumcenter in xi-eta coordiante system (IN/OUT)
 *  RETURN VALUE
 *    NONE
 *******/
void eg_TriCircumcenter3D (double a[3], double b[3], double c[3], 
                        double circumcenter[3], double *xi, double *eta)
{
  double xba, yba, zba, xca, yca, zca;
  double balength, calength;
  double xcrossbc, ycrossbc, zcrossbc;
  double denominator;
  double xcirca, ycirca, zcirca;
  double p1[2], p2[2], p3[2];

  /* Use coordinates relative to point `a' of the triangle. */
  xba = b[0] - a[0];
  yba = b[1] - a[1];
  zba = b[2] - a[2];
  xca = c[0] - a[0];
  yca = c[1] - a[1];
  zca = c[2] - a[2];
  /* Squares of lengths of the edges incident to `a'. */
  balength = xba * xba + yba * yba + zba * zba;
  calength = xca * xca + yca * yca + zca * zca;
  
  /* Cross product of these edges. */
#ifdef EXACT
  /* Use orient2d() from http://www.cs.cmu.edu/~quake/robust.html     */
  /*   to ensure a correctly signed (and reasonably accurate) result, */
  /*   avoiding any possibility of division by zero.                  */
  p1[0] = b[1]; p1[1] = b[2];
  p2[0] = c[1]; p2[1] = c[2];
  p3[0] = a[1]; p3[1] = a[2];
  /*xcrossbc = orient2d(b[1], b[2], c[1], c[2], a[1], a[2]); mod Romek*/
  xcrossbc = orient2d(p1, p2, p3);
  
  p1[0] = b[2]; p1[1] = b[0];
  p2[0] = c[2]; p2[1] = c[0];
  p3[0] = a[2]; p3[1] = a[0];
  /*ycrossbc = orient2d(b[2], b[0], c[2], c[0], a[2], a[0]); mod Romek */
  ycrossbc = orient2d(p1, p2, p3);
  
  p1[0] = b[0]; p1[1] = b[1];
  p2[0] = c[0]; p2[1] = c[1];
  p3[0] = a[0]; p3[1] = a[1];
  /* zcrossbc = orient2d(b[0], b[1], c[0], c[1], a[0], a[1]); mod Romek*/
  zcrossbc = orient2d(p1, p2, p3);
#else
  /* Take your chances with floating-point roundoff. */
  xcrossbc = yba * zca - yca * zba;
  ycrossbc = zba * xca - zca * xba;
  zcrossbc = xba * yca - xca * yba;
#endif

  /* Calculate the denominator of the formulae. */
  denominator = 0.5 / (xcrossbc * xcrossbc + ycrossbc * ycrossbc +
                       zcrossbc * zcrossbc);

  /* Calculate offset (from `a') of circumcenter. */
  xcirca = ((balength * yca - calength * yba) * zcrossbc -
            (balength * zca - calength * zba) * ycrossbc) * denominator;
  ycirca = ((balength * zca - calength * zba) * xcrossbc -
            (balength * xca - calength * xba) * zcrossbc) * denominator;
  zcirca = ((balength * xca - calength * xba) * ycrossbc -
            (balength * yca - calength * yba) * xcrossbc) * denominator;
  circumcenter[0] = xcirca;
  circumcenter[1] = ycirca;
  circumcenter[2] = zcirca;

  if (xi != (double *) NULL) {
    /* To interpolate a linear function at the circumcenter, define a     */
    /*   coordinate system with a xi-axis directed from `a' to `b' and    */
    /*   an eta-axis directed from `a' to `c'.  The values for xi and eta */
    /*   are computed by Cramer's Rule for solving systems of linear      */
    /*   equations.                                                       */

    /* There are three ways to do this calculation - using xcrossbc, */
    /*   ycrossbc, or zcrossbc.  Choose whichever has the largest    */
    /*   magnitude, to improve stability and avoid division by zero. */
    if (((xcrossbc >= ycrossbc) ^ (-xcrossbc > ycrossbc)) &&
        ((xcrossbc >= zcrossbc) ^ (-xcrossbc > zcrossbc))) {
      *xi = (ycirca * zca - zcirca * yca) / xcrossbc;
      *eta = (zcirca * yba - ycirca * zba) / xcrossbc;
    } else if ((ycrossbc >= zcrossbc) ^ (-ycrossbc > zcrossbc)) {
      *xi = (zcirca * xca - xcirca * zca) / ycrossbc;
      *eta = (xcirca * zba - zcirca * xba) / ycrossbc;
    } else {
      *xi = (xcirca * yca - ycirca * xca) / zcrossbc;
      *eta = (ycirca * xba - xcirca * yba) / zcrossbc;
    }
  }
}
