/********************************************************************
 eg_lib.c  --
 
 Jelle Muylle <jelle@sect.mce.hw.ac.uk>
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: eg_lib.c,v 1.13 2002/10/16 19:05:09 peteri Exp $ */

#define VERSION "$Revision: 1.13 $"

#include <stdio.h>
#include <math.h>

#include "eg_lib.h"

#define EG_MAX_TOL  3           /* number of defined tolerances */

/* indices of tolerances in 'tolerances' array */
#define EG_ZERO_DIST    0       /* a distance assumed to be zero */
#define EG_ZERO_AREA    1       /* a relative area difference assumed to be zero */
#define EG_PAR_LINE     2       /* tolerance indicating two lines are parallel */

/* identification string */
static char rcsid[] = "$Id: eg_lib.c,v 1.13 2002/10/16 19:05:09 peteri Exp $";

/****d*
* NAME
*   Vector operations
* DESCRIPTION
*   Two macros are defined for the cross and dot product of
*   two vector of length three.
* SYNOPSIS
*/
/**/ #define CROSS3D(result, v1, v2) \
/**/     (result)[0] = (v1)[1]*(v2)[2] - (v1)[2]*(v2)[1]; \
/**/     (result)[0] = (v1)[2]*(v2)[0] - (v1)[0]*(v2)[2]; \
/**/     (result)[0] = (v1)[0]*(v2)[1] - (v1)[1]*(v2)[0];
/**/ #define DOT3D(v1, v2) \
/**/     ((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2])
/*
******/

/* global variables */
static double tolerances[EG_MAX_TOL];

/* Shewchuk's predicates: see predicates.c for details */
extern void exactinit();
extern double orient2d(double *pa, double *pb, double *pc);
extern double orient3d(double *pa, double *pb, double *pc, double *pd);
extern double incircle(double *pa, double *pb, double *pc, double *pd);
extern double insphere(double *pa, double *pb, double *pc, double *pd,
                       double *pe);

/* functions referenced in this file only */
static int project5(double *axis, double *t1, double *t2, double *t3,
                    double *s1, double *s2);

static int project6(double *axis, double *t1, double *t2, double *t3,
                    double *z1, double *z2, double *z3);

/* initialized */
static int eg_initialized=FALSE;

/****f*
* NAME
*   eg_Init
* DESCRIPTION
*   Initializes the eg_lib. In particular, it initializes Shewchuk's exact arithmetic
*   and sets default values for the tolerances.
* ARGUMENTS
*   none
* RETURN VALUE
*   none
* NOTES       
*   It is safe to call this function more than once.
*   Initialization is done only once
******/
void eg_Init()
{
  char *dummy;
  if (eg_initialized == FALSE)
  {
    eg_initialized = TRUE;
    exactinit();
    eg_SetTolerance(EG_ZERO_DIST_TOL, 1.0e-8);
    eg_SetTolerance(EG_ZERO_AREA_TOL, 1.0e-8);
    eg_SetTolerance(EG_PAR_LINE_TOL, 0.00001);
  
    /* this is to prevent the compiler to optimize the rcsid
       string away
     */
     dummy = rcsid;
  }
}                               /* end of eg_Init */

/****f*
* NAME
*   eg_Version
* DESCRIPTION
*   Prints version information.
* ARGUMENTS
*   none
* RETURN VALUE
*   none
******/
void eg_Version()
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  printf("\neg_lib version %s (%s %s)\n", VERSION, __TIME__, __DATE__);
}                               /* end of eg_Version */

/****f*
* NAME
*   eg_SetTolerance
* DESCRIPTION
*   Sets a particular tolerance to a specified value.
* ARGUMENTS
*   flag - the tolerance index as defined by the macros
*   tol - the new value for the tolerance
* RETURN VALUE
*   none
* SEE ALSO
*   Tolerance indices
******/
void eg_SetTolerance(int flag, double tol)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  if (flag == EG_ZERO_DIST_TOL)
    tolerances[EG_ZERO_DIST] = tol;
  else if (flag == EG_ZERO_AREA_TOL)
    tolerances[EG_ZERO_AREA] = tol;
  else if (flag == EG_PAR_LINE_TOL)
    tolerances[EG_PAR_LINE] = tol;
  else
    eg_GeneralError("eg_SetTolerance: invalid tolerance flag !");

}                               /* end of eg_SetTolerance */

/****f*
* NAME
*   eg_GetTolerance
* DESCRIPTION
*   Obtain the current value of a particular tolerance.
* ARGUMENTS
* 	flag - the tolerance index as defined by the macros
* RETURN VALUE
*   the current value of the chosen tolerance
* SEE ALSO
*   Tolerance indices
******/
double eg_GetTolerance(int flag)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  if (flag == EG_ZERO_DIST_TOL)
    return (tolerances[EG_ZERO_DIST]);
  else if (flag == EG_ZERO_AREA_TOL)
    return (tolerances[EG_ZERO_AREA]);
  else if (flag == EG_PAR_LINE_TOL)
    return (tolerances[EG_PAR_LINE]);
  else
  {
    eg_GeneralError("eg_GetTolerance: invalid tolerance flag !");
  }
  /* make everybody happy */
  return(0.0);
}                               /* end of eg_GetTolerance */

/****f*
* NAME
*   eg_PointDist3D
* DESCRIPTION
*   Determines the distance between two points in 3D.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
* RETURN VALUE
*   the Euclidean distance between the two points
******/
double eg_PointDist3D(double *pa, double *pb)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return sqrt(eg_PointSqrDist3D(pa,pb));
}                               /* end of eg_PointDist3D */

/****f*
* NAME
*   eg_PointSqrDist3D
* DESCRIPTION
*   Calculates the squared distance between two points in 3D.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
* RETURN VALUE
*   the square of the Euclidean distance between the two points
******/
double eg_PointSqrDist3D(double *pa, double *pb)
{
  register double d;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  d = (pb[0] - pa[0]) * (pb[0] - pa[0]);
  d += (pb[1] - pa[1]) * (pb[1] - pa[1]);
  d += (pb[2] - pa[2]) * (pb[2] - pa[2]);

  return (d);
}                               /* end of eg_PointSqrDist3D */

/****f*
* NAME
*   eg_PointDist2D
* DESCRIPTION
*   Determines the distance between two points in 2D.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
* RETURN VALUE
*   the Euclidean distance between the two points in 2D
******/
double eg_PointDist2D(double *pa, double *pb)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return sqrt(eg_PointSqrDist2D(pa,pb));
}                               /* end of eg_PointDist2D */

/****f*
* NAME
*   eg_PointSqrDist2D
* DESCRIPTION
*   Calculates the squared distance between two points in 2D.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
* RETURN VALUE
*   the square of the Euclidean distance between the two points in 2D
******/
double eg_PointSqrDist2D(double *pa, double *pb)
{
  register double d;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  d = (pb[0] - pa[0]) * (pb[0] - pa[0]);
  d += (pb[1] - pa[1]) * (pb[1] - pa[1]);

  return (d);
}                               /* end of eg_PointSqrDist2D */

/****f*
 * NAME
 *   eg_PointProjInsideSegment
 * DESCRIPTION
 *   Determines whether the projection of a point to the line of the
 *   segment is between the segment points or outside.
 * ARGUMENTS
 *   sb - the coordinates of the first point of the segment
 *   se - the coordinates of the second point of the segment
 *   p  - the coordinates of the point
 ******/
int eg_PointProjInsideSegment(double *sb, double *se, double *p)
{
  double v1[3], v2[3], seg[3];
  double d1, d2;

  seg[0] = sb[0] - se[0];
  seg[1] = sb[1] - se[1];
  seg[2] = sb[2] - se[2];

  v1[0] = sb[0] - p[0];
  v1[1] = sb[1] - p[1];
  v1[2] = sb[2] - p[2];
  
  v2[0] = se[0] - p[0];
  v2[1] = se[1] - p[1];
  v2[2] = se[2] - p[2];
  
  d1 = DOT3D(seg, v1);
  d2 = DOT3D(seg, v2);

  if(   (d1 > 0 && d2 < 0)
     || (d1 < 0 && d2 > 0) )
  {
    return 1;
  }
  else
  {
    return 0;
  }
} /* end of eg_PointProjInsideSegment */


/****f*
* NAME
*   eg_PointLineDist2D
* DESCRIPTION
*  Determines the distance in 2D between a line segment(\textbf{lb},\textbf{le}) and a 
*  point (\textbf{p}). As a result the \textbf{mu} parameter of the closest point to \textbf{p} 
*  (\textit{cp}) on the segment is also given so that 
*
*  \textit{cp}=\textbf{lb}+\textbf{mu}$\cdot$(\textbf{le}-\textbf{lb})
* ARGUMENTS
*   lb - the coordinates of the first point of the line
*   le - the coordinates of the second point of the line
*   p - the coordinates of the point
*   mu - a return parameter as described above
* RETURN VALUE
*   the distance as described above, except when \textbf{lb} coincides
*   with \textbf{le}, then -1.0 is returned.
******/
double eg_PointLineDist2D(double *lb, double *le, double *p, double *mu)
{
  register double nom;
  register double denom;
  double cp[2];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  denom = (le[0] - lb[0]) * (le[0] - lb[0]);
  denom += (le[1] - lb[1]) * (le[1] - lb[1]);

  if (denom < tolerances[EG_ZERO_DIST])
  {
    return -1.0;
  }

  nom = (p[0] - lb[0]) * (le[0] - lb[0]) + (p[1] - lb[1]) * (le[1] - lb[1]);

  *mu = nom / denom;

  if (fabs(eg_Orient2D(lb, le, p)) < tolerances[EG_ZERO_DIST] * denom)
  {
    return 0.0;
  }
  else
  {
    cp[0] = lb[0] + (*mu) * (le[0] - lb[0]);
    cp[1] = lb[1] + (*mu) * (le[1] - lb[1]);

    return eg_PointDist2D(cp, p);
  }
}                               /* end of eg_PointLineDist2D */

/****f*
* NAME
*   eg_PointLineDist3D
* DESCRIPTION
*  Determines the distance in 3D between a line segment(\textbf{lb},\textbf{le}) and a 
*  point (\textbf{p}). As a result the \textbf{mu} parameter of the closest point to \textbf{p} 
*  (\textit{cp}) on the segment is also given so that 
*
*  \textit{cp}=\textbf{lb}+\textbf{mu}$\cdot$(\textbf{le}-\textbf{lb})
* ARGUMENTS
*   lb - the coordinates of the first point of the line
*   le - the coordinates of the second point of the line
*   p - the coordinates of the point
*   mu - a return parameter as described above
* RETURN VALUE
*   the distance as described above, except when \textbf{lb} coincides
*   with \textbf{le}, then -1.0 is returned.
******/
double eg_PointLineDist3D(double *lb, double *le, double *p, double *mu)
{
  register double nom = 0.0;
  register double denom = 0.0;
  double cp[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  denom = (le[0] - lb[0]) * (le[0] - lb[0]);
  denom += (le[1] - lb[1]) * (le[1] - lb[1]);
  denom += (le[2] - lb[2]) * (le[2] - lb[2]);

  if (denom < tolerances[EG_ZERO_DIST])
  {
    return -1.0;
  }

  nom = (p[0] - lb[0]) * (le[0] - lb[0]);
  nom += (p[1] - lb[1]) * (le[1] - lb[1]);
  nom += (p[2] - lb[2]) * (le[2] - lb[2]);

  *mu = nom / denom;

  cp[0] = lb[0] + (*mu) * (le[0] - lb[0]);
  cp[1] = lb[1] + (*mu) * (le[1] - lb[1]);
  cp[2] = lb[2] + (*mu) * (le[2] - lb[2]);

  return eg_PointDist3D(cp, p);
}                               /* end of eg_PointLineDist3D */

/****f*
* NAME
*   eg_LineLineDist3D
* DESCRIPTION
*   Returns the perpendicular distance between two lines.
* ARGUMENTS
*   lb1 - coordinates of first point of first line
*   le1 - coordinates of second point of first line
*   lb2 - coordinates of first point of second line
*   le2 - coordinates of second point of second line
*   mu1 - return value (see notes)
*   mu2 - return value (see notes)
*   lli_res - return value (see notes)
* RETURN VALUE
*   the distance as described above. 
* NOTES
*   The distance is always positive except in case when EG_INVALID
*   would be returned by a call to eg_LineLineIntersect3D for these
*   two line segments.
*   Such a call is also executed and \textbf{mu1} and \textbf{mu2} 
*   are returned as in eg_LineLineIntersect3D.
*   except when lines are parallel, which can be detected
*   as the result of the eg_LineLineIntersect3D is put into
*   \textbf{lli_res}.
*   In case of parallel, the mu's reflect the projection of
*   line 2 on line 1 in line 1 local coordinates.
******/
double eg_LineLineDist3D(double *lb1, double *le1, 
                         double *lb2, double *le2, 
                         double *mu1, double *mu2, int *lli_res)
{
  double p1[3], p2[3];
  int s;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  *lli_res = eg_LineLineIntersect3D(lb1, le1, lb2, le2, mu1, mu2);

  /* EG_INVALID */
  if (*lli_res == EG_INVALID)
    return -1.0;

  /* EG_SKEWED */
  if (*lli_res == EG_SKEWED)
  {
    for (s = 0; s < 3; s++)
    {
      p1[s] = lb1[s] + (*mu1) * (le1[s] - lb1[s]);
      p2[s] = lb2[s] + (*mu2) * (le2[s] - lb2[s]);
    }
    return eg_PointDist3D(p1, p2);
  }

  /* EG_PARALLEL */
  if (*lli_res == EG_PARALLEL)
  {
    eg_PointLineDist3D(lb1, le1, lb2, mu1);
    return eg_PointLineDist3D(lb1, le1, le2, mu2);
  }

  /* EG_INTERSECT */
  if (*lli_res == EG_INTERSECT)
    return 0.0;

  else
    eg_GeneralError
      ("eg_LineLineDist3D: unexpected result from eg_LineLineIntersect3D");

  return 0.0;                   /* to keep the compiler happy */
}                               /* end of eg_LineLineDist3D */

/****f*
* NAME
*   eg_PointPlaneDist
* DESCRIPTION
*   Returns the distance \textbf{d} between a point and a plane in space.
* ARGUMENTS
*   p - the coordinates of the point
*   pa - the coordinates of the first point defining the plane
*   pb - the coordinates of the second point defining the plane
*   pc - the coordinates of the third point defining the plane
*   d - returned distance as defined above
*   tli_res - return value (see notes)
*   gennum - return value (see notes)
*   tricrds - return value (see notes)
* RETURN VALUE
*   EG_INVALID for invalid triangle specification, EG_SUCCESS else
* NOTES
*  At the same time the normal on the plane is also determined
*  and the line through the point parallel with the normal is
*  constructed. The result of a TriangleLineIntersect3D between
*  this line and the triangle defining the plane is also returned
*  by return values \textbf{tli_res}, \textbf{gennum} and \textbf{tricrds}.
******/
int eg_PointPlaneDist(double *p, double *pa, 
                      double *pb, double *pc, double *d,
                      int *tli_res, int *gennum, 
                      double *tricrds)
{
  double n[3];
  double p2[3];
  double is[3];
  double mu;
  int i;

  /* determine normal */
  if (eg_TriangNormal3D(pa,pb,pc,n)!=EG_SUCCESS)
    return EG_INVALID;
  
  /* make normal line through p */
  for (i=0;i<3;i++)
    p2[i]=p[i]+n[i];

  /* the triangle-line intersection */
  (*tli_res)=eg_TriangLineIntersect3D(pa,pb,pc,p,p2,is,gennum,tricrds,&mu);
  if ((*tli_res)==EG_INVALID)
    return EG_INVALID;
  
  /* and the distance calculation */
  (*d)=eg_PointDist3D(p,is);
  
  return EG_SUCCESS;
}

/****f*
* NAME
*   eg_LineLineIntersect2D
* DESCRIPTION
*   Determines whether two lines in 2D are parallel or intersecting.
*   If intersecting the intersection point is returned.
* ARGUMENTS
*   lb1 - coordinates of the first point of the first line
*   le1 - coordinates of the second point of the first line
*   lb2 - coordinates of the first point of the second line
*   le2 - coordinates of the second point of the second line
*   mu1 - return value (see notes)
*   mu2 - return value (see notes)
* RETURN VALUE
*   EG_INVALID for invalid line specification (eg. \textbf{lb1}=\textbf{le1}.
*   EG_INTERSECT if the lines intersect and EG_PARALLEL	if the lines 
*   are parallel.
* NOTES
*   In order to find the intersection point (li) the result 
*   parameters (mu) should be used like:
*
*   \noindent li=lb+mu*(le-lb)
******/
int eg_LineLineIntersect2D(double *lb1, double *le1, 
                           double *lb2, double *le2,
                           double *mu1, double *mu2)
{
  register double denom;
  register double nom;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  if (eg_PointDist2D(lb1, le1) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }
  if (eg_PointDist2D(lb2, le2) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  denom = (le2[1] - lb2[1]) * (le1[0] - lb1[0]);
  denom -= (le2[0] - lb2[0]) * (le1[1] - lb1[1]);

  if (fabs(denom) < tolerances[EG_PAR_LINE] * tolerances[EG_PAR_LINE])
  {
    return EG_PARALLEL;
  }

  nom = (le2[0] - lb2[0]) * (lb1[1] - lb2[1]) - 
        (le2[1] - lb2[1]) * (lb1[0] - lb2[0]);
  *mu1 = nom / denom;

  nom = (le1[0] - lb1[0]) * (lb1[1] - lb2[1]) - 
        (le1[1] - lb1[1]) * (lb1[0] - lb2[0]);
  *mu2 = nom / denom;

  return EG_INTERSECT;
}                               /* end of eg_LineLineIntersect2D */

/****f*
* NAME
*   eg_LineLineIntersect3D
* DESCRIPTION
*   Determines whether two lines in 3D are parallel, skewed or intersecting.
*   If skewed or intersecting the parameters
*   of the points were the shortest distance between the lines is made
*   is also returned.
* ARGUMENTS
*   lb1 - coordinates of the first point of the first line
*   le1 - coordinates of the second point of the first line
*   lb2 - coordinates of the first point of the second line
*   le2 - coordinates of the second point of the second line
*   mu1 - return value (see notes)
*   mu2 - return value (see notes)
* RETURN VALUE
*   EG_INVALID for invalid line specification (eg. \textbf{lb1}=\textbf{le1}.
*   EG_INTERSECT if the lines intersect and EG_PARALLEL	if the lines 
*   are parallel. EG_SKEWED if the lines are not parallel and do 
*   not intersect.
* NOTES
*   In order to find the intersection point or point of closest distance 
*   (li) the result, the parameters (mu) should be used like:
*
*   \noindent li=lb+mu*(le-lb)
******/
int eg_LineLineIntersect3D(double *lb1, double *le1, 
                           double *lb2, double *le2,
                           double *mu1, double *mu2)
{
  double l1;
  double l2;
  double numer;
  double denom;
  double p13[3], p43[3], p21[3];
  double d1343, d4321, d1321, d4343, d2121;
  double li1[3], li2[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  p13[0] = lb1[0] - lb2[0];
  p43[0] = le2[0] - lb2[0];
  p21[0] = le1[0] - lb1[0];

  p13[1] = lb1[1] - lb2[1];
  p43[1] = le2[1] - lb2[1];
  p21[1] = le1[1] - lb1[1];

  p13[2] = lb1[2] - lb2[2];
  p43[2] = le2[2] - lb2[2];
  p21[2] = le1[2] - lb1[2];


  l1 = p21[0] * p21[0];
  l2 = p43[0] * p43[0];

  l1 += p21[1] * p21[1];
  l2 += p43[1] * p43[1];

  l1 += p21[2] * p21[2];
  l2 += p43[2] * p43[2];

  if (l1 < tolerances[EG_ZERO_DIST] * tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }
  if (l2 < tolerances[EG_ZERO_DIST] * tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  d1343 = p13[0] * p43[0] + p13[1] * p43[1] + p13[2] * p43[2];
  d4321 = p43[0] * p21[0] + p43[1] * p21[1] + p43[2] * p21[2];
  d1321 = p13[0] * p21[0] + p13[1] * p21[1] + p13[2] * p21[2];
  d4343 = p43[0] * p43[0] + p43[1] * p43[1] + p43[2] * p43[2];
  d2121 = p21[0] * p21[0] + p21[1] * p21[1] + p21[2] * p21[2];

  denom = d2121 * d4343 - d4321 * d4321;
  if (EG_ABS(denom) < tolerances[EG_PAR_LINE] * tolerances[EG_PAR_LINE])
  {
    return EG_PARALLEL;
  }
  numer = d1343 * d4321 - d1321 * d4343;

  *mu1 = numer / denom;
  *mu2 = (d1343 + d4321 * (*mu1)) / d4343;

  li1[0] = lb1[0] + (*mu1) * (le1[0] - lb1[0]);
  li2[0] = lb2[0] + (*mu2) * (le2[0] - lb2[0]);

  li1[1] = lb1[1] + (*mu1) * (le1[1] - lb1[1]);
  li2[1] = lb2[1] + (*mu2) * (le2[1] - lb2[1]);

  li1[2] = lb1[2] + (*mu1) * (le1[2] - lb1[2]);
  li2[2] = lb2[2] + (*mu2) * (le2[2] - lb2[2]);

  if (eg_PointDist3D(li1, li2) < tolerances[EG_ZERO_DIST])
  {
    return EG_INTERSECT;
  }

  return EG_SKEWED;
}                               /* end of eg_LineLineIntersect3D */

/****f*
* NAME
*   eg_TriangLineIntersect3D
* DESCRIPTION
*   Determines the intersection between a triangular face and a line
*   segment in 3D. The triangular face is defined by the three sets
*   of its corner node coordinates. The segment is defined
*   by its beginning and ending point.
* ARGUMENTS
*   t1 - coordinates of the first point of the triangle
*   t2 - coordinates of the second point of the triangle
*   t3 - coordinates of the third point of the triangle
*   lb - coordinates of the first point of the line
*   le - coordinates of the second point of the line
*   is - return value (see return values)
*   gennum - return value (see return values)
*   tricrds - return value (see return values)
*   mu - return value (see notes)
* RETURN VALUE
*   \begin{description}
*   \item[EG_INVALID]	the line was defined invalidly, 
*   (eg. \textbf{lb}=\textbf{le} or the triangular face is invalid 
*   when two or more of its corner nodes collide. 
*   \item[EG_PARALLEL] the line through the segment is parallel with
*  	the plane through the face.
*   \end{description}
*   For all other return values an intersection point was determined 
*   between the line through the segment and the plane through the 
*   face. The coordinates of this point are stored in \textbf{is}. This 
*   point can also be obtained from the trilinear coords of the face
*   whereby 
*
*   \noindent \textbf{is}=\textbf{tricrds}[0]*\textbf{t1}+\textbf{tricrds}[1]*\textbf{t2}+\textbf{tricrds}[2]*\textbf{t3}
*   and from the segment parameter \textbf{mu}
*
*   \noindent \textbf{is}=\textbf{lb}+\textbf{mu}*(\textbf{le}-\textbf{lb})
*   The status of this intersection point is further described by the
*   following return values.
*   \begin{description}
*   \item[EG_INSIDE] the intersection point lies completely inside 
*   the triangular face
*   \item[EG_OUTSIDE]	it lies completely outside
*   \item[EG_ONVERTEX] the intersection point collides with one of 
*   the vertices of the triangular face. In this case the vertex 
*   index will be stored in \textbf{gennum}
*   \item[EG_ONEDGE] the intersection point falls on one of the
*		triangular face's edges. The edge's index will be stored in 
*   \textbf{gennum}.
*   \end{description}
******/
int eg_TriangLineIntersect3D(double *t1, double *t2, double *t3,
                             double *lb, double *le,
                             double *is, int *gennum, 
                             double *tricrds, double *mu)
{
  register int i;
  double normal[3];
  double dif1[3], dif2[3];
  double nom, denom;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  /* validity checking */
  if (eg_PointDist3D(lb, le) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  /* determine normal on the face */
  if (eg_TriangNormal3D(t1, t2, t3, normal) == EG_INVALID)
  {
    return EG_INVALID;
  }

  /* determine parallel or not */
  for (i = 0; i < 3; i++)
  {
    dif1[i] = t1[i] - lb[i];
    dif2[i] = le[i] - lb[i];
  }
  denom = eg_VectDotProd(3, dif2, normal);
  nom = eg_VectDotProd(3, dif1, normal);
  if (EG_ABS(denom) < tolerances[EG_PAR_LINE] * eg_PointDist3D(lb, le))
  {
    return EG_PARALLEL;
  }
  (*mu) = nom / denom;

  /* the intersection point */
  for (i = 0; i < 3; i++)
  {
    is[i] = lb[i] + (*mu) * (le[i] - lb[i]);
  }

  return eg_TriangPointIntersect3D(t1, t2, t3, is, gennum, tricrds);

}                               /* end of eg_TriangLineIntersect3D  */

int eg_TriangPointIntersect3D(double *t1, double *t2, double *t3,
                              double *p, int *gennum, double *tricrds)
{
  double st, s1, s2, s3;
  int q1, q2, q3;
  double mu2;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  /* determine trilinear coordinates */
  st = eg_TriangArea3D(t1, t2, t3);
  s1 = eg_TriangArea3D(p, t2, t3);
  s2 = eg_TriangArea3D(p, t3, t1);
  s3 = eg_TriangArea3D(p, t1, t2);
  tricrds[0] = s1 / st;
  tricrds[1] = s2 / st;
  tricrds[2] = s3 / st;

  /* determine intersection point's status */
  q1 = (s1 < tolerances[EG_ZERO_DIST] * eg_PointDist3D(t2, t3) / 2.0);
  q2 = (s2 < tolerances[EG_ZERO_DIST] * eg_PointDist3D(t3, t1) / 2.0);
  q3 = (s3 < tolerances[EG_ZERO_DIST] * eg_PointDist3D(t1, t2) / 2.0);

  if (q1)
  {
    if (q2)
    {
      (*gennum) = 2;
      return EG_ONVERTEX;
    }
    else if (q3)
    {
      (*gennum) = 1;
      return EG_ONVERTEX;
    }
    (*gennum) = 0;
#ifdef XXX
    res = eg_LineLineIntersect3D(lb, le, t2, t3, &mu1, &mu2);
#ifdef EG_DEBUG
    if (res != EG_INTERSECT)
      eg_GeneralWarning
        ("eg_lib: discrepancy between LineLineIntersect & LineTriangIntersect");
#endif
#endif
    if ((mu2 < 1.0) && (mu2 > 0.0))
      return EG_ONEDGE;
    else
      return EG_OUTSIDE;
  }
  else if (q2)
  {
    if (q3)
    {
      (*gennum) = 0;
      return EG_ONVERTEX;
    }
    (*gennum) = 1;
#ifdef XXX
    res = eg_LineLineIntersect3D(lb, le, t3, t1, &mu1, &mu2);
#ifdef EG_DEBUG
    if (res != EG_INTERSECT)
      eg_GeneralWarning
        ("eg_lib: discrepancy between LineLineIntersect & LineTriangIntersect");
#endif
#endif
    if ((mu2 < 1.0) && (mu2 > 0.0))
      return EG_ONEDGE;
    else
      return EG_OUTSIDE;
  }
  else if (q3)
  {
    (*gennum) = 2;
#ifdef XXX
    res = eg_LineLineIntersect3D(lb, le, t1, t2, &mu1, &mu2);
#ifdef EG_DEBUG
    if (res != EG_INTERSECT)
      eg_GeneralWarning
        ("eg_lib: discrepancy between LineLineIntersect & LineTriangIntersect");
#endif
#endif
    if ((mu2 < 1.0) && (mu2 > 0.0))
      return EG_ONEDGE;
    else
      return EG_OUTSIDE;
  }

  if (EG_ABS((st - s1 - s2 - s3) / st) < tolerances[EG_ZERO_AREA])
  {
    return EG_INSIDE;
  }
  else
  {
    return EG_OUTSIDE;
  }
}                               /* end of eg_TriangPointIntersect3D  */

/****f*
* NAME
*   eg_TriangNormal3D
* DESCRIPTION
*   Determines the normal on a triangular face in 3D. The normal is
*   oriented towards the half space from which the face can be seen as 
*   oriented anticlockwise. The normal is normalized in length.
* RETURN VALUE
*   EG_INVALID if any two of the face's corner nodes collide. 
*   EG_SUCCESS if operation succesfull
* ARGUMENTS
*   t1 - coordinates of the first point of the triangle
*   t2 - coordinates of the second point of the triangle
*   t3 - coordinates of the third point of the triangle
*   normal - the returned normal
******/
int eg_TriangNormal3D(double *t1, double *t2, double *t3, double *normal)
{
  register double ln;
  double dif1[3], dif2[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  /* validity checking */
  if (eg_PointDist3D(t1, t2) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }
  if (eg_PointDist3D(t2, t3) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }
  if (eg_PointDist3D(t3, t1) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  dif1[0] = t2[0] - t1[0];
  dif2[0] = t3[0] - t1[0];

  dif1[1] = t2[1] - t1[1];
  dif2[1] = t3[1] - t1[1];

  dif1[2] = t2[2] - t1[2];
  dif2[2] = t3[2] - t1[2];

  eg_VectCrossVect(dif1, dif2, normal);

  /* normalizing */
  ln = normal[0] * normal[0];
  ln += normal[1] * normal[1];
  ln += normal[2] * normal[2];
  ln = sqrt(ln);

  normal[0] /= ln;
  normal[1] /= ln;
  normal[2] /= ln;

  return EG_SUCCESS;

}                               /* end of eg_TriangNormal3D */

/****f*
* NAME
*   eg_TriangArea3D
* DESCRIPTION
*   Determines the surface area of a triangular face in 3D.
* ARGUMENTS
*   t1 - coordinates of the first point of the triangle
*   t2 - coordinates of the second point of the triangle
*   t3 - coordinates of the third point of the triangle
* RETURN VALUE
*   the surface area.
* NOTES
*   the return value is always positive.
******/
double eg_TriangArea3D(double *t1, double *t2, double *t3)
{
  register double ln;
  double dif1[3], dif2[3];
  double normal[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  dif1[0] = t2[0] - t1[0];
  dif2[0] = t3[0] - t1[0];

  dif1[1] = t2[1] - t1[1];
  dif2[1] = t3[1] - t1[1];

  dif1[2] = t2[2] - t1[2];
  dif2[2] = t3[2] - t1[2];

  eg_VectCrossVect(dif1, dif2, normal);

  /* determining length */
  ln = normal[0] * normal[0];
  ln += normal[1] * normal[1];
  ln += normal[2] * normal[2];

  return 0.5 * sqrt(ln);

}                               /* end of eg_TriangArea3D */

/****f*
* NAME
*   eg_PolyArea2D
* DESCRIPTION
*   Determines the surface area of a polygon in 2D.
* ARGUMENTS
*  	vert - the coordinates of the vertices of the polygon
*   nvert - number of vertices defining the polygon   
* RETURN VALUE
*   the surface area.
* NOTES
*   the return value is always positive.
******/
double eg_PolyArea2D(double **vert, int nvert)
{
  register int i;
  register int j;
  register double sum = 0.0;
  register double pa;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  if (nvert < 3)
  {
    return -1.0;
  }

  for (i = nvert - 1, j = 0; j < nvert; i = j, j++)
  {
    sum += vert[i][0] * vert[j][1] - vert[j][0] * vert[i][1];
  }
  pa=fabs(0.5 * sum);
  if (pa<tolerances[EG_ZERO_AREA])
    return 0.0;
  return (pa);
}                               /* end of eg_PolyArea2D */

/****f*
* NAME
*   eg_TriangArea2D
* DESCRIPTION
*   Determines the surface area of a triangular face in 2D.
* ARGUMENTS
*   t1 - coordinates of the first point of the triangle
*   t2 - coordinates of the second point of the triangle
*   t3 - coordinates of the third point of the triangle
* RETURN VALUE
*   the surface area.
* NOTES
*   the return value is always positive.
******/
double eg_TriangArea2D(double *t1, double *t2, double *t3)
{
  register double a;
  register double pa;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  a = t1[0] * t2[1] - t1[1] * t2[0];
  a += t2[0] * t3[1] - t2[1] * t3[0];
  a += t3[0] * t1[1] - t3[1] * t1[0];

  pa=fabs(0.5 * a);
  if (pa<tolerances[EG_ZERO_AREA])
    return 0.0;
  return (pa);
}                               /* end of eg_TriangArea2D */

/****f*
* NAME
*   eg_PolyCentroid2D
* DESCRIPTION
*   Determines the centroid (and surface area) of a polygon in 2D.
* ARGUMENTS
*  	vert - the coordinates of the vertices of the polygon
*   nvert - number of vertices defining the polygon   
*   centroid - the returned coordinates of the centroid
*   area - the returned surface area
* RETURN VALUE
*   EG_INVALID if the input is invalid (eg. \textbf{nvert}<3). 
*   EG_SUCCESS else.
******/
int eg_PolyCentroid2D(double **vert, int nvert, 
                      double *centroid, double *area)
{
  register int i;
  register int j;
  register double a;
  register double sum = 0.0;
  register double xtmp = 0.0;
  register double ytmp = 0.0;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  if (nvert < 3)
  {
    return EG_INVALID;          /* invalid number of vertices */
  }

  for (i = nvert - 1, j = 0; j < nvert; i = j, j++)
  {
    a = vert[i][0] * vert[j][1] - vert[j][0] * vert[i][1];
    sum += a;
    xtmp += (vert[j][0] + vert[i][0]) * a;
    ytmp += (vert[j][1] + vert[i][1]) * a;
  }

  *area = 0.5 * sum;

  if (EG_ABS(sum) < tolerances[EG_ZERO_AREA])
  {
    return EG_INVALID;
  }

  centroid[0] = xtmp / (3.0 * sum);
  centroid[1] = ytmp / (3.0 * sum);

  return EG_SUCCESS;
}                               /* end of eg_PolyCentroid2D */

/****f*
* NAME
*   eg_TetrahBarycenter
* DESCRIPTION
*   Calculates the coordinates of a tetrahedron's barycenter.
* ARGUMENTS
*   v1 - coordinates of the first vertex of the tetrahedron.
*   v2 - coordinates of the first vertex of the tetrahedron.
*   v3 - coordinates of the first vertex of the tetrahedron.
*   v4 - coordinates of the first vertex of the tetrahedron.
*   barycenter - the returned coordinates of tetrahedron barycenter
* RETURN VALUE
*   none
******/
void eg_TetrahBarycenter(double *v1, double *v2, 
                         double *v3, double *v4,
                         double *barycenter)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  barycenter[0] = 0.25 * (v1[0] + v2[0] + v3[0] + v4[0]);
  barycenter[1] = 0.25 * (v1[1] + v2[1] + v3[1] + v4[1]);
  barycenter[2] = 0.25 * (v1[2] + v2[2] + v3[2] + v4[2]);
  return;
}                               /* end of eg_TetrahBarycenter */

/****f*
* NAME
*   eg_TetrahVolume
* DESCRIPTION
*   Calculates the colume of a tetrahedron.
* ARGUMENTS
*   v1 - coordinates of the first vertex of the tetrahedron.
*   v2 - coordinates of the first vertex of the tetrahedron.
*   v3 - coordinates of the first vertex of the tetrahedron.
*   v4 - coordinates of the first vertex of the tetrahedron.
* RETURN VALUE
*   the volume of tetrahedron
* NOTES
*   This function is fast but not robust.
*   For robust calculation of tetrahedron volume
*   use eg_Orient3D which gives as output signed volume
*   of tetrahedron multiplied by six.
******/
double eg_TetrahVolume(double *v1, double *v2, double *v3, double *v4)
{
  double v1dx, v2dx, v3dx;
  double v1dy, v2dy, v3dy;
  double v1dz, v2dz, v3dz;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  v1dx = v1[0] - v4[0];
  v1dy = v1[1] - v4[1];
  v1dz = v1[2] - v4[2];

  v2dx = v2[0] - v4[0];
  v2dy = v2[1] - v4[1];
  v2dz = v2[2] - v4[2];

  v3dx = v3[0] - v4[0];
  v3dy = v3[1] - v4[1];
  v3dz = v3[2] - v4[2];

  return EG_ABS(v1dx * (v2dy * v3dz - v2dz * v3dy)
               + v2dx * (v3dy * v1dz - v3dz * v1dy)
               + v3dx * (v1dy * v2dz - v1dz * v2dy))/6.0;
}                               /* end of eg_TetrahVolume */

/****f*
* NAME
*   eg_Orient2D
* DESCRIPTION
*   This function just forwards call to Shewchuk's orient2d
*   which is an adaptive and robust exact 2D orientation test.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
*   pc - the coordinates of the third point
* RETURN VALUE
*   Return a positive value if the points pa, pb, and pc occur
*   in counterclockwise order; a negative value if they occur
*   in clockwise order; and zero if they are collinear.  The 
*   result is also a rough approximation of twice the signed
*   area of the triangle defined by the three points.
******/
double eg_Orient2D(double *pa, double *pb, double *pc)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return orient2d(pa, pb, pc);
}                               /* end of eg_Orient2D */

/****f*
* NAME
*   eg_Orient3D
* DESCRIPTION
*   This function just forwards call to Shewchuk's orient3d 
*   which is an adaptive and robust exact 3D orientation test.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
*   pc - the coordinates of the third point
*   pd - the coordinates of the fourth point
* RETURN VALUE
*   Return a positive value if the point pd lies below the
*   plane passing through pa, pb, and pc; "below" is defined so
*   that pa, pb, and pc appear in counterclockwise order when
*   viewed from above the plane.  Returns a negative value if
*   pd lies above the plane.  Returns zero if the points are
*   coplanar.  The result is also a rough approximation of six
*   times the signed volume of the tetrahedron defined by the
*   four points.
******/
double eg_Orient3D(double *pa, double *pb, double *pc, double *pd)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return orient3d(pa, pb, pc, pd);
}                               /* end of eg_Orient3D */

/****f*
* NAME
*  eg_InCircle
* DESCRIPTION
*  This function just forwards call to Shewchuk's incircle.
*  which is an adaptive and robust exact 2D in circle test.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
*   pc - the coordinates of the third point
*   pd - the coordinates of the fourth point
* RETURN VALUE
*  Return a positive value if the point pd lies inside the
*  circle passing through pa, pb, and pc; a negative value if
*  it lies outside; and zero if the four points are cocircular.
*  The points pa, pb, and pc must be in counterclockwise
*  order, or the sign of the result will be reversed.
******/
double eg_InCircle(double *pa, double *pb, double *pc, double *pd)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return incircle(pa, pb, pc, pd);
}                               /* end of eg_InCircle */

/****f*
* NAME
*  eg_InSphere
* DESCRIPTION
*  This function just forwards call to Shewchuk's insphere.
*  which is an adaptive and robust exact 3D in sphere test.
* ARGUMENTS
*   pa - the coordinates of the first point
*   pb - the coordinates of the second point
*   pc - the coordinates of the third point
*   pd - the coordinates of the fourth point
*   pe - the coordinates of the fifth point
* RETURN VALUE
*   Return a positive value if the point pe lies inside the
*   sphere passing through pa, pb, pc, and pd; a negative value
*   if it lies outside; and zero if the five points are
*   cospherical.  The points pa, pb, pc, and pd must be ordered
*   so that they have a positive orientation (as defined by
*   eg_Orient3D()), or the sign of the result will be reversed.
******/
double eg_InSphere(double *pa, double *pb, double *pc, double *pd, 
                   double *pe)
{
  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  return insphere(pa, pb, pc, pd, pe);
}                               /* end of eg_InSphere */


/* Tolerance for alpha and beta in segment-triangle intersection 
 * This tolerance is crucial in determining type of intersection.
 * alpha and beta can bee seen as relative distances from vertex 0
 * along triangle sides:
 *    for vertex 0  alpha=0, beta=0;
 *    for vertex 1  alpha=1, beta=0;
 *    for vertex 2  alpha=0, beta=1; 
 */
#define ab_eps 1.0e-4
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 eg_V1_TriangLineIntersect3D -- 

   ALTERNATIVE IMPLEMENTATION OF TRIANGLE LINE INTERSECT3D

   SEE: eg_TriangLineIntersect3D

   CHANGES: tricrds is not holding trilinear coordinates.
            tricrds[0] and tricrds[1] are holding parameters
            of intersection point related to side vectors (t2-t1) 
            and (t3-t1), ie
            is = tricrds[0] * (t2-t1) + tricrds[1] * (t3-t1) + t1
 
            The value of tricrds[2] is meaningless.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int eg_V1_TriangLineIntersect3D(double *t1, double *t2, double *t3,
                                double *lb, double *le,
                                double *is, int *gennum,
                                double *tricrds, double *mu)
{
  register double alpha;
  register double beta;
  double normal[3], ray[3];     /* triangle normal and segment direction */
  double n_ray[3];              /* normalized ray vector */
  double u[3], v[3];            /* vectors along triangle sides */
  double normalDt1, normalDray, normalDlb;
  double normal_length, ray_length;
  double parallel, d;
  short ix[3];                  /* for axis indices */
  double normalCorr, alphaTol, betaTol, abTol;
  double s1, s2, s3, st;

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  /* initialize alpha and beta (any value >1.0 and <0.0) */
  alpha = beta = -10.0;

  *gennum = 100;                /* again dummy value (proper ones are 0, 1, 2 */



  /* validity checking */
  if (eg_PointDist3D(lb, le) < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  /* calculate alongside vectors u, v and segment direction vector ray */
  u[0] = t2[0] - t1[0];
  v[0] = t3[0] - t1[0];
  ray[0] = le[0] - lb[0];


  u[1] = t2[1] - t1[1];
  v[1] = t3[1] - t1[1];
  ray[1] = le[1] - lb[1];

  u[2] = t2[2] - t1[2];
  v[2] = t3[2] - t1[2];
  ray[2] = le[2] - lb[2];

  /* calculate plane normal */
  eg_VectCrossVect(u, v, normal);
  /* find axis most orthogonal to the triangle plane;
   * ix[0] will hold index of that and and ix[1], ix[2]
   * are indices of other axes
   */
  ix[0] = 0;
  /* use 'd' to hold the absolute value of maximal component of 'normal' */
  d = EG_ABS(normal[0]);
  if (d < EG_ABS(normal[1]))
  {
    ix[0] = 1;
    d = EG_ABS(normal[1]);
  }
  if (d < EG_ABS(normal[2]))
  {
    ix[0] = 2;
    d = EG_ABS(normal[2]);
  }
  switch (ix[0])
  {
    case (0):
      {
        ix[1] = 1;
        ix[2] = 2;
        break;
      }
    case (1):
      {
        ix[1] = 0;
        ix[2] = 2;
        break;
      }
    case (2):
      {
        ix[1] = 0;
        ix[2] = 1;
        break;
      }
    default:
      {
        eg_GeneralError("eg_V1_TriangLineIntersect3D: Internale error %s %d\n",
                       "Invalid value of index ix[0] : ", ix[0]);
      }
  }

  /* check if triangle point are not colinear */
  if (d < tolerances[EG_ZERO_DIST])
  {
    return EG_INVALID;
  }

  /* Calculate length of ix[0] compontent of 'normal' vector.
     I will need it later on 
   */
  normalCorr = EG_ABS(normal[ix[0]]);

  /* calculate dot products */
  normalDt1 = -(normal[0] * t1[0] + normal[1] * t1[1] + normal[2] * t1[2]);
  normalDray = normal[0] * ray[0] + normal[1] * ray[1] + normal[2] * ray[2];
  normalDlb = normal[0] * lb[0] + normal[1] * lb[1] + normal[2] * lb[2];

  /* normalize 'normal' and 'ray' vectors and check if they are parallel */
  /* but as I need later on ray new vector is introduced n_ray */
  normal_length = normal[0] * normal[0];
  ray_length = ray[0] * ray[0];

  normal_length += normal[1] * normal[1];
  ray_length += ray[1] * ray[1];

  normal_length += normal[2] * normal[2];
  ray_length += ray[2] * ray[2];

  normal_length = sqrt(normal_length);
  ray_length = sqrt(ray_length);

  /* normalize vectors 'normal' and 'n_ray' */
  normal[0] = normal[0] / normal_length;
  n_ray[0] = ray[0] / ray_length;

  normal[1] = normal[1] / normal_length;
  n_ray[1] = ray[1] / ray_length;

  normal[2] = normal[2] / normal_length;
  n_ray[2] = ray[2] / ray_length;

  /* check if triangle and segment are parallel */
  parallel =
    normal[0] * n_ray[0] + normal[1] * n_ray[1] + normal[2] * n_ray[2];

  if (EG_ABS(parallel) < tolerances[EG_PAR_LINE])
  {
    return EG_PARALLEL;
  }

  /* calculate the parameter mu corresponding to intersection point */
  *mu = -(normalDt1 + normalDlb) / normalDray;

  /* calculate coordinates of intersestion point */
  is[0] = lb[0] + (*mu) * ray[0];
  is[1] = lb[1] + (*mu) * ray[1];
  is[2] = lb[2] + (*mu) * ray[2];

  /* map intersection point and vectors u, v on plane ix[2]ix[3] */

  /* CAUTION: 
   *       intersection point has now coordinates (u[0], v[0], 0)
   *       u vector has coordinates (u[1], v[1], 0)
   *       v vector has coordinates (u[2], v[2], 0)
   * This is a bit inconsisten notation but the code works and
   * I beter will not change it
   */

  u[0] = is[ix[1]] - t1[ix[1]];
  v[0] = is[ix[2]] - t1[ix[2]];
  u[1] = t2[ix[1]] - t1[ix[1]];
  u[2] = t3[ix[1]] - t1[ix[1]];
  v[1] = t2[ix[2]] - t1[ix[2]];
  v[2] = t3[ix[2]] - t1[ix[2]];



  /* solve system (2x2) of equation for alpha and beta:
     | u(1) u(2) | |alpha|    |u(0)|
     | --------- | | --- |  = |----|
     | v(1) v(2) | |beta |    |v(0)|
   */
  if (EG_ABS(u[1]) <= 1.0e-20)
  {
    beta = u[0] / u[2];
    if (beta >= 0.0 && beta <= 1.0);
    {
      alpha = (v[0] - beta * v[2]) / v[1];
    }
  }
  else
  {
    beta = (v[0] * u[1] - u[0] * v[1]) / (v[2] * u[1] - u[2] * v[1]);
    if (beta >= 0.0 && beta <= 1.0);
    {
      alpha = (u[0] - beta * u[2]) / u[1];
    }
  }

  /* fill tricoordinates array 
     tricrds[0] = alpha;
     tricrds[1] = beta; */

  /* determine trilinear coordinates */
  st = eg_TriangArea3D(t1, t2, t3);
  s1 = eg_TriangArea3D(is, t2, t3);
  s2 = eg_TriangArea3D(is, t3, t1);
  s3 = eg_TriangArea3D(is, t1, t2);
  tricrds[0] = s1 / st;
  tricrds[1] = s2 / st;
  tricrds[2] = s3 / st;



  /* calculate the tolerance corection coefficients */
  /* The correced tolerances for alpha, beta and (1-alpha-beta) 
   * This corrected tolreances are equivalent to  
   * distances between intersection point and triangle segments
   */
  /* length ov vector v (see CAUTION * above) */
  betaTol = tolerances[EG_ZERO_DIST] * sqrt(u[1] * u[1] + v[1] * v[1]);
  /* length of vector u */
  alphaTol = tolerances[EG_ZERO_DIST] * sqrt(u[2] * u[2] + v[2] * v[2]);
  /* length of vectror (v-u) */
  abTol = tolerances[EG_ZERO_DIST] * sqrt((u[2] - u[1]) * (u[2] - u[1])
                                          + (v[2] - v[1]) * (v[2] - v[1]));

  /* printf ("%f %f %f %f %f %f\n", alphaTol, betaTol, abTol, normalCorr, alpha,
     beta); */
  /* check if and where segment intersects triangle */
  if (normalCorr * alpha < -alphaTol
      || normalCorr * beta < -betaTol
      || abTol < normalCorr * (alpha + beta - 1))
  {
    return EG_OUTSIDE;
  }
  else
  {
    if (EG_ABS(normalCorr * alpha) <= alphaTol)
    {
      if (EG_ABS(normalCorr * beta) <= betaTol)
      {
        *gennum = 0;
        return EG_ONVERTEX;
      }
      else if (normalCorr * EG_ABS(1.0 - beta) <= betaTol)
      {
        *gennum = 2;
        return EG_ONVERTEX;
      }
      else
      {
        *gennum = 1;
        return EG_ONEDGE;
      }
    }
    else if (normalCorr * EG_ABS(beta) <= betaTol)
    {
      if (normalCorr * EG_ABS(1.0 - alpha) < alphaTol)
      {
        *gennum = 1;
        return EG_ONVERTEX;
      }
      else
      {
        *gennum = 2;
        return EG_ONEDGE;
      }
    }
    else if (normalCorr * EG_ABS(1 - (alpha + beta)) < abTol)
    {
      *gennum = 0;
      return EG_ONEDGE;
    }
    else
    {
      return EG_INSIDE;
    }
  }
}                               /* end of eg_V1_TriangLineIntersect3D */

/****f*
* NAME
*   eg_DoSegmentTriangleIntersect
* DESCRIPTION
*   Check if segment and triangle intersect.
* ARGUMENTS
*   t1 - coordinates of the first point of the triangle
*   t2 - coordinates of the second point of the triangle
*   t3 - coordinates of the third point of the triangle
*   lb - coordinates of the first point of the segment
*   le - coordinates of the second point of the segment
* RETURN VALUE
*   1 if segment and triangle intersect, 0 otherwise
******/
int eg_DoSegmentTriangleIntersect(double *t1, double *t2, double *t3,
                                  double *s1, double *s2)
{
  double T1[3], T2[3], T3[3];
  double S1[3], S2[3];
  double e1[3], e2[3], e3[3];
  double f1[3];
  double ef11[3], ef21[3], ef31[3];
  double n1[3];
  double g1[3], g2[3], g3[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  T1[0] = 0.0;
  T1[1] = 0.0;
  T1[2] = 0.0;
  T2[0] = t2[0] - t1[0];
  T2[1] = t2[1] - t1[1];
  T2[2] = t2[2] - t1[2];
  T3[0] = t3[0] - t1[0];
  T3[1] = t3[1] - t1[1];
  T3[2] = t3[2] - t1[2];

  S1[0] = s1[0] - t1[0];
  S1[1] = s1[1] - t1[1];
  S1[2] = s1[2] - t1[2];
  S2[0] = s2[0] - t1[0];
  S2[1] = s2[1] - t1[1];
  S2[2] = s2[2] - t1[2];

  e1[0] = T2[0] - T1[0];
  e1[1] = T2[1] - T1[1];
  e1[2] = T2[2] - T1[2];
  e2[0] = T3[0] - T2[0];
  e2[1] = T3[1] - T2[1];
  e2[2] = T3[2] - T2[2];
  e3[0] = T1[0] - T3[0];
  e3[1] = T1[1] - T3[1];
  e3[2] = T1[2] - T3[2];

  f1[0] = S2[0] - S1[0];
  f1[1] = S2[1] - S1[1];
  f1[2] = S2[2] - S1[2];

  CROSS3D(n1, e1, e2);          /* normal to the triangle */

  CROSS3D(g1, e1, n1);
  CROSS3D(g2, e2, n1);
  CROSS3D(g3, e3, n1);

  CROSS3D(ef11, e1, f1);
  CROSS3D(ef21, e2, f1);
  CROSS3D(ef31, e3, f1);

  /* test projections */
  if (project5(n1, T1, T2, T3, S1, S2) == 0)
    return 0;

  if (project5(ef11, T1, T2, T3, S1, S2) == 0)
    return 0;
  if (project5(ef21, T1, T2, T3, S1, S2) == 0)
    return 0;
  if (project5(ef31, T1, T2, T3, S1, S2) == 0)
    return 0;

  if (project5(g1, T1, T2, T3, S1, S2) == 0)
    return 0;
  if (project5(g2, T1, T2, T3, S1, S2) == 0)
    return 0;
  if (project5(g3, T1, T2, T3, S1, S2) == 0)
    return 0;

  return 1;
}                               /* end of eg_DoSegmentTriangleIntersect */

/****f*
* NAME
*   eg_DoTriangleTriangleIntersect
* DESCRIPTION
*   Check if two triangles intersect.
* ARGUMENTS
*   t1 - coordinates of the first point of the first triangle
*   t2 - coordinates of the second point of the first triangle
*   t3 - coordinates of the third point of the first triangle
*   z1 - coordinates of the first point of the second triangle
*   z2 - coordinates of the second point of the second triangle
*   z3 - coordinates of the third point of the second triangle
* RETURN VALUE
*   1 if the triangles intersect, 0 otherwise
******/
int eg_DoTriangleTriangleIntersect(double *t1, double *t2, double *t3,
                                   double *z1, double *z2, double *z3)
{
  double T1[3], T2[3], T3[3];
  double Z1[3], Z2[3], Z3[3];
  double e1[3], e2[3], e3[3];
  double f1[3], f2[3], f3[3];
  double ef11[3], ef12[3], ef13[3];
  double ef21[3], ef22[3], ef23[3];
  double ef31[3], ef32[3], ef33[3];
  double n1[3];
  double m1[3];
  double g1[3], g2[3], g3[3];
  double h1[3], h2[3], h3[3];

  if (eg_initialized == FALSE)
    eg_GeneralError("eg_lib: eg_lib not initialized");
  
  T1[0] = 0.0;
  T1[1] = 0.0;
  T1[2] = 0.0;
  T2[0] = t2[0] - t1[0];
  T2[1] = t2[1] - t1[1];
  T2[2] = t2[2] - t1[2];
  T3[0] = t3[0] - t1[0];
  T3[1] = t3[1] - t1[1];
  T3[2] = t3[2] - t1[2];

  Z1[0] = z1[0] - t1[0];
  Z1[1] = z1[1] - t1[1];
  Z1[2] = z1[2] - t1[2];
  Z2[0] = z2[0] - t1[0];
  Z2[1] = z2[1] - t1[1];
  Z2[2] = z2[2] - t1[2];
  Z3[0] = z3[0] - t1[0];
  Z3[1] = z3[1] - t1[1];
  Z3[2] = z3[2] - t1[2];

  e1[0] = T2[0] - T1[0];
  e1[1] = T2[1] - T1[1];
  e1[2] = T2[2] - T1[2];
  e2[0] = T3[0] - T2[0];
  e2[1] = T3[1] - T2[1];
  e2[2] = T3[2] - T2[2];
  e3[0] = T1[0] - T3[0];
  e3[1] = T1[1] - T3[1];
  e3[2] = T1[2] - T3[2];

  f1[0] = Z2[0] - Z1[0];
  f1[1] = Z2[1] - Z1[1];
  f1[2] = Z2[2] - Z1[2];
  f2[0] = Z3[0] - Z2[0];
  f2[1] = Z3[1] - Z2[1];
  f2[2] = Z3[2] - Z2[2];
  f3[0] = Z1[0] - Z3[0];
  f3[1] = Z1[1] - Z3[1];
  f3[2] = Z1[2] - Z3[2];

  CROSS3D(n1, e1, e2);          /* normal to the first triangle triangle */
  CROSS3D(m1, e1, e2);          /* normal to the first triangle triangle */

  CROSS3D(g1, e1, n1);
  CROSS3D(g2, e2, n1);
  CROSS3D(g3, e3, n1);

  CROSS3D(h1, f1, m1);
  CROSS3D(h2, f2, m1);
  CROSS3D(h3, f3, m1);

  CROSS3D(ef11, e1, f1);
  CROSS3D(ef21, e1, f2);
  CROSS3D(ef31, e1, f3);
  CROSS3D(ef21, e2, f1);
  CROSS3D(ef22, e2, f2);
  CROSS3D(ef23, e2, f3);
  CROSS3D(ef31, e3, f1);
  CROSS3D(ef32, e3, f2);
  CROSS3D(ef33, e3, f3);

  /* test projections */
  if (project6(n1, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(m1, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;

  if (project6(ef11, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef12, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef13, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef21, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef22, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef23, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef31, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef32, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(ef33, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;

  if (project6(g1, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(g2, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(g3, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(h1, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(h2, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;
  if (project6(h3, T1, T2, T3, Z1, Z2, Z3) == 0)
    return 0;

  return 1;
}                               /* end of eg_DoSegmentTriangleIntersect */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 project5 --
        Check if projection of segment and triange edges on specified
        axis overlap
 Return value:
        0 if projections do not overlap
        1 otherwise
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int project5(double *axis, double *t1, double *t2, double *t3,
                    double *s1, double *s2)
{
  double T1, T2, T3;
  double S1, S2;
  double mx1, mx2, mn1, mn2;

  T1 = DOT3D(axis, t1);
  T2 = DOT3D(axis, t2);
  T3 = DOT3D(axis, t3);
  S1 = DOT3D(axis, s1);
  S2 = DOT3D(axis, s2);

  mx1 = EG_MAX3(T1, T2, T3);
  mn1 = EG_MIN3(T1, T2, T3);
  mx2 = EG_MAX(S1, S2);
  mn2 = EG_MIN(S1, S2);

  if (mn1 > mx2)
    return 0;
  if (mn2 > mx1)
    return 0;

  return 1;
}                               /* end of project5 */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 project6 --
        Check if projection of any edges of two triangles overlap
       
 Return value:
        0 if projections do not overlap
        1 otherwise
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
static int project6(double *axis, double *t1, double *t2, double *t3,
                    double *z1, double *z2, double *z3)
{
  double T1, T2, T3;
  double Z1, Z2, Z3;
  double mx1, mx2, mn1, mn2;

  T1 = DOT3D(axis, t1);
  T2 = DOT3D(axis, t2);
  T3 = DOT3D(axis, t3);
  Z1 = DOT3D(axis, z1);
  Z2 = DOT3D(axis, z2);
  Z3 = DOT3D(axis, z3);

  mx1 = EG_MAX3(T1, T2, T3);
  mn1 = EG_MIN3(T1, T2, T3);
  mx2 = EG_MAX3(Z1, Z2, Z3);
  mn2 = EG_MIN3(Z1, Z2, Z3);

  if (mn1 > mx2)
    return 0;
  if (mn2 > mx1)
    return 0;

  return 1;
}                               /* end of project6 */


/* Modifications:
 * $Log: eg_lib.c,v $
 * Revision 1.13  2002/10/16 19:05:09  peteri
 * - dependence on e_lib is removed
 *
 * Revision 1.12  2002/02/05 12:45:16  romek
 * Identification string added.
 *
 * Revision 1.11  2002/02/05 12:24:07  romek
 * Doc fix.
 *
 * Revision 1.10  2002/02/05 12:12:04  peteri
 * - new function to determine whether point falls between segment end points
 *   if it is projected onto the segment
 *
 * Revision 1.9  2001/10/07 23:21:05  romek
 * eg_TetrahVolume: fixed bug in calculating volume
 *
 * Revision 1.8  2001/08/28 10:51:20  romek
 * Added CVS keywords for Id and Log messages
 *
 */

