/********************************************************************
 eg_lib.h  --
        The E Geometry Library: eg_lib

 Jelle Muylle <jelle@sect.mce.hw.ac.uk>
 Roman Putanowicz <romek@sect.mce.hw.ac.uk>
 SECT Research Group, Heriot-Watt University, Edinburgh    
 ********************************************************************/

/* $Id: eg_lib.h,v 1.10 2002/10/16 19:05:10 peteri Exp $ */

/****m*
* NAME
*   eg_lib
* DESCRIPTION
*   This is a geometric function library based on and for use with
*   the SECT Research Group's e_lib library.
*
*   The library contains functions and macros for:
*   \begin{itemize}
*   \item tolerance management
*   \item distances between points, lines and planes
*   \item normal and surface functions
*   \item various metric properties
*   \item intersection functions
*   \item orientation functions
*   \end{itemize}
*
*   The main aim of the library is to provide consistent function
*   results among different geometric evaluations.
*   The library is far from complete, but additions are made whenever
*   they are deemed necessary.
* NOTES
*   The orientation functions provided are basically wrappers for the
*   public domain code by J.R. Shewchuk, \textit{Routines for Arbitrary 
*   Precision Floating-point Arithmetic and Fast Robust Geometric 
*   Predicates}
* COPYRIGHT
*   Jelle Muylle \texttt{jelle@sect.mce.hw.ac.uk}, \\
*   Roman Putanowicz \texttt{romek@sect.mce.hw.ac.uk}, 2001
******/

#ifndef EG_LIB_H
#define EG_LIB_H

#include <stdlib.h>

#define EG_LINE_MAX  400               /* maximum output line length */

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

/****d*
* NAME
*   Result macros
* DESCRIPTION
*   Several macros are defined for the return values
*   of the geometric functions.
* SYNOPSIS
*/
/**/ #define EG_INVALID      FALSE 
/**/ #define EG_SUCCESS      TRUE
/**/ #define EG_PARALLEL     10
/**/ #define EG_INTERSECT    11
/**/ #define EG_SKEWED       12
/**/ #define EG_INSIDE       13
/**/ #define EG_OUTSIDE      14
/**/ #define EG_ONVERTEX     15
/**/ #define EG_ONEDGE       16 
/**/ #define EG_NOTREACH     17
/**/ 
/**/ #define EG_NOT_CONVEX                101
/**/ #define EG_NOT_CONVEX_DEGENERATE     102
/**/ #define EG_CONVEX_DEGENERATE         103
/**/ #define EG_CONVEX_CCW                104
/**/ #define EG_CONVEX_CW                 105
/*
******/

/****d*
* NAME
*   Tolerance indices
* DESCRIPTION
*   Macros to indicate which tolerance is to be set or retrieved
*   when using the tolerance management functions.
* SYNOPSIS
*/
/**/ #define EG_ZERO_DIST_TOL   1
/**/ #define EG_ZERO_AREA_TOL   2
/**/ #define EG_PAR_LINE_TOL    3
/*
******/

/* Miscellaneous macros */
#define EG_ABS(X)         (((X) > 0) ? (X) : (-(X)))
#define EG_MIN(A,B)       (((A) < (B)) ? (A) : (B))
#define EG_MAX(A,B)       (((A) > (B)) ? (A) : (B))
#define EG_MIN3(A,B,C)    EG_MIN(EG_MIN(A,B),C)
#define EG_MAX3(A,B,C)    EG_MAX(EG_MAX(A,B),C)
#define EG_MIN4(A,B,C,D)  EG_MIN(EG_MIN3(A,B,C),D)
#define EG_MAX4(A,B,C,D)  EG_MAX(EG_MAX3(A,B,C),D)
#define EG_SQUARE(X)      ((X)*(X))

/* Initialization */
void eg_Init();
void eg_Version();

/* Tolerance management */
void eg_SetTolerance(int flag, double tol);
double eg_GetTolerance(int flag);

/* Distance between points and lines */
double eg_PointDist2D(double *pa, double *pb);
double eg_PointDist3D(double *pa, double *pb);
double eg_PointSqrDist2D(double *pa, double *pb);
double eg_PointSqrDist3D(double *pa, double *pb);
int    eg_PointProjInsideSegment(double *sb, double *se, double *p);
double eg_PointLineDist2D(double *lb, double *le, double *p, double *mu);
double eg_PointLineDist3D(double *lb, double *le, double *p, double *mu);

/* Distance between lines and lines */
double eg_LineLineDist3D(double *lb1, double *le1, double *lb2, 
    				   double *le2, double *mu1, double *mu2, int *lli_res);

/* Distance between point and plane (in 3D ofcourse) */
int eg_PointPlaneDist(double *p, double *pa, double *pb, double *pc,
                      double *d, int *tli_res, int *gennum, double *tricrds);

/* Normal and surface functions */
int eg_TriangNormal3D(double *t1, double *t2, double *t3, double *normal);
double eg_TriangArea2D(double *t1, double *t2, double *t3);
double eg_PolyArea2D(double **vert, int nvert);
double eg_TriangArea3D(double *t1, double *t2, double *t3);

/* Centroid functions */
int eg_PolyCentroid2D(double **vert, int nvert,
    				  double *centroid,double *area);

/* Various metric properties */
/* Tetrahedron barycenter */
void eg_TetrahBarycenter (double *v1, double *v2, double *v3, double *v4,
                double *barycenter);
/* Tetrahedron volume */
double eg_TetrahVolume (double *v1, double *v2, double *v3, double *v4);

/* Intersection functions */
int eg_LineLineIntersect2D(double *lb1, double *le1, double *lb2, 
    				   double *le2, double *mu1, double *mu2);
int eg_LineLineIntersect3D(double *lb1, double *le1, double *lb2, 
    				   double *le2, double *mu1, double *mu2);
int eg_TriangLineIntersect3D(double *t1, double *t2, double *t3,
                             double *lb, double *le,
                             double *is, int *gennum, 
                             double *tricrds, double *mu);
int eg_TriangPointIntersect3D (double *t1, double *t2, double *t3,
    					   	   double *p, int *gennum, double *tricrds);
/* Alternative implementation  of line  triangle intersection */
int eg_V1_TriangLineIntersect3D(double *t1, double *t2, double *t3,
                             double *lb, double *le,
                             double *is, int *gennum, 
                             double *tricrds, double *mu);
/* Checking existence of intersection */
int eg_DoSegmentTriangleIntersect (double *t1, double *t2, double *t3, 
                                   double *s1, double *s2); 
int eg_DoTriangleTriangleIntersect (double *t1, double *t2, double *t3, 
                                    double *z1, double *z2, double *z3); 

/* Orientation functions */
double eg_Orient2D(double *pa, double *pb, double *pc);
double eg_Orient3D(double *pa, double *pb, double *pc, double *pd);

/* InCircle and InSphere */
double eg_InCircle(double *pa, double *pb, double *pc, double *pd);
double eg_InSphere(double *pa, double *pb, double *pc, double *pd, double *pe);


/* circumcenter functions */
void eg_TetCircumcenter (double a[3], double b[3], double c[3], double d[3], 
                         double circumcenter[3], double *xi, double *eta, double *zeta);
void eg_TriCircumcenter2D (double a[2], double b[2], double c[2], 
                           double circumcenter[2], double *xi, double *eta);
void eg_TriCircumcenter3D (double a[3], double b[3], double c[3], 
                           double circumcenter[3], double *xi, double *eta);

/* Error handling routines */
void eg_GeneralWarning(char *Warning, ...);
void eg_GeneralError(char *ErrorMessage, ...);
     
/* Math routines */
double eg_VectDotProd(unsigned int n, double *x, double *y);
void eg_VectCrossVect(double *a, double *b, double *c);

#endif /* EG_LIB_H */
           

/* Modifications:
 * $Log: eg_lib.h,v $
 * Revision 1.10  2002/10/16 19:05:10  peteri
 * - dependence on e_lib is removed
 *
 * Revision 1.9  2002/08/04 13:14:20  peteri
 * - consistent naming and some formatting
 *
 * Revision 1.8  2002/07/19 14:09:28  romek
 * Changes to add functions to calculate circumcenter of tetrahderon and triangle.
 *
 * Revision 1.7  2002/02/05 12:12:05  peteri
 * - new function to determine whether point falls between segment end points
 *   if it is projected onto the segment
 *
 * Revision 1.6  2001/08/28 10:51:34  romek
 * Added CVS keywords for Id and Log messages
 *
 */
