
#include "eps.h"

#ifdef SUBDOM_BOUND
#include "boundary.h"
#endif


/* #################################################################### */

/* return 0 if error
   return 1 if success 
*/

int e_Write_EPS_File(MESH *mesh, char *filename, OPTIONS *options, 
                     int NPoly, POLYGON *Polygons, int *Order, int error)
{
  char            epsfile[E_LINE_MAX];
  FILE           *fp;
  double          str;
  int             i, j, abs_error = 0;
  int             minNLayerTriang2, minNLayerQuad3;


  /* check the mesh substructures */
  switch(options->type)
  {
    case EPS_PLAINMESH:
      break;
    case EPS_SUBDOMAIN:
      if(mesh->SubStructSet[E_SS_NUM_DECOMP] != TRUE)
      {
        if(error)
        {
          e_GeneralWarning("NO subdomain structure, plotting plain mesh !");
          options->type = EPS_PLAINMESH;
        }
        else
          return 0;
      }
      break;
    case EPS_ERROR:
      if(mesh->SubStructSet[E_SS_NUM_FEERROR] != TRUE)
      {
        if(error)
        {
          e_GeneralWarning("NO finite element error structure, plotting plain mesh !");
          options->type = EPS_PLAINMESH;
        }
        else
          return 0;
      }

      /* if automatic error, determine max */
      if(!(options->user_value))
      {
        if(mesh->FEError.FEErrors[0] < 0)
        {
          abs_error = 1;
          e_GeneralWarning("Error value smaller than zero.");
          e_GeneralWarning("Absolute values will be used.");
          e_GeneralWarning("-- Message is shown only once. --");
        }

        options->max_error = fabs(mesh->FEError.FEErrors[0]);
        for(i = 1; i < mesh->TotalNElems; i++)
        {
          if((mesh->FEError.FEErrors[i] < 0) && (abs_error == 0))
          {
            abs_error = 1;
            e_GeneralWarning("Error value smaller than zero.");
            e_GeneralWarning("Absolute values will be used.");
            e_GeneralWarning("-- Message is shown only once. --");
          }
          if(options->max_error < fabs(mesh->FEError.FEErrors[i]))
            options->max_error = fabs(mesh->FEError.FEErrors[i]);
        }
      }
      break;
    case EPS_STRESS:
      if(mesh->SubStructSet[E_SS_NUM_STRESS] != TRUE)
      {
        if(error)
        {
          e_GeneralWarning("NO stress structure, plotting plain mesh !");
          options->type = EPS_PLAINMESH;
          break;
        }
        else
          return 0;
      }

      /* check that stress point exists */
      if(mesh->NElemsLink1 && mesh->NStressPointsLink1 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for LINK1 !");
        else
          return 0;
      }

      if(mesh->NElemsLink2 && mesh->NStressPointsLink2 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for LINK2 !");
        else
          return 0;
      }

      if(mesh->NElemsLink3 && mesh->NStressPointsLink3 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for LINK3 !");
        else
          return 0;
      }

      if(mesh->NElemsLink4 && mesh->NStressPointsLink4 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for LINK4 !");
        else
          return 0;
      }

      if(mesh->NElemsLink5 && mesh->NStressPointsLink5 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for LINK5 !");
        else
          return 0;
      }

      if(mesh->NElemsTriang1 && mesh->NStressPointsTriang1 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for TRIANG1 !");
        else
          return 0;
      }

      if(mesh->NElemsTriang2)
      {
        if(mesh->NStressPointsTriang2 < options->stress_point)
        {
          if(error)
            e_GeneralError("Stress point does not exist for TRIANG2 !");
          else
            return 0;
        }

        /* we need material, that structure contains the number of layers */
        if(!mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
        {
          if(error)
            e_GeneralError("Material data is not available, but needed for TRIANG2 stress plotting !");
          else
            return 0;
        }

        minNLayerTriang2 = mesh->CompMatsTriang2[0]->NLayers;
        for(i = 1; i < mesh->NElemsTriang2; i++)
          if(minNLayerTriang2 > mesh->CompMatsTriang2[i]->NLayers)
            minNLayerTriang2 = mesh->CompMatsTriang2[i]->NLayers;

        if(minNLayerTriang2 < options->stress_layer)
        {
          if(error)
            e_GeneralError("Stress layer does not exist for TRIANG2 !");
          else
            return 0;
        }
      }

      if(mesh->NElemsTriang3 && mesh->NStressPointsTriang3 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for TRIANG3 !");
        else
          return 0;
      }

      if(mesh->NElemsTriang4 && mesh->NStressPointsTriang4 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for TRIANG4 !");
        else
          return 0;
      }

      if(mesh->NElemsTriang5 && mesh->NStressPointsTriang5 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for TRIANG5 !");
        else
          return 0;
      }

      if(mesh->NElemsQuad1 && mesh->NStressPointsQuad1 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for QUAD1 !");
        else
          return 0;
      }

      if(mesh->NElemsQuad3)
      {
        if(mesh->NStressPointsQuad3 < options->stress_point)
        {
          if(error)
            e_GeneralError("Stress point does not exist for QUAD3 !");
          else
            return 0;
        }

        /* we need material, that structure contains the number of layers */
        if(!mesh->SubStructSet[E_SS_NUM_MAT_PROPS])
        {
          if(error)
            e_GeneralError("Material data is not available, but needed for QUAD3 stress plotting !");
          else
            return 0;
        }

        minNLayerQuad3 = mesh->CompMatsQuad3[0]->NLayers;
        for(i = 1; i < mesh->NElemsQuad3; i++)
          if(minNLayerQuad3 > mesh->CompMatsQuad3[i]->NLayers)
            minNLayerQuad3 = mesh->CompMatsQuad3[i]->NLayers;

        if(minNLayerQuad3 < options->stress_layer)
        {
          if(error)
            e_GeneralError("Stress layer does not exist for QUAD3 !");
          else
            return 0;
        }
      }

      if(mesh->NElemsTetrah1 && mesh->NStressPointsTetrah1 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for TETRAH1 !");
        else
          return 0;
      }

      if(mesh->NElemsBlock1 && mesh->NStressPointsBlock1 < options->stress_point)
      {
        if(error)
          e_GeneralError("Stress point does not exist for BLOCK1 !");
        else
          return 0;
      }


      /* if automatic stress, determine max */
      if(!(options->user_value))
      {
        options->max_total_stress = 0.0;

        /* absolute maximum stress value */
        if(mesh->NElemsLink1)
        {
          for (i = 0; i < mesh->NElemsLink1; i++)
          {
            str = mesh->Stresses.StressesLink1[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsLink2)
        {
          for (i = 0; i < mesh->NElemsLink2; i++)
          {
            str = mesh->Stresses.StressesLink2[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsLink3)
        {
          for (i = 0; i < mesh->NElemsLink3; i++)
          {
            str = mesh->Stresses.StressesLink3[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsLink4)
        {
          for (i = 0; i < mesh->NElemsLink4; i++)
          {
            str = mesh->Stresses.StressesLink4[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsLink5)
        {
          for (i = 0; i < mesh->NElemsLink5; i++)
          {
            str = mesh->Stresses.StressesLink5[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTriang1)
        {
          for (i = 0; i < mesh->NElemsTriang1; i++)
          {
            str = mesh->Stresses.StressesTriang1[i]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTriang2)
        {
          for (i = 0; i < mesh->NElemsTriang2; i++)
          {
            str = mesh->Stresses.StressesTriang2[i]
                                                [options->stress_layer-1]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTriang3)
        {
          for (i = 0; i < mesh->NElemsTriang3; i++)
          {
            str = mesh->Stresses.StressesTriang3[i]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTriang4)
        {
          for (i = 0; i < mesh->NElemsTriang4; i++)
          {
            str = mesh->Stresses.StressesTriang4[i]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTriang5)
        {
          for (i = 0; i < mesh->NElemsTriang5; i++)
          {
            str = mesh->Stresses.StressesTriang5[i]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsQuad1)
        {
          for (i = 0; i < mesh->NElemsQuad1; i++)
          {
            str = mesh->Stresses.StressesQuad1[i]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsQuad3)
        {
          for (i = 0; i < mesh->NElemsQuad3; i++)
          {
            str = mesh->Stresses.StressesQuad3[i]
                                              [options->stress_layer-1]
                                              [options->stress_point-1]
                                              [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }

        if(mesh->NElemsTetrah1)
        {
          for (i = 0; i < mesh->NElemsTetrah1; i++)
          {
            str = mesh->Stresses.StressesTetrah1[i]
                                                [options->stress_point-1]
                                                [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }
        if(mesh->NElemsBlock1)
        {
          for (i = 0; i < mesh->NElemsBlock1; i++)
          {
            str = mesh->Stresses.StressesBlock1[i]
                                               [options->stress_point-1]
                                               [options->stress_type-1];
            if(options->max_total_stress < fabs(str))
              options->max_total_stress = fabs(str);
          }
        }
      }
      break;
    default:
      if(error)
      {
        e_GeneralWarning("Unknown substructure, plotting plain mesh !");
        options->type = EPS_PLAINMESH;
      }
      else
        return 0;
  }

  /* open the EPS file */
  strcpy(epsfile, filename);
  strcat(epsfile, ".eps");

  if((fp = fopen(epsfile, "w")) == NULL)
  {
    if(error)
    {
      e_FileError(filename, "Cannot open EPS file !");
    }
    else
      return 0;
  }

  options->fp = fp;

  /* EPS header definiton, mainly bounding box */
  fprintf(fp, "%%!PS-Adobe-2.0 EPSF-2.0\n");
  fprintf(fp, "%%%%Creator: SECT-EPS-Creator \n");
  fprintf(fp, "%%%%BoundingBox: %f %f %f %f\n",
          -options->margin, 
          -options->margin, 
          options->picture_size_x+options->margin, 
          options->picture_size_y+options->margin );
  fprintf(fp, "%%EndComments\n");
  
  /* to make the EPS file short, simplify the commands */
  fprintf(fp, "\n/l {lineto} def");
  fprintf(fp, "\n/m {moveto} def");
  fprintf(fp, "\n/s {stroke} def");
  fprintf(fp, "\n/c {setrgbcolor} def");
  fprintf(fp, "\n/p {closepath fill} def");
  fprintf(fp, "\n/w {1.0 1.0 1.0 setrgbcolor} def");
  fprintf(fp, "\n/b {0.0 0.0 0.0 setrgbcolor} def");

  /* subdomain font definitons */
  fprintf(fp, "\n/tt /Times-Roman findfont %f scalefont def", 
          options->subdom_t_size);

  fprintf(fp, "\n/tm /Times-Roman findfont %f scalefont def",
          E_MIN(options->picture_size_x, options->picture_size_y) *
          options->number_size);

  /* end of general definitions */
  fprintf(fp, "\n\n");

  /* line thickness */
  fprintf(fp, "\n0.00 setlinewidth\n");

  /* color change for subdomains not always */
  if(options->type == EPS_SUBDOMAIN && options->hidden_line == 0)
  {
    int *nelempersub;
    int  index, subdom;
  
    nelempersub = e_alloc_i_vect(mesh->Decomp.NSubDoms);
    for(i = 0; i < NPoly; i++)
    {
      index = Polygons[i].mesh_index +
              mesh->FirstNetIndex[Polygons[i].elem_type] - 1;
      nelempersub[mesh->Decomp.SubDoms[index]-1]++;
    }

    for(i = 0; i < mesh->Decomp.NSubDoms-1; i++)
    {
      nelempersub[i+1] += nelempersub[i];
    }

    for(i = 0; i < NPoly; i++)
    {
      index = Polygons[i].mesh_index +
              mesh->FirstNetIndex[Polygons[i].elem_type] - 1;
      subdom = mesh->Decomp.SubDoms[index]-1;
      Order[nelempersub[subdom]-1] = index;
      nelempersub[subdom]--;
    }
    e_free(nelempersub);

  }

  /* draw out filled elements */
  for(i = 0; i < NPoly; i++)
  {
    draw_polygon(mesh, Polygons[Order[i]], *options, error);
  }

  /* if no hidden line, then now plot the element boundary */
  if(   (!(options->hidden_line))
     && (   (options->type == EPS_PLAINMESH)
         || (   (options->type != EPS_PLAINMESH)
             && (options->border) ) ) )
  {
    if(options->bcolor || options->type == EPS_PLAINMESH)
      fprintf(options->fp, "\nb"); /* black */
    else
      fprintf(options->fp, "\nw"); /* white */

    /* draw out the boundary of the element */
    for(i = 0; i < NPoly; i++)
    {
      if(   (   Polygons[i].no_vertices > 2 
             && options->type != EPS_PLAINMESH)
         || (options->type == EPS_PLAINMESH) )
      {
        draw_simple(Polygons[i], *options);
        fprintf(options->fp, " s");
      }
    }
  }

  /* element numbering */
  if(options->element_number)
  {
    double centre[2];
  
    for(j = 0; j < NPoly; j++)
    {
      centre[0] = 0.0;
      centre[1] = 0.0;
      for(i = 0; i < Polygons[j].no_vertices; i++)
      {
        centre[0] += Polygons[j].screen_coords[i][0];
        centre[1] += Polygons[j].screen_coords[i][1];
      }
      centre[0] /= Polygons[j].no_vertices;
      centre[1] /= Polygons[j].no_vertices;
 
      fprintf(options->fp, "\n%f %f m", centre[0],
              options->picture_size_y-centre[1]);
      fprintf(options->fp, "\ntm setfont (%d) show", Polygons[j].index+1);
    }
  }

#ifdef SUBDOM_BOUND
  if(!Subdom_Boundary(mesh, Polygons, options))
    return 0;
#endif

  fprintf(fp,"\nshowpage\n\n");

  fclose(fp);
  
  if(error)
  {
    fprintf(stderr, "\n\nEPS file generation is succesfull. \n\n");
  }
  else
  {
    return 1;
  }

  /* */
  return 1;
}




