#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "e_lib.h"


#include "eps.h"

/****************************************************************************\
                                Painter's algorithm

What does it do ?
-----------------

Draws the polygons in such an order, that if a polygon is covered by another
one, then first it draws the covered then overpaints the one which is on top
of the previous one.


Assumptions:
------------
  - Only 1D, 2D elements in the mesh (For 3D elements the projected outline
    should be determined as a convex hull for further analysis.)
  - Quad elements are planes ! If they are not plane their convex hull must
    be determined by a proper alogrithm. Furthermore the face of the quad
    element cannot be used as a separating plane in that case.


How does the method work ??
---------------------------

1: Sort the elements according to the determined priorities. The first
   element in the list is printed first, therefore the elements et the
   beginning of the list will be (mot likely) covered by other elements.
   Mikado sorting is used which searches the list in two directions.
   It removes all elements which is not covered by any other element
   and puts them at the end of the ordered list. On the other hand it
   removes all elements which are covered by another element but not
   covering any other element and puts them at front of the plotting 
   list. Then repeats the search, but not considering the already
   removed elements.

2: The priorities of the elements is determined on the fly. If several
   Mikado iterations are necessary, then certain elements are compared
   with each other several times. This is a strategic decision. The original
   implementation used lots of storage rather than do computation and it could
   not sort 10000 elements. The current implementation calculates everything
   and does not store anything except the most important data and it
   can sort 10000 element in less than 20 seconds.
   To determine the priority of two elements the following algorithm is used:

   Step 1:
      Checks, whether the 2D bounding boxes of the projected polygons overlap.
      If they are far away from each other, then there is no reason for
      further checking since they cannot cover each other.

   Step 2:
      This step checks the convex hulls of the projected polygonss. The
      convex hulls of the two polygons can intersect, one can contain the
      other or they can be disjoint.
        
        - If the convex hulls intersect each other, then the z
          coordinate of the intersection point is determined (in the
          projected space) and whichever has a smaller value is closer
          to the eye. However in certain cases the intersection points
          can be very close to each other in the projected space,
          in this case the order of the polygons is stored and the
          decision is delayed. This means, that either Step 3 can decide
          the order of the polygons, or the algorithm has to rely on this
          "critical" intersection.
            
        - The convex hull of one of the polygons contains the other
          polygon's convex hull. At this stage, it can be determined
          whether an element is totally obscured or not.
            
        - If the convex hulls of the elements are disjoint, then no
          ordering should be done.

   Step 3:
      This step is executed only if there was a critical intersection
      between the convex hulls of the polygons in Step 2. In this case
      several sub-steps can be used to determine the order of the polygons:
        
        - If the polygons are so far away from each other in 3D space that the 
          circumscribed spheres do not touch or intersect each other then
          there is no need for separation planes. The distance of the centre
          of the polygons from the projection centre determine the order of
          the polygons.
        
        - If the circumscribed sphere test fails, then the program determines
          the bounding box around the polygons. (Maximum X, Y, Z and minimum
          X, Y, Z coordinates are determined.) Then the program tries to use
          the faces of the bounding box as separating plane between two
          polygons. First it takes the bounding box of the first polygon and
          checks whether the nodes of the other polygon are all outside. If
          this is not true, then it tries vice-versa, taking the bounding box
          of the second polygon and checks with the nodes of the first polygon.

        - If the faces of the bounding box are not separating planes then we
          have one last chance, that the face of the polygon itself is a
          separating plane between the two polygons. In this case it checks
          only 2D elements (triangles, quads, but no links).
        
        When a separating plane is determined three data are stored: the
        normal, a constant (a dot product of the normal vector and
        a point which is on the plane) and an indicator. The indicator
        determines, that the first polygon (first argument to the function)
        is on which side of the plane. If the polygon is on the same side
        as the normal then the value is one otherwise minus one.
   Step 4:
      It is possible, that no separating plane is found. In this case
      the order determined by the critical intersection is used.

\**************************************************************************/


#define INFINITE 1e20
#define EPS      1e-8


#define NOT_FOUND         0
/* 1 is the separation plane */
#define USE_SOAP_BUBBLES  2
#define NONE              -10

#define DISJOINT            1
#define INTERSECTION        2
#define POLY2_INSIDE_POLY1  3
#define POLY1_INSIDE_POLY2  4



double Hidden_Proj_center[3];
double Hidden_Distance;



#define sign(x) ((x) ? (((x) < 0) ? -1 : 1) : 0)


#define SubVec2D(ab, a, b) \
        ((ab)[0] = (b)[0] - (a)[0], (ab)[1] = (b)[1] - (a)[1])

#define SubVec3D(ab, a, b)          \
        ((ab)[0] = (b)[0] - (a)[0], \
         (ab)[1] = (b)[1] - (a)[1], \
         (ab)[2] = (b)[2] - (a)[2]  )



#define DotProduct2D(a, b) \
        ((a)[0]*(b)[0]+(a)[1]*(b)[1])

#define DotProduct3D(a, b) \
        ((a)[0] * (b)[0] + (a)[1] * (b)[1] + (a)[2] * (b)[2])




#define LinearComb2D(r, p, pq, l) \
        ((r)[0] = (p)[0] + (l)*(pq)[0], \
         (r)[1] = (p)[1] + (l)*(pq)[1]    )

#define LinearComb3D(r, p, pq, l) \
        ((r)[0] = (p)[0] + (l)*(pq)[0], \
         (r)[1] = (p)[1] + (l)*(pq)[1], \
         (r)[2] = (p)[2] + (l)*(pq)[2]    )

#define Normal2D(n, v) \
        ((n)[0] = -(v)[1], (n)[1] = (v)[0])


#define Length(x) sqrt(DotProduct3D(x, x))


#define WhichSide(point, plane) \
        sign(DotProduct3D((plane)->normal, point) - (plane)->cnst)

#define T2(w, x) \
        ((w) = (x) / (Hidden_Distance - (x)))


#define IsOverlap(min1, max1, min2, max2)              \
  ( (   (min1)[0] > (max2)[0] || (max1)[0] < (min2)[0] \
     || (min1)[1] > (max2)[1] || (max1)[1] < (min2)[1] \
     || (E_ABS((min1)[0] - (max2)[0]) < EPS)           \
     || (E_ABS((max1)[0] - (min2)[0]) < EPS)           \
     || (E_ABS((min1)[1] - (max2)[1]) < EPS)           \
     || (E_ABS((max1)[1] - (min2)[1]) < EPS)           \
    ) ? 0 : 1                                          \
  )


/* function definitions */
static void CreateSepPlane(PLANE *s, POLYGON *obj1, POLYGON *obj2);



/* ######################################################################## */
/*                      Projected convex hull checking                      */
/* ######################################################################## */


/* Whether a polygon is inside another one. */
static int PointInsidePolygon(double *q, int n, double **poly,
                              int *polygon_behind)
{
  double *p = *poly;
  double *r, *s;
  double  dist_r, dist_s;
  double  p0[3];
  double  normal[2], cnst;
  int     i;
  double  t;
  double  w1, w2;
  double  z[2];
  
  
  normal[0] = q[1] - p[1];
  normal[1] = p[0] - q[0];

  cnst = DotProduct2D(p, normal);
  
  r = poly[1];
  s = NULL;
  
  dist_r = DotProduct2D(r, normal) - cnst;
  dist_s = 0;
  
  /* search for a side whose two end points are on different sides
     of the p-q line, where q is the point "supposed" to be inside
     and p is the first point of the polygon */
  for(i = 2; i < n; i++)
  {
    s = poly[i];
    dist_s = DotProduct2D(s, normal) - cnst;
    if(sign(dist_r) != sign(dist_s))
      break;
    r = s;
    dist_r = dist_s;
  }
  
  /* all the point on the same side */
  if(i == n)
    return 0;

  /* determine the intersection of the p-q line and the side of the
     polygon, it will be point P(0) */
  t = dist_r / (dist_r - dist_s);
  for(i = 0; i < 3; i++)
    p0[i] = r[i] + t * (s[i] - r[i]);
  
  i = E_ABS(p0[0] - p[0]) > E_ABS(p0[1] - p[1]) ? 0 : 1;
  t = (q[i] - p[i]) / (p0[i] - p[i] + EPS);

  /* point Q must lie between point P and P(0) */
  if( (E_ABS(t) < EPS) || (E_ABS(t-1) < EPS) ||
      !(t > 0 && t < 1.0))
    return 0;

  /* the point is inside the polygon */
  
  /* 3 dimension */
  T2(z[0], q[2]);
  T2(w1, p[2]);
  T2(w2, p0[2]);
  z[1] = (1 - t) * w1 + t * w2;
  
  *polygon_behind = ((z[0] < z[1]) ? 1 : 0);
  
  return 1;
}




static int CheckPolygonInside(int n1, double **hull1,
                              int n2, double **hull2,
                              int *polygon_behind)
{
  int i, j;
  
  /* Get a node which is not equal to a node of hull2.
     It is important, because we are checking for signs,
     and zero is not a proper answer in this case! */
  for(i = 0; i < n1; i++)
  {
    /* check with all nodes */
    for(j = 0; j < n2; j++)
    {
      if(E_ABS(hull1[i][0] - hull2[j][0]) < EPS &&
         E_ABS(hull1[i][1] - hull2[j][1]) < EPS &&
         E_ABS(hull1[i][2] - hull2[j][2]) < EPS    )
        break;
    }
    if(j == n2) /* not equal to any of the nodes, probably we found one */
    {
      int ns, nx;
      double check, cc;
      double normal[3];

      /* check whether the found node is not on the side of
         the polygon */
      ns = 0;
      for(j = 0; j < n2; j++)
      {
        nx = n2 - j - 1;
        normal[0] = hull1[i][1] - hull2[ns][1];
        normal[1] = hull2[ns][0] - hull1[i][0];
        cc = DotProduct2D(hull2[ns], normal);
        check = DotProduct2D(hull2[nx], normal) - cc;
        if(E_ABS(check) < EPS)
          break;
        ns = nx;
      }
      if(j == n2)
        break;
    }
  }
  if(i != n1)
  {
    return(PointInsidePolygon(hull1[i], n2, hull2, polygon_behind));
  }

  return 0;
}





/* Intersection of two lines. */
static int IntersectSegments_2D(double *a1, double *b1, double *d1,
                                double *a2, double *b2, double *d2,
                                double *z)
{
  double  ne[2], ab[2];
  double  d_ne, ab_ne, t1, t2;
  
  Normal2D(ne, d2);
  d_ne = DotProduct2D(d1, ne);
  SubVec2D(ab, a1, a2);
  ab_ne = DotProduct2D(ab, ne);
    
  /* Are they parallel? */
  if(E_ABS(d_ne) < EPS)
  {
    if(E_ABS(ab_ne) > EPS)
      return(0);

    t1 = 0.0;
  }
  else
  {
    t1 = ab_ne / d_ne;
  }
  
  if(   (E_ABS(t1) < EPS)
     || (E_ABS(t1-1) < EPS)
     || !(t1 > 0 && t1 < 1.0) )
      return(0);
    
  /* use a non zero value, one of them should be non zero otherwise */
  if(E_ABS(d2[0]) > EPS)
      t2 = (a1[0] + t1 * d1[0] - a2[0]) / d2[0];
  else if(E_ABS(d2[1]) > EPS)
      t2 = (a1[1] + t1 * d1[1] - a2[1]) / d2[1];
  else
      return(0); /* the edge goes through the projection center */
  
  if(   (E_ABS(t2) < EPS)
     || (E_ABS(t2-1) < EPS)
     || !(t2 > 0 && t2 < 1.0))
      return(0);
  
  /* z values of the projected intersection point */
  z[0] = a1[2] + t1 * (b1[2] - a1[2]);
  z[1] = a2[2] + t2 * (b2[2] - a2[2]);

  /* FIXME: do I really need this, since a1,a2,b1,b2 are already
     properly projected */
  /* projection rays are not parallel, perspective */
  if(Hidden_Distance < 1000)
  {
    T2(z[0], z[0]);
    T2(z[1], z[1]);
  }
  
  return(1);
}


/* Check whether two convex hulls intersect or overlap. */
static int RelativePosHulls(int n1, double **hull1,
                            int n2, double **hull2,
                            double z[2], int *polygon_behind,
                            int *delay_decision)
{
  double **a1, **b1;
  double **a2, **b2;
  double **hi1 = hull1 + n1;
  double **hi2 = hull2 + n2;
  double   dz, previous_dz; /* distance of the intersection point */
  double   d1[3], d2[3];
  int      intersection;
  
  *delay_decision = 0;

  /* go through all sides in hull1 */
  for(a1 = hull1, b1 = a1+1; a1 < hi1; a1++, b1++)
  {
    if(b1 == hi1)
      b1 = hull1;

    /* vector pointing from first point to second on the side */
    SubVec2D(d1, *a1, *b1);

    /* go through all sides in hull2 */
    for(a2 = hull2, b2 = a2+1; a2 < hi2; a2++, b2++)
    {
      
      if(b2 == hi2)
        b2 = hull2;
      /* vector pointing from first point to second on the side */
      SubVec2D(d2, *a2, *b2);

      
      intersection = IntersectSegments_2D(*a1, *b1, d1, *a2, *b2, d2, z);
      dz = z[1] - z[0];

      if(intersection)
      {
        if(E_ABS(dz) > EPS)
        {
          return INTERSECTION;
        }
        else
        {
          /* This is bad for the final priority decision !
             The difference between the z coordinates is quite small.
             If such a case occured, compare the results. If they have
             the same sign, it is all right, otherwise wait with
             the decision. */
          if(*delay_decision)
          {
            if(sign(dz) == sign(previous_dz))
                return INTERSECTION;
            else /* rely on a better result */
                if(E_ABS(dz) > E_ABS(previous_dz))
                    previous_dz = dz;
          }
          else
          {
            *delay_decision = 1;
            previous_dz = dz;
          }
        }
      }
    } /* for (a2) */
  } /* for (a1) */

  if(*delay_decision) /* Still??? Bad luck! */
  {
    return INTERSECTION;
  }

  /* check that one of the polygons is totally covered by the other */
  if(CheckPolygonInside(n1, hull1, n2, hull2, polygon_behind))
      return POLY1_INSIDE_POLY2;

  if(CheckPolygonInside(n2, hull2, n1, hull1, polygon_behind))
      return POLY2_INSIDE_POLY1;

  /* There is no relation between the convex hulls! */

  return DISJOINT;
}


/* final decision: intersect convex hulls and compare the
   z-values of the two space points that correspond to the
   first intersection point */
static int IntersectPolygonHulls(int first, POLYGON *elem1, POLYGON *elem2,
                                 int *delay_decision)
{
  double z[2];
  int    polygon_behind = 0;
  
  switch(RelativePosHulls(elem1->no_vertices, elem1->screen_coords,
                          elem2->no_vertices, elem2->screen_coords,
                          z, &polygon_behind, delay_decision))
  {
    case INTERSECTION:
    {
      if(z[0] > z[1])
          return elem1->index;
      else
          return elem2->index;
    }
    case POLY2_INSIDE_POLY1:
    {
      if(polygon_behind)
      {
        elem2->obscured = 1;
        return elem1->index;
      }
      else
      {
        elem1->obscured = 1;
        return elem2->index;
      }
    }
    case POLY1_INSIDE_POLY2:
    {
      if(polygon_behind)
      {
        elem1->obscured = 1;
        return elem2->index;
      }
      else
      {
        elem2->obscured = 1;
        return elem1->index;
      }
    }
    case DISJOINT:
    {
      return NONE;
    }
    default:
    {
      e_GeneralError("IntersectPolygonHulls: Internal error ");
    }
  }

  /* never reached */
  return NONE;
}


/* ######################################################################## */
/*                          Using separator plane                           */
/* ######################################################################## */


/* Distance between object centre and projection centre.
   Not real distance, but since the same style will be compared
   it does not matter. */
static double CentreFromEye(POLYGON *obj)
{
  double r[3];

  SubVec3D(r, obj->proj_center, Hidden_Proj_center);
  
  return(DotProduct3D(r,r));
}



static int BoundingBoxSepPlane(int side, PLANE *s,
                               POLYGON *obj1, POLYGON *obj2)
{
  /* corner nodes */
  int    corner[8][3] = { {0,0,0}, {1,0,0}, {1,1,0}, {0,1,0}, 
                          {0,0,1}, {1,0,1}, {1,1,1}, {0,1,1}  }; 
  int    normal[6][3] = { {0,0,-1}, /* bottom face */
                          {0,-1,0}, {1,0,0}, {0,1,0}, {-1,0,0}, /* sides */
                          {0,0,1} }; /* top face */
  double  v[3];
  int    *nn;
  double  cnst, fcnst;
  int     i, j, k, equal;
  double  tt;


  for(i = 0; i < 6; i++)
  {
    nn = normal[i];

    /* position vector on the face of the box */
    for(k = 0; k < 3; k++)
    {
      v[k] = 0.0;
      if(nn[k] == 1)
        v[k] = obj1->box_max[k];
      else if(nn[k] == -1)
        v[k] = obj1->box_min[k];
    }
    /* constant of the plane */
    fcnst = DotProduct3D(nn, v);

    /* check the nodes of the element, that not all of them
    lies on the bounding face */
    for(k = 0; k < obj1->no_vertices; k++)
    {
      cnst = DotProduct3D(nn, obj1->world_coords[k])-fcnst;
      if(E_ABS(cnst) > EPS)
        break;
    }
    /* all nodes are on the face, not good for this kind of checking */
    if(k == obj1->no_vertices)
      continue;
      
    equal = 0;
    for(j = 0; j < 8; j++)
    {
      /* coords of the bounding box of other object */
      v[0] = (corner[j][0] ? obj2->box_max[0] : obj2->box_min[0]);
      v[1] = (corner[j][1] ? obj2->box_max[1] : obj2->box_min[1]);
      v[2] = (corner[j][2] ? obj2->box_max[2] : obj2->box_min[2]);

      /* check that all nodes of the other bounding box are on the same
         side of the face */
      tt = DotProduct3D(nn, v)-fcnst;
    
      if(sign(tt) == sign(cnst))
        break;
      if(tt == 0)
        equal++;
    }

    /* if all points of the bounding box are on the same side of the face */
    if(j == 8 && equal != 8)
    {
      s->info = side;
      s->cnst = fcnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = nn[k];
      return 1;
    }
  }

  return 0;
}

static int FaceSepPlane(PLANE *s, POLYGON *obj1, POLYGON *obj2)
{
  int     pos, neg, i, k;
  double  cnst;
  
  if(obj1->no_vertices > 2) /* only >2D elements */
  {
    pos = 0; neg = 0;
    for(i = 0; i < obj2->no_vertices; i++)
    {
      cnst = DotProduct3D(obj1->local_plane.normal, obj2->world_coords[i]) - 
             obj1->local_plane.cnst;
      if(cnst > 0)
        pos++;
      else if(cnst < 0)
        neg++;
    }
    if(   (pos == 0 && neg != 0)
       || (pos != 0 && neg == 0) )
    {
      s->info = (pos ? -1 : 1);
      s->cnst = obj1->local_plane.cnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = obj1->local_plane.normal[k];
      return 1;
    }
  }

  return 0;
}





static void CreateSepPlane(PLANE *s, POLYGON *obj1, POLYGON *obj2)
{
  double dist[3];

  /* they are far away from each other */
  SubVec3D(dist, obj1->center, obj2->center);
  if(Length(dist) > (obj1->max_rad + obj2->max_rad))
  {
    s->info = USE_SOAP_BUBBLES;
    return;
  }
  
  /* check obj1 bounding box faces with obj2 nodes */
  if(BoundingBoxSepPlane(-1, s, obj1, obj2) != 0)
      return;

  /* check obj2 bounding box faces with obj1 nodes */
  if(BoundingBoxSepPlane(1, s, obj2, obj1) != 0)
      return;
  
  /* check obj1 face with obj2 nodes */
  if(FaceSepPlane(s, obj1, obj2) != 0)
      return;
  
  /* check obj2 face with obj1 nodes */
  if(FaceSepPlane(s, obj2, obj1) != 0)
      return;

  s->info = NOT_FOUND;
}


/* ####################################################################### */
/*                 Sorting the elements according to priority              */
/* ####################################################################### */


/* This function determines which of the two elements is closer to the
   projection centre (should be drawn later), or there is no ordering
   between the two. */
static int WhichPolygonIsInFront(POLYGON *elem1, POLYGON *elem2)
{
  int         first = NONE;
  PLANE       sep_plane;
  int         delay_decision = 0;
  
  /* one of them is already covered, no need for further check */
  if(elem1->obscured || elem2->obscured)
    return NONE;
  
  /* Do the bounding boxes overlap ? Uses projected coordinates */
  if(!IsOverlap(elem1->proj_min, elem1->proj_max,
                elem2->proj_min, elem2->proj_max))
    return NONE; /* bounding boxes do not overlap, no ordering */

  /* intersect the convex hull of the polygons */
  first = IntersectPolygonHulls(first, elem1, elem2, &delay_decision);

  /* if there was a critical intersection decision is delayed, and
     we try to determine the order by separation planes */
  if(delay_decision == 0)
      return first;

  /* projected z coords are far away ? */
  if(   (elem1->proj_min[2] > elem2->proj_max[2])
     || (E_ABS(elem1->proj_min[2] - elem2->proj_max[2]) < EPS) )
  {
    return(elem1->index);
  }
  else if(elem2->proj_min[2] > elem1->proj_max[2])
  {
    return(elem2->index);
  }

  /* try with some kind of separating planes */
  CreateSepPlane(&sep_plane, elem1, elem2);

  /* circumscribing spheres do not intersect, do not overlap each other */
  if(sep_plane.info == USE_SOAP_BUBBLES)
  {
    /* distance from the eye position */
    if(CentreFromEye(elem1) < CentreFromEye(elem2))
    {
      return(elem1->index);
    }
    else
    {
      return(elem2->index);
    }
  }
  /* there is a separating plane between the two elements */
  else if(sep_plane.info != NOT_FOUND)
  {
    /* check that the eye position is on the same side as the element
    with the smaller index, can the other elem obscure the first one ? */
    if(WhichSide(Hidden_Proj_center, &sep_plane) == sep_plane.info)
    { 
      return(elem1->index);
    }
    else /* if 0, tough luck */
    { 
      return(elem2->index);
    }
  }

  /* no separating plane was found, only the critical intersection is left
     as a last resort :-( */
  /* fprintf(stderr, "\nCritical intersection is used !"); */
  
  return(first);
}



/*
 * Go through the list of polygons, check their relation to each other.
 */
static int MikadoSelect(int n, POLYGON **res_obj)
{
  int  i, j;
  int  index1, index2;
  int  priority;
  
  for(i = 0; i < n; i++)
  {
    index1 = res_obj[i]->index;

    /* FIXME: this is a restriction, so links are the last */
    if(res_obj[i]->no_vertices == 2)
      break;

    for(j = i+1; j < n; j++)
    {
      index2 = res_obj[j]->index;

      /* decide which element is in front of the other or
         they have no relation */
      priority = WhichPolygonIsInFront(res_obj[i], res_obj[j]);
      if(priority == index1)
      {
        /* mark that "i" cannot be obscured and "j" cannot be on the front */
        res_obj[i]->is_in_back = 0;
        res_obj[j]->is_in_front = 0;
      }
      else if(priority == index2)
      {
        /* mark that "j" cannot be obscured and "i" cannot be on the front */
        res_obj[i]->is_in_front = 0;
        res_obj[j]->is_in_back = 0;
      }
    }
  }
  
  return 0;
}



int sort_objects(int *order, int n, POLYGON *poly)
{
  POLYGON **res_obj;
  POLYGON **obj;
  int    i, j, k, prev_k, i1, i2;
  int    i_min, i_max;
  int    mikado_iter = 0;
  
  /* initially all of the polygons are unsorted */
  res_obj = e_calloc(n, sizeof(POLYGON *));
  for(i = 0; i < n; i++)
  {
    res_obj[i] = poly+i;
    /* mark them, so they can be in front or back */
    res_obj[i]->is_in_front = res_obj[i]->is_in_back = 1;
  }

  i_min = 0;
  i_max = n;
  k = n;

  while(k)
  {
    /* sort the element from front and from backwards as well */
    MikadoSelect(k, res_obj);
    mikado_iter++;
    
    /* put the objects into the ordered list */
    prev_k = k; k = 0; i1 = i_min; i2 = i_max;
    for(i = i1, obj = res_obj; i < i2; i++, obj++)
    {
      j = (*obj)->index;
      
      if((*obj)->is_in_back)
        order[i_min++] = j;
      else if((*obj)->is_in_front)
        order[--i_max] = j;
      else
      {
        /* it was covered and it was covering some object, try again */
        res_obj[k] = (*obj);
        res_obj[k]->is_in_front = res_obj[k]->is_in_back = 1;
        k++;
      }
    }
    /* no element could be sorted at this stage,
       this is a Gordian knot */
    if(k == prev_k)
    {
      e_free(res_obj);
      return 0;
    }
  }

  /* fprintf(stderr, "\nNumber of sorting iterations %d", mikado_iter); */
  
  e_free(res_obj);
  return 1;
  
}










/* ######################################################################## */
/*                            Setup before sorting                          */
/* ######################################################################## */


#define Min_max(min, max, x)         \
        {                            \
          if((x) < min) (min) = (x); \
          if((x) > max) (max) = (x); \
        }


#define Min_max_vec(min_v, max_v, v)               \
        {                                          \
          Min_max((min_v)[0], (max_v)[0], (v)[0]); \
          Min_max((min_v)[1], (max_v)[1], (v)[1]); \
          Min_max((min_v)[2], (max_v)[2], (v)[2]); \
        }

#define copy_vec(r, v) \
        ((r)[0] = (v)[0], (r)[1] = (v)[1], (r)[2] = (v)[2])



void setup_proj_center(double azim, double elev, double dist, double *target)
{
  double a, c, x, e;

  Hidden_Proj_center[0] = 1;
  Hidden_Proj_center[1] = 1;
  Hidden_Proj_center[2] = 1;

  a = azim;
  /* protect the tangent calculation */
  if(a == 0 || a == M_PI)
    a += 1e-14;

  e = elev;

  /* elevation angle */
  while(e > 2*M_PI || e < -2*M_PI)
  {
    e += 2 * M_PI * (e > 0 ? -1 : 1);
  }
  if(e < 0)
    e += 2*M_PI;


  /* determine projection centre */
  if(e > 0.5*M_PI && e < 1.5*M_PI)
  {
    a += M_PI;
  }

  /* the azimuth angle */
  while(a > 2*M_PI || a < -2*M_PI)
  {
    a += 2 * M_PI * (a > 0 ? -1 : 1);
  }
  if(a < 0)
    a += 2*M_PI;


  Hidden_Proj_center[2] *= dist * sin(e);

/* this is only valid if the elevation rotation is
   around the y-axis */
/*
  if(a > 0.5*M_PI && a < 1.5*M_PI)
  {
    Hidden_Proj_center[0] = -1;
  }

  if(a > M_PI && a < 2*M_PI )
  {
    Hidden_Proj_center[1] = -1;
  }
*/


  if(a > M_PI && a < 2*M_PI)
  {
    Hidden_Proj_center[0] = -1;
  }

  if(   (a > 0 && a < 0.5*M_PI)
     || (a > 1.5*M_PI && a < 2*M_PI) )
  {
    Hidden_Proj_center[1] = -1;
  }

  a = a - M_PI/2;
  c = E_ABS(tan(a));
  
  x = sqrt((dist * dist - Hidden_Proj_center[2] * Hidden_Proj_center[2]) /
           (c * c + 1));
  
  Hidden_Proj_center[0] *= x;
  Hidden_Proj_center[1] *= c * x;

  Hidden_Proj_center[0] += target[0];
  Hidden_Proj_center[1] += target[1];
  Hidden_Proj_center[2] += target[2];

  Hidden_Distance = dist;

}




void setup_hidden_line1(int NPoly, POLYGON *pool)
{
  int i, j;

  double v1[3];
  double v2[3];
  double a;
  double r[3];

  for(i = 0; i < NPoly; i++)
  {

    /* setup the plane of the element */
    if(pool[i].no_vertices > 2)
    {
      SubVec3D(v1, pool[i].world_coords[0], pool[i].world_coords[1]);
      SubVec3D(v2, pool[i].world_coords[2], pool[i].world_coords[1]);

      pool[i].local_plane.normal[0] = v1[1] * v2[2] - v2[1] * v1[2];
      pool[i].local_plane.normal[1] = v2[0] * v1[2] - v1[0] * v2[2];
      pool[i].local_plane.normal[2] = v1[0] * v2[1] - v2[0] * v1[1];
      a = sqrt(pool[i].local_plane.normal[0] * pool[i].local_plane.normal[0] +
               pool[i].local_plane.normal[1] * pool[i].local_plane.normal[1] +
               pool[i].local_plane.normal[2] * pool[i].local_plane.normal[2]);
      pool[i].local_plane.normal[0] /= a;
      pool[i].local_plane.normal[1] /= a;
      pool[i].local_plane.normal[2] /= a;
    
      /* calculate the dot product of the normal and a point on the plane */
      pool[i].local_plane.cnst = DotProduct3D(pool[i].local_plane.normal, 
                                              pool[i].world_coords[0]     );
    }


    /* determine max radius for soap bubble  */
    /* and create 3D bounding boxes */
    SubVec3D(r, pool[i].world_coords[0], pool[i].center);
    pool[i].max_rad = Length(r);

    copy_vec(pool[i].box_min, pool[i].world_coords[0]);
    copy_vec(pool[i].box_max, pool[i].world_coords[0]);

    for(j = 1; j < pool[i].no_vertices; j++)
    {
      SubVec3D(r, pool[i].world_coords[j], pool[i].center);
      if( pool[i].max_rad < Length(r) )
      {
        pool[i].max_rad = Length(r);
      }

      Min_max_vec(pool[i].box_min, 
                  pool[i].box_max, 
                  pool[i].world_coords[j]);

    }
  }
}

/* projects the world points to the screen */
void setup_hidden_line2(int NPoly, POLYGON *pool)
{
  int i, j;
  double rotated[3];

  for(i = 0; i < NPoly; i++)
  {
    pool[i].obscured = 0;

    /* create 2D projected data */
    pool[i].proj_center[0] = 0.0;
    pool[i].proj_center[1] = 0.0;
    pool[i].proj_center[2] = 0.0;
    for(j = 0; j < pool[i].no_vertices; j++)
    {
      /* projection */
      p_3Dworld_to_screen(rotated, pool[i].world_coords[j]);
      
      /* center of rotated element */
      pool[i].proj_center[0] += rotated[0];
      pool[i].proj_center[1] += rotated[1];
      pool[i].proj_center[2] += rotated[2];
    }
    pool[i].proj_center[0] /= pool[i].no_vertices;
    pool[i].proj_center[1] /= pool[i].no_vertices;
    pool[i].proj_center[2] /= pool[i].no_vertices;

    /* projected contour */
    copy_vec(pool[i].proj_min, pool[i].screen_coords[0]);
    copy_vec(pool[i].proj_max, pool[i].screen_coords[0]);
    for(j = 1; j < pool[i].no_vertices; j++)
      Min_max_vec(pool[i].proj_min, 
                  pool[i].proj_max, 
                  pool[i].screen_coords[j]);

  }

}





