#include <assert.h>
#include "findmpr.h"
#include "nodelem.h"
#include "getvertices.h"

static double SegmentLength (MESH *mesh, int v1, int v2);

  
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 FindNodalMpr --
        Find nodal mesh parameters by examining lengts of the edges.

 This function visits all nodes and for each one calculates the max
 edge length adjacent to the vertex. The mesh parameter at that 
 vertex is set to the (1 + 0.1)*(max. edge length).

 CAUTION: it is assumed tha the mesh data structure has allocated
 array for the nodal mesh parameter.
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FindNodalMpr (MESH *mesh)
{
  int elem, nvert, i, node;
  int *vertices;
  int *nodeElemAdjLen = NULL;
  int **nodeElemAdj = NULL;
  int *vflag;
  double max_elength, elength;
  
  assert (mesh->MeshParam.NodalMeshParams != NULL);
  assert (mesh != NULL);

  /* allocate memory for adjacency matrices */
  nodeElemAdj    = e_init_frag_i_matr (mesh->NMeshPoints);
  nodeElemAdjLen = e_alloc_i_vect (mesh->NMeshPoints);
  vflag = e_alloc_i_vect (mesh->NMeshPoints);
  
  /* find nodes to elements adjacency of i-th input mesh */
  e_InitMeshPointElemConnection (mesh, nodeElemAdjLen, nodeElemAdj);
    
  for (node=0; node < mesh->NMeshPoints; node++)
  {
     vflag[node] = -1; /* set it to initial value */
  }

  /* for each node */
  for (node=0; node < mesh->NMeshPoints; node++)
  {
     max_elength = -1.0;
     
     /* for each element adjacent to the node */
     for (elem=0; elem <nodeElemAdjLen[node]; elem++)
     {
        /* get vertices of the element */
        e_GetElementVertices(nodeElemAdj[node][elem]-1, mesh, &nvert, &vertices);
        for (i=0; i<nvert; i++)
        {
          /* check vertex flag */
          if (vflag[vertices[i]-1] == node) 
          {
            /* node has not been visted */
            continue;
          }
          else  /* calculate edge length and compare with maximum length */
          {
            elength = SegmentLength(mesh, node, vertices[i]-1);
            if (elength > max_elength) 
            {
              max_elength = elength;
            }
            /* flag node as visited */
            vflag[vertices[i]-1] = node;
          } 
        } 
     }
     /* set meshparam to slightly biger value than the maximal edge length */
     mesh->MeshParam.NodalMeshParams[node] = max_elength + 0.01*max_elength;
  }   
  e_free ((void *) nodeElemAdjLen); 
  e_free ((void *) vflag);
  e_free_frag_i_matr (mesh->NNodes, nodeElemAdj); 
} /* end of FindNodalMpr */


/* calculate length of a segment between vertices v1 and v2 
   v1 and v2 should be specifed as 0 based.
 */
static double SegmentLength (MESH *mesh, int v1, int v2)
{
  double dx, dy, dz;
  double length;

  assert (v1 >=0 && v1 <mesh->NMeshPoints);
  assert (v2 >=0 && v2 <mesh->NMeshPoints);
  
  dx = mesh->MeshPointCoords[v1][0] - mesh->MeshPointCoords[v2][0];
  dy = mesh->MeshPointCoords[v1][1] - mesh->MeshPointCoords[v2][1];
  dz = mesh->MeshPointCoords[v1][2] - mesh->MeshPointCoords[v2][2];
  length = sqrt (dx*dx + dy*dy + dz*dz); 

  return length;
} /* end of SegmentLength */
