/* FILE trf.c
 *
 * Transform meshes
 *
 */

#include <math.h>
#include <e_lib.h>

#ifdef WIN32
#define strcasecmp(s1,s2) strcmpi(s1,s2)
#endif

#define  KW_NTRFS       "NTRFS"
#define  KW_ROT         "ROTATE"
#define	 KW_SCALE       "SCALE"
#define	 KW_MOVE        "MOVE"
#define  KW_L_ROT       "L_ROTATE"
#define	 KW_L_SCALE     "L_SCALE"
#define	 KW_L_MOVE	"L_MOVE"
#define  KW_A_ROT       "A_ROTATE"


#define  MODE_ROT     0
#define  MODE_SCALE   1
#define  MODE_MOVE    2
#define  MODE_A_ROT   3

#define  E_PI		3.14159265358979

void transform(int mode,double dx,double dy,double dz,double *coords);
void a_rotate(double dx,double dy,double dz,double w,double **crds,int n);

int main(int argc, char *argv[])
{
  char    trffn[E_LINE_MAX];
  char    meshfn[E_LINE_MAX];
  MESH    *mesh;
  FILE    *trffp;
  int	    nd, nc, i, trf, ntrf;
  char    line[E_LINE_MAX];
  char    word[E_LINE_MAX];
  int	    linenumber;
  double  dx, dy, dz, w;
  int     trf_geom=FALSE;
  int     mode;
  int			sn,en;
  
  if (argc<3)
  {
    printf("\nUsage: trf  mesh transformation [-N]\n");
    printf("   mesh           : name of the mesh definition file\n");
    printf("   transformation : name of the transformations file\n");
    printf("    -N            : also transform the geometric model\n");
    printf("All files should be specified without extension!\n\n");
    exit(1);
  }

  strcpy(trffn, argv[2]);
  strcat(trffn, ".trf");
  e_GetMeshData(&mesh, argv[1]);
  
  if (argc>3)
      if (strcasecmp(argv[3],"-N")==0)
      {
        trf_geom=TRUE;
        printf("\nGeometric model is also transformed\n");
      }
  
  if (trf_geom)
      e_GetGeomModel(mesh, argv[1]);
  
  if ((trffp=fopen(trffn, "r"))==NULL)
  {
    e_FileError(trffn, "cannot open transformations file");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  linenumber=0;
  if (!e_GetNextLine(line, &linenumber, trffn, trffp))
  {
    e_FileLineError(linenumber, trffn, "cannot read number of transformations keyword (1)");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  if (sscanf(line, "%s", word)!=1)
  {
    e_FileLineError(linenumber, trffn, "cannot read number of transformations keyword (2)");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  if (strcasecmp(KW_NTRFS, word)!=0)
  {
    e_FileLineError(linenumber, trffn, "cannot read number of transformations keyword (3)");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  if (sscanf(line, "%*s %d", &ntrf)!=1)
  {
    e_FileLineError(linenumber, trffn, "cannot read number of transformations");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  if (ntrf<1)
  {
    e_FileLineError(linenumber, trffn, "number of transformations should be > 0");
    e_FreeMeshData(mesh);
    fclose(trffp);
    exit(1);
  }
  
  for (trf=0;trf<ntrf;trf++)
  {
    if (!e_GetNextLine(line, &linenumber,trffn, trffp))
    {
      e_FileLineError(linenumber, trffn, "cannot read transformation keyword");
      e_FreeMeshData(mesh);
      fclose(trffp);
      exit(1);
    }
    if (sscanf(line, "%s", word)!=1)
    {
      e_FileLineError(linenumber, trffn, "cannot read transformation keyword");
      e_FreeMeshData(mesh);
      fclose(trffp);
      exit(1);
    }
    if (!((strcasecmp(KW_ROT, word)==0)||
    	  (strcasecmp(KW_SCALE, word)==0)||
          (strcasecmp(KW_MOVE, word)==0)||
          (strcasecmp(KW_L_ROT, word)==0)||
          (strcasecmp(KW_L_SCALE, word)==0)||
          (strcasecmp(KW_L_MOVE, word)==0)||
          (strcasecmp(KW_A_ROT, word)==0)))
    {
      e_FileLineError(linenumber, trffn, "unknown transformation keyword");
      e_FreeMeshData(mesh);
      fclose(trffp);
      exit(1);
    }
    if (sscanf(line, "%*s %lf %lf %lf", &dx, &dy, &dz)!=3)
    {
      e_FileLineError(linenumber, trffn, "cannot read transformation parameters");
      e_FreeMeshData(mesh);
      fclose(trffp);
      exit(1);
    }
    
    sn=0;
    en=mesh->NMeshPoints-1;
    
    if ((strcasecmp(KW_L_MOVE, word)==0)||
    	(strcasecmp(KW_L_ROT, word)==0)||
        (strcasecmp(KW_L_SCALE, word)==0))
    {
      if (sscanf(line, "%*s %*lf %*lf %*lf %d %d", &sn, &en)!=2)
      {
        e_FileLineError(linenumber, trffn, "cannot read transformation node limits");
        e_FreeMeshData(mesh);
        fclose(trffp);
        exit(1);
      }
      if ((sn<1)||(sn>mesh->NMeshPoints))
          sn=0;
      else
          sn--;
      if ((en<1)||(en>mesh->NMeshPoints))
          en=mesh->NMeshPoints-1;
      else
          en--;
    }
    
    if (strcasecmp(KW_A_ROT, word)==0)
    {
      if (sscanf(line, "%*s %*lf %*lf %*lf %lf", &w)!=1)
      {
        e_FileLineError(linenumber, trffn, "cannot read 4th axis rotation value");
        e_FreeMeshData(mesh);
        fclose(trffp);
        exit(1);
      }
      if ((w<-360)||(w>360))
      {
        e_FileLineError(linenumber, trffn, "rotation angles should be in [-360,360]");
        e_FreeMeshData(mesh);
        fclose(trffp);
        exit(1);
      }	
      mode=MODE_A_ROT;
    }
    
    if ((strcasecmp(KW_MOVE, word)==0)||
    	(strcasecmp(KW_L_MOVE, word)==0))
        mode=MODE_MOVE;    
    else if ((strcasecmp(KW_SCALE, word)==0)||
             (strcasecmp(KW_L_SCALE, word)==0))
        mode=MODE_SCALE;
    else if ((strcasecmp(KW_ROT, word)==0)||
             (strcasecmp(KW_L_ROT, word)==0))
    {
      if ((dx<-360)||(dx>360)||(dy<-360)||(dy>360)||(dz<-360)||(dz>360))
      {
        e_FileLineError(linenumber, trffn, "rotation angles should be in [-360,360]");
        e_FreeMeshData(mesh);
        fclose(trffp);
        exit(1);
      }
      mode=MODE_ROT;
    }
    
    if (mode==MODE_A_ROT)
    {
      a_rotate(dx,dy,dz,w,mesh->MeshPointCoords,mesh->NMeshPoints);
      if (trf_geom)
      {
        a_rotate(dx,dy,dz,w,mesh->Geom.EndPointCoords,mesh->Geom.NEndPoints);
        for (nc=0;nc<mesh->Geom.NNurbsCurves;nc++)
            a_rotate(dx,dy,dz,w,mesh->Geom.NurbsCurves[nc].CntlPoints,mesh->Geom.NurbsCurves[nc].Degree+1);
      }
    }
    else
    {
      for (nd=sn;nd<=en;nd++)
          transform(mode,dx,dy,dz,mesh->MeshPointCoords[nd]);
      if (trf_geom)
      {
        for (nd=0;nd<mesh->Geom.NEndPoints;nd++)
            transform(mode,dx,dy,dz,mesh->Geom.EndPointCoords[nd]);
        for (nc=0;nc<mesh->Geom.NNurbsCurves;nc++)
            for (nd=0;nd<mesh->Geom.NurbsCurves[nc].Degree+1;nd++)
                transform(mode,dx,dy,dz,mesh->Geom.NurbsCurves[nc].CntlPoints[nd]);
      }   
    }
  }
  
  fclose(trffp);
  strcpy(meshfn, argv[1]);
  strcat(meshfn, ".trf");
  e_PutMeshData(mesh, meshfn);
  if (trf_geom)
      e_PutGeomModel(mesh, meshfn);
  e_FreeMeshData(mesh);

  return 0;
}

void transform(int mode,double dx,double dy,double dz,double *coords)
{
  double  cx, cy, cz, sx, sy, sz;
  double  tx, ty, tz;
  
  if (mode==MODE_MOVE)
  {
    coords[0]+=dx;
    coords[1]+=dy;
    coords[2]+=dz;
  }
  else if (mode==MODE_SCALE)
  {
    coords[0]*=dx;
    coords[1]*=dy;
    coords[2]*=dz;
  }
  else if (mode==MODE_ROT)
  {
    cx=cos(dx*E_PI/180);
    sx=sin(dx*E_PI/180);
    cy=cos(dy*E_PI/180);
    sy=sin(dy*E_PI/180);
    cz=cos(dz*E_PI/180);
    sz=sin(dz*E_PI/180);
    /** rotation round X **/
    ty=coords[1]*cx - coords[2]*sx;
    tz=coords[1]*sx + coords[2]*cx;
    coords[1]=ty;
    coords[2]=tz;
    /** rotation round Y **/
    tx=coords[0]*cy + coords[2]*sy;
    tz=-coords[0]*sy + coords[2]*cy;
    coords[0]=tx;
    coords[2]=tz;
    /** rotation round Z **/
    tx=coords[0]*cz - coords[1]*sz;
    ty=coords[0]*sz + coords[1]*cz;
    coords[0]=tx;
    coords[1]=ty;
  }
  else if (mode==MODE_A_ROT)
  {
    e_GeneralError("MODE_A_ROT is not implemented");
  }
}

void a_rotate(double dx,double dy,double dz,double w,double **crds,int n)
{
  /* rotate around arbitrary axis */
  double **Rx,**Ry,**Rz,x,y,**Rfi,**Atmp,**Btmp;
  double  *vec, *rvec, d, fi; 
  int      i;

  Rx=e_alloc_cont_d_matr(4,4);
  Ry=e_alloc_cont_d_matr(4,4);
  Rz=e_alloc_cont_d_matr(4,4);
  Rfi=e_alloc_cont_d_matr(4,4);
  Atmp=e_alloc_cont_d_matr(4,4);
  Btmp=e_alloc_cont_d_matr(4,4);
  vec=e_alloc_d_vect(4);
  rvec=e_alloc_d_vect(4);

  fi = (w * E_PI)/180;
  
  d = sqrt(dx*dx +dy*dy +dz*dz);
  dx/=d;
  dy/=d;
  dz/=d;

  d = sqrt(dy*dy + dz*dz);
  
  Rx[0][0] = 1;
  Rx[0][1] = 0;
  Rx[0][2] = 0;
  Rx[0][3] = 0;

  Rx[1][0] = 0;
  if(d == 0)
  {
    Rx[1][1] = 1;
    Rx[1][2] = 0;
  }
  else
  {
    Rx[1][1] = dz/d;
    Rx[1][2] = dy/d;
  }
  Rx[1][3] = 0;  

  Rx[2][0] = 0;
  if(d == 0)
  {
    Rx[2][1] = 0;
    Rx[2][2] = 1;
  }
  else
  {
    Rx[2][1] = -dy/d;
    Rx[2][2] = dz/d;
  }
  Rx[2][3] = 0;

  Rx[3][0] = 0;
  Rx[3][1] = 0;
  Rx[3][2] = 0;
  Rx[3][3] = 1;
  
  Ry[0][0] = d;
  Ry[0][1] = 0;
  Ry[0][2] = dx;
  Ry[0][3] = 0;

  Ry[1][0] = 0;
  Ry[1][1] = 1;
  Ry[1][2] = 0;
  Ry[1][3] = 0;
  
  Ry[2][0] = -dx;
  Ry[2][1] = 0;
 
  Ry[2][2] = d;
  Ry[2][3] = 0;

  Ry[3][0] = 0;
  Ry[3][1] = 0;
  Ry[3][2] = 0;
  Ry[3][3] = 1;

  Rz[0][0] = cos(fi);
  Rz[0][1] = sin(fi);
  Rz[0][2] = 0;
  Rz[0][3] = 0;

  Rz[1][0] = -sin(fi);
  Rz[1][1] = cos(fi);
  Rz[1][2] = 0;
  Rz[1][3] = 0;  

  Rz[2][0] = 0;
  Rz[2][1] = 0;
  Rz[2][2] = 1;
  Rz[2][3] = 0;

  Rz[3][0] = 0;
  Rz[3][1] = 0;
  Rz[3][2] = 0;
  Rz[3][3] = 1;

  e_MatrMult(4, Rx, Ry, Btmp);
  e_MatrMult(4, Btmp, Rz, Atmp);
   
  if(d != 0)
  {
    x = Rx[1][1];
    y = Rx[1][2];
    
    Rx[1][1] = x/(x*x+y*y);
    Rx[1][2] = -y/(x*x+y*y);
    Rx[2][1] = y/(x*x+y*y);
    Rx[2][2] = x/(x*x+y*y);
  }
  if(dx != 0)
  {
    x = Ry[0][0];
    y = Ry[0][2];
    
    Ry[0][0] = x/(x*x+y*y);
    Ry[0][2] = -y/(x*x+y*y);
    Ry[2][0] = y/(x*x+y*y);
    Ry[2][2] = x/(x*x+y*y);
  }
  
  e_MatrMult(4, Atmp, Ry, Btmp);
  e_MatrMult(4, Btmp, Rx, Rfi);

  for (i=0;i<n;i++)
  {
    vec[0]=crds[i][0];
    vec[1]=crds[i][1];
    vec[2]=crds[i][2];
    vec[3]=1.0;
    e_VectMatrMult( 4, &vec[0], Rfi, &rvec[0]);
    crds[i][0]=rvec[0];
    crds[i][1]=rvec[1];
    crds[i][2]=rvec[2];
  }

  e_free_cont_d_matr(4,Rx);
  e_free_cont_d_matr(4,Ry);
  e_free_cont_d_matr(4,Rz);
  e_free_cont_d_matr(4,Rfi);
  e_free_cont_d_matr(4,Atmp);
  e_free_cont_d_matr(4,Btmp);
  e_free(vec);
  e_free(rvec);
}

/* end of file */
