/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              GetParameters()                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"

#ifndef __E_T800SAL__
#include <stdio.h>
#endif

#ifdef __E_T800__
#include <chan.h>
#endif

#include <string.h>
#include <e_lib.h>

#include "getparam.h"


/* ////////////////////////////////////////////////////////////////////////// */
/* GetParameters: To read in the RSB parameters. Called by "input".           */

/* The return value is 1 if no input errors occur, else 0. */


int GetParameters(int   TotalNElems,
                  int  *NSubdomains, 
                  int  *NStartMinNode,
                  int  *LinkGrab,
                  int **ElemsPerSubdomain,
                  char *InputFile          )
{
   if ((FullPathName = OpenInputFile(InputFile,"prm",&fp)) == NULL)
   {
      return 0;
   }
   else if ((GetSubdomainData(NSubdomains))                &&
            (GetMinNodeNumber(NStartMinNode))              &&
            (GetLinkGrabOption(LinkGrab))                  &&
            (GetIfUniformOption(TotalNElems, *NSubdomains, 
                                ElemsPerSubdomain))           )
   {
      e_free(FullPathName);
      fclose(fp);
      return 1;
   }
   else
   {
      e_free(FullPathName);
      fclose(fp);
      return 0;
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* OpenInputFile: To open an input file. The extension "extension" is added
   to the filename argument.                                                  */

/* The return value is a pointer to the full file pathname if the file is
   opened successfully, else NULL.                                            */


static char *OpenInputFile(char *FileName, char *extension, FILE **fp)
{
   int FullPathNameLength;
   char *FullPathName;


   FullPathNameLength = strlen(FileName) + strlen(extension) + 1;
   FullPathName = e_alloc_c_vect(FullPathNameLength + 1);

   if (FullPathName == NULL)
      return NULL;

   strcpy(FullPathName,FileName);
   strcat(FullPathName,".");
   strcat(FullPathName,extension);

   if ((*fp = fopen(FullPathName,"r")) == NULL)
   {
      e_FileError(FullPathName,"cannot open file");
      e_free(FullPathName);
      return NULL;
   }
   else
      return FullPathName;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetData: To read in the input data. */

/* The return value is 1 if the data are read successfully, else 0. */


static int GetSubdomainData(int *NSubdomains)
{
   char word[E_LINE_MAX + 1];        /* potential keyword */

   LineNumber = 0;

   if (   !GetKeyword(line,word,&LineNumber,FullPathName,fp)
       || !WordMatchesKeyword(KW_NSUBDOMAINS,word)
       || !GetNSubdomains(NSubdomains)                      )
   {
      return 0;
   }
   else
      return 1;
}

static int GetMinNodeNumber(int *NStartMinNode)
{
   char word[E_LINE_MAX + 1];        /* potential keyword */

   if (   !GetKeyword(line,word,&LineNumber,FullPathName,fp)
       || !WordMatchesKeyword(KW_STARTMINNODE,word)
       || !GetStartMinNode(NStartMinNode)                    )
   {
      return 0;
   }
   else
      return 1;
}

static int GetLinkGrabOption(int *LinkGrab)
{
   char word[E_LINE_MAX + 1];        /* potential keyword */

   if (   !GetKeyword(line,word,&LineNumber,FullPathName,fp)
       || !WordMatchesKeyword(KW_LINKGRAB,word)
       || !GetLinkGrab(LinkGrab)                             )
   {
      return 0;
   }
   else
      return 1;
}


static int GetIfUniformOption(int TotalNElems, int NSubdomains,
                              int **ElemsPerSubdomain)
{
   char word[E_LINE_MAX + 1];        /* potential keyword */

   if (   !GetKeyword(line,word,&LineNumber,FullPathName,fp)
       || !WordMatchesKeyword(KW_UNIFORM,word)
       || !GetIfUniform(TotalNElems, NSubdomains, ElemsPerSubdomain) )
   {
      return 0;
   }
   else if (e_GetLine(line,E_LINE_MAX,&LineNumber,fp) != EOF)
   {
      e_FileLineError(LineNumber,FullPathName,"EOF is expected");
      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* WordMatchesKeyword: To return whether a word is a specified keyword. */

/* The return value is 1 if the word matches the keyword, else 0. */


static int WordMatchesKeyword(char *Keyword, char *word)
{
   if (strcmp(Keyword,word) != 0)
   {
      if (IsAKeyword(word))
         e_FileLineError(LineNumber,FullPathName,"keyword out of sequence");
      else
         e_FileLineError(LineNumber,FullPathName,"undefined keyword");

      return 0;
   }
   else
      return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* Get... input functions: To read in data corresponding to a keyword. */

/* The return value is 1 if the data are read successfully, else 0. */

static int GetNSubdomains(int *NSubdomains)
{
   if (sscanf(line,"%*s %d",NSubdomains) == 1)
   {
      if (*NSubdomains < 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined number of subdomains");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read number of subdomains");
      return 0;
   }
   return 1;
}

static int GetStartMinNode(int *NStartMinNode)
{
   if (sscanf(line,"%*s %d", NStartMinNode) == 1)
   {
      if(*NStartMinNode < 1)
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined number of starting node");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read number of starting node");
      return 0;
   }
   return 1;
}


static int GetLinkGrab(int *LinkGrab)
{
   if (sscanf(line,"%*s %d", LinkGrab) == 1)
   {
      if((*LinkGrab != 0) &&
         (*LinkGrab != 1)    )
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined link-grab option");
         return 0;
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read link-grab option");
      return 0;
   }
   return 1;
}


static int GetIfUniform(int TotalNElems, int NSubdomains,
                        int **ElemsPerSubdomain)
{
   int Uniform, remainder, k;
   char *check;

   if (sscanf(line,"%*s %d", &Uniform) == 1)
   {
      if((Uniform != 0) &&
         (Uniform != 1)    )
      {
         e_FileLineError(LineNumber,FullPathName,
                         "ill-defined uniformity option");
         return 0;
      }
      if((*ElemsPerSubdomain = e_alloc_i_vect(NSubdomains)) == NULL)
        return 0;
      if(Uniform == 1)
      {
        remainder = TotalNElems % NSubdomains;

        for (k = 0; k < remainder; k++)
          (*ElemsPerSubdomain)[k] = TotalNElems/NSubdomains + 1;

        for ( ; k < NSubdomains; k++)
          (*ElemsPerSubdomain)[k] = TotalNElems/NSubdomains;
      }
      else
      {
        for(k = 0; k < NSubdomains; k++)
        {
          printf("\nNo of elements in Subdomain %d: ", k+1);
          check = fgets(line, E_LINE_MAX, stdin);
          if(   (check == NULL)
             || (sscanf(line, "%d", &((*ElemsPerSubdomain)[k])) != 1) )
            e_GeneralError("Can not read number of elements !");
        }
        for(k = 0; k < NSubdomains; k++)
          TotalNElems -= (*ElemsPerSubdomain)[k];
        if(TotalNElems != 0)
          e_GeneralError("Wrong total number of elements !");
      }
   }
   else
   {
      e_FileLineError(LineNumber,FullPathName,
                      "cannot read uniformity option");
      return 0;
   }
   return 1;
}



/* ////////////////////////////////////////////////////////////////////////// */
/* GetKeyword: To read in a potential keyword. */

/* The return value is 1 if a word is read successfully, else 0. */


static int GetKeyword(char *line, char *word, int *LineNumber,
                                  char *FullPathName, FILE *fp)
{
   if (!e_GetNextLine(line,LineNumber,FullPathName,fp))
      return 0;
   else if (sscanf(line,"%s",word) == 1)
      return 1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsAKeyword: To return whether a word is a keyword. */

/* The return value is 1 if "word" is a keyword, else 0. */

static int IsAKeyword(char *word)
{
   if(strcmp(KW_NSUBDOMAINS,word) == 0)
      return TRUE;
   else if(strcmp(KW_STARTMINNODE,word) == 0)
      return TRUE;
   else if(strcmp(KW_LINKGRAB,word) == 0)
      return TRUE;
   else if(strcmp(KW_UNIFORM,word) == 0)
      return TRUE;
   else
      return FALSE;
}

