/* 
   This file implements a sort-of heuristics for the shortest Hamiltonian 
   path around a node. The merge of several circles is really 
   quick-and-dirty implementation. Moreover, the whole program has not
   been optimised, it simply works. Sorry, I have no more time.
*/


/* 
There is to a special feature in the program !!!
If one edge is shared by more than two triangles there is a problem with
the original algorithm. (This can happen in the case of a frame structure, where
a wall can start from a concrete slab.)

               | <- wall
               |
               |____ <- slab
               |
               |


If the program detects that around the current node there is an edge
which is shared by more than two triangles, then switches a variable
to one (true). If the variable true, two things happen:
  - At the edge, all adjacent elements will be registered in such a way, which
tells to the program that they have two neighbours. Unfortunately one connection
is not stored, since there is not enough space for that in the structure. This
means that it is possible to jump from one element to the other in one direction
but not backwards.
  - The one-way jumping possibility causes the problem, that we cannot formulate
circles. To avoid this, we apply a trick. (It is allowed, since we are already sure
that there cannot be only _one_ circle, there should be more.) In this case those
link elements which has only one neighbour (a triangle) will be resetted as such,
which will indicate that they have no neighbour. Since when we write back the 
adjacency information to the original structure we do not delete connections which
are not existent in the local information, the resetted adjacency will not be lost.
By the reset there will be min 3 free nodes in the graph (corresponding to the links)
and now they can be connected as necessary, forming the circle. (This method handles
those cases as well when extra links are connected to the current node.)
Bingo !
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <e_lib.h>
#include "graph.h"

#define MAX_CIRCLE 16

typedef struct {
  int index;
  double dist;
} DISTANCE;

static int GCut; /* see comments above */


static int adj_sort_function(const void *a, const void *b);
static void assemble_graph_circle(int node, int *nlocal_adj, int **local_adj,
                                  int *degree, int **adjacency,
                                  int *no_node_elem, int **node_elem);
static void element_type(MESH *mesh, int elem, 
                         int ***nodes, int *no_nodes_in_elem, int *elem_index);
static void create_sphere(MESH *mesh, int node, double **centre,
                          int *no_node_elem, int **node_elem);
static int e_sort_function(const void *a, const void *b);
static double calc_distance(int from, int to, double **centre);

static void sort_points_by_distance(int at_node, int from_elem,
                                    double **centre, DISTANCE *distance,
                                    int *no_node_elem, int **node_elem);
static void merge_circles(int no_circle, int *circle_start, 
                          int **local_adj, double **centre);
static void write_circle_to_adjacency(int node, int *nlocal_adj, int **local_adj, 
                                      int *degree, int **adjacency,
                                      int *no_node_elem, int **node_elem);
static void setcircularadjacency(MESH *mesh, int *Degree, int **Adjacency,
                                 int *no_node_elem, int **node_elem);



static int adj_sort_function(const void *a, const void *b)
{
  if( *((int *) a) < *((int *) b) )
    return(-1);
  if( *((int *) a) > *((int *) b) )
    return(1);
  return(0);
}

void SetCircAdjacency(MESH *mesh, int *degree, int **adjacency,
                      int *no_node_elem, int **node_elem)
{
  int i;

  GV_handled = e_alloc_i_vect(mesh->TotalNElems);

  SetDualAdjacency(mesh, degree, adjacency,
                   no_node_elem, node_elem);

  setcircularadjacency(mesh, degree, adjacency, no_node_elem, node_elem);

  for(i = 0; i < mesh->TotalNElems; i++)
  {
    qsort(adjacency[i], degree[i], sizeof(int), adj_sort_function);
  }

  e_free(GV_handled);
}



void assemble_graph_circle(int node, int *nlocal_adj, int **local_adj,
                           int *degree, int **adjacency,
                           int *no_node_elem, int **node_elem)
{
  int j, i_adj, i_elem;
  int elem;

  GCut = 0;
  for(j = 0; j < no_node_elem[node]; j++)
  {
    elem = node_elem[node][j]-1;
    for(i_adj = 0; i_adj < degree[elem]; i_adj++)
    {
      for(i_elem = j+1; i_elem < no_node_elem[node]; i_elem++)
      {
        if(adjacency[elem][i_adj] == node_elem[node][i_elem])
        {
          if(nlocal_adj[j] != 2)
          {
            local_adj[j][nlocal_adj[j]] = i_elem;
            nlocal_adj[j]++;
          }
          else /* see comments above */
          {
            e_GeneralWarning("Element is adjacent to more than two element");
            GCut = 1;
          }

          if(nlocal_adj[i_elem] != 2)
          {
            local_adj[i_elem][nlocal_adj[i_elem]] = j;
            nlocal_adj[i_elem]++;
          }
          else /* see comments above */
          {
            e_GeneralWarning("Element is adjacent to more than two element");
            GCut = 1;
          }

          break;
        }
      }
    }
  }

  /* see comments above */
  if(GCut)
  {
    for(j = 0; j < no_node_elem[node]; j++)
    {
      if(nlocal_adj[j] == 1)
      {
        local_adj[j][0] = 0;
        nlocal_adj[j]--;
      }
    }
  }

}



static void element_type(MESH *mesh, int elem, 
                         int ***nodes, int *no_nodes_in_elem, int *elem_index)
{


/* determine element type */
  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK1]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK1] <= elem    )

  {
    (*nodes) = mesh->NodesLink1;
    (*no_nodes_in_elem) = 2;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_LINK1];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK2]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK2] <= elem    )

  {
    (*nodes) = mesh->NodesLink2;
    (*no_nodes_in_elem) = 2;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_LINK2];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK3]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK3] <= elem    )

  {
    (*nodes) = mesh->NodesLink3;
    (*no_nodes_in_elem) = 2;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_LINK3];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK4]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK4] <= elem    )

  {
    (*nodes) = mesh->NodesLink4;
    (*no_nodes_in_elem) = 2;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_LINK4];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_LINK5]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_LINK5] <= elem    )

  {
    (*nodes) = mesh->NodesLink5;
    (*no_nodes_in_elem) = 2;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_LINK5];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3] <= elem    )

  {
    (*nodes) = mesh->NodesTriang3;
    (*no_nodes_in_elem) = 3;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG3];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4] <= elem    )

  {
    (*nodes) = mesh->NodesTriang4;
    (*no_nodes_in_elem) = 3;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG4];
  }

  if(mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5]         &&
     mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5] <= elem    )

  {
    (*nodes) = mesh->NodesTriang5;
    (*no_nodes_in_elem) = 3;
    (*elem_index) = elem - mesh->FirstNetIndex[E_ELEM_TYPE_TRIANG5];
  }

}


static void create_sphere(MESH *mesh, int node, double **centre,
                          int *no_node_elem, int **node_elem)
{
  int   j, i_cent, i_dim;
  int **nodes;
  int   no_nodes_in_elem, elem_index;
  double length;

  for(j = 0; j < no_node_elem[node]; j++)
  {
    /* get element type and other info */
    element_type(mesh, node_elem[node][j], &(nodes),
                 &(no_nodes_in_elem), &(elem_index));

    /* determine the centre of the element - 
       would not be necessary with links */
    centre[j][0] = 0.00;
    centre[j][1] = 0.00;
    centre[j][2] = 0.00;
    for(i_cent = 0; i_cent < no_nodes_in_elem; i_cent++)
    {
      for(i_dim = 0; i_dim < 3; i_dim++)
      {
        centre[j][i_dim] +=
          mesh->MeshPointCoords[nodes[elem_index][i_cent]-1][i_dim];
      }
    }

    /* calculate the distance between the node and the element centre */
    length = 0.00;
    for(i_dim = 0; i_dim < 3; i_dim++)
    {
      centre[j][i_dim] = (centre[j][i_dim] / (double)no_nodes_in_elem) - 
                         mesh->MeshPointCoords[node][i_dim];
      length += E_SQUARE(centre[j][i_dim]);
    }

    /* normalise the distance */
    length = sqrt(length);
    for(i_dim = 0; i_dim < 3; i_dim++)
      centre[j][i_dim] /= length;
  }

}


static int e_sort_function(const void *a, const void *b)
{
  if( ((DISTANCE *)a)[0].dist < ((DISTANCE *)b)[0].dist )
    return(-1);
  if( ((DISTANCE *)a)[0].dist > ((DISTANCE *)b)[0].dist )
    return(1);
  return(0);
}


static double calc_distance(int from, int to, double **centre)
{
  int i_dim;
  double distance = 0.00;

  for(i_dim = 0; i_dim < 3; i_dim++)
    distance += E_SQUARE(centre[from][i_dim] - centre[to][i_dim]);
  distance = sqrt(distance);

  return distance;
}



static void sort_points_by_distance(int at_node, int from_elem,
                                    double **centre, DISTANCE *distance,
                                    int *no_node_elem, int **node_elem)
{
  int i_dist, i_dim, index = 0;

  for(i_dist = 0; i_dist < no_node_elem[at_node]; i_dist++)
  {
    if(i_dist != from_elem)
    {
      distance[index].index = i_dist;
      distance[index].dist  = calc_distance(from_elem, i_dist, centre);
      index++;
    }
  }

  qsort(distance, no_node_elem[at_node]-1, sizeof(DISTANCE), 
        e_sort_function);
}


static void merge_circles(int no_circle, int *circle_start, 
                          int **local_adj, double **centre)
{
  int i;
  int max_circle, max_1, max_2, min_circle, min_1, min_2;
  int start, current, old, next;
  int other_max, index_1, index_2;
  double max_length, big_length, length;
  double length_1, length_2, length_11, length_12, length_21, length_22;
  short  rotate;

  while(no_circle > 1)
  {
    /* search for the longest connection to be cut */
    max_circle = -1;
    max_length = -1;
    for(i = 0; i < no_circle; i++)
    {
      start = circle_start[i];
      current = start;
      old = -1;
      do
      {
        next = local_adj[current][0];
        if(next == old)
          next = local_adj[current][1];
        length = calc_distance(current, next, centre);
        if(max_length < length)
        {
          max_length = length;
          max_circle = i;
          max_1 = current;
          max_2 = next;
        }
        old = current;
        current = next;
      }
      while (current != start);
    }

    /* search for the second longest length */
    big_length = -1;
    /* search for closest node in all other circles, 
       at both end of the longest edge */
    for(i = 0; i < no_circle; i++)
    {
      if(i != max_circle)
      {
        start = circle_start[i];
        current = start;
        old = -1;
        do
        {
          next = local_adj[current][0];
          if(next == old)
            next = local_adj[current][1];
          length = calc_distance(current, next, centre);
          if(big_length < length)
          {
            big_length = length;
            min_circle = i;
            min_1 = current;
            min_2 = next;
          }
          old = current;
          current = next;
        }
        while (current != start);
      }
    }

    rotate = 0;

    length_11 = calc_distance(min_1, max_1, centre);
    length_12 = calc_distance(min_2, max_2, centre);
    length_1  = length_11 + length_12;

    length_21 = calc_distance(min_1, max_2, centre);
    length_22 = calc_distance(min_2, max_1, centre);
    length_2  = length_21 + length_22;

    /* we want connect max_1 -> min_1 and max_2 -> min_2 */
    /* select the shorter new connection */
    if(length_2 < length_1)
    {
      other_max = min_1;
      min_1 = min_2;
      min_2 = other_max;
      rotate = !rotate;
    }

    /* but if the first combination contains a closing arc, 
       that is better, so reverse the exchange */
    if(((E_ABS(length_11 - 2.0) < 1e-6) ||
        (E_ABS(length_12 - 2.0) < 1e-6)    ) && rotate)
    {
      other_max = min_1;
      min_1 = min_2;
      min_2 = other_max;
      rotate = !rotate;
    }

    /* first combination shorter, but second contains a closing arc,
       if it was not exchanged then do it */
    if(((E_ABS(length_21 - 2.0) < 1e-6) ||
        (E_ABS(length_22 - 2.0) < 1e-6)    ) && !rotate)
    {
      other_max = min_1;
      min_1 = min_2;
      min_2 = other_max;
      rotate = !rotate;
    }

    /* cut and connect the two circle */
    if(local_adj[max_1][0] == max_2)
      index_1 = 0;
    else if(local_adj[max_1][1] == max_2)
      index_1 = 1;

    if(local_adj[max_2][0] == max_1)
      index_2 = 0;
    else if(local_adj[max_2][1] == max_1)
      index_2 = 1;

    local_adj[max_1][index_1] = min_1;
    local_adj[max_2][index_2] = min_2;

    if(local_adj[min_1][0] == min_2)
      index_1 = 0;
    else if(local_adj[min_1][1] == min_2)
      index_1 = 1;

    if(local_adj[min_2][0] == min_1)
      index_2 = 0;
    else if(local_adj[min_2][1] == min_1)
      index_2 = 1;

    local_adj[min_1][index_1] = max_1;
    local_adj[min_2][index_2] = max_2;

    /* get out the second circle and reduce their number */
    for(i = 0; i < no_circle; i++)
    {
      if(i > min_circle)
        circle_start[i-1] = circle_start[i];
    }
    no_circle--;
  }


}


static void write_circle_to_adjacency(int node, int *nlocal_adj, int **local_adj, 
                                      int *degree, int **adjacency,
                                      int *no_node_elem, int **node_elem)
{
  int i_node, elem, i_s;

  for(i_node = 0; i_node < no_node_elem[node]; i_node++)
  {
    elem = node_elem[node][i_node]-1;

    /* theoretically it could not find itself in local_adj */

    /* check first neighbour */
    /* this is only when a link is skipped */
    if(local_adj[i_node][0] >= 0)
    {
      for(i_s = 0; i_s < degree[elem]; i_s++)
        if(adjacency[elem][i_s] == node_elem[node][local_adj[i_node][0]])
          break;
      if(i_s == degree[elem])
      {
        adjacency[elem] = e_incr_i_vect( degree[elem],
                                         adjacency[elem], 1 );
        adjacency[elem][degree[elem]] = node_elem[node][local_adj[i_node][0]];
        /* another crazy check */
        if(adjacency[elem][degree[elem]] == elem+1)
          e_GeneralError("\nTries to store itself: %d, node %d", 
                         elem+1, node+1);

        degree[elem]++;
      }
    }

    /* check second neighbour */
    /* this is only when a link is skipped */
    if(local_adj[i_node][1] >= 0)
    {

      for(i_s = 0; i_s < degree[elem]; i_s++)
        if(adjacency[elem][i_s] == node_elem[node][local_adj[i_node][1]])
          break;
      if(i_s == degree[elem])
      {
        adjacency[elem] = e_incr_i_vect( degree[elem],
                                         adjacency[elem], 1 );
        adjacency[elem][degree[elem]] = node_elem[node][local_adj[i_node][1]];

        if(adjacency[elem][degree[elem]] == elem+1)
          e_GeneralError("\nTries to store itself: %d, node %d", 
                         elem+1, node+1);

        degree[elem]++;
      }

    }


  }
}


static void setcircularadjacency(MESH *mesh, int *Degree, int **Adjacency,
                                 int *no_node_elem, int **node_elem)
{
  int       i, j, k, kk, ic, counter;
  int     **local_adj;
  int      *nlocal_adj;
  double  **centre;
  DISTANCE *distance;
  int       no_link, current_elem, old_elem, new_elem;
  int       non_part, unhandled, no_circle;
  double    dist_1, dist_2;
  int       insert_elem, neighbour_1, neighbour_2;
  char      message[E_LINE_MAX];
  int       circle_start[MAX_CIRCLE];
  int      *circle_check;
  int       start;


  no_link = mesh->NElemsLink1 + mesh->NElemsLink2 + mesh->NElemsLink3 +
            mesh->NElemsLink4 + mesh->NElemsLink5;

  local_adj     = e_alloc_cont_i_matr(mesh->TotalNElems, 2);
  nlocal_adj    = e_alloc_i_vect(mesh->TotalNElems);
  centre        = e_alloc_cont_d_matr(mesh->TotalNElems, 3);
  distance      = e_calloc(mesh->TotalNElems, sizeof(DISTANCE));
  circle_check  = e_alloc_i_vect(mesh->TotalNElems);

  for(i = 0; i < mesh->NMeshPoints; i++)
  {

    /* simple check whether any unhandled/unfinished link */
    counter = 0;
    for(ic = 0; ic < no_node_elem[i]; ic++)
    {
      if(GV_handled[node_elem[i][ic]-1] == 0)
        counter++;
    }

    /* unfinished elements or only one element -> skip */
    if((counter == 0) || (no_node_elem[i] < 2))
      continue;

    no_circle = 0;
    unhandled = 0;

    /* clear */
    memset(nlocal_adj,   0, sizeof(int)*mesh->TotalNElems);
    memset(local_adj[0], 0, sizeof(int)*mesh->TotalNElems*2);
    memset(circle_check, 0, sizeof(int)*mesh->TotalNElems);


    /* when I am an idiot I need THIS for DEBUG_INFINITY level :-) */

    {
      int kk;
      for(kk = 0; kk < no_node_elem[i]; kk++)
      {
        local_adj[kk][0] = -1;
        local_adj[kk][1] = -1;
      }
    }

    /* assemble the already existing graph */
    assemble_graph_circle(i, nlocal_adj, local_adj,
                          Degree, Adjacency, no_node_elem, node_elem);

#if 0
    /* count not fully connected elements */
    for(j = 0; j < no_node_elem[i]; j++)
    {
      if(nlocal_adj[j] != 2)
        unhandled++;
    }
#endif

    /* there is/are unhandled element(s) and 
       there are more than 1 element at the meshpoint*/
/*
    if((unhandled != 0) && (no_node_elem[i] > 1))
*/
    {

      /* only if there was no cut off */

      if(GCut == 0)
      {

        /* check for any existing closed circle, these are not considered before
           and might play an important role later */
        while(1)
        {
          /* find any element with two connection and not checked yet */
          for(j = 0; j < no_node_elem[i]; j++)
          {
            if((nlocal_adj[j] == 2) && (circle_check[j] == 0))
            {
              start = j;
              circle_start[no_circle] = j;
              circle_check[j] = 1;
              current_elem = local_adj[j][0];
              if(current_elem == j)
                current_elem = local_adj[j][1];
              old_elem = j;
              break;
            }
          }

          /* no such element, finish this search */
          if(j == no_node_elem[i])
            break;

          /* search until elem has two connection and circle is not closed */
          while((nlocal_adj[current_elem] == 2) && (current_elem != start))
          {
            circle_check[current_elem] = 1;
            new_elem = local_adj[current_elem][0];
            if(new_elem == old_elem)
              new_elem = local_adj[current_elem][1];
            old_elem = current_elem;
            current_elem = new_elem;
          }

          /* if circle was closed warning, and count it */
          if(current_elem == start)
          {
            e_GeneralWarning("Extra circle detected !");
            no_circle++;
          }
        }

      }


      /* create geometric information around the node */
      create_sphere(mesh, i, centre, no_node_elem, node_elem);

      /* unfortunately we are redoing the circle, because there is
      a special case, when there is a circle and one unhandled element.
      In that case the algorithm would not detect the existing circle. */
      non_part = no_node_elem[i];

      /* select the current element, the starting element
         which is not part of a circle */
      for(j = 0; j < no_node_elem[i]; j++)
      {
        if(circle_check[j] == 0)
        {
          current_elem = j;
          break;
        }
      }

      /* go until only one or no element left out */
      while(non_part > 1)
      {
        circle_start[no_circle] = current_elem;
        old_elem = -1;
        do
        {
          if(nlocal_adj[current_elem] == 2)
          {
            new_elem = local_adj[current_elem][0];
            if(new_elem == old_elem)
              new_elem = local_adj[current_elem][1];
            old_elem = current_elem;
          }
          else
          {
            /* unfinished link */
            if(node_elem[i][current_elem] <= no_link)
            {
              sort_points_by_distance(i, current_elem, centre, distance,
                                      no_node_elem, node_elem);

              /* selection of the next neighbour by two strategy */
              for(k = 0; k < no_node_elem[i]-1; k++)
              {
                kk = distance[k].index;
                /* link and has less than two neighbour and does connect back
                    to the current_elem */
                if((node_elem[i][kk] <= no_link)               &&
                   ((nlocal_adj[kk] == 0)                    ||
                    ((nlocal_adj[kk] == 1)              &&
                     (local_adj[kk][0] != current_elem)    )    )    )
                {
                  local_adj[current_elem][nlocal_adj[current_elem]] = kk;
                  local_adj[kk][nlocal_adj[kk]] = current_elem;
                  nlocal_adj[current_elem]++;
                  nlocal_adj[kk]++;
                  break;
                }
              }

              if(k == no_node_elem[i]-1)
              {
                for(k = 0; k < no_node_elem[i]-1; k++)
                {
                  kk = distance[k].index;
                  /* finally, link and has less than two neighbour */
                  if((nlocal_adj[kk] != 2)            && 
                     (node_elem[i][kk] <= no_link)    )
                  {
                    local_adj[current_elem][nlocal_adj[current_elem]] = kk;
                    local_adj[kk][nlocal_adj[kk]] = current_elem;
                    nlocal_adj[current_elem]++;
                    nlocal_adj[kk]++;
                    break;
                  }
                }
                /* link and we could not put it into any other circle and could not
                   find any other element to connect */
                if(k == no_node_elem[i]-1)
                {
                  e_GeneralWarning("Single element (%d) without connection !", 
                                   current_elem);

                  /* There are more than one element at the node, but we could not
                     place this link, that means that there are closed circles
                     at the node (one or more). Since we already detected the closed
                     circles here it should be decreased by one, because we set
                     current_elem to another element which will go through a circle ! */
                  /* when we break it will be incremented outside of the loop, 
                     so decrease now to keep the correct number of circles */
                  if(no_circle)
                    no_circle--;

                  /* Theoretically we can break, because we already counted the circles
                     and there is no way that we find an available element, that is
                     why we are here */
                  break;

#if 0
                  /* this is an old version where we go through another circle */
                  current_elem++;


                  circle_start[no_circle] = current_elem;
                  if(nlocal_adj[current_elem] != 2)
                    e_GeneralError("End could not be found ! Impossible !");
                  kk = local_adj[current_elem][0];
#endif
                }
              }

              old_elem = current_elem;
              new_elem = kk;
            }
            else
            {
              printf("\n\nCurrent elem: %d", current_elem);
              printf("\nnode : %d   no: %d", i+1, no_node_elem[i]);
              e_GeneralError("Triangle without two neighbours on the sides !");
            }
          }
          current_elem = new_elem;
        } while (current_elem != circle_start[no_circle]);

        no_circle++;

        if(no_circle == MAX_CIRCLE)
        {
          e_GeneralError("Maximum number of circles exceeded. Increase MAX_CIRCLE.");
        }

        non_part = 0;
        for(j = 0; j < no_node_elem[i]; j++)
        {
          if(nlocal_adj[j] != 2)
          {
            non_part++;
            if(non_part == 1)
              current_elem = j;
          }
        }
      }


      if(no_circle == 0)
      {
        sprintf(message, "No circle at node %d !", i);
        e_GeneralError(message);
      }

      /* there are more than one circle, merge them */
      if(no_circle != 1)
      {
        sprintf(message, "Number of circle at node %d: %d", i+1, no_circle);
        e_GeneralWarning(message);

        merge_circles(no_circle, circle_start, local_adj, centre);
      }


      /* by now there is only one circle, insert the left-out element
         which is stored in current_elem */
      if(non_part == 1)
      {
        if(nlocal_adj[current_elem] != 0)
        {
          /* it was a warning, but it cannot happen, because we are cutting the links
             to form a separate circle */
          sprintf(message, "Single left-out element %d, skipped with connection",
                  node_elem[i][current_elem]);
          e_GeneralError(message);

#if 0
          local_adj[current_elem][1] = -1;
          nlocal_adj[current_elem] = 2;
#endif

        }
        else
        {
          /* only there was no cut off */
          if(GCut == 0)
          {
            sprintf(message, "Single left-out element: %d", node_elem[i][current_elem]);
            e_GeneralWarning(message);

            sort_points_by_distance(i, current_elem, centre, distance,
                                    no_node_elem, node_elem);

            /* check that we do not connect to ourself !!! */
            kk = distance[0].index;
            neighbour_1 = local_adj[kk][0];
            neighbour_2 = local_adj[kk][1];

            dist_1  = calc_distance(neighbour_1,  current_elem, centre);
            dist_1 += calc_distance(current_elem, kk,           centre);
            dist_1 += calc_distance(kk,           neighbour_2,  centre);

            dist_2  = calc_distance(neighbour_2,  current_elem, centre);
            dist_2 += calc_distance(current_elem, kk,           centre);
            dist_2 += calc_distance(kk,           neighbour_1,  centre);

            insert_elem = (dist_1 < dist_2 ? neighbour_1 : neighbour_2);

            local_adj[current_elem][0] = kk;
            local_adj[current_elem][1] = insert_elem;
            nlocal_adj[current_elem] = 2;

            if(local_adj[kk][0] == insert_elem)
              local_adj[kk][0] = current_elem;
            else
              local_adj[kk][1] = current_elem;

            if(local_adj[insert_elem][0] == kk)
              local_adj[insert_elem][0] = current_elem;
            else
              local_adj[insert_elem][1] = current_elem;
          }
          else
          {
            /* link is totally skipped, nothing we can do */
            nlocal_adj[current_elem] = 2;
            local_adj[current_elem][0] = -1;
            local_adj[current_elem][1] = -1;
          }
        }
      }

      /* !!! sanity check !!! */
      for(j = 0; j < no_node_elem[i]; j++)
      {
        if((node_elem[i][j] <= no_link) && (nlocal_adj[j] != 2))
          e_GeneralError("Left-out element");
      }

      write_circle_to_adjacency(i, nlocal_adj, local_adj, Degree, Adjacency,
                                no_node_elem, node_elem);

    }
  }

  /* free up locally allocated memory */
  e_free_cont_i_matr(mesh->NMeshPoints, local_adj);
  e_free(nlocal_adj);
  e_free_cont_d_matr(mesh->NMeshPoints, centre);
  e_free(distance);
  e_free(circle_check);

}



