#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "ga.h"

/* to understand the code see

  David E. Goldberg
	Genetic Algorithms in Search, Optimization,
	and Machine Learning
  ISBN 0-201-15767-5

*/

/***********************************************/
/***********************************************/
/***********************************************/
/*****   Nothing should be changed below   *****/
/***********************************************/
/***********************************************/
/***********************************************/


#define L_O_CHROM_TYPE  (sizeof(CHROM_TYPE)*8)

/* for scaling the fitnesses */
#define FMULTIPLE       2.00

/* when we are doing the scaling of the fitnesses, the difference
   can be zero. To avoid division by zero we use this small constant */
#define ABS_ZERO        1e-14

typedef CHROM_TYPE *CHROMOSOME;

typedef struct
{
  CHROMOSOME chrom;
  double fitness;
  int parent1;
  int parent2;
  int xsite;
} INDIVIDUAL;

typedef INDIVIDUAL *POPULATION;


/* --------------- functions ----------------- */

double p_rand();
int p_flip(double probability);
long p_rnd(long low, long high);
long p_select_roulette_wheel(long popsize, double sumfitness,
			       POPULATION pop);
void p_crossover(CHROMOSOME parent1, CHROMOSOME parent2,
		   CHROMOSOME child1, CHROMOSOME child2,
		   long lchrom, long *ncross, long *nmutation,
		   long *jcross, double pcross, double pmutation);
CHROM_TYPE p_mutation(CHROM_TYPE alleleval, double pmutation, long *nmutation);
void p_generation(POPULATION oldpop, POPULATION newpop,
		    long popsize, double sumfitness,
		    long lchrom, long *ncross, long *nmutation,
		    long *jcross, double pcross, double pmutation,
		    long npars, DESIGNPARAMS param);
void p_decode_params(long lchrom, CHROMOSOME chrom,
		       long npars, DESIGNPARAMS param);
void p_prescale(double umax, double uavg, double umin,
		  double *a, double *b);
void p_scalepop(long popsize, double max, double avg, double min,
		  double *sumfitness, POPULATION pop);
void p_statistics(long popsize, double *max, double *avg,
		    double *min, double *sumfitness,
		    POPULATION pop, long *max_number);
void p_initpop(long popsize, long lchrom, POPULATION oldpop,
	         long npars, DESIGNPARAMS param);
void p_print_stat(long gen, double max, double min, double avg, POPULATION pop,
      long popsize, long lchrom, long ncross, long nmutation, int printing);


/* ----------------------- */
/* random number generator */
/* ----------------------- */

double p_rand()
{
  return (double) rand() / (double) RAND_MAX;
}				/* end of p_rand */


/* ------------------------------------------ */
/* genetic algorithm functions                */
/* ------------------------------------------ */



int p_flip(double probability)	/* fair coin flipping */
{
  if (probability == 1.00)
    return 1;
  else
    return (p_rand() <= probability);
}				/* end of p_flip */

long p_rnd(long low, long high)
{
  if (low >= high)
    return low;
  else
  {
    return (long) (p_rand() * ((double) (high) - (double) (low))) + low;
  }
}				/* end of p_rnd */


long p_select_roulette_wheel(long popsize, double sumfitness,
			       POPULATION pop)
{
  double rand, partsum;
  long j;

  partsum = 0.00;
  j = 0;
  rand = p_rand() * sumfitness;	/* wheel point calculation */
  do				/* find the wheel slot */
  {
    partsum += pop[j].fitness;
    j++;
  } while ((partsum < rand) && (j != popsize));
  return (j - 1);		/* return the individual number */
}				/* end of p_select_roulette_wheel */


void p_crossover(CHROMOSOME parent1, CHROMOSOME parent2,
		   CHROMOSOME child1, CHROMOSOME child2,
		   long lchrom, long *ncross, long *nmutation,
		   long *jcross, double pcross, double pmutation)
{
  long j, n_bits, rest_chrom, i;
  char mask1, mask2;

  if (p_flip(pcross))
  {
    (*jcross) = p_rnd(0, lchrom - 1);	/* cross between 1 and length-1 */
    (*ncross)++;		/* increment crossover counter */
  }
  else
    (*jcross) = lchrom;

  for (j = 0; j < ((*jcross) / L_O_CHROM_TYPE); j++)
  {
    child1[j] = p_mutation(parent1[j], pmutation, nmutation);
    child2[j] = p_mutation(parent2[j], pmutation, nmutation);
  }

/* if there is crossover */
  if ((*jcross) != lchrom)
  {
    n_bits = (*jcross) % L_O_CHROM_TYPE;

    mask1 = 0;
    for (i = 0; i < n_bits; i++)
      mask1 |= 1 << i;

    mask2 = 0;
    for (i = n_bits; i < L_O_CHROM_TYPE; i++)
      mask2 |= 1 << i;

    child1[j] = (parent1[j] & mask1) | (parent2[j] & mask2);
    child2[j] = (parent2[j] & mask1) | (parent1[j] & mask2);

    child1[j] = p_mutation(child1[j], pmutation, nmutation);
    child2[j] = p_mutation(child2[j], pmutation, nmutation);

    rest_chrom = (lchrom / L_O_CHROM_TYPE) + ((lchrom % L_O_CHROM_TYPE) ? 1 : 0);

    for (j = ((*jcross) / L_O_CHROM_TYPE) + 1; j < rest_chrom; j++)
    {
      child1[j] = p_mutation(parent2[j], pmutation, nmutation);
      child2[j] = p_mutation(parent1[j], pmutation, nmutation);
    }
  }
  else
  {
    if (lchrom % L_O_CHROM_TYPE)
    {
      child1[j] = p_mutation(parent1[j], pmutation, nmutation);
      child2[j] = p_mutation(parent2[j], pmutation, nmutation);
    }
  }
}				/* end of p_crossover */


CHROM_TYPE p_mutation(CHROM_TYPE alleleval, double pmutation, long *nmutation)
{
  int i;
  CHROM_TYPE mask;

  for (i = 0; i < L_O_CHROM_TYPE; i++)
  {
    if (p_flip(pmutation))
    {
      (*nmutation)++;
      mask = 1 << i;
      alleleval ^= mask;
    }
  }
  return alleleval;
}				/* end of p_mutation */

/* -------------------------------------------------------- */


void p_generation(POPULATION oldpop, POPULATION newpop,
		    long popsize, double sumfitness,
		    long lchrom, long *ncross, long *nmutation,
		    long *jcross, double pcross, double pmutation,
		    long npars, DESIGNPARAMS param)
{
  long mate1, mate2, j, i, len;

  for (j = 0; j < popsize; j += 2)
  {
    if(popsize-1 == j)
    {
/* selection for the last */
      mate1 = p_select_roulette_wheel(popsize, sumfitness, oldpop);

      len = (lchrom / L_O_CHROM_TYPE) + 
            ((lchrom % L_O_CHROM_TYPE) ? 1 : 0);

      for (i = 0; i < len; i++)
      {
        newpop[j].chrom[i] = p_mutation(oldpop[mate1].chrom[i], pmutation, nmutation);
      }
      p_decode_params(lchrom, newpop[j].chrom, npars, param);
      newpop[j].fitness = objective_function(npars, param);
      newpop[j].parent1 = j;
      newpop[j].parent2 = j;
      newpop[j].xsite = lchrom;
    }
    else
    {
/* selection for crossover and mutation */
      mate1 = p_select_roulette_wheel(popsize, sumfitness, oldpop);
      mate2 = p_select_roulette_wheel(popsize, sumfitness, oldpop);

/* crossover and mutation */
      p_crossover(oldpop[mate1].chrom, oldpop[mate2].chrom,
		  newpop[j].chrom, newpop[j + 1].chrom,
		  lchrom, ncross, nmutation, jcross, pcross, pmutation);
/* decode & evaluate */
      p_decode_params(lchrom, newpop[j].chrom, npars, param);
      newpop[j].fitness = objective_function(npars, param);
      newpop[j].parent1 = mate1;
      newpop[j].parent2 = mate2;
      newpop[j].xsite = (*jcross);

      p_decode_params(lchrom, newpop[j + 1].chrom, npars, param);
      newpop[j + 1].fitness = objective_function(npars, param);
      newpop[j + 1].parent1 = mate1;
      newpop[j + 1].parent2 = mate2;
      newpop[j + 1].xsite = (*jcross);
    }
  }				/* end of for */
}				/* end of p_generation */


void p_decode_params(long lchrom, CHROMOSOME chrom,
		       long npars, DESIGNPARAMS param)
{
  long j, k, jposition;
  double powerof2;
  CHROM_TYPE chrom_part;

  jposition = 0;
  for (j = 0; j < npars; j++)
  {
    powerof2 = 1;
    param[j].parameter = 0.00;
    if (param[j].lparam)
    {
      for (k = 0; k < param[j].lparam; k++)
      {
	if (!(jposition % L_O_CHROM_TYPE))
	  chrom_part = chrom[jposition / L_O_CHROM_TYPE];

	if (chrom_part & (1 << (jposition % L_O_CHROM_TYPE)))
	  param[j].parameter += powerof2;
	powerof2 *= 2;
	jposition++;
      }
    }
    param[j].parameter = param[j].minparam +
      (param[j].maxparam - param[j].minparam) /
      (pow(2.00, param[j].lparam) - 1) * param[j].parameter;
  }
}				/* end of p_decode_params */


void p_prescale(double umax, double uavg, double umin,
		  double *a, double *b)
{
  double delta;

  if (umin > ((FMULTIPLE * uavg - umax) / (FMULTIPLE - 1.00)))
  {
    delta = umax - uavg;
    if (delta < ABS_ZERO)
      delta = ABS_ZERO;
    *a = (FMULTIPLE - 1.00) * uavg / delta;
    *b = uavg * (umax - FMULTIPLE * uavg) / delta;
  }
  else
  {
    delta = uavg - umin;
    if (delta < ABS_ZERO)
      delta = ABS_ZERO;
    *a = uavg / delta;
    *b = -umin * uavg / delta;
  }
}				/* end of p_prescale */

void p_scalepop(long popsize, double max, double avg, double min,
		  double *sumfitness, POPULATION pop)
{
  long j;
  double a, b;

  p_prescale(max, avg, min, &a, &b);
  (*sumfitness) = 0.00;
  for (j = 0; j < popsize; j++)
  {
    pop[j].fitness = a * pop[j].fitness + b;
    (*sumfitness) += pop[j].fitness;
  }
}				/* end of p_scalepop */


void p_statistics(long popsize, double *max, double *avg,
		    double *min, double *sumfitness,
		    POPULATION pop, long *max_number)
{
  long j;

  (*sumfitness) = pop[0].fitness;
  (*min) = pop[0].fitness;
  (*max) = pop[0].fitness;
  (*max_number) = 0;

  for (j = 1; j < popsize; j++)
  {
    (*sumfitness) += pop[j].fitness;
    if (pop[j].fitness > (*max))
    {
      (*max) = pop[j].fitness;
      (*max_number) = j;
    }
    if (pop[j].fitness < (*min))
      (*min) = pop[j].fitness;
  }
  (*avg) = (*sumfitness) / (double) popsize;
}				/* end of p_statistics */


void p_initpop(long popsize, long lchrom, POPULATION oldpop,
	         long npars, DESIGNPARAMS param)
{
  long j, k, i, no_chrom;
  CHROM_TYPE mask;

  no_chrom = (lchrom / L_O_CHROM_TYPE) + ((lchrom % L_O_CHROM_TYPE) ? 1 : 0);

  for (j = 0; j < popsize; j++)
  {
    for (k = 0; k < no_chrom; k++)
      for (i = 0; i < L_O_CHROM_TYPE; i++)
	if (p_flip(0.5))
	{
	  mask = 1 << i;
	  oldpop[j].chrom[k] ^= mask;
	}
    p_decode_params(lchrom, oldpop[j].chrom, npars, param);
    oldpop[j].fitness = objective_function(npars, param);
  }
}				/* end of p_initpop */


void genetic_algorithm(long popsize, double pcross, double pmutation,
		  long npars, DESIGNPARAMS param, long maxgen, int scaling, int printing)
{
  long gen, lchrom, ncross = 0, nmutation = 0, jcross;
  double max, avg, min, sumfitness;
  long i, alloc_chrom, max_number;
  POPULATION oldpop, newpop, temppop;

  oldpop = (POPULATION) calloc(popsize, sizeof(INDIVIDUAL));
  newpop = (POPULATION) calloc(popsize, sizeof(INDIVIDUAL));
  lchrom = 0;
  for (i = 0; i < npars; i++)
    lchrom += param[i].lparam;
  alloc_chrom = (lchrom / L_O_CHROM_TYPE) + ((lchrom % L_O_CHROM_TYPE) ? 1 : 0);

  for (i = 0; i < popsize; i++)
  {
    oldpop[i].chrom = (CHROMOSOME) calloc(alloc_chrom, L_O_CHROM_TYPE);
    newpop[i].chrom = (CHROMOSOME) calloc(alloc_chrom, L_O_CHROM_TYPE);
  }

/* genetic algorithm */

  gen = 0;
  p_initpop(popsize, lchrom, oldpop, npars, param);
  p_statistics(popsize, &max, &avg, &min, &sumfitness, oldpop, &max_number);

  p_print_stat(gen, max, min, avg, oldpop, popsize, lchrom, ncross, nmutation, printing);

  do
  {
    gen++;
    if(scaling)
    {
      p_scalepop(popsize, max, avg, min, &sumfitness, oldpop);
    }
    p_generation(oldpop, newpop, popsize, sumfitness, lchrom,
		 &ncross, &nmutation, &jcross, pcross, pmutation,
		 npars, param);
    p_statistics(popsize, &max, &avg, &min, &sumfitness, newpop, &max_number);

    p_print_stat(gen, max, min, avg, newpop, popsize, lchrom, ncross, nmutation, printing);

    temppop = newpop;
    newpop = oldpop;
    oldpop = temppop;

  } while (gen < maxgen);

/* get the best individual */
  p_decode_params(lchrom, oldpop[max_number].chrom, npars, param);

/* cleanup */

  for (i = 0; i < popsize; i++)
  {
    free(oldpop[i].chrom);
    free(newpop[i].chrom);
  }
  free(oldpop);
  free(newpop);

}				/* end of genetic_algorithm */

void p_print_stat(long gen, double max, double min, double avg, POPULATION pop,
       long popsize, long lchrom, long ncross, long nmutation, int printing)
{
  long i, j, k, no_chrom;
  CHROM_TYPE mask;

  if (printing > 0)
  {
    printf("\nIn %d generation --- Max = %lf Avg = %lf Min = %lf",
      gen, max, avg, min);
  }

  if (printing > 1)
  {
    printf("\nNo. crossover = %ld    No. mutation = %ld", ncross, nmutation);
  }

  if (printing > 2)
  {
    no_chrom = (lchrom / L_O_CHROM_TYPE) + ((lchrom % L_O_CHROM_TYPE) ? 1 : 0);

    for (j = 0; j < popsize; j++)
    {
      printf("\n%d) ", j);

      for (k = 0; k < no_chrom; k++)
	for (i = 0; i < L_O_CHROM_TYPE; i++)
	{
	  if (lchrom == k * L_O_CHROM_TYPE + i)
	    break;
	  mask = 1 << i;
	  printf("%d", pop[j].chrom[k] & mask ? 1 : 0);
	}

      printf("  - raw fitness = %lf - cross point = %d - parent 1=%d 2=%d", 
         pop[j].fitness, pop[j].xsite, pop[j].parent1, pop[j].parent2);
    }
  }
}				/* end of p_print_stat */
