#include <stdio.h>
#include <math.h>
#include <e_lib.h>

#include "graph.h"
#include "strip_create.h"
#include "subdom_graph.h"

MESH           *G_master;
char            G_projectname[E_LINE_MAX];
int             G_no_one_connect;

/* Dual graph of the mesh */
int            *NoNodeElem;
int           **NodeElem;
int            *Degree;
int           **Adjacency;

/* Subdomain graph of the mesh */
int            *SubdomGraphDegree;
int           **SubdomGraph;
int           **SubdomGraphRef;

int            *TwoWaySubdomGraphDegree;
int           **TwoWaySubdomGraph;
int           **TwoWaySubdomGraphRef;


static void     write_weighted_graph_file(char *filename, MESH * mesh,
                                          int *degree, int **adjacency,
                                          int **edge_weight);

static void     reorganise_graph(int no, int *degree,
                                 int **adjacency, int **edge_weight,
                                 int *twodegree,
                                 int **twoadjacency,
                                 int **twoedge_weight);

/* -------------------------------------------------------------------- */


static void reorganise_graph(int no, int *degree,
                             int **adjacency, int **edge_weight,
                             int *twodegree,
                             int **twoadjacency,
                             int **twoedge_weight)
{
  int             i, j, other;

  for(i = 0; i < no; i++)
  {
    for(j = 0; j < degree[i]; j++)
    {
      if((!G_no_one_connect) && (edge_weight[i][j] == 1))
        continue;
      twoadjacency[i][twodegree[i]] = adjacency[i][j];
      twoedge_weight[i][twodegree[i]] = edge_weight[i][j];
      twodegree[i]++;
      other = adjacency[i][j] - 1;
      twoadjacency[other][twodegree[other]] = i + 1;
      twoedge_weight[other][twodegree[other]] = edge_weight[i][j];
      twodegree[other]++;
    }
  }
}




static void write_weighted_graph_file(char *filename, MESH * mesh,
                                      int *degree, int **adjacency,
                                      int **edge_weight)
{
  char            graph_file[E_LINE_MAX];
  FILE           *fp;
  int             i, j, edges, k;
  int            *strip_weight;

  strcpy(graph_file, filename);
  strcat(graph_file, ".str.graph");

  if((fp = fopen(graph_file, "w")) == NULL)
  {
    e_GeneralError("Can not open strip-graph file for write !");
  }

  printf("\n\n Graph file: %s\n\n", graph_file);

#ifdef __WEIGHTED_VERTEX_EDGE__
  /* count edges */
  edges = 0;
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    edges += degree[i];
  }

  /* output number of vertex and edges */
  fprintf(fp, "%d %d 11", mesh->Decomp.NSubDoms, edges / 2);

  /* count elements in strips, weight of a strip */
  strip_weight = e_alloc_i_vect(mesh->Decomp.NSubDoms);
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    strip_weight[mesh->Decomp.SubDoms[i] - 1]++;
  }

  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    /* output vertex weight */
    fprintf(fp, "\n%d ", strip_weight[i]);
    for(j = 0; j < degree[i]; j++)
    {
      fprintf(fp, "%d %d ", adjacency[i][j], edge_weight[i][j]-1);
    }
  }
#endif

#ifdef __WEIGHTED_VERTEX__
  /* count edges */
  edges = 0;
  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    for(j = 0; j < degree[i]; j++)
    {
      edges += edge_weight[i][j];
    }
  }

  /* output number of vertex and edges */
  fprintf(fp, "%d %d 10", mesh->Decomp.NSubDoms, edges / 2);

  /* count elements in strips, weight of a strip */
  strip_weight = e_alloc_i_vect(mesh->Decomp.NSubDoms);
  for(i = 0; i < mesh->TotalNElems; i++)
  {
    strip_weight[mesh->Decomp.SubDoms[i] - 1]++;
  }

  for(i = 0; i < mesh->Decomp.NSubDoms; i++)
  {
    /* output vertex weight */
    fprintf(fp, "\n%d ", strip_weight[i]);
    for(j = 0; j < degree[i]; j++)
    {
      for(k = 0; k < edge_weight[i][j]; k++)
        fprintf(fp, "%d ", adjacency[i][j]);
    }
  }
#endif

  e_free(strip_weight);
  fclose(fp);
}


main(int argc, char *argv[])
{
  char            line[E_LINE_MAX];
  char           *check;

  strcpy(G_projectname, "");
  if(argc > 1)
  {
    strcpy(G_projectname, argv[1]);
  }

  if(!strlen(G_projectname))
  {
    printf("\n\nType in the input filename (.mdf) !");
    printf("\nGive no extension : ");
    check = fgets(line, E_LINE_MAX, stdin);
    if((check == NULL)
       || (sscanf(line, "%s", &G_projectname) != 1))
      e_GeneralError("Can not read filename!");
  }

  printf("\nThe projectname is taken as: %s", G_projectname);
  printf("\n    This is the name of the input .MDF file.");

  printf("\nDo you want the one-node connections be");
  printf("\nrepresented in the graph as well: [0/1]\n");
  check = fgets(line, E_LINE_MAX, stdin);
  if((check == NULL)
     || (sscanf(line, "%d", &G_no_one_connect) != 1))
    e_GeneralError("Can not read option!");
  if(!((G_no_one_connect == 0) || (G_no_one_connect == 1)))
    e_GeneralError("Invalid option !");

  e_GetMeshData(&G_master, G_projectname);
  if((G_master->NElemsLink1 +
      G_master->NElemsLink2 +
      G_master->NElemsLink3 +
      G_master->NElemsLink4 +
      G_master->NElemsLink5 +
      G_master->NElemsTriang3 +
      G_master->NElemsTriang4 +
      G_master->NElemsTriang5) != G_master->TotalNElems)
    e_GeneralError("Mesh contains unsupported element(s) !");

/* Initialise the dual graph, 
   if there is an out-of-plane link, QUIT */
  NoNodeElem = e_alloc_i_vect(G_master->NNodes);
  NodeElem = e_init_frag_i_matr(G_master->NNodes);
  Degree = e_alloc_i_vect(G_master->TotalNElems);
  Adjacency = e_init_frag_i_matr(G_master->TotalNElems);

  e_InitMeshPointElemConnection(G_master, NoNodeElem, NodeElem);
  SetDualAdjacency(G_master, Degree, Adjacency,
                   NoNodeElem, NodeElem);

  /* creating the strip information */
  Create_Strips(G_master, Degree, Adjacency);
  printf("\n\n    Number of strips: %d", G_master->Decomp.NSubDoms);

  /* creating the subdomain graph */
  SubdomGraphDegree = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  SubdomGraph = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms,
                                    G_master->Decomp.NSubDoms - 1);
  SubdomGraphRef = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms,
                                       G_master->Decomp.NSubDoms - 1);
  Create_Subdomain_Graph(G_master, SubdomGraphDegree, SubdomGraph,
                         SubdomGraphRef);

  /* reorganise the subdomain graph information, because it contains
     only one way edges */
  TwoWaySubdomGraphDegree = e_alloc_i_vect(G_master->Decomp.NSubDoms);
  TwoWaySubdomGraph = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms,
                                          G_master->Decomp.NSubDoms - 1);
  TwoWaySubdomGraphRef = e_alloc_cont_i_matr(G_master->Decomp.NSubDoms,
                                             G_master->Decomp.NSubDoms - 1);
  reorganise_graph(G_master->Decomp.NSubDoms,
                   SubdomGraphDegree, SubdomGraph, SubdomGraphRef,
                   TwoWaySubdomGraphDegree,
                   TwoWaySubdomGraph,
                   TwoWaySubdomGraphRef);

  write_weighted_graph_file(G_projectname, G_master,
                            TwoWaySubdomGraphDegree, TwoWaySubdomGraph,
                            TwoWaySubdomGraphRef);

}



