#include "e_lib.h"

#define GRAPH_ALLOC_UNIT 16

void e_ReadGraphFile(char *filename, int *NoNodes, 
                     int **Degree, int ***Adjacency,
                     int **NodeWeights, int ***EdgeWeights)
{
  char     graph_file[E_LINE_MAX];
  char     line[E_LINE_MAX];
  char    *check, *p_line;
  FILE    *fp;
  int      line_count = 0;

  int      no_nodes, no_edges, type, args;
  int      i, j;

  /* read in the graph file */
  sprintf(graph_file, "%s.graph", filename);  
  if((fp = fopen(graph_file, "r")) == NULL)
  {
    sprintf(line, "Cannot open \"%s\" file for reading !", graph_file);
    e_GeneralError(line);
  }

  line_count++;
  check = fgets(line, E_LINE_MAX, fp);
  if(check == NULL)
    e_FileLineError(line_count, filename, 
                    "Can not read header in the graph file !");

  args = sscanf(line, "%d %d %d", &no_nodes, &no_edges, &type);
  if(args < 2 || args > 3)
    e_GeneralError("Invalid header in the graph file !");
  if(args == 2)
    type = 0;

  *NoNodes = no_nodes;

  /* allocate the structures */
  *Degree = e_alloc_i_vect(no_nodes);
  *Adjacency = e_alloc_frag_i_matr(no_nodes, GRAPH_ALLOC_UNIT);
  *NodeWeights = e_alloc_i_vect(no_nodes);
  *EdgeWeights = e_alloc_frag_i_matr(no_nodes, GRAPH_ALLOC_UNIT);

  /* fill up the structures */
  for(i = 0; i < no_nodes; i++)
  {
    line_count++;
    check = fgets(line, E_LINE_MAX, fp);
    if(check == NULL)
      e_FileLineError(line_count, filename, "Can not read line !");
    
    p_line = line;
    /* user specified vertex weight */
    if(type == 10 || type == 11)
    {
      p_line = GetStrArg(p_line, "%d", &((*NodeWeights)[i]));
    }
    else
    {
      (*NodeWeights)[i] = 1;
    }
    
    j = 0;
    while(1)
    {
      p_line = GetStrArg(p_line, "%d", &((*Adjacency)[i][j]));
      /* nothing to read, quit */
      if(p_line == NULL)
        break;

      /* if weighted edges */
      if(type == 1 || type == 11)
      {
        p_line = GetStrArg(p_line, "%d", &((*EdgeWeights)[i][j]));
        if(p_line == NULL)
          e_FileLineError(line_count, filename, "Cannot read edge weight !");
      }
      else
      {
        (*EdgeWeights)[i][j] = 1;
      }
      
      j++;
      (*Adjacency)[i] = e_incr_i_vect(j, (*Adjacency)[i], GRAPH_ALLOC_UNIT);
      (*EdgeWeights)[i] = e_incr_i_vect(j, (*EdgeWeights)[i], GRAPH_ALLOC_UNIT);
      
    }
    (*Degree)[i] = j;
  }

  fclose(fp);

}

void e_WriteGraphFile(char *filename, int type, int NoNodes, 
                      int *Degree, int **Adjacency,
                      int *NodeWeights, int **EdgeWeights)
{
  char     graph_file[E_LINE_MAX];
  char     line[E_LINE_MAX];
  char    *check, *p_line;
  FILE    *fp;
  int      line_count = 0;

  int      no_edges;
  int      i, j;

  if(type != 0 && type != 1 && type != 10 && type != 11)
    e_GeneralError("Invalid graph type !");

  /* read in the graph file */
  sprintf(graph_file, "%s.graph", filename);  
  if((fp = fopen(graph_file, "w")) == NULL)
  {
    sprintf(line, "Cannot open \"%s\" file for writing !", graph_file);
    e_GeneralError(line);
  }

  no_edges = 0;
  for(i = 0; i < NoNodes; i++)
    no_edges += Degree[i];
  no_edges /= 2;

  fprintf(fp, "%d %d ", NoNodes, no_edges);
  if(type != 0)
    fprintf(fp, " %d", type);

  for(i = 0; i < NoNodes; i++)
  {
    fprintf(fp, "\n");
    /* if the nodes are weighted */
    if(type == 10 || type == 11)
    {
      if(NodeWeights == NULL)
        e_GeneralError("Graph node weights are not specified !");

      fprintf(fp, "%d ", NodeWeights[i]);
    }

    for(j = 0; j < Degree[i]; j++)
    {
      fprintf(fp, "%d ", Adjacency[i][j]);
      if(type == 1 || type == 11)
      {
        if(EdgeWeights == NULL)
          e_GeneralError("Graph edge weights are not specified !");

        fprintf(fp, "%d ", EdgeWeights[i][j]);
      }
    }
  }

  fclose(fp);
}


void e_WritePTNFile(char *filename, int NoNodes, int *Subdom)
{
  char   out_file[E_LINE_MAX];
  char   line[E_LINE_MAX];
  FILE  *fp;
  int    i;

  sprintf(out_file, "%s.graph.ptn", filename);
  if((fp = fopen(out_file, "w")) == NULL)
  {
    sprintf(line, "Cannot open \"%s\" file for writing !", out_file);
    e_GeneralError(line);
  }

  for(i = 0; i < NoNodes; i++)
    fprintf(fp, "%d\n", Subdom[i]-1);

  fclose(fp);
}


void e_ReadPTNFile(char *filename, int NoNodes, 
                   int *NSubdoms, int *Subdoms)
{
  char   out_file[E_LINE_MAX];
  char   line[E_LINE_MAX];
  FILE  *fp;
  int    i, linecount;

  sprintf(out_file, "%s.graph.ptn", filename);
  if((fp = fopen(out_file, "r")) == NULL)
  {
    sprintf(line, "Cannot open \"%s\" file for reading !", out_file);
    e_GeneralError(line);
  }
  
  (*NSubdoms) = -1;

  for(i = 0; i < NoNodes; i++)
  {
    linecount = 0;

    e_GetLine(line, E_LINE_MAX, &linecount, fp);
    if(sscanf(line, "%d", &(Subdoms[i])) != 1)
      e_FileLineError(linecount, out_file,
                      "Could not read partition number !");

    /* value between 1-n */
    Subdoms[i]++;

    /* checking the validity */
    if(Subdoms[i] <= 0)
      e_FileLineError(linecount, out_file,
                      "Invalid partition number !");

    if(Subdoms[i] > (*NSubdoms))
      (*NSubdoms) = Subdoms[i];
  }

  fclose(fp);
}


