#include "e_lib.h"

#define SEGREF_SMALL_MPR 1e-6
#define SEGREF_SMALL_CRV_DELTAT 1e-6
#define SEGREF_SMALL_CRV_LENGTH 1e-8
#define SEGREF_MAX_ITER 50

#define CRVREF_VERSION "$Revision: 1.2 $"

#define SIGN(a) ((a>0.0)?+1:((a<0.0)?-1:0))

static double crv_getlength(NURBSCURVE * curve, double start_t, double end_t,
                            double accuracy);
static void crv_getcoords(NURBSCURVE * curve, double t, double *crds);
static void crv_length2param(NURBSCURVE * curve, double prev_t,
                             double *next_t, double estimate_t, int sign,
                             double length, double accuracy);

int e_CurveRefine(NURBSCURVE * curve,
                  double start_t,
                  double start_mpr,
                  double end_t,
                  double end_mpr,
                  double ***seg_crd,
                  double **seg_mpr,
                  double **seg_t, 
                  int generate_nodes, int mode)
{
  int n;
  int nbar;
  double s;
  double sbar;
  double sl;                    /* length increment along curve */
  double slt;                   /* corresponding curve param */
  double estimate_t;            /* estimate for slt */
  double alfa;
  double alfabar;
  double ab_pow;
  double beta;
  double mpr;
  double accuracy;              /* accuracy for curve interpolation */
  int i;

  accuracy = E_MIN(start_mpr, end_mpr);

  /* checking the input */
  if ((mode != E_SEGREF_NONE) && (mode != E_SEGREF_ODD)
      && (mode != E_SEGREF_EVEN))
    e_GeneralError("e_CurveRefine: unrecognized mode!");
  if ((start_mpr < SEGREF_SMALL_MPR) || (end_mpr < SEGREF_SMALL_MPR))
    e_GeneralError
      ("e_CurveRefine: mesh parameters should not be negative or very small!");

  /* first the length */
  s = crv_getlength(curve, start_t, end_t, accuracy);

  if ((start_mpr >= s) && (end_mpr >= s))
  {
    if (!generate_nodes)
      return 2;
    else
    {
      (*seg_mpr) = e_alloc_d_vect(2);
      (*seg_crd) = e_alloc_cont_d_matr(2, 3);
      (*seg_t) = e_alloc_d_vect(2);
      (*seg_mpr)[0] = start_mpr;
      (*seg_mpr)[1] = end_mpr;
      (*seg_t)[0] = start_t;
      (*seg_t)[1] = end_t;
      crv_getcoords(curve, start_t, (*seg_crd)[0]);
      crv_getcoords(curve, end_t, (*seg_crd)[1]);
    }
  }
  if ((start_mpr < s) && (end_mpr < s))
  {
    /* alfa=(s-start_mpr)/(s-end_mpr); */
    alfa = (s - start_mpr + end_mpr) / s;
    if (E_ABS(alfa - 1.0) < SEGREF_SMALL_MPR)
      n = (int) ceil(s / start_mpr) - 1;
    else
    {
      /* n=(int)ceil(log(end_mpr/start_mpr)/log(alfa)); */
      n = (int) ceil(log(end_mpr / start_mpr) / log(alfa) - 1.0);
    }
    nbar = n;
    if (mode == E_SEGREF_EVEN)
    {
      if (n % 2 != 0)
        nbar = n + 1;
    }
    else if (mode == E_SEGREF_ODD)
    {
      if (n % 2 == 0)
        nbar = n + 1;
    }
    if (!generate_nodes)
      return nbar + 2;

    /* generate the nodes */
    alfabar = pow(end_mpr / start_mpr, 1.0 / (double) nbar);
    if (E_ABS(alfabar - 1.0) < SEGREF_SMALL_MPR)
      sbar = (nbar + 1) * start_mpr;
    else
      sbar =
        (start_mpr * (1.0 - pow(alfabar, (double) (nbar + 1)))) / (1.0 -
                                                                   alfabar);
    beta = s / sbar;
    (*seg_mpr) = e_alloc_d_vect(nbar + 2);
    (*seg_crd) = e_alloc_cont_d_matr(nbar + 2, 3);
    (*seg_t) = e_alloc_d_vect(nbar + 2);

    if (start_mpr < end_mpr)
    {
      (*seg_mpr)[0] = start_mpr;
      (*seg_t)[0] = start_t;
      crv_getcoords(curve, start_t, (*seg_crd)[0]);

      (*seg_mpr)[1 + 0] = start_mpr + (end_mpr - start_mpr) *
        beta * start_mpr / s;
      estimate_t = start_t + (end_t - start_t) * beta * start_mpr / s;
      sl = beta * start_mpr;
      crv_length2param(curve, start_t, &slt, estimate_t,
                       SIGN(end_t - start_t), sl, accuracy);
      crv_getcoords(curve, slt, (*seg_crd)[1]);
      (*seg_t)[1 + 0] = slt;

      ab_pow = alfabar;
      for (i = 1; i < nbar; i++)
      {
        (*seg_mpr)[1 + i] = (*seg_mpr)[i] + (end_mpr - start_mpr) *
          beta * start_mpr * ab_pow / s;
        estimate_t = slt + (end_t - start_t) * beta * start_mpr * ab_pow / s;
        sl *= alfabar;
        crv_length2param(curve, slt, &slt, estimate_t, SIGN(end_t - start_t),
                         sl, accuracy);
        crv_getcoords(curve, slt, (*seg_crd)[1 + i]);
        (*seg_t)[1 + i] = slt;
        ab_pow *= alfabar;
      }

      (*seg_mpr)[1 + nbar] = end_mpr;
      (*seg_t)[1 + nbar] = end_t;
      crv_getcoords(curve, end_t, (*seg_crd)[1 + nbar]);
    }
    else
    {
      (*seg_mpr)[1 + nbar] = end_mpr;
      (*seg_t)[1 + nbar] = end_t;
      crv_getcoords(curve, end_t, (*seg_crd)[1 + nbar]);

      (*seg_mpr)[nbar] = end_mpr + (start_mpr - end_mpr) * beta * end_mpr / s;
      estimate_t = end_t + (start_t - end_t) * beta * end_mpr / s;
      sl = beta * end_mpr;
      crv_length2param(curve, end_t, &slt, estimate_t, SIGN(start_t - end_t),
                       sl, accuracy);
      crv_getcoords(curve, slt, (*seg_crd)[nbar]);
      (*seg_t)[nbar] = slt;

      ab_pow = 1.0 / alfabar;
      for (i = nbar - 1; i > 0; i--)
      {
        (*seg_mpr)[i] = (*seg_mpr)[i + 1] + (start_mpr - end_mpr) *
          beta * end_mpr / s * ab_pow;
/*            estimate_t+=(start_t-end_t)* */
        estimate_t = slt + (start_t - end_t) * beta * end_mpr / s * ab_pow;
        sl /= alfabar;
        crv_length2param(curve, slt, &slt, estimate_t, SIGN(start_t - end_t),
                         sl, accuracy);
        crv_getcoords(curve, slt, (*seg_crd)[i]);
        (*seg_t)[i] = slt;
        ab_pow /= alfabar;
      }

      (*seg_mpr)[0] = start_mpr;
      (*seg_t)[0] = start_t;
      crv_getcoords(curve, start_t, (*seg_crd)[0]);
    }
  }
  else                          /* one of the mesh params is >= length */
  {
    sbar = 0.0;                 /* now used as running length */
    n = 0;
    mpr = E_MIN(start_mpr, end_mpr);
    while (sbar < s)
    {
      sbar += mpr;
      n++;
      mpr = E_MIN(start_mpr, end_mpr) + E_ABS(end_mpr - start_mpr) * sbar / s;
    }
    n--;
    nbar = n;
    if (mode == E_SEGREF_EVEN)
    {
      if (n % 2 != 0)
        nbar = n + 1;
    }
    else if (mode == E_SEGREF_ODD)
    {
      if (n % 2 == 0)
        nbar = n + 1;
    }
    if (!generate_nodes)
      return nbar + 2;
    if (nbar > n)
    {
      mpr = E_MIN(start_mpr, end_mpr) + E_ABS(end_mpr - start_mpr) * sbar / s;
      sbar += mpr;
    }
    beta = s / sbar;

    (*seg_mpr) = e_alloc_d_vect(nbar + 2);
    (*seg_t) = e_alloc_d_vect(nbar + 2);
    (*seg_crd) = e_alloc_cont_d_matr(nbar + 2, 3);

    if (start_mpr < end_mpr)
    {
      (*seg_mpr)[0] = start_mpr;
      (*seg_t)[0] = start_t;
      crv_getcoords(curve, start_t, (*seg_crd)[0]);

      sbar = 0.0;
      mpr = start_mpr;
      sbar += mpr;

      sl = beta * mpr;
      estimate_t = start_t + (end_t - start_t) / s * beta * mpr;
      crv_length2param(curve, start_t, &slt, estimate_t,
                       SIGN(end_t - start_t), sl, accuracy);
      crv_getcoords(curve, slt, (*seg_crd)[1]);
      (*seg_mpr)[1 + 0] = start_mpr + (end_mpr - start_mpr) / s * beta * mpr;
      (*seg_t)[1 + 0] = slt;

      for (i = 1; i < nbar; i++)
      {
        mpr = start_mpr + (end_mpr - start_mpr) * sbar / s;
        sbar += mpr;
        sl = beta * mpr;
        estimate_t = start_t + (end_t - start_t) * beta * sbar / s;
        crv_length2param(curve, slt, &slt, estimate_t, SIGN(end_t - start_t),
                         sl, accuracy);
        crv_getcoords(curve, slt, (*seg_crd)[1 + i]);
        (*seg_mpr)[1 + i] = start_mpr + (end_mpr - start_mpr) *
          beta * sbar / s;
        (*seg_t)[1 + i] = slt;
      }

      (*seg_mpr)[1 + nbar] = end_mpr;
      (*seg_t)[1 + nbar] = end_t;
      crv_getcoords(curve, end_t, (*seg_crd)[1 + nbar]);
    }
    else
    {
      (*seg_mpr)[1 + nbar] = end_mpr;
      (*seg_t)[1 + nbar] = end_t;
      crv_getcoords(curve, end_t, (*seg_crd)[1 + nbar]);

      sbar = 0.0;
      mpr = end_mpr;
      sbar += mpr;

      sl = beta * mpr;
      estimate_t = end_t + (start_t - end_t) / s * beta * mpr;
      crv_length2param(curve, end_t, &slt, estimate_t, SIGN(start_t - end_t),
                       sl, accuracy);
      crv_getcoords(curve, slt, (*seg_crd)[nbar]);
      (*seg_mpr)[nbar] = end_mpr + (start_mpr - end_mpr) / s * beta * mpr;
      (*seg_t)[nbar] = slt;

      for (i = nbar - 1; i > 0; i--)
      {
        mpr = end_mpr + (start_mpr - end_mpr) * sbar / s;
        sbar += mpr;
        sl = beta * mpr;
        estimate_t = end_t + (start_t - end_t) * beta * sbar / s;
        crv_length2param(curve, slt, &slt, estimate_t, SIGN(start_t - end_t),
                         sl, accuracy);
        crv_getcoords(curve, slt, (*seg_crd)[i]);
        (*seg_t)[i] = slt;
        (*seg_mpr)[i] = end_mpr + (start_mpr - end_mpr) * beta * sbar / s;
      }

      (*seg_mpr)[0] = start_mpr;
      (*seg_t)[0] = start_t;
      crv_getcoords(curve, start_t, (*seg_crd)[0]);
    }
  }

  return nbar + 2;
}

static double crv_getlength(NURBSCURVE * curve, double start_t, double end_t,
                            double accuracy)
{
  /* return the length of the curve with number curveno
     between parameter values start_t and end_t */

  double st_crd[3];             /* coords of starting point of curve segment */
  double end_crd[3];            /* coords of ending point of curve segment */
  double prev_crd[3];           /* coords of previous interpolation point */
  double curr_crd[3];           /* coords of current interpolation point */
  double s;                     /* length of cord */
  double length;                /* length of curve segment */
  int i;                        /* counters */
  int n;                        /* no of interpolation points */
  int p;                        /* interpolation counter */
  double t;                     /* curve parameter of interpolation point */

  crv_getcoords(curve, start_t, st_crd);
  crv_getcoords(curve, end_t, end_crd);

  s = 0.0;
  for (i = 0; i < 3; i++)
    s += (end_crd[i] - st_crd[i]) * (end_crd[i] - st_crd[i]);
  s = sqrt(s);

  n = E_MAX(10, 10 * (int) (s / accuracy));

  length = 0.0;
  t = start_t;
  for (i = 0; i < 3; i++)
    prev_crd[i] = st_crd[i];
  for (p = 0; p < n; p++)
  {
    t += (end_t - start_t) / (double) n;
    crv_getcoords(curve, t, curr_crd);
    s = 0.0;
    for (i = 0; i < 3; i++)
    {
      s += (curr_crd[i] - prev_crd[i]) * (curr_crd[i] - prev_crd[i]);
      prev_crd[i] = curr_crd[i];
    }
    length += sqrt(s);
  }

  return length;
}

static void crv_getcoords(NURBSCURVE * curve, double t, double *crds)
{
  /* return the coordinates of the point on the 
     curve with number curveno
     at parameter values t */

  double a, b, c, d, e, f, g, h;
  int i;                        /* the 3 coord directions */
  double tt;
  double N, D;                  /* nominator and denominator */

  tt = 1.0 - t;
  for (i = 0; i < 3; i++)
  {
    /* this is strange! I thought that CntlPoints had only the coords of the control
       points, but it seems to contain also the endpoint coordinates of the curve!?! */
       
    a = curve->CntlPoints[0][i] * /* endpoint 1 */
      curve->CntlPoints[0][3];  /* weight of endpoint 1 */
    b = curve->CntlPoints[1][i] * /* cntlpoint 1 */
      curve->CntlPoints[1][3] * 3; /* weight of cntlpoint 1 */
    c = curve->CntlPoints[2][i] * /* cntlpoint 2 */
      curve->CntlPoints[2][3] * 3; /* weight of cntlpoint 2 */
    d = curve->CntlPoints[3][i] * /* endpoint 2 */
      curve->CntlPoints[3][3];  /* weight of endpoint 2 */
    e = curve->CntlPoints[0][3];
    f = curve->CntlPoints[1][3] * 3;
    g = curve->CntlPoints[2][3] * 3;
    h = curve->CntlPoints[3][3];

    N = a * tt * tt * tt + b * t * tt * tt + c * t * t * tt + d * t * t * t;
    D = e * tt * tt * tt + f * t * tt * tt + g * t * t * tt + h * t * t * t;
    if (fabs(D) < 1e-30)
      e_GeneralError("e_curve_refine: crv_getcoords has zero denominator");
    crds[i] = N / D;
  }
}

static void crv_length2param(NURBSCURVE * curve, double prev_t,
                             double *next_t, double estimate_t, int sign,
                             double length, double accuracy)
{
  /* return the parameter next_t of the point which is length away
     from the point with parameter prev_t on the curve with number curveno
     in the direction of sign (+1/0/-1) */

  double curr_length;
  double curr_t;
  int it;

  /* sanity check on input */
  if (SIGN(estimate_t - prev_t) != sign)
    e_GeneralError("e_CurveRefine: crv_length2param estimate on wrong side");

  curr_t = estimate_t;
  curr_length = crv_getlength(curve, prev_t, curr_t, accuracy);
  it = 0;

  while ((E_ABS(curr_length - length) / length > SEGREF_SMALL_CRV_LENGTH)
         && (it < SEGREF_MAX_ITER))
/*    while ((E_ABS(curr_t-prev_t)>SEGREF_SMALL_CRV_DELTAT)&&(it<SEGREF_MAX_ITER)) */
  {
    curr_t = prev_t + (curr_t - prev_t) * length / curr_length;
    curr_length = crv_getlength(curve, prev_t, curr_t, accuracy);
    it++;
  }

  if (it == SEGREF_MAX_ITER)
    e_GeneralWarning("e_CurveRefine: crv_length2param iteration unfinished");

  if (SIGN(curr_t - prev_t) != sign)
  {
    printf
      ("prev_t=%f length=%f curr_t=%f estim_t=%f curr_l=%f sign=%d\n",
       prev_t, length, curr_t, estimate_t, curr_length, sign);
    e_GeneralError("e_CurveRefine: crv_length2param iteration failed (*1)");
  }

  if (E_ABS(curr_length - length) > 0.1)
    e_GeneralError("e_CurveRefine: crv_length2param iteration failed (*2)");

  (*next_t) = curr_t;
}

int e_GetCurve(MESH *mesh, int node1, int node2, NURBSCURVE ** curve,
                      double *t1, double *t2)
{
  int Number1, Number2;
  int i, j;

  Number1 = mesh->NNurbsCurves[node1 - 1];
  Number2 = mesh->NNurbsCurves[node2 - 1];


  for (i = 0; i < Number1; i++)
  {
    for (j = 0; j < Number2; j++)
    {
      if (mesh->NurbsCurves[node1 - 1][i] == mesh->NurbsCurves[node2 - 1][j])
      {
        *curve = mesh->NurbsCurves[node1 - 1][i];
        *t1 = mesh->NurbsCurveParams[node1 - 1][i];
        *t2 = mesh->NurbsCurveParams[node2 - 1][j];
        return 1;
      }
    }
  }

  return 0;
}
