/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       Miscellaneous E-Lib Functions                        */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

#include "e_pltfrm.h"
#include <stdio.h>
#include <string.h>
#include <e_lib.h>

#include "miscel.h"


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              e_SetAdjacency()                              */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To deterine the degree vector and adjacency matrix of a mesh. Two elements
   are said to be adjacent if they share an interface. The return value is 1
   if the operation is successful, else 0.                                    */

/* NOTE: It is assumed that the mesh is composed of only one element type.    */


int e_SetAdjacency(MESH *mesh, int **Degree, int ***AdjElems)
{
   int i,j,k,l,r;
   int *AdjacentElements;       /* elements sharing a mesh-point with the
                                   current element                        */ 
   int ElementIndex;            /* element index */
   int NAdjacentElements;       /* number of elements sharing a mesh-point
                                   with the current element                */ 
   int **NeighElements;         /* elements shared by a mesh-point */ 
   int NIpE;                    /* number of interfaces per element */
   int *NNeighElements;         /* number of elements shared by a mesh-point */
   int *NTimesAdjacent;         /* number of times an element shares
                                   mesh-points with another element  */ 
   int NVpE;                    /* number of vertices per element */
   int NVpI;                    /* number of vertices per interface */
   int VectorMax;               /* maximum vector length */
   int **Vertices;              /* vertex node indices */


   if (!e_IsOfOneElementType(mesh))
   {
      e_GeneralError("e_SetAdjacency: mesh contains more than one element type");
      return 0;
   }
   else if (e_LinkElementsOnly(mesh))
   {
      e_GeneralError("e_SetAdjacency: mesh composed of link elements");
      return 0;
   }
   else if (  mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3
            + mesh->NElemsTetrah1
            + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_SetAdjacency: unsupported element type");
      return 0;
   }

   NIpE = e_GetNInterfacesPerElement(mesh);
   NVpE = e_GetNVerticesPerElement(mesh);
   NVpI = e_GetNVerticesPerInterface(mesh);
   Vertices = e_GetVertices(mesh);

   *Degree = NULL;
   *AdjElems = NULL;
   NNeighElements = NULL;
   NeighElements = NULL;

   if (               (*Degree = e_alloc_i_vect(mesh->TotalNElems)) == NULL
       || (*AdjElems = e_alloc_frag_i_matr(mesh->TotalNElems,NIpE)) == NULL
       ||      (NNeighElements = e_alloc_i_vect(mesh->NMeshPoints)) == NULL
       ||   (NeighElements = e_init_frag_i_matr(mesh->NMeshPoints)) == NULL)
   {
      e_FreeAdjacency(mesh,*Degree,*AdjElems);
      e_free(NNeighElements);
      e_free_frag_i_matr(mesh->NMeshPoints,NeighElements);
      return 0;
   }

   VectorMax = -1;
   AdjacentElements = NULL;
   NTimesAdjacent = NULL;

   for (i = 0; i < mesh->TotalNElems; i++)
   {
      ElementIndex = i + 1;
      NAdjacentElements = 0;

      for (j = 0; j < NVpE; j++)
      {
         k = Vertices[i][j] - 1;

         for (l = 0; l < NNeighElements[k]; l++)
         {
            for (r = 0; r < NAdjacentElements; r++)
            {
               if (AdjacentElements[r] == NeighElements[k][l]) break;
            }

            if (r == NAdjacentElements)
            {
               if (r > VectorMax)
               {
                  VectorMax = r;

                  if (   (AdjacentElements
                             = e_incr_i_vect(r,AdjacentElements,1)) == NULL
                      || (NTimesAdjacent
                               = e_incr_i_vect(r,NTimesAdjacent,1)) == NULL)
                  {
                     e_FreeAdjacency(mesh,*Degree,*AdjElems);
                     e_free(NNeighElements);
                     e_free_frag_i_matr(mesh->NMeshPoints,NeighElements);
                     e_free(AdjacentElements);
                     e_free(NTimesAdjacent);
                     return 0;
                  }
               }

               AdjacentElements[r] = NeighElements[k][l];
               NAdjacentElements++;
            }

            NTimesAdjacent[r]++;
         }

         if ((NeighElements[k] = e_incr_i_vect(NNeighElements[k],
                                               NeighElements[k],
                                               NVECTOR_MAX       )) == NULL)
         {
            e_FreeAdjacency(mesh,*Degree,*AdjElems);
            e_free(NNeighElements);
            e_free_frag_i_matr(mesh->NMeshPoints,NeighElements);
            e_free(AdjacentElements);
            e_free(NTimesAdjacent);
            return 0;
         }

         NeighElements[k][NNeighElements[k]++] = ElementIndex;
      }

      for (r = 0; r < NAdjacentElements; r++)
      {
         if (NTimesAdjacent[r] == NVpI)
         {
            k = AdjacentElements[r] - 1;

            if ((*Degree)[i] == NIpE || (*Degree)[k] == NIpE)
            {
               e_GeneralError("e_SetAdjacency: an element has more adjacent elements than interfaces");
               e_FreeAdjacency(mesh,*Degree,*AdjElems);
               e_free(NNeighElements);
               e_free_frag_i_matr(mesh->NMeshPoints,NeighElements);
               e_free(AdjacentElements);
               e_free(NTimesAdjacent);
               return 0;
            }

            (*AdjElems)[i][(*Degree)[i]++] = AdjacentElements[r];
            (*AdjElems)[k][(*Degree)[k]++] = ElementIndex;
         }

         NTimesAdjacent[r] = 0;
      }
   }

   e_free(NNeighElements);
   e_free_frag_i_matr(mesh->NMeshPoints,NeighElements);
   e_free(AdjacentElements);
   e_free(NTimesAdjacent);

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeAdjacency: To free memory allocated by the adjacency calculation. */


void e_FreeAdjacency(MESH *mesh, int *Degree, int **AdjElems)
{
   e_free(Degree);
   e_free_frag_i_matr(mesh->TotalNElems,AdjElems);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                              e_SetAdjElems()                               */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* e_SetAdjElems: To determine the adjacency matrix of a mesh such that each
   adjacent element of a given element is associated with the appropriate
   shared interface. If a planar element (i + 1) with nodes A, B and C has an
   element (k + 1) adjacent to its first edge AB, then AdjElems[i][0] = k + 1.
   If there was no element sharing edge AB then AdjElems[i][0] would equal 0.
   For tetrahedral elements, the j-th interface is the one opposite the j-th
   vertex node.                                                               */

/* The return value is 1 if the adjacency matrix is set successfully, else 0. */


int e_SetAdjElems(MESH *mesh, int ***AdjElems)
{
   int i,j,k,m;
   int *Degree;
   int **Vertices;
   int ElemType;
   int *InterfVertices;
   int NIpE;
   int NVpE;
   int NVpI;
   int *pTemp1;
   int *pTemp2;

   char *ElemFound;


   if (!e_SetAdjacency(mesh,&Degree,AdjElems))
      return 0;

   ElemType = e_GetElementType(mesh);
   NVpE = e_GetNVerticesPerElement(mesh);
   NIpE = e_GetNInterfacesPerElement(mesh);
   NVpI = e_GetNVerticesPerInterface(mesh);
   Vertices = e_GetVertices(mesh);

   InterfVertices = NULL;
   pTemp1 = NULL;
   ElemFound = NULL;

   if (   (InterfVertices = e_alloc_i_vect(NVpE + 1)) == NULL
       ||             (pTemp1 = e_alloc_i_vect(NIpE)) == NULL
       ||          (ElemFound = e_alloc_c_vect(NIpE)) == NULL)
   {
      e_FreeAdjacency(mesh,Degree,*AdjElems);
      e_free(InterfVertices);
      e_free(pTemp1);
      e_free(ElemFound);
      return 0;
   }

   if (   ElemType == E_ELEM_TYPE_TRIANG1
       || ElemType == E_ELEM_TYPE_TRIANG2
       || ElemType == E_ELEM_TYPE_TRIANG3
       || ElemType == E_ELEM_TYPE_TRIANG4
       || ElemType == E_ELEM_TYPE_TRIANG5
       || ElemType == E_ELEM_TYPE_QUAD1
       || ElemType == E_ELEM_TYPE_QUAD3  )
   {
      for (i = 0; i < mesh->TotalNElems; i++)
      {
         for (j = 0; j < NVpE; j++)
            InterfVertices[j] = Vertices[i][j];

         InterfVertices[NVpE] = Vertices[i][0];
         memset(ElemFound,0,NIpE*sizeof(char));

         for (k = 0; k < NIpE; k++)
            pTemp1[k] = GetAdjElem(NVpE,NVpI,Vertices,&InterfVertices[k],
                                       Degree[i],(*AdjElems)[i],ElemFound);
         pTemp2 = (*AdjElems)[i];
         (*AdjElems)[i] = pTemp1;
         pTemp1 = pTemp2;
      }
   }
   else if (ElemType == E_ELEM_TYPE_TETRAH1)
   {
      for (i = 0; i < mesh->TotalNElems; i++)
      {
         memset(ElemFound,0,NIpE*sizeof(char));

         for (k = 0; k < NIpE; k++)
         {
            m = 0;

            for (j = 0; j < NVpE; j++)
            {
               if (j != k)
                  InterfVertices[m++] = Vertices[i][j];
            }

            pTemp1[k] = GetAdjElem(NVpE,NVpI,Vertices,InterfVertices,
                                   Degree[i],(*AdjElems)[i],ElemFound);
         }

         pTemp2 = (*AdjElems)[i];
         (*AdjElems)[i] = pTemp1;
         pTemp1 = pTemp2;
      }
   }
   else
   {
      e_GeneralError("e_SetAdjElems: mesh composed of BLOCK1 elements");
      e_FreeAdjacency(mesh,Degree,*AdjElems);
      e_free(InterfVertices);
      e_free(pTemp1);
      e_free(ElemFound);
      return 0;
   }

   e_free(Degree);
   e_free(InterfVertices);
   e_free(pTemp1);
   e_free(ElemFound);

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetAdjElem: To return the index of the adjacent element which shares a
   specified interface. If no such adjacent element exists, the return value
   is 0.                                                                      */


static int GetAdjElem(int NVpE, int NVpI, int **Vertices,
                      int *InterfVertices, int Degree, int *AdjElems,
                      char *ElemFound                                )
{
   int j,k,l,m;
   int NCommonVertices;


   for (j = 0; j < Degree; j++)
   {
      if (ElemFound[j])
         continue;

      k = AdjElems[j] - 1;
      NCommonVertices = 0;

      for (l = 0; l < NVpE; l++)
      {
         for (m = 0; m < NVpI; m++)
         {
            if (InterfVertices[m] == Vertices[k][l])
               NCommonVertices++;
         }
      }

      if (NCommonVertices == NVpI)
      {
         ElemFound[j] = TRUE;
         return AdjElems[j];
      }
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeAdjElems: To free the memory allocated by "e_SetAdjElems". */


void e_FreeAdjElems(MESH *mesh, int **AdjElems)
{
   e_free_frag_i_matr(mesh->TotalNElems,AdjElems);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           e_SetNodalMeshParams()                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To determine the nodal mesh parameters of a mesh from its element mesh
   parameters. The return value is 1 if the operation is successful, else 0.  */


int e_SetNodalMeshParams(MESH *mesh)
{
   int k;
   int *NNeighElems;


   if (   !mesh->SubStructSet[E_SS_NUM_MESHPARAM]
       || !mesh->MeshParam.ElemMeshParamsSet     )
   {
      e_GeneralError("e_SetNodalMeshParams: element mesh parameters not set");
      return 0;
   }
   else if (  mesh->NElemsLink1
            + mesh->NElemsLink2
            + mesh->NElemsLink3
            + mesh->NElemsLink4
            + mesh->NElemsLink5
            + mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3
            + mesh->NElemsTetrah1
            + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_SetNodalMeshParams: unsupported element type");
      return 0;
   }

   mesh->MeshParam.NodalMeshParams = NULL;
   NNeighElems = NULL;

   if (   (mesh->MeshParam.NodalMeshParams
                       = e_alloc_d_vect(mesh->NMeshPoints)) == NULL
       || (NNeighElems = e_alloc_i_vect(mesh->NMeshPoints)) == NULL)
   {
      e_free(mesh->MeshParam.NodalMeshParams);
      e_free(NNeighElems);
      return 0;
   }

   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_LINK1,E_NVPE_LINK1,
                              mesh->NElemsLink1,mesh->NodesLink1,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_LINK2,E_NVPE_LINK2,
                              mesh->NElemsLink2,mesh->NodesLink2,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_LINK3,E_NVPE_LINK3,
                              mesh->NElemsLink3,mesh->NodesLink3,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_LINK4,E_NVPE_LINK4,
                              mesh->NElemsLink4,mesh->NodesLink4,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_LINK5,E_NVPE_LINK5,
                              mesh->NElemsLink5,mesh->NodesLink5,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TRIANG1,E_NVPE_TRIANG1,
                              mesh->NElemsTriang1,mesh->NodesTriang1,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TRIANG2,E_NVPE_TRIANG2,
                              mesh->NElemsTriang2,mesh->NodesTriang2,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TRIANG3,E_NVPE_TRIANG3,
                              mesh->NElemsTriang3,mesh->NodesTriang3,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TRIANG4,E_NVPE_TRIANG4,
                              mesh->NElemsTriang4,mesh->NodesTriang4,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TRIANG5,E_NVPE_TRIANG5,
                              mesh->NElemsTriang5,mesh->NodesTriang5,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_QUAD1,E_NVPE_QUAD1,
                              mesh->NElemsQuad1,mesh->NodesQuad1,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_QUAD3,E_NVPE_QUAD3,
                              mesh->NElemsQuad3,mesh->NodesQuad3,NNeighElems);
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_TETRAH1,E_NVPE_TETRAH1,
                              mesh->NElemsTetrah1,mesh->NodesTetrah1,
                              NNeighElems                             );
   SetElemTypeNodalMeshParams(mesh,E_ELEM_TYPE_BLOCK1,E_NVPE_BLOCK1,
                              mesh->NElemsBlock1,mesh->NodesBlock1,NNeighElems);

   for (k = 0; k < mesh->NMeshPoints; k++)
   {
      if (NNeighElems[k] > 0)
         mesh->MeshParam.NodalMeshParams[k] /= NNeighElems[k];
   }

   e_free(NNeighElems);

   mesh->MeshParam.NodalMeshParamsSet = TRUE;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetElemTypeNodalMeshParams: To determine the contribution of a given
   element type to the nodal mesh parameters.                                 */


static void SetElemTypeNodalMeshParams(MESH *mesh, int ElemType,
                                       int NVpE, int NElems,
                                       int **Nodes, int *NNeighElems)
{
   int i,j,k;
   
   double *ElemMeshParams;
   
   
   if (ElemType == E_ELEM_TYPE_LINK1)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsLink1;
   else if (ElemType == E_ELEM_TYPE_LINK2)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsLink2;
   else if (ElemType == E_ELEM_TYPE_LINK3)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsLink3;
   else if (ElemType == E_ELEM_TYPE_LINK4)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsLink4;
   else if (ElemType == E_ELEM_TYPE_LINK5)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsLink5;
   else if (ElemType == E_ELEM_TYPE_TRIANG1)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTriang1;
   else if (ElemType == E_ELEM_TYPE_TRIANG2)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTriang2;
   else if (ElemType == E_ELEM_TYPE_TRIANG3)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTriang3;
   else if (ElemType == E_ELEM_TYPE_TRIANG4)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTriang4;
   else if (ElemType == E_ELEM_TYPE_TRIANG5)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTriang5;
   else if (ElemType == E_ELEM_TYPE_QUAD1)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsQuad1;
   else if (ElemType == E_ELEM_TYPE_QUAD3)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsQuad3;
   else if (ElemType == E_ELEM_TYPE_TETRAH1)
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsTetrah1;
   else
      ElemMeshParams = mesh->MeshParam.ElemMeshParamsBlock1;
   
   for (i = 0; i < NElems; i++)
   {
      for (j = 0; j < NVpE; j++)
      {
         k = Nodes[i][j] - 1;
         mesh->MeshParam.NodalMeshParams[k] += ElemMeshParams[i];
         NNeighElems[k]++;
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                           e_GetNCutInterfaces()                            */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To return the number of cut interfaces in a decomposed mesh. The return
   value is -1 if the operation is not successful.                            */

/* NOTE: It is assumed that the mesh is composed of only one element type.    */


int e_GetNCutInterfaces(MESH *mesh)
{
   int i,j,k;
   int *Degree;                 /* element degree vector */
   int **AdjElems;              /* element adjacency matrix */
   int NCutInterfaces;          /* number of cut interfaces */
   int SubdomainIndex;          /* subdomain index */


   if (!mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      e_GeneralError("e_GetNCutInterfaces: decomposition data not set");
      return -1;
   }
   else if (!e_IsOfOneElementType(mesh))
   {
      e_GeneralError("e_GetNCutInterfaces: mesh contains more than one element type");
      return -1;
   }
   else if (e_LinkElementsOnly(mesh))
   {
      e_GeneralError("e_GetNCutInterfaces: mesh composed of link elements");
      return -1;
   }
   else if (  mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3
            + mesh->NElemsTetrah1
            + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_GetNCutInterfaces: unsupported element type");
      return -1;
   }

   if (!e_SetAdjacency(mesh,&Degree,&AdjElems))
      return -1;

   NCutInterfaces = 0;
   
   for (i = 0; i < mesh->TotalNElems; i++)
   {
      SubdomainIndex = mesh->Decomp.SubDoms[i];
       
      for (j = 0; j < Degree[i]; j++)
      {
         k = AdjElems[i][j] - 1;
         
         if (mesh->Decomp.SubDoms[k] != SubdomainIndex)
            NCutInterfaces++;
      }
   }
   
   NCutInterfaces /= 2;

   e_FreeAdjacency(mesh,Degree,AdjElems);
                                                    
   return NCutInterfaces;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          e_GetNInterSubDomNodes()                          */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To determine the number and weighted number of inter-subdomain nodes where,
   if a node is shared by "n" subdomains, then its weighting is "n - 1". The
   return value is 1 if the operation is successful, else 0.                  */


int e_GetNInterSubDomNodes(MESH *mesh, int *NInterSubDomNodes,
                                int *WeightedNInterSubDomNodes)
{
   int i,j,k;
   int **AdjacentElements;      /* elements adjacent to each node */
   int *AdjacentSubdomains;     /* subdomains adjacent to a node */
   int *NAdjacentElements;      /* number of elements adjacent to each node */
   int NAdjacentSubdomains;     /* number of subdomains adjacent to a node */


   if (!mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      e_GeneralError("e_GetNInterSubDomNodes: decomposition data not set");
      return 0;
   }
   else if (  mesh->NElemsLink1
            + mesh->NElemsLink2
            + mesh->NElemsLink3
            + mesh->NElemsLink4
            + mesh->NElemsLink5
            + mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3
            + mesh->NElemsTetrah1
            + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_GetNInterSubDomNodes: unsupported element type");
      return 0;
   }

   NAdjacentElements = NULL;
   AdjacentElements = NULL;
   AdjacentSubdomains = NULL;

   if (             (NAdjacentElements = e_alloc_i_vect(mesh->NNodes)) == NULL
       ||        (AdjacentElements = e_init_frag_i_matr(mesh->NNodes)) == NULL
       || (AdjacentSubdomains = e_alloc_i_vect(mesh->Decomp.NSubDoms)) == NULL)
   {
      e_free(NAdjacentElements);
      e_free_frag_i_matr(mesh->NNodes,AdjacentElements);
      e_free(AdjacentSubdomains);
      return 0;
   }

   if (   !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_LINK1,E_NNPE_LINK1,mesh->NElemsLink1,
               mesh->NodesLink1,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_LINK2,E_NNPE_LINK2,mesh->NElemsLink2,
               mesh->NodesLink2,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_LINK3,E_NNPE_LINK3,mesh->NElemsLink3,
               mesh->NodesLink3,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_LINK4,E_NNPE_LINK4,mesh->NElemsLink4,
               mesh->NodesLink4,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_LINK5,E_NNPE_LINK5,mesh->NElemsLink5,
               mesh->NodesLink5,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TRIANG1,E_NNPE_TRIANG1,mesh->NElemsTriang1,
               mesh->NodesTriang1,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TRIANG2,E_NNPE_TRIANG2,mesh->NElemsTriang2,
               mesh->NodesTriang2,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TRIANG3,E_NNPE_TRIANG3,mesh->NElemsTriang3,
               mesh->NodesTriang3,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TRIANG4,E_NNPE_TRIANG4,mesh->NElemsTriang4,
               mesh->NodesTriang4,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TRIANG5,E_NNPE_TRIANG5,mesh->NElemsTriang5,
               mesh->NodesTriang5,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_QUAD1,E_NNPE_QUAD1,mesh->NElemsQuad1,
               mesh->NodesQuad1,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_QUAD3,E_NNPE_QUAD3,mesh->NElemsQuad3,
               mesh->NodesQuad3,NAdjacentElements,AdjacentElements   )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_TETRAH1,E_NNPE_TETRAH1,mesh->NElemsTetrah1,
               mesh->NodesTetrah1,NAdjacentElements,AdjacentElements       )
       || !SetElemTypeNInterSubDomNodes
              (mesh,E_ELEM_TYPE_BLOCK1,E_NNPE_BLOCK1,mesh->NElemsBlock1,
               mesh->NodesBlock1,NAdjacentElements,AdjacentElements     )   )
   {
      e_free(NAdjacentElements);
      e_free_frag_i_matr(mesh->NNodes,AdjacentElements);
      e_free(AdjacentSubdomains);
      return 0;
   }

   *NInterSubDomNodes = 0;
   *WeightedNInterSubDomNodes = 0;

   for (k = 0; k < mesh->NNodes; k++)
   {
      for (i = 0; i < mesh->Decomp.NSubDoms; i++)
         AdjacentSubdomains[i] = 0;

      for (j = 0; j < NAdjacentElements[k]; j++)
      {
         i = AdjacentElements[k][j] - 1;
         AdjacentSubdomains[mesh->Decomp.SubDoms[i] - 1] = 1;
      }

      NAdjacentSubdomains = 0;

      for (i = 0; i < mesh->Decomp.NSubDoms; i++)
         NAdjacentSubdomains += AdjacentSubdomains[i];

      if (NAdjacentSubdomains > 1)
         (*NInterSubDomNodes)++;

      *WeightedNInterSubDomNodes += NAdjacentSubdomains - 1;
   }

   e_free(NAdjacentElements);
   e_free_frag_i_matr(mesh->NNodes,AdjacentElements);
   e_free(AdjacentSubdomains);
                                      
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetElemTypeNInterSubDomNodes: To determine the contribution to the numbers
   of intersubdomain nodes for a given element type.                          */


static int SetElemTypeNInterSubDomNodes(MESH *mesh,
                                        int ElemType, int NNpE,
                                        int NElems, int **Nodes,
                                        int *NAdjacentElements,
                                        int **AdjacentElements  )
{
   int i,j,k;
   int NetIndex;
   
   
   NetIndex = mesh->FirstNetIndex[ElemType];
   
   for (i = 0; i < NElems; i++)
   {
      for (j = 0; j < NNpE; j++)
      {
         k = Nodes[i][j] - 1;

         AdjacentElements[k] = e_incr_i_vect(NAdjacentElements[k],
                                             AdjacentElements[k],
                                             NVECTOR_MAX          );
         if (AdjacentElements[k] == NULL)
            return 0;
         
         AdjacentElements[k][NAdjacentElements[k]++] = NetIndex;
      }

      NetIndex++;
   }

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          e_GetInterSubDomNodes()                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To determine the inter-subdomain nodes and the subdomains to which they
   are adjacent. The return value is 1 if the operation is successful,
   else 0.                                                                    */


int e_GetInterSubDomNodes(MESH *mesh, int *NInterSubdomainNodes,
                             int **InterSubdomainNodes, int **NNodeSubdomains,
                                                         int ***NodeSubdomains)
{
   int j,k,l;
   int *NNodeSubs;              /* number of subdomains adjacent to each node */
   int **NodeSubs;              /* indicates subdomains adjacent to each node */
   int **TempNodeSubdomains;    /* temporary "NodeSubdomains" */


   if (!mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      e_GeneralError("e_GetInterSubDomNodes: decomposition data not set");
      return 0;
   }
   else if (  mesh->NElemsLink1
            + mesh->NElemsLink2
            + mesh->NElemsLink3
            + mesh->NElemsLink4
            + mesh->NElemsLink5
            + mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3
            + mesh->NElemsTetrah1
            + mesh->NElemsBlock1  != mesh->TotalNElems)
   {
      e_GeneralError("e_GetInterSubDomNodes: unsupported element type");
      return 0;
   }

   *NodeSubdomains = NULL;
   NNodeSubs = NULL;
   NodeSubs = NULL;

   if (   (*NodeSubdomains
               = e_alloc_frag_i_matr(mesh->NNodes,
                                     mesh->Decomp.NSubDoms)) == NULL
       ||         (NNodeSubs = e_alloc_i_vect(mesh->NNodes)) == NULL
       || (NodeSubs
               = e_alloc_frag_i_matr(mesh->NNodes,
                                     mesh->Decomp.NSubDoms)) == NULL)
   {
      e_free_frag_i_matr(mesh->NNodes,*NodeSubdomains);
      e_free(NNodeSubs);
      e_free_frag_i_matr(mesh->NNodes,NodeSubs);
      return 0;
   }

   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_LINK1,E_NNPE_LINK1,
                               mesh->NElemsLink1,mesh->NodesLink1,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_LINK2,E_NNPE_LINK2,
                               mesh->NElemsLink2,mesh->NodesLink2,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_LINK3,E_NNPE_LINK3,
                               mesh->NElemsLink3,mesh->NodesLink3,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_LINK4,E_NNPE_LINK4,
                               mesh->NElemsLink4,mesh->NodesLink4,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_LINK5,E_NNPE_LINK5,
                               mesh->NElemsLink5,mesh->NodesLink5,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TRIANG1,E_NNPE_TRIANG1,
                               mesh->NElemsTriang1,mesh->NodesTriang1,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TRIANG2,E_NNPE_TRIANG2,
                               mesh->NElemsTriang2,mesh->NodesTriang2,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TRIANG3,E_NNPE_TRIANG3,
                               mesh->NElemsTriang3,mesh->NodesTriang3,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TRIANG4,E_NNPE_TRIANG4,
                               mesh->NElemsTriang4,mesh->NodesTriang4,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TRIANG5,E_NNPE_TRIANG5,
                               mesh->NElemsTriang5,mesh->NodesTriang5,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_QUAD1,E_NNPE_QUAD1,
                               mesh->NElemsQuad1,mesh->NodesQuad1,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_QUAD3,E_NNPE_QUAD3,
                               mesh->NElemsQuad3,mesh->NodesQuad3,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_TETRAH1,E_NNPE_TETRAH1,
                               mesh->NElemsTetrah1,mesh->NodesTetrah1,NodeSubs);
   SetElemTypeInterSubDomNodes(mesh,E_ELEM_TYPE_BLOCK1,E_NNPE_BLOCK1,
                               mesh->NElemsBlock1,mesh->NodesBlock1,NodeSubs);
                               
   *NInterSubdomainNodes = 0;

   for (k = 0; k < mesh->NNodes; k++)
   {
      for (j = 0; j < mesh->Decomp.NSubDoms; j++)
      {
         if (NodeSubs[k][j] > 0)
            (*NodeSubdomains)[k][NNodeSubs[k]++] = j + 1;
      }

      if (NNodeSubs[k] > 1)
         (*NInterSubdomainNodes)++;
   }

   *InterSubdomainNodes = NULL;
   *NNodeSubdomains = NULL;
   TempNodeSubdomains = NULL;

   if (   (*InterSubdomainNodes
                  = e_alloc_i_vect(*NInterSubdomainNodes)) == NULL
       || (*NNodeSubdomains
                  = e_alloc_i_vect(*NInterSubdomainNodes)) == NULL
       || (TempNodeSubdomains
              = e_init_frag_i_matr(*NInterSubdomainNodes)) == NULL)
   {
      e_free(*InterSubdomainNodes);
      e_free(*NNodeSubdomains);
      e_free_frag_i_matr(mesh->NNodes,*NodeSubdomains);
      e_free(NNodeSubs);
      e_free_frag_i_matr(mesh->NNodes,NodeSubs);
      e_free_frag_i_matr(*NInterSubdomainNodes,TempNodeSubdomains);
      return 0;
   }

   *NInterSubdomainNodes = 0;

   for (k = 0; k < mesh->NNodes; k++)
   {
      if (NNodeSubs[k] > 1)
      {
         l = (*NInterSubdomainNodes)++;
         (*InterSubdomainNodes)[l] = k + 1;
         (*NNodeSubdomains)[l] = NNodeSubs[k];

         if ((TempNodeSubdomains[l] = e_alloc_i_vect(NNodeSubs[k])) == NULL)
         {
            e_free(*InterSubdomainNodes);
            e_free(*NNodeSubdomains);
            e_free_frag_i_matr(mesh->NNodes,*NodeSubdomains);
            e_free(NNodeSubs);
            e_free_frag_i_matr(mesh->NNodes,NodeSubs);
            e_free_frag_i_matr(*NInterSubdomainNodes,TempNodeSubdomains);
            return 0;
         }

         for (j = 0; j < NNodeSubs[k]; j++)
            TempNodeSubdomains[l][j] = (*NodeSubdomains)[k][j];
      }
   }

   e_free_frag_i_matr(mesh->NNodes,*NodeSubdomains);
   e_free(NNodeSubs);
   e_free_frag_i_matr(mesh->NNodes,NodeSubs);

   *NodeSubdomains = TempNodeSubdomains;

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* SetElemTypeInterSubDomNodes: To determine the contribution to the
   of intersubdomain nodes data for a given element type.                     */


static void SetElemTypeInterSubDomNodes(MESH *mesh, int ElemType,
                                        int NNpE, int NElems,
                                        int **Nodes, int **NodeSubs)
{
   int i,j,k,l;
   int NetIndex;
   
   
   NetIndex = mesh->FirstNetIndex[ElemType] - 1;   
   
   for (i = 0; i < NElems; i++)
   {
      k = mesh->Decomp.SubDoms[NetIndex++] - 1;

      for (j = 0; j < NNpE; j++)
      {
         l = Nodes[i][j] - 1;
         NodeSubs[l][k] = 1;
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_FreeInterSubDomNodes: To free memory allocated by
   "e_GetInterSubDomNodes()".                                                 */


void e_FreeInterSubDomNodes(int NInterSubdomainNodes, int *InterSubdomainNodes,
                                     int *NNodeSubdomains, int **NodeSubdomains)
{
   e_free(InterSubdomainNodes);
   e_free(NNodeSubdomains);
   e_free_frag_i_matr(NInterSubdomainNodes,NodeSubdomains);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                       e_SubDomBoundaries2PolyPolygon()                     */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To determine the indices of the mesh-points which lie on the mesh
   subdomain boundaries in the order needed by the Windows "PolyPolygon" GDI
   function. The return value is the total number of points to be plotted by
   "PolyPolygon" if the operation is successful, else 0.                      */

/* NOTE: It is assumed that the mesh is composed solely of planar elements of
         one element type.                                                    */


int e_SubDomBoundaries2PolyPolygon(MESH *mesh, int **PolygonVertices, 
                                   int **NPolygonVertices, int *NPolygons)
{
   int i,j,k,l;
   int ***BoundaryPoints;
   int **NBoundaryPoints;
   int *NConnectedRegions;   
   int NSubDoms;
   int TotalNPoints;

   char Decomposed;


   if (!e_PlanarElementsOnly(mesh))
   {
      e_GeneralError("e_SubDomBoundaries2PolyPolygon: mesh contains non-planar elements");
      return 0;
   }
   else if (!e_IsOfOneElementType(mesh))
   {
      e_GeneralError("e_SubDomBoundaries2PolyPolygon: mesh contains more than one element type");
      return 0;
   }
   else if (  mesh->NElemsTriang1
            + mesh->NElemsTriang2
            + mesh->NElemsTriang3
            + mesh->NElemsTriang4
            + mesh->NElemsTriang5
            + mesh->NElemsQuad1
            + mesh->NElemsQuad3   != mesh->TotalNElems)
   {
      e_GeneralError("e_SubDomBoundaries2PolyPolygon: unsupported element type");
      return 0;
   }

   if (mesh->SubStructSet[E_SS_NUM_DECOMP])
   {
      Decomposed = TRUE;
   }
   else 
   {
      if ((mesh->Decomp.SubDoms = e_alloc_i_vect(mesh->TotalNElems)) == NULL)
         return 0;

      for (i = 0; i < mesh->TotalNElems; i++)
         mesh->Decomp.SubDoms[i] = 1;

      mesh->Decomp.NSubDoms = 1;
      mesh->SubStructSet[E_SS_NUM_DECOMP] = TRUE;
      Decomposed = FALSE;
   }

   g_Mesh = mesh;
   g_NMeshPoints = mesh->NMeshPoints;
   g_TotalNElems = mesh->TotalNElems;
   NSubDoms = mesh->Decomp.NSubDoms;

   if (!CalcDpA())
   {
      if (!Decomposed) e_FreeDecompData(mesh);

      return 0;
   }

   NConnectedRegions = NULL;
   NBoundaryPoints = NULL;
   BoundaryPoints = NULL;
   
   if (      (NConnectedRegions = e_alloc_i_vect(NSubDoms)) == NULL
       ||  (NBoundaryPoints = e_init_frag_i_matr(NSubDoms)) == NULL
       || (BoundaryPoints
              = (int ***)e_calloc(NSubDoms,sizeof(int **))) == NULL)
   {
      FreeCalcDpA();
      e_free(NConnectedRegions);
      e_free(NBoundaryPoints);
      e_free(BoundaryPoints);

      if (!Decomposed) e_FreeDecompData(mesh);

      return 0;
   }
   
   *NPolygons = 0;
   TotalNPoints = 0;
   
   for (i = 0; i < NSubDoms; i++)
   {
      if (!GetSubdomainBoundary(&NConnectedRegions[i],&NBoundaryPoints[i],
                                                  &BoundaryPoints[i],i + 1))
      {
         for (j = 0; j < i; j++)
            e_free_frag_i_matr(NConnectedRegions[j],BoundaryPoints[j]);
         
         FreeCalcDpA();
         e_free(NConnectedRegions);         
         e_free_frag_i_matr(i,NBoundaryPoints);
         e_free(BoundaryPoints);

         if (!Decomposed) e_FreeDecompData(mesh);

         return 0;
      }
      
      *NPolygons += NConnectedRegions[i];
      
      for (k = 0; k < NConnectedRegions[i]; k++)
         TotalNPoints += NBoundaryPoints[i][k] + 1;
   }
   
   *PolygonVertices = NULL;
   *NPolygonVertices = NULL;
   
   if (   (*PolygonVertices = e_alloc_i_vect(TotalNPoints)) == NULL
       ||  (*NPolygonVertices = e_alloc_i_vect(*NPolygons)) == NULL)
   {      
      for (i = 0; i < NSubDoms; i++)
         e_free_frag_i_matr(NConnectedRegions[i],BoundaryPoints[i]);
         
      FreeCalcDpA();
      e_free(*PolygonVertices);
      e_free(*NPolygonVertices);
      e_free(NConnectedRegions);         
      e_free_frag_i_matr(NSubDoms,NBoundaryPoints);
      e_free(BoundaryPoints);

      if (!Decomposed) e_FreeDecompData(mesh);

      return 0;
   }
         
   for (i = 0; i < NSubDoms; i++)
   {
      for (k = 0; k < NConnectedRegions[i]; k++)
      {
         for (l = 0; l < NBoundaryPoints[i][k]; l++)
            *(*PolygonVertices)++ = BoundaryPoints[i][k][l];
                    
         **PolygonVertices = *(*PolygonVertices - NBoundaryPoints[i][k]);
         (*PolygonVertices)++;
         *(*NPolygonVertices)++ = NBoundaryPoints[i][k] + 1;
      }
   }
   
   *PolygonVertices -= TotalNPoints;   
   *NPolygonVertices -= *NPolygons;   
   
   for (i = 0; i < NSubDoms; i++)
      e_free_frag_i_matr(NConnectedRegions[i],BoundaryPoints[i]);
         
   FreeCalcDpA();
   e_free(NConnectedRegions);         
   e_free_frag_i_matr(NSubDoms,NBoundaryPoints);
   e_free(BoundaryPoints);

   if (!Decomposed) e_FreeDecompData(mesh);

   return TotalNPoints;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* CalcDpA: To calculate the degree vector and connectivity matrix. The
   return value is 1 if the operation is successful, else 0.                  */


static int CalcDpA()
{
   int i,j;
   int ElementIndex;
   int vertex0,vertex1,vertex2;
   int **Vertices;


   Vertices = e_GetVertices(g_Mesh);
   g_NVpE = e_GetNVerticesPerElement(g_Mesh);

   g_Degree = NULL;
   g_Adjacency = NULL;
   g_InterfaceMeshPoints = NULL;
   g_NAdjacentMeshPoints = NULL;
   g_InterfaceElements1 = NULL;
   g_InterfaceElements2 = NULL;
      
   if (                    (g_Degree = e_alloc_i_vect(g_TotalNElems)) == NULL
       ||   (g_Adjacency = e_alloc_frag_i_matr(g_TotalNElems,g_NVpE)) == NULL
       || (g_InterfaceMeshPoints = e_init_frag_i_matr(g_NMeshPoints)) == NULL
       ||     (g_NAdjacentMeshPoints = e_alloc_i_vect(g_NMeshPoints)) == NULL
       ||  (g_InterfaceElements1 = e_init_frag_i_matr(g_NMeshPoints)) == NULL
       ||  (g_InterfaceElements2 = e_init_frag_i_matr(g_NMeshPoints)) == NULL)
   {
      FreeCalcDpA();
      return 0;
   }
  
   for (i = 0; i < g_TotalNElems; i++)
   {
      ElementIndex = i + 1;
      vertex2 = Vertices[i][0];
      vertex0 = vertex2;

      for (j = 1; j <= g_NVpE; j++)
      {
         vertex1 = vertex2;
         vertex2 = (j < g_NVpE) ? Vertices[i][j] : vertex0;

         if (!UpdateDpA(ElementIndex,vertex1,vertex2))
         {
            FreeCalcDpA();
            return 0;
         }
      }
   }
                                                       
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateDpA: To update the degree vector and connectivity matrix. The return
   value is 1 if the operation is successful, else 0.                         */


static int UpdateDpA(int ElementIndex, int vertex1, int vertex2)
{
   int i,j,k,l;
   int TempVertex;


   if (vertex1 > vertex2)
   {                                   
      TempVertex = vertex1;
      vertex1 = vertex2;
      vertex2 = TempVertex;
   }
   
   i = ElementIndex - 1;
   l = vertex1 - 1; 
   
   for (j = 0; j < g_NAdjacentMeshPoints[l]; j++)
      if (vertex2 == g_InterfaceMeshPoints[l][j]) break;
      
   if (j == g_NAdjacentMeshPoints[l])
   {  
      if (   (g_InterfaceMeshPoints[l]
                 = e_incr_i_vect(g_NAdjacentMeshPoints[l],
                                 g_InterfaceMeshPoints[l],
                                 NVECTOR_MAX              )) == NULL
          || (g_InterfaceElements1[l]
                 = e_incr_i_vect(g_NAdjacentMeshPoints[l],
                                 g_InterfaceElements1[l],
                                 NVECTOR_MAX             ))  == NULL
          || (g_InterfaceElements2[l]
                 = e_incr_i_vect(g_NAdjacentMeshPoints[l],
                                 g_InterfaceElements2[l],
                                 NVECTOR_MAX             ))  == NULL)
      {
         return 0;
      }
      
      g_InterfaceMeshPoints[l][g_NAdjacentMeshPoints[l]] = vertex2;
      g_InterfaceElements1[l][g_NAdjacentMeshPoints[l]] = ElementIndex;
      g_NAdjacentMeshPoints[l]++;
   }
   else
   {  
      k = g_InterfaceElements1[l][j] - 1;
      g_InterfaceElements2[l][j] = ElementIndex;
      g_Adjacency[i][g_Degree[i]] = g_InterfaceElements1[l][j];
      g_Adjacency[k][g_Degree[k]] = ElementIndex;
      g_Degree[i]++;
      g_Degree[k]++;
   }
      
   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* FreeCalcDpA: To free the memory associated with the connectivity
   calculation.                                                               */


static void FreeCalcDpA()
{
   e_free(g_Degree);
   e_free_frag_i_matr(g_TotalNElems,g_Adjacency);
   e_free(g_NAdjacentMeshPoints);
   e_free_frag_i_matr(g_NMeshPoints,g_InterfaceMeshPoints);
   e_free_frag_i_matr(g_NMeshPoints,g_InterfaceElements1);
   e_free_frag_i_matr(g_NMeshPoints,g_InterfaceElements2);
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetSubdomainBoundary: To determine the indices of the mesh-points which
   lie on a given inter-subdomain boundary in the order they appear on the
   boundary. The return value is 1 if the operation is successful, else 0.    */


static int GetSubdomainBoundary(int *NConnectedRegions, int **NBoundaryPoints,
                                     int ***BoundaryPoints, int SubdomainIndex)
{
   int i,j,k;
   int CurrentEndPoint;
   int NewStartingInterface;
   int StartPoint;
   int *TempBoundaryPoints;

   char NotAllConnectedRegionsFound;


   g_NBoundaryInterfaces = 0;

   for (k = 0; k < g_NMeshPoints; k++)
   {
      for (j = 0; j < g_NAdjacentMeshPoints[k]; j++)
      {
         if (IsABoundaryInterface(k,j,SubdomainIndex))
            g_NBoundaryInterfaces++;
      }
   }

   g_BoundaryMeshPoints1 = NULL;
   g_BoundaryMeshPoints2 = NULL;

   if (   (g_BoundaryMeshPoints1
              = e_alloc_i_vect(g_NBoundaryInterfaces)) == NULL
       || (g_BoundaryMeshPoints2
              = e_alloc_i_vect(g_NBoundaryInterfaces)) == NULL)
   {
      e_free(g_BoundaryMeshPoints1);
      e_free(g_BoundaryMeshPoints2);
      return 0;
   }

   g_NBoundaryInterfaces = 0;

   for (k = 0; k < g_NMeshPoints; k++)
   {
      for (j = 0; j < g_NAdjacentMeshPoints[k]; j++)
      {
         if (IsABoundaryInterface(k,j,SubdomainIndex))
         {
            g_BoundaryMeshPoints1[g_NBoundaryInterfaces] = k + 1;
            g_BoundaryMeshPoints2[g_NBoundaryInterfaces]
               = g_InterfaceMeshPoints[k][j];
            g_NBoundaryInterfaces++;
         }
      }
   }

   *NConnectedRegions = 0;

   if (g_NBoundaryInterfaces == 0)
   {
      e_free(g_BoundaryMeshPoints1);
      e_free(g_BoundaryMeshPoints2);
      *BoundaryPoints = NULL;
      return 1;
   }

   *NBoundaryPoints = NULL;
   *BoundaryPoints = NULL;
   g_InterfaceNotStored = NULL;

   if (   (*NBoundaryPoints
                   = e_incr_i_vect(0,*NBoundaryPoints,1)) == NULL
       || (*BoundaryPoints
               = e_incr_frag_i_matr(0,*BoundaryPoints,1)) == NULL
       || ((*BoundaryPoints)[0]
               = e_alloc_i_vect(2*g_NBoundaryInterfaces)) == NULL
       || (g_InterfaceNotStored
                 = e_alloc_c_vect(g_NBoundaryInterfaces)) == NULL)
   {
      e_free(g_BoundaryMeshPoints1);
      e_free(g_BoundaryMeshPoints2);
      e_free(*NBoundaryPoints);
      e_free_frag_i_matr(1,*BoundaryPoints);
      e_free(g_InterfaceNotStored);
      return 0;
   }

   for (k = 0; k < g_NBoundaryInterfaces; k++)
      g_InterfaceNotStored[k] = TRUE;

   StartPoint = g_BoundaryMeshPoints1[0];
   CurrentEndPoint = g_BoundaryMeshPoints2[0];
   (*BoundaryPoints)[0][0] = StartPoint;
   (*BoundaryPoints)[0][1] = CurrentEndPoint;
   (*NBoundaryPoints)[0] = 2;
   g_InterfaceNotStored[0] = FALSE;
   NotAllConnectedRegionsFound = TRUE;

   while (NotAllConnectedRegionsFound)
   {
      UpdateBoundary(StartPoint,&CurrentEndPoint,*NConnectedRegions,
                                    *NBoundaryPoints,*BoundaryPoints);

      if (CurrentEndPoint == StartPoint)
      {
         (*NConnectedRegions)++;
         NewStartingInterface = GetNewStartingInterface();

         if (NewStartingInterface > 0)
         {
            if (   (*NBoundaryPoints
                       = e_incr_i_vect
                            (*NConnectedRegions,*NBoundaryPoints,1)) == NULL
                || (*BoundaryPoints
                        = e_incr_frag_i_matr
                             (*NConnectedRegions,*BoundaryPoints,1)) == NULL
                || ((*BoundaryPoints)[*NConnectedRegions]
                          = e_alloc_i_vect(2*g_NBoundaryInterfaces)) == NULL)
            {
               e_free(g_BoundaryMeshPoints1);
               e_free(g_BoundaryMeshPoints2);
               e_free(*NBoundaryPoints);

               if (*NBoundaryPoints == NULL)
                  e_free_frag_i_matr(*NConnectedRegions,*BoundaryPoints);
               else
                  e_free_frag_i_matr(*NConnectedRegions + 1,*BoundaryPoints);

               e_free(g_InterfaceNotStored);
               return 0;
            }

            k = NewStartingInterface - 1;
            StartPoint = g_BoundaryMeshPoints1[k];
            CurrentEndPoint = g_BoundaryMeshPoints2[k];
            (*BoundaryPoints)[*NConnectedRegions][0] = StartPoint;
            (*BoundaryPoints)[*NConnectedRegions][1] = CurrentEndPoint;
            (*NBoundaryPoints)[*NConnectedRegions] = 2;
            g_InterfaceNotStored[k] = FALSE;
         }
         else
         {
            NotAllConnectedRegionsFound = FALSE;
         }
      }
   }

   for (i = 0; i < *NConnectedRegions; i++)
   {
      if ((TempBoundaryPoints
              = e_alloc_i_vect((*NBoundaryPoints)[i])) == NULL)
      {
         e_free(g_BoundaryMeshPoints1);
         e_free(g_BoundaryMeshPoints2);
         e_free(*NBoundaryPoints);
         e_free_frag_i_matr(*NConnectedRegions,*BoundaryPoints);
         e_free(g_InterfaceNotStored);
         return 0;
      }

      for (k = 0; k < (*NBoundaryPoints)[i]; k++)
         TempBoundaryPoints[k] = (*BoundaryPoints)[i][k];

      e_free((*BoundaryPoints)[i]);

      (*BoundaryPoints)[i] = TempBoundaryPoints;
   }

   e_free(g_BoundaryMeshPoints1);
   e_free(g_BoundaryMeshPoints2);
   e_free(g_InterfaceNotStored);

   return 1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* IsABoundaryInterface: To determine whether an interfaces lies on a given
   subdomain boundary.                                                        */


static int IsABoundaryInterface(int k, int j, int SubdomainIndex)
{
   int ElementIndex1,ElementIndex2;
   int subdomain1,subdomain2;


   ElementIndex1 = g_InterfaceElements1[k][j];
   ElementIndex2 = g_InterfaceElements2[k][j];

   subdomain1 = g_Mesh->Decomp.SubDoms[ElementIndex1 - 1];
   subdomain2 = (ElementIndex2 > 0) ?
                g_Mesh->Decomp.SubDoms[ElementIndex2 - 1] : 0;

   if (subdomain1 == subdomain2)
      return 0;
   else if (   subdomain1 == SubdomainIndex
            || subdomain2 == SubdomainIndex)
      return 1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* UpdateBoundary: To update the subdomain boundary data.                     */


static void UpdateBoundary(int StartPoint, int *CurrentEndPoint,
                           int NConnectedRegions, int *NBoundaryPoints,
                                                   int **BoundaryPoints)
{
   int k;


   for (k = 1; k < g_NBoundaryInterfaces; k++)
   {
      if (g_InterfaceNotStored[k])
      {
         if (g_BoundaryMeshPoints1[k] == *CurrentEndPoint)
         {
            if (g_BoundaryMeshPoints2[k] != StartPoint)
            {
               BoundaryPoints[NConnectedRegions]
                             [NBoundaryPoints[NConnectedRegions]]
                  = g_BoundaryMeshPoints2[k];
               *CurrentEndPoint = g_BoundaryMeshPoints2[k];
               NBoundaryPoints[NConnectedRegions]++;
               g_InterfaceNotStored[k] = FALSE;
            }
            else
            {
               *CurrentEndPoint = StartPoint;
               g_InterfaceNotStored[k] = FALSE;
            }

            return;
         }
         else if (g_BoundaryMeshPoints2[k] == *CurrentEndPoint)
         {
            if (g_BoundaryMeshPoints1[k] != StartPoint)
            {
               BoundaryPoints[NConnectedRegions]
                             [NBoundaryPoints[NConnectedRegions]]
                  = g_BoundaryMeshPoints1[k];
               *CurrentEndPoint = g_BoundaryMeshPoints1[k];
               NBoundaryPoints[NConnectedRegions]++;
               g_InterfaceNotStored[k] = FALSE;
            }
            else
            {
               *CurrentEndPoint = StartPoint;
               g_InterfaceNotStored[k] = FALSE;
            }

            return;
         }
      }
   }
}


/* ////////////////////////////////////////////////////////////////////////// */
/* GetNewStartingInterface: To return the index of an inter-subdomain
   interface which has not already be found. The return value is 0 if no such
   interface exists.                                                          */


static int GetNewStartingInterface()
{
   int k;


   for (k = 0; k < g_NBoundaryInterfaces; k++)
   {
      if (g_InterfaceNotStored[k]) return k + 1;
   }

   return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                    e_FreeSubDomBoundaries2PolyPolygon()                    */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* To free memory allocated by "e_SubDomBoundaries2PolyPolygon()". */


void e_FreeSubDomBoundaries2PolyPolygon(int *PolygonVertices,
                                        int *NPolygonVertices)
{
   e_free(PolygonVertices);
   e_free(NPolygonVertices);
}


/* ////////////////////////////////////////////////////////////////////////// */
/*                                                                            */
/*                          Mesh Type Specific Data                           */
/*                                                                            */
/* ////////////////////////////////////////////////////////////////////////// */

/* ////////////////////////////////////////////////////////////////////////// */
/* The following functions return element properties of a mesh assumed to be
   composed of one element type.                                              */


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetElementType: To return the element type.  */


int e_GetElementType(MESH *mesh)
{
   if (mesh->NElemsLink1 > 0)
      return E_ELEM_TYPE_LINK1;
   else if (mesh->NElemsLink2 > 0)
      return E_ELEM_TYPE_LINK2;
   else if (mesh->NElemsLink3 > 0)
      return E_ELEM_TYPE_LINK3;
   else if (mesh->NElemsLink4 > 0)
      return E_ELEM_TYPE_LINK4;
   else if (mesh->NElemsLink5 > 0)
      return E_ELEM_TYPE_LINK5;
   else if (mesh->NElemsTriang1 > 0)
      return E_ELEM_TYPE_TRIANG1;
   else if (mesh->NElemsTriang2 > 0)
      return E_ELEM_TYPE_TRIANG2;
   else if (mesh->NElemsTriang3 > 0)
      return E_ELEM_TYPE_TRIANG3;
   else if (mesh->NElemsTriang4 > 0)
      return E_ELEM_TYPE_TRIANG4;
   else if (mesh->NElemsTriang5 > 0)
      return E_ELEM_TYPE_TRIANG5;
   else if (mesh->NElemsQuad1 > 0)
      return E_ELEM_TYPE_QUAD1;
   else if (mesh->NElemsQuad3 > 0)
      return E_ELEM_TYPE_QUAD3;
   else if (mesh->NElemsTetrah1 > 0)
      return E_ELEM_TYPE_TETRAH1;
   else if (mesh->NElemsBlock1 > 0)
      return E_ELEM_TYPE_BLOCK1;
   else
      return -1;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNInterfacesPerElement: To return the number of interfaces per
   element.                                                                   */


int e_GetNInterfacesPerElement(MESH *mesh)
{
   if (mesh->NElemsTriang1 > 0)
      return E_NIPE_TRIANG1;
   else if (mesh->NElemsTriang2 > 0)
      return E_NIPE_TRIANG2;
   else if (mesh->NElemsTriang3 > 0)
      return E_NIPE_TRIANG3;
   else if (mesh->NElemsTriang4 > 0)
      return E_NIPE_TRIANG4;
   else if (mesh->NElemsTriang5 > 0)
      return E_NIPE_TRIANG5;
   else if (mesh->NElemsQuad1 > 0)
      return E_NIPE_QUAD1;
   else if (mesh->NElemsQuad3 > 0)
      return E_NIPE_QUAD3;
   else if (mesh->NElemsTetrah1 > 0)
      return E_NIPE_TETRAH1;
   else if (mesh->NElemsBlock1 > 0)
      return E_NIPE_BLOCK1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNVerticesPerElement: To return the number of vertices per element. */


int e_GetNVerticesPerElement(MESH *mesh)
{
   if (mesh->NElemsLink1 > 0)
      return E_NVPE_LINK1;
   else if (mesh->NElemsLink2 > 0)
      return E_NVPE_LINK2;
   else if (mesh->NElemsLink3 > 0)
      return E_NVPE_LINK3;
   else if (mesh->NElemsLink4 > 0)
      return E_NVPE_LINK4;
   else if (mesh->NElemsLink5 > 0)
      return E_NVPE_LINK5;
   else if (mesh->NElemsTriang1 > 0)
      return E_NVPE_TRIANG1;
   else if (mesh->NElemsTriang2 > 0)
      return E_NVPE_TRIANG2;
   else if (mesh->NElemsTriang3 > 0)
      return E_NVPE_TRIANG3;
   else if (mesh->NElemsTriang4 > 0)
      return E_NVPE_TRIANG4;
   else if (mesh->NElemsTriang5 > 0)
      return E_NVPE_TRIANG5;
   else if (mesh->NElemsQuad1 > 0)
      return E_NVPE_QUAD1;
   else if (mesh->NElemsQuad3 > 0)
      return E_NVPE_QUAD3;
   else if (mesh->NElemsTetrah1 > 0)
      return E_NVPE_TETRAH1;
   else if (mesh->NElemsBlock1 > 0)
      return E_NVPE_BLOCK1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNNodesPerElement: To return the number of nodes per element. */


int e_GetNNodesPerElement(MESH *mesh)
{
   if (mesh->NElemsLink1 > 0)
      return E_NNPE_LINK1;
   else if (mesh->NElemsLink2 > 0)
      return E_NNPE_LINK2;
   else if (mesh->NElemsLink3 > 0)
      return E_NNPE_LINK3;
   else if (mesh->NElemsLink4 > 0)
      return E_NNPE_LINK4;
   else if (mesh->NElemsLink5 > 0)
      return E_NNPE_LINK5;
   else if (mesh->NElemsTriang1 > 0)
      return E_NNPE_TRIANG1;
   else if (mesh->NElemsTriang2 > 0)
      return E_NNPE_TRIANG2;
   else if (mesh->NElemsTriang3 > 0)
      return E_NNPE_TRIANG3;
   else if (mesh->NElemsTriang4 > 0)
      return E_NNPE_TRIANG4;
   else if (mesh->NElemsTriang5 > 0)
      return E_NNPE_TRIANG5;
   else if (mesh->NElemsQuad1 > 0)
      return E_NNPE_QUAD1;
   else if (mesh->NElemsQuad3 > 0)
      return E_NNPE_QUAD3;
   else if (mesh->NElemsTetrah1 > 0)
      return E_NNPE_TETRAH1;
   else if (mesh->NElemsBlock1 > 0)
      return E_NNPE_BLOCK1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNVerticesPerInterface: To return the number of vertices per
   interface.                                                                 */


int e_GetNVerticesPerInterface(MESH *mesh)
{
   if (mesh->NElemsTriang1 > 0)
      return E_NVPI_TRIANG1;
   else if (mesh->NElemsTriang2 > 0)
      return E_NVPI_TRIANG2;
   else if (mesh->NElemsTriang3 > 0)
      return E_NVPI_TRIANG3;
   else if (mesh->NElemsTriang4 > 0)
      return E_NVPI_TRIANG4;
   else if (mesh->NElemsTriang5 > 0)
      return E_NVPI_TRIANG5;
   else if (mesh->NElemsQuad1 > 0)
      return E_NVPI_QUAD1;
   else if (mesh->NElemsQuad3 > 0)
      return E_NVPI_QUAD3;
   else if (mesh->NElemsTetrah1 > 0)
      return E_NVPI_TETRAH1;
   else if (mesh->NElemsBlock1 > 0)
      return E_NVPI_BLOCK1;
   else
      return 0;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetNodes: To return the mesh nodes. */


int **e_GetNodes(MESH *mesh)
{
   if (mesh->NElemsLink1 > 0)
      return mesh->NodesLink1;
   else if (mesh->NElemsLink2 > 0)
      return mesh->NodesLink2;
   else if (mesh->NElemsLink3 > 0)
      return mesh->NodesLink3;
   else if (mesh->NElemsLink4 > 0)
      return mesh->NodesLink4;
   else if (mesh->NElemsTriang1 > 0)
      return mesh->NodesTriang1;
   else if (mesh->NElemsTriang2 > 0)
      return mesh->NodesTriang2;
   else if (mesh->NElemsTriang3 > 0)
      return mesh->NodesTriang3;
   else if (mesh->NElemsTriang4 > 0)
      return mesh->NodesTriang4;
   else if (mesh->NElemsTriang5 > 0)
      return mesh->NodesTriang5;
   else if (mesh->NElemsQuad1 > 0)
      return mesh->NodesQuad1;
   else if (mesh->NElemsQuad3 > 0)
      return mesh->NodesQuad3;
   else if (mesh->NElemsTetrah1 > 0)
      return mesh->NodesTetrah1;
   else if (mesh->NElemsBlock1 > 0)
      return mesh->NodesBlock1;
   else
      return NULL;
}


/* ////////////////////////////////////////////////////////////////////////// */
/* e_GetVertices: To return the mesh vertices. */


int **e_GetVertices(MESH *mesh)
{
   return e_GetNodes(mesh);
}
