#include "eps.h"

/************************************************************************************\
                                Painter's algorithm

What does it do ?
-----------------

Draws the polygons in such an order, that if a polygon is covered by another one,
then first it draws the covered then overpaints the one which is on top of the previous 
one.


Assumptions:
------------
  - Only 1D, 2D elements in the mesh (For 3D elements the projected outline should be
    determined as a convex hull for further analysis.)
  - Quad elements are planes !


How does the method work ??
---------------------------

Step 1: Consider the real 3D mesh. I tries to put "separating planes" between
        the elements.
        
        - If the elements are so far away from each other in 3D that the 
        circumscribed spheres do not touch or intersect each other then there is
        no need for planes. The elements are marked, that the position of the
        center of the spheres will determine the order.
        
        - If the circumscribed sphere test fails, then the program determines
        the bounding box around the elements. (Maximum X, Y, Z and minimum X, Y, Z
        coordinates are determined.) Then the program tries to use the faces of the
        bounding box as separating plane between two elements. First it takes the
        bounding box of the first element and checks whether the nodes of the other element
        are all outside. If this is not true, then it tries vice-versa, taking the
        bounding box of the second element and checks with the nodes of the first element.

        - If the faces of the bounding box are not separating planes then we have one last
        chance, that the face itself is a separating plane between the two elements.
        In this case it checks only 2D elements (triangles, quads, but no links).
        
        When a separating plane is determined, it is stored with the element which has
        smaller global index. Three data are stored about the plane: the normal, a constant 
        (a dot product of the normal vector and a point which is on the plane) and an 
        indicator. The indicator determines, that the element which stores the separating
        plane is on which side of the plane. If the element is on the same side as the normal
        then the value is one otherwise minus one.
        
        It is possible, that no separating plane is found. No panic, there are other ways
        to figure out the order of the elements.
        
Step 2: After the determination of the separating planes, the mesh elements are projected
        into a 2D plane. (This is how it will appear on the scree or paper.) The 2D bounding
        box is determined again for each element. If only the azimuth and the elevation angles
        are given, then the projection centre (the position of the eye where we are looking from)
        should be determines as well.

Step 3: The program tries to determine the priority of the elements.
        There are two ways:
        - The first is simpler, it does not use separating planes.
        - In the second run, it utilises the already determined separating planes.
        
        The steps are the following:
        
        - Checks, whether the 2D bounding boxes overlap. If they are far away from each
        other, then there is no reason for further checking since they cannot cover each
        other.
        
        - If the 2D bounding boxes overlap or intersect each other, then the program checks
        whether the projected z coordinates of the bounding box are far away from each other.
        (Basically checks, whether the projected max and min coordinates do not overlap in the z
        direction.) If it is true, then the program either returns with the index of the 
        closer element or jumps ahead for further checking.
        
        - If the bounding boxes overlap in some way in the z direction then reads the 
        information from the separating plane.
            
            - If the 3D circumscribing spheres do not intersect or overlap, then at this stage
            the distance between the projection centre and the centre of the element is 
            determined for both elements. Whichever element is closer is taken and depending 
            whether further processing is required it jumps ahead or returns with the index.
            
            - A separating plane was determined earlier, then now the program checks whether
            the projection centre is on the same side as the element. Depending on result of
            this test it can determine which element of the two is first. If further processing
            is required the program jumps ahead.
            
        - This step checks the convex hulls of the projected elements. The convex hulls of the
        two element can intersect, one can contain the other or they can be disjoint.
        
            - If the convex hulls intersect each other, then the z coordinate of the intersection
            point is determined (in the projected space) and whichever has a smaller value is
            closer to the eye.
            
            - The convex hull of one of the elements contains the other element's convex hull.
            At this stage, it can be determined whether an element is totally obscured or not.
            If there was no selection beforehand, which is closer to the eye, then according
            to the z coordinates a selection can be made.
            
            - If the convex hulls of the elements are disjoint, then no ordering can be done.
            
Step 4: Sort the elements according to the determined priorities. Mikado sorting is used.
        It searches in two directions. It removes all elements which is not covered by any
        other element and puts them at the end of the plotting list. On the other hand
        it removes all elements which are covered and puts them at front of the plotting 
        list. Then repeats the search, but not considering the already removed elements.


\************************************************************************************/


#define INFINITE 1e20
#define EPS      1e-8

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define NOT_FOUND         0
/* 1 is the separation plane */
#define USE_SOAP_BUBBLES  2
#define NONE              -10

#define DISJOINT            1
#define INTERSECTION        2
#define POLY2_INSIDE_POLY1  3
#define POLY1_INSIDE_POLY2  4


static int  Point_behind_polygon;
double Hidden_Proj_center[3];
double Hidden_Distance;



#define subt_vec2(ab, a, b) \
        ((ab)[0] = (b)[0] - (a)[0], (ab)[1] = (b)[1] - (a)[1])

#define dot_product2(a, b) \
        ((a)[0]*(b)[0]+(a)[1]*(b)[1])

#define linear_comb2(r, p, pq, l) \
        ((r)[0] = (p)[0] + l*(pq)[0], \
         (r)[1] = (p)[1] + l*(pq)[1]    )

#define parallel_z(v) \
        (E_ABS((v)[0]) < EPS && E_ABS((v)[1]) < EPS)

#define normal_vec2(n, v) \
        ((n)[0] = -(v)[1], (n)[1] = (v)[0])

#define subt_vec(ab, a, b) \
        ((ab)[0] = (b)[0] - (a)[0], \
         (ab)[1] = (b)[1] - (a)[1], \
         (ab)[2] = (b)[2] - (a)[2]    )

#define dot_product(a, b) \
        ((a)[0] * (b)[0] + (a)[1] * (b)[1] + (a)[2] * (b)[2])

#define linear_comb(r, p, pq, l) \
        ((r)[0] = (p)[0] + l*(pq)[0], \
         (r)[1] = (p)[1] + l*(pq)[1], \
         (r)[2] = (p)[2] + l*(pq)[2]    )

#define Length(x) sqrt(dot_product(x, x))

#define sign(x) (x ? ((x < 0) ? (-1) : (1)) : 0)

#define which_side(point, plane) \
        sign(dot_product((plane)->normal, point) - (plane)->cnst)

#define T2(w, x) \
        ((w) = (x) / (Hidden_Distance - (x)))

/* function definitions */
static int overlap(double *min1, double *max1, double *min2, double *max2);
static double average_dist(POLYGON *obj);
static int inside_poly(double q[], int n, double **poly, double *z);

static double intersect_2Dlines(double *a, double *d, double *b, double *e);
static int relative_pos_hulls(double z[2], int n1, double **hull1,
                              int n2, double **hull2);
static int which_elem_first(POLYGON *elem1, POLYGON *elem2, int plane);
static void create_sep_planes(PLANE *s, POLYGON obj1, POLYGON obj2);
static int mikado_select(int n, POLYGON **res_obj);


/* ############################################################################ */
/*                            Element priority determination                    */
/* ############################################################################ */


/* Given bounding boxes, whether they overlap or not. */
static int overlap(double *min1, double *max1, double *min2, double *max2)
{
  if(min1[0] > max2[0] || max1[0] < min2[0] ||
     min1[1] > max2[1] || max1[1] < min2[1]    )
    return 0;
  else
    return 1;
}


/* Distance between object centre and projection centre.
   Not real distance, but since the same style will be compared
   it does not matter. */
static double average_dist(POLYGON *obj)
{
  double r[3];

  subt_vec(r, obj->center, Hidden_Proj_center);
  
  return(dot_product(r,r));
}


/* Whether a polygon is inside another one. */
static int inside_poly(double *q, int n, double **poly, double *z)
{
  double *p = *poly;
  double *r, *s;
  double  dist_r, dist_s;
  double  p0[3];
  double  normal[2], cnst;
  int     i;
  double  t;
  double  w1, w2;
  
  normal[0] = q[1] - p[1];
  normal[1] = p[0] - q[0];

  cnst = dot_product2(p, normal);
  
  r = poly[1];
  
  dist_r = dot_product2(r, normal) - cnst;
  
  for(i = 2; i < n; i++)
  {
    s = poly[i];
    dist_s = dot_product2(s, normal) - cnst;
    if(sign(dist_r) != sign(dist_s))
      break;
    r = s;
    dist_r = dist_s;
  }
  
  if(i == n) /* all the point on the same side */
    return 0;

  t = dist_r / (dist_r - dist_s);
  for(i = 0; i < 3; i++)
    p0[i] = r[i] + t * (s[i] - r[i]);
  
  i = E_ABS(p0[0] - p[0]) > E_ABS(p0[1] - p[1]) ? 0 : 1;
  t = (q[i] - p[i]) / (p0[i] - p[i] + EPS);

  if( (E_ABS(t) < EPS) || (E_ABS(t-1) < EPS) ||
      !(t > 0 && t < 1.0))
    return 0;
    
  /* 3 dimension */
  T2(z[0], q[2]);
  T2(w1, p[2]);
  T2(w2, p0[2]);
  z[1] = (1 - t) * w1 + t * w2;
  Point_behind_polygon = ((z[0] < z[1]) ? 1 : 0);
  
  return 1;
}

/* Intersection of two lines. */
static double intersect_2Dlines(double *a, double *d, double *b, double *e)
{
  double  ne[2], ab[2];
  double  d_ne, ab_ne;
  
  normal_vec2(ne, e);
  d_ne = dot_product2(d, ne);
  subt_vec2(ab, a, b);
  ab_ne = dot_product2(ab, ne);
    
  /* Are they parallel? */
  if(E_ABS(d_ne) < EPS)
  {
    if(E_ABS(ab_ne) < EPS)
      return(0);
    else
      return(INFINITE);
  }
  else
  {
    return(ab_ne / d_ne);
  }
}


/* Check whether two convex hulls intersect or overlap. */
static int relative_pos_hulls(double z[2], int n1, double **hull1,
                              int n2, double **hull2)
{
  double **a1, **b1;
  double **a2, **b2;
  double **hi1 = hull1 + n1;
  double **hi2 = hull2 + n2;
  double   t1, t2; /* parameters for the intersection point */
  double   dz, previous_dz; /* distance of the intersection point */
  int      delay_decision = 0; /* if the difference is small */
  double   d1[3], d2[3];
  int      i, j;


  for(a1 = hull1, b1 = a1+1; a1 < hi1; a1++, b1++)
  {
    if(b1 == hi1)
      b1 = hull1;
    subt_vec2(d1, *a1, *b1);
    for(a2 = hull2, b2 = a2+1; a2 < hi2; a2++, b2++)
    {
      int parallel;
      
      if(b2 == hi2)
        b2 = hull2;
      subt_vec2(d2, *a2, *b2);
      t1 = intersect_2Dlines(*a1, d1, *a2, d2);
      if( (E_ABS(t1) < EPS) || (E_ABS(t1-1) < EPS) ||
          !(t1 > 0 && t1 < 1.0))
        continue;
    
      /* use a non zero value, one of them should be non zero otherwise */
      if(E_ABS(d2[0]) > EPS)
        t2 = ((*a1)[0] + t1 * d1[0] - (*a2)[0]) / d2[0];
      else if(E_ABS(d2[1]) > EPS)
        t2 = ((*a1)[1] + t1 * d1[1] - (*a2)[1]) / d2[1];
      else
        continue; /* the edge goes through the projection center */
        
      if( (E_ABS(t2) < EPS) || (E_ABS(t2-1) < EPS) ||
          !(t2 > 0 && t2 < 1.0))
        continue;
      
      /* z values of the projected intersection point */
      z[0] = (*a1)[2] + t1 * ((*b1)[2] - (*a1)[2]);
      z[1] = (*a2)[2] + t2 * ((*b2)[2] - (*a2)[2]);
      
      if(Hidden_Distance < 1000) /* projection rays are not parallel, perspective */
      {
        T2(z[0], z[0]);
        T2(z[1], z[1]);
      }
      
      dz = z[1] - z[0];
      
      if(E_ABS(dz) > 1e-2)
        return INTERSECTION;
      else
      {
        /* This is bad for the final priority decision !
        The difference between the z coordinates is quite small.
        If such a case occured, compare the results. If they have
        the same sign, it is all right, otherwise wait with
        the decision. */
        if(delay_decision)
        {
          if(sign(dz) == sign(previous_dz))
            return INTERSECTION;
          else /* rely on a better result */
            if(E_ABS(dz) > E_ABS(previous_dz))
              previous_dz = dz;
        }
        else
        {
          delay_decision = 1;
          previous_dz = dz;
        }
      }
    } /* for (a2) */
  } /* for (a1) */

  if(delay_decision) /* Still??? Bad luck! */
  {
    e_GeneralWarning("Critical intersection of elements! ");
    return INTERSECTION;
  }
  
  /* Get a node which is not equal to a node of hull2.
     It is important, because we are checking for signs,
     and zero is not a proper answer in this case! */
  for(i = 0; i < n1; i++)
  {
    /* check with all nodes */
    for(j = 0; j < n2; j++)
    {
      if(E_ABS(hull1[i][0] - hull2[j][0]) < EPS &&
         E_ABS(hull1[i][1] - hull2[j][1]) < EPS &&
         E_ABS(hull1[i][2] - hull2[j][2]) < EPS    )
        break;
    }
    if(j == n2) /* not equal to any of the nodes, probably we found one */
    {
      int ns, nx;
      double check, cc;
      double normal[3];

      /* check whether the found node is not on the side of
         the polygon */
      ns = 0;
      for(j = 0; j < n2; j++)
      {
        nx = n2 - j - 1;
        normal[0] = hull1[i][1] - hull2[ns][1];
        normal[1] = hull2[ns][0] - hull1[i][0];
        cc = dot_product2(hull2[ns], normal);
        check = dot_product2(hull2[nx], normal) - cc;
        if(E_ABS(check) < EPS)
          break;
        ns = nx;
      }
      if(j == n2)
        break;
    }
  }
  if(i != n1)
  {
    if(inside_poly(hull1[i], n2, hull2, z))
      return POLY1_INSIDE_POLY2;
  }

  /* Get a node which is not equal to a node of hull1. */
  for(i = 0; i < n2; i++)
  {
    for(j = 0; j < n1; j++)
    {
      if(E_ABS(hull2[i][0] - hull1[j][0]) < EPS &&
         E_ABS(hull2[i][1] - hull1[j][1]) < EPS &&
         E_ABS(hull2[i][2] - hull1[j][2]) < EPS    )
        break;
    }
    if(j == n1)
    {
      int ns, nx;
      double check, cc;
      double normal[3];

      /* check whether the found node is not on the side of
         the polygon */
      ns = 0;
      for(j = 0; j < n1; j++)
      {
        nx = n1 - j - 1;
        normal[0] = hull2[i][1]  - hull1[ns][1];
        normal[1] = hull1[ns][0] - hull2[i][0];
        cc = dot_product2(hull1[ns], normal);
        check = dot_product2(hull1[nx], normal) - cc;
        if(E_ABS(check) < EPS)
          break;
        ns = nx;
      }
      if(j == n1)
        break;
    }
  }
  if(i != n2)
  {
    if(inside_poly(hull2[i], n1, hull1, z))
      return POLY2_INSIDE_POLY1;
  }

  /* There is no relation between the convex hulls! */

  return DISJOINT;
}


/* This function determines which of the two elements is closer to the projection
   centre (should be drawn later), or there is no ordering between the two. */
static int which_elem_first(POLYGON *elem1, POLYGON *elem2, int plane)
{
  int         first = NONE;
  PLANE  *sep_plane;
  double      z[2];

  /* one of them is already covered, no need for further check */
  if(elem1->obscured || elem2->obscured)
    return NONE;
  
  /* Do the bounding boxes overlap ? Uses projected coordinates */
  if(!overlap(elem1->box_min, elem1->box_max,
              elem2->box_min, elem2->box_max))
    return NONE; /* bounding boxes do not overlap, no ordering */

  /* rotated z coords are far away ? */
  if(elem1->box_min[2] > elem2->box_max[2])
  {
    first = elem1->index; 
    if(plane) 
      goto intersect_outlines; 
    else 
      return first; 
  }
  else if(elem2->box_min[2] > elem1->box_max[2])
  { 
    first = elem2->index; 
    if(plane) 
      goto intersect_outlines; 
    else
      return first;
  }

  /* try with some kind of separating planes */
  sep_plane = elem1->sep_plane + (elem2->index - elem1->index - 1);

  /* circumscribing spheres do not intersect overlap each other */
  if(sep_plane->info == USE_SOAP_BUBBLES)
  {
    /* distance from the eye position */
    if(average_dist(elem1) < average_dist(elem2))
    { 
      first = elem1->index; 
      if(plane) 
        goto intersect_outlines; 
      else 
        return first; 
    }
    else
    { 
      first = elem2->index; 
      if(plane) 
        goto intersect_outlines; 
      else
        return first;
    }
  }
  /* there is a separating plane betweet the two elements */
  else if(sep_plane->info != NOT_FOUND)
  {
    /* check that the eye position is on the same side as the element
    with the smaller index, can the other elem obscure the first one ? */
    if(which_side(Hidden_Proj_center, sep_plane) == sep_plane->info)
    { 
      first = elem1->index; 
      if(plane) 
        goto intersect_outlines; 
      else 
        return first; 
    }
    else /* if 0, tough luck */
    { 
      first = elem2->index; 
      if(plane) 
        goto intersect_outlines; 
      else
        return first;
    }
  }

  /* final decision: intersect convex hulls and compare the
  z-values of the two space points that correspond to the
  first intersection point */

  intersect_outlines:
  
  switch(relative_pos_hulls(z, elem1->no_vertices, elem1->screen_coords,
                            elem2->no_vertices, elem2->screen_coords))
  {
    case INTERSECTION:
      if(z[0] > z[1])
        return elem1->index;
      else
        return elem2->index;
    case POLY2_INSIDE_POLY1:
      if(first == elem1->index) /* and always convex !!! */
      {
        elem2->obscured = 1;
        return first;
      }
      if(first == NONE)
      {
        if(Point_behind_polygon)
          return elem1->index;
        else
          return elem2->index;
      }
      return first;
    case POLY1_INSIDE_POLY2:
      if(first == elem2->index) /* and always convex !!! */
      {
        elem1->obscured = 1;
        return first;
      }
      if(first == NONE)
      {
        if(Point_behind_polygon)
          return elem2->index;
        else
          return elem1->index;
      }
      return first;
    case DISJOINT:
      return NONE;
  }

}


/* ############################################################################ */
/*                   Sorting the elements according to priority                 */
/* ############################################################################ */


static int mikado_select(int n, POLYGON **res_obj)
{
  int  i, j;
  int  index, index1, index2;

  for(i = 0; i < n; i++)
  {
    res_obj[i]->is_in_front = res_obj[i]->is_in_back = 1;
  }
  
  for(i = 0; i < n; i++)
  {
    index1 = res_obj[i]->index;

    if(res_obj[i]->no_vertices == 2)
      break;

    for(j = i+1; j < n; j++)
    {
      index2 = res_obj[j]->index;
      
      if(res_obj[i]->sep_plane[index2-index1-1].priority == index1)
      {
        res_obj[i]->is_in_back = 0;
        res_obj[j]->is_in_front = 0;
      }
      else if(res_obj[i]->sep_plane[index2-index1-1].priority == index2)
      {
        res_obj[i]->is_in_front = 0;
        res_obj[j]->is_in_back = 0;
      }
    }
  }
      
  /* check whether search was successfull */
  for(i = 0; i < n; i++)
  {
    if(res_obj[i]->is_in_front || res_obj[i]->is_in_back)
      return 1;
  }
  return 0;

}

int sort_objects(int *order, int n, POLYGON *poly, int plane)
{
  POLYGON **res_obj;
  POLYGON **obj;
  int    i, j, k, i1, i2;
  int    i_min, i_max;


  /* determine priorities */
  for(i = 0; i < n; i++)
  {
    if(poly[i].no_vertices == 2)
      break;
#ifdef __DEBUG__
    if(plane)
      printf("\n");
#endif
    for(j = i+1; j < n; j++)
    {
      poly[i].sep_plane[j-i-1].priority = which_elem_first(poly+i, poly+j, plane);
#ifdef __DEBUG__
      if(plane)
        printf("%d ", poly[i].sep_plane[j-i-1].priority);
#endif
    }
  }

  /* do the sorting */

  res_obj = e_calloc(n, sizeof(POLYGON *));
  for(i = 0; i < n; i++)
    res_obj[i] = poly+i;

  i_min = 0;
  i_max = n;
  k = n;

  while(k)
  {
    if(!mikado_select(k, res_obj))
    {
      if(plane)
      {
        for(i = 0; i < k; i++)
        {
          printf("\nelem %d", (res_obj[i])->index);
        }
        e_GeneralWarning("Gordian knot while sorting!");
        return 0;
      }
      else
        return 0;
    }
    k = 0; i1 = i_min; i2 = i_max;
    for(i = i1, obj = res_obj; i < i2; i++, obj++)
    {
      j = (*obj)->index;
      
      if((*obj)->is_in_back)
        order[i_min++] = j;
      else if((*obj)->is_in_front)
        order[--i_max] = j;
      else
        res_obj[k++] = (*obj);
    }
/*    printf("\n-- res: %d", k); */
  }

  e_free(res_obj);
  return 1;
  
}


/* ############################################################################ */
/*                            Setup before sorting                              */
/* ############################################################################ */
/* All geometry checking is done in real 3D  */

#define Min_max(min, max, x) \
        if(x < min) min = x; \
        if(x > max) max = x;

#define Min_max_vec(min_v, max_v, v) { \
        Min_max((min_v)[0], (max_v)[0], (v)[0]); \
        Min_max((min_v)[1], (max_v)[1], (v)[1]); \
        Min_max((min_v)[2], (max_v)[2], (v)[2]); \
        }

#define copy_vec(r, v) \
        ((r)[0] = (v)[0], (r)[1] = (v)[1], (r)[2] = (v)[2])


void setup_proj_center(double azim, double elev, double dist, double *target)
{
  double a, c, x, e, pi;

  Hidden_Proj_center[0] = 1;
  Hidden_Proj_center[1] = 1;
  Hidden_Proj_center[2] = 1;

  a = azim;
  /* protect the tangent calculation */
  if(a == 0 || a == M_PI)
    a += 1e-14;

  e = elev;

  /* elevation angle */
  while(e > 2*M_PI || e < -2*M_PI)
  {
    e += 2 * M_PI * (e > 0 ? -1 : 1);
  }
  if(e < 0)
    e += 2*M_PI;


  /* determine projection centre */
  if(e > 0.5*M_PI && e < 1.5*M_PI)
  {
    a += M_PI;
  }

  /* the azimuth angle */
  while(a > 2*M_PI || a < -2*M_PI)
  {
    a += 2 * M_PI * (a > 0 ? -1 : 1);
  }
  if(a < 0)
    a += 2*M_PI;


  Hidden_Proj_center[2] *= dist * sin(e);


  if(a > M_PI && a < 2*M_PI)
  {
    Hidden_Proj_center[0] = -1;
  }

  if(   (a > 0 && a < 0.5*M_PI)
     || (a > 1.5*M_PI && a < 2*M_PI) )
  {
    Hidden_Proj_center[1] = -1;
  }

  a = a - M_PI/2;
  c = E_ABS(tan(a));
  
  x = sqrt((dist * dist - Hidden_Proj_center[2] * Hidden_Proj_center[2]) / (c * c + 1));
  
  Hidden_Proj_center[0] *= x;
  Hidden_Proj_center[1] *= c * x;

  Hidden_Proj_center[0] += target[0];
  Hidden_Proj_center[1] += target[1];
  Hidden_Proj_center[2] += target[2];

  Hidden_Distance = dist;

}


static void create_sep_planes(PLANE *s, POLYGON obj1, POLYGON obj2)
{
  double dist[3];
  int    corner[8][3] = { 0,0,0, 1,0,0, 1,1,0, 0,1,0, 
                          0,0,1, 1,0,1, 1,1,1, 0,1,1  }; /* corner nodes */
  int    normal[6][3] = { 0,0,-1, /* bottom face */
                          0,-1,0, 1,0,0, 0,1,0, -1,0,0, /* sides */
                          0,0,1 }; /* top face */
  double  v[3], v1[3], v2[3];
  int    *nn;
  double  cnst, fcnst;
  double  face_normal[3];
  int     i, j, k, pos, neg, equal;
  double  tt;

  /* they are far away from each other */
  subt_vec(dist, obj1.center, obj2.center);
  if(Length(dist) > (obj1.max_rad + obj2.max_rad))
  {
    s->info = USE_SOAP_BUBBLES;
    return;
  }

  /* check whether faces of the bounding boxes are separating planes */
  
  /* check obj1 bounding box faces with obj2 nodes */
  for(i = 0; i < 6; i++)
  {
    nn = normal[i];

    /* position vector on the face of the box */
    for(k = 0; k < 3; k++)
    {
      v[k] = 0.0;
      if(nn[k] == 1)
        v[k] = obj1.box_max[k];
      else if(nn[k] == -1)
        v[k] = obj1.box_min[k];
    }
    fcnst = dot_product(nn, v);

    /* check the nodes of the element, that not all of them
    lies on the bounding face */
    for(k = 0; k < obj1.no_vertices; k++)
    {
      cnst = dot_product(nn, obj1.world_coords[k])-fcnst;
      if(E_ABS(cnst) > EPS)
        break;
    }
    /* all nodes are on the face, not good for this kind of checking */
    if(k == obj1.no_vertices)
      continue;
      
    equal = 0;
    for(j = 0; j < 8; j++)
    {
      /* coords of the bounding box of other object */
      v[0] = (corner[j][0] ? obj2.box_max[0] : obj2.box_min[0]);
      v[1] = (corner[j][1] ? obj2.box_max[1] : obj2.box_min[1]);
      v[2] = (corner[j][2] ? obj2.box_max[2] : obj2.box_min[2]);
      
      tt = dot_product(nn, v)-fcnst;
    
      if(sign(tt) == sign(cnst))
        break;
      if(tt == 0)
        equal++;
    }
    
    if(j == 8 && equal != 8)
    {
      s->info = -1;
      s->cnst = fcnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = nn[k];
      return;
    }
  }

  /* check obj2 bounding box faces with obj1 nodes */
  for(i = 0; i < 6; i++)
  {
    nn = normal[i];

    /* position vector on the face of the box */
    for(k = 0; k < 3; k++)
    {
      v[k] = 0.0;
      if(nn[k] == 1)
        v[k] = obj2.box_max[k];
      else if(nn[k] == -1)
        v[k] = obj2.box_min[k];
    }
    fcnst = dot_product(nn, v);

    /* check the nodes of the element, that not all of them
    lies on the bounding face */
    for(k = 0; k < obj2.no_vertices; k++)
    {
      cnst = dot_product(nn, obj2.world_coords[k])-fcnst;
      if(E_ABS(cnst) > EPS)
        break;
    }
    if(k == obj2.no_vertices)
      continue;
    
    
    equal = 0;
    for(j = 0; j < 8; j++)
    {
      /* coords of the bounding box of other object */
      v[0] = (corner[j][0] ? obj1.box_max[0] : obj1.box_min[0]);
      v[1] = (corner[j][1] ? obj1.box_max[1] : obj1.box_min[1]);
      v[2] = (corner[j][2] ? obj1.box_max[2] : obj1.box_min[2]);

      tt = dot_product(nn, v)-fcnst;
    
      if(sign(tt) == sign(cnst))
        break;
      if(tt == 0)
        equal++;
    }
    
    if(j == 8 && equal != 8)
    {
      s->info = 1;
      s->cnst = fcnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = nn[k];
      return;
    }
  }


  /* last chance, the face of the element is a separating plane */
  if(obj1.no_vertices > 2) /* only >2D elements */
  {
    pos = 0; neg = 0;
    for(i = 0; i < obj2.no_vertices; i++)
    {
      cnst = dot_product(obj1.local_plane.normal, obj2.world_coords[i]) - 
             obj1.local_plane.cnst;
      if(cnst > 0)
        pos++;
      else if(cnst < 0)
        neg++;
    }
    if(   (pos == 0 && neg != 0)
       || (pos != 0 && neg == 0) )
    {
      s->info = (pos ? -1 : 1);
      s->cnst = obj1.local_plane.cnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = obj1.local_plane.normal[k];
      return;
    }
  }


  if(obj2.no_vertices > 2) /* only >2D elements */
  {
    pos = 0; neg = 0;
    for(i = 0; i < obj1.no_vertices; i++)
    {
      cnst = dot_product(obj2.local_plane.normal, obj1.world_coords[i]) - 
             obj2.local_plane.cnst;
      if(cnst > 0)
        pos++;
      else if(cnst < 0)
        neg++;
    }
    if(   (pos == 0 && neg != 0)
       || (pos != 0 && neg == 0) )
    {
      s->info = (pos ? 1 : -1);
      s->cnst = obj2.local_plane.cnst;
      for(k = 0; k < 3; k++)
        s->normal[k] = obj2.local_plane.normal[k];
      return;
    }
  }

  s->info = NOT_FOUND;
}




void setup_hidden_line1(int NPoly, POLYGON *pool)
{
  double v[3];
  int i, j;

  double v1[3];
  double v2[3];
  double a;
  double r[3];
  double rotated[3];


  for(i = 0; i < NPoly; i++)
  {

    /* setup the plane of the element */
    if(pool[i].no_vertices > 2)
    {
      subt_vec(v1, pool[i].world_coords[0], pool[i].world_coords[1]);
      subt_vec(v2, pool[i].world_coords[2], pool[i].world_coords[1]);

      pool[i].local_plane.normal[0] = v1[1] * v2[2] - v2[1] * v1[2];
      pool[i].local_plane.normal[1] = v2[0] * v1[2] - v1[0] * v2[2];
      pool[i].local_plane.normal[2] = v1[0] * v2[1] - v2[0] * v1[1];
      a = sqrt(pool[i].local_plane.normal[0] * pool[i].local_plane.normal[0] +
               pool[i].local_plane.normal[1] * pool[i].local_plane.normal[1] +
               pool[i].local_plane.normal[2] * pool[i].local_plane.normal[2]);
      pool[i].local_plane.normal[0] /= a;
      pool[i].local_plane.normal[1] /= a;
      pool[i].local_plane.normal[2] /= a;
    
      /* calculate the dot product of the normal and a point on the plane */
      pool[i].local_plane.cnst = dot_product(pool[i].local_plane.normal, 
                                             pool[i].world_coords[0]           );
    }


    /* determine max radius for soap bubble  */
    /* and create 3D bounding boxes */
    subt_vec(r, pool[i].world_coords[0], pool[i].center);
    pool[i].max_rad = Length(r);

    copy_vec(pool[i].box_min, pool[i].world_coords[0]);
    copy_vec(pool[i].box_max, pool[i].world_coords[0]);

    for(j = 1; j < pool[i].no_vertices; j++)
    {
      subt_vec(r, pool[i].world_coords[j], pool[i].center);
      if( pool[i].max_rad < Length(r) )
      {
        pool[i].max_rad = Length(r);
      }

      Min_max_vec(pool[i].box_min, 
                  pool[i].box_max, 
                  pool[i].world_coords[j]);

    }
  }

  for(i = 0; i < NPoly; i++)
  {
    if(pool[i].no_vertices != 2)
    {
      pool[i].sep_plane = e_calloc(NPoly-i-1, sizeof(PLANE));

      for(j = i+1; j < NPoly; j++)
      {
        create_sep_planes(&(pool[i].sep_plane[j-i-1]),
                          pool[i], pool[j] );
      }
    }
  }
}


void setup_hidden_line2(int NPoly, POLYGON *pool)
{
  int i, j;
  double rotated[3];

  for(i = 0; i < NPoly; i++)
  {
    pool[i].obscured = 0;

    /* create 2D projected data */
    pool[i].center[0] = 0.0;
    pool[i].center[1] = 0.0;
    pool[i].center[2] = 0.0;
    for(j = 0; j < pool[i].no_vertices; j++)
    {
      /* projection */
      p_3Dworld_to_screen(rotated, pool[i].world_coords[j]);
      
      /* center of rotated element */
      pool[i].center[0] += rotated[0];
      pool[i].center[1] += rotated[1];
      pool[i].center[2] += rotated[2];
    }
    pool[i].center[0] /= pool[i].no_vertices;
    pool[i].center[1] /= pool[i].no_vertices;
    pool[i].center[2] /= pool[i].no_vertices;

    /* projected contour */
    copy_vec(pool[i].box_min, pool[i].screen_coords[0]);
    copy_vec(pool[i].box_max, pool[i].screen_coords[0]);
    for(j = 1; j < pool[i].no_vertices; j++)
      Min_max_vec(pool[i].box_min, 
                  pool[i].box_max, 
                  pool[i].screen_coords[j]);

  }

}





